/*
 *  Copyright (c) Members of the EGEE Collaboration. 2004.
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  GLite Data SRM Util CPP: ExponentialBackoffLogics
 *
 *  Authors: Zsolt Molnar <zsolt.molnar@cern.ch - http://www.zsoltmolnar.hu>
 *  Version info: $Id: BackoffLogicExecutor.h,v 1.2 2009/07/30 17:40:34 molnarzs Exp $
 *  Release: $Name:  $
 *
 */
#ifndef GLITE_DATA_SRM_UTIL_CPP_BACKOFF_LOGIC_EXECUTOR_H_
#define GLITE_DATA_SRM_UTIL_CPP_BACKOFF_LOGIC_EXECUTOR_H_

#include "glite/data/srm/util/types.h"
#include "glite/data/srm/util/details/IBackoffLogic.h"

using glite::data::srm::util::details::IBackoffLogic;
using glite::data::srm::util::SrmStatus;

struct BackoffLogicExecutor
{
	template <typename Request>
	static SrmStatus executeAsync(Request& req)
	{
		IBackoffLogic& logic = *(req.getBackoffLogic());
		logic.reset();

		do {
			req.executeOneCall();

			if (req.status.code == SrmStatus::SRM_INTERNAL_ERROR) {
				int eWT = req.estimatedWaitTime();

				if (eWT > 0) {
					logic.setSleepTime(eWT);
                } else if (eWT == 0) {
                    // This case means that the request timed out on the server side. See:
                    // https://sdm.lbl.gov/srm-wg/doc/SRM.v2.2.html
                    logic.forceTimeout();
                    break;
                }

				logic.sleep();
			} else {
				break;
			}
		}
		while (!logic.timeout());

		SrmStatus result;

		if (logic.timeout()) {
			req.abort();
			result = SrmStatus(SrmStatus::SRM_REQUEST_TIMED_OUT, "Request timeout (internal error or too long processing), request aborted");
		} else {
			result = SrmStatus(req.status.code, req.status.message);
		}

		return result;
	}

	template <typename Request>
	static SrmStatus updateAsyncUntilQueued(Request& req)
	{
		while (!req.backoff());
		return req.status;
	}

	template <typename Request>
	static SrmStatus executeSync(Request& req)
	{
		IBackoffLogic& logic = *(req.getBackoffLogic());
		logic.reset();
		SrmStatus result;

		do {
			result = req.executeOneCall();

			if (result.code == SrmStatus::SRM_INTERNAL_ERROR) {
				int eWT = req.estimatedWaitTime();

				if (eWT > 0) {
					logic.setSleepTime(eWT);
                } else if (eWT == 0) {
                    // This case means that the request timed out on the server side. See:
                    // https://sdm.lbl.gov/srm-wg/doc/SRM.v2.2.html
                    logic.forceTimeout();
                    break;
                }

				logic.sleep();
			} else {
				break;
			}
		}
		while (!logic.timeout());

		if (logic.timeout())
			result = SrmStatus(SrmStatus::SRM_REQUEST_TIMED_OUT, "Request timeout due to internal error");

		return result;
	}
};

#endif //GLITE_DATA_SRM_UTIL_CPP_EXPONENTIAL_BACKOFF_LOGIC_EXECUTOR_H_
