/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004.
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  GLite Data SRM Util CPP : ErrorPatterns class
 *
 *  Authors: Paolo Tedesco <paolo.tedesco@cern.ch>
 *  Version info: $Id: ErrorPatterns.cpp,v 1.2 2007/11/05 10:36:52 ptedesco Exp $
 *  Release: $Name:  $
 *
 */

#include "ErrorPatterns.h"

using namespace glite::data::agents;
using namespace glite::data::srm::util;

namespace{
// constants
// non-SRM error codes
static const std::string GsiAuthenticationErrorCategory  = "GSI_AUTHENTICATION_ERROR";
static const std::string TcpConnectionErrorCategory      = "TCP_CONNECTION_ERROR";
} // namespace
/**
 * Singleton instance
 */
ErrorPatterns& ErrorPatterns::instance(){
    static ErrorPatterns s_instance;
    return s_instance;
}
/**
 * return true if given message corresponds to a GSI authentication error 
 * pattern
 * @param msg [IN] the message to match
 * @return true if match found, false otherwise
 */
bool ErrorPatterns::isGsiAuthenticationError(const std::string& msg){
    return m_patterns.checkCategory(msg,GsiAuthenticationErrorCategory);
}
/**
 * return true if given message corresponds to a TCP connection error
 * pattern
 * @param msg [IN] the message to match
 * @return true if match found, false otherwise
 */
bool ErrorPatterns::isTcpConnectionError(const std::string& msg){
    return m_patterns.checkCategory(msg,TcpConnectionErrorCategory);
}
/**
 * Translate given message to an SrmStatus code applying internal patterns.
 * @param msg [IN] the message to match
 * @return SrmStatus code (SRM_FAILURE if no match found).
 */
SrmStatus ErrorPatterns::findSrmError(const std::string& msg){
    SrmStatus result;
    result.code = SrmStatus::SRM_FAILURE;
    // Set the message
    result.message = msg;
    // Find the correct Error
    SrmCategoriesMap::const_iterator it = m_srmCategoriesMap.begin();
    for (;it != m_srmCategoriesMap.end();++it){
        const std::string& category = it->first;
        if (m_patterns.checkCategory(msg,category)){
            result.code = it->second;
            break;
        }
    }
    return result;
}

/**
 * constructor
 */
ErrorPatterns::ErrorPatterns():m_patterns("glite-data-srm-util-cpp.patterns.xml"){
    m_srmCategoriesMap["SRM_SUCCESS"]                   = SrmStatus::SRM_SUCCESS;
    m_srmCategoriesMap["SRM_FAILURE"]                   = SrmStatus::SRM_FAILURE;
    m_srmCategoriesMap["SRM_AUTHENTICATION_FAILURE"]    = SrmStatus::SRM_AUTHENTICATION_FAILURE;
    m_srmCategoriesMap["SRM_AUTHORIZATION_FAILURE"]     = SrmStatus::SRM_AUTHORIZATION_FAILURE;
    m_srmCategoriesMap["SRM_INVALID_REQUEST"]           = SrmStatus::SRM_INVALID_REQUEST;
    m_srmCategoriesMap["SRM_INVALID_PATH"]              = SrmStatus::SRM_INVALID_PATH;
    m_srmCategoriesMap["SRM_FILE_LIFETIME_EXPIRED"]     = SrmStatus::SRM_FILE_LIFETIME_EXPIRED;
    m_srmCategoriesMap["SRM_SPACE_LIFETIME_EXPIRED"]    = SrmStatus::SRM_SPACE_LIFETIME_EXPIRED;
    m_srmCategoriesMap["SRM_EXCEED_ALLOCATION"]         = SrmStatus::SRM_EXCEED_ALLOCATION;
    m_srmCategoriesMap["SRM_NO_USER_SPACE"]             = SrmStatus::SRM_NO_USER_SPACE;
    m_srmCategoriesMap["SRM_NO_FREE_SPACE"]             = SrmStatus::SRM_NO_FREE_SPACE;
    m_srmCategoriesMap["SRM_DUPLICATION_ERROR"]         = SrmStatus::SRM_DUPLICATION_ERROR;
    m_srmCategoriesMap["SRM_NON_EMPTY_DIRECTORY"]       = SrmStatus::SRM_NON_EMPTY_DIRECTORY;
    m_srmCategoriesMap["SRM_TOO_MANY_RESULTS"]          = SrmStatus::SRM_TOO_MANY_RESULTS;
    m_srmCategoriesMap["SRM_INTERNAL_ERROR"]            = SrmStatus::SRM_INTERNAL_ERROR;
    m_srmCategoriesMap["SRM_FATAL_INTERNAL_ERROR"]      = SrmStatus::SRM_FATAL_INTERNAL_ERROR;
    m_srmCategoriesMap["SRM_NOT_SUPPORTED"]             = SrmStatus::SRM_NOT_SUPPORTED;
    m_srmCategoriesMap["SRM_REQUEST_QUEUED"]            = SrmStatus::SRM_REQUEST_QUEUED;
    m_srmCategoriesMap["SRM_REQUEST_INPROGRESS"]        = SrmStatus::SRM_REQUEST_INPROGRESS;
    m_srmCategoriesMap["SRM_REQUEST_SUSPENDED"]         = SrmStatus::SRM_REQUEST_SUSPENDED;
    m_srmCategoriesMap["SRM_ABORTED"]                   = SrmStatus::SRM_ABORTED;
    m_srmCategoriesMap["SRM_RELEASED"]                  = SrmStatus::SRM_RELEASED;
    m_srmCategoriesMap["SRM_FILE_PINNED"]               = SrmStatus::SRM_FILE_PINNED;
    m_srmCategoriesMap["SRM_FILE_IN_CACHE"]             = SrmStatus::SRM_FILE_IN_CACHE;
    m_srmCategoriesMap["SRM_SPACE_AVAILABLE"]           = SrmStatus::SRM_SPACE_AVAILABLE;
    m_srmCategoriesMap["SRM_LOWER_SPACE_GRANTED"]       = SrmStatus::SRM_LOWER_SPACE_GRANTED;
    m_srmCategoriesMap["SRM_DONE"]                      = SrmStatus::SRM_DONE;
    m_srmCategoriesMap["SRM_PARTIAL_SUCCESS"]           = SrmStatus::SRM_PARTIAL_SUCCESS;
    m_srmCategoriesMap["SRM_REQUEST_TIMED_OUT"]         = SrmStatus::SRM_REQUEST_TIMED_OUT;
    m_srmCategoriesMap["SRM_LAST_COPY"]                 = SrmStatus::SRM_LAST_COPY;
    m_srmCategoriesMap["SRM_FILE_BUSY"]                 = SrmStatus::SRM_FILE_BUSY;
    m_srmCategoriesMap["SRM_FILE_LOST"]                 = SrmStatus::SRM_FILE_LOST;
    m_srmCategoriesMap["SRM_FILE_UNAVAILABLE"]          = SrmStatus::SRM_FILE_UNAVAILABLE;
    m_srmCategoriesMap["SRM_CUSTOM_STATUS"]             = SrmStatus::SRM_CUSTOM_STATUS;
}
/**
 * destructor
 */
ErrorPatterns::~ErrorPatterns(){
}


