/*
 *  Copyright (c) Members of the EGEE Collaboration. 2004.
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  GLite Data SRM Util CPP: Stat Class
 *
 *  Authors: Zsolt Molnar <zsolt.molnar@cern.ch - http://www.zsoltmolnar.hu>
 *  Version info: $Id: ExponentialBackoffLogic.cpp,v 1.2 2009/07/30 17:40:34 molnarzs Exp $
 *  Release: $Name:  $
 *
 */
#include "ExponentialBackoffLogic.h"
#include <unistd.h>

void ExponentialBackoffLogic_DefaultSleep::sleep(const unsigned int seconds)
{
	::sleep(seconds);
};

unsigned int ExponentialBackoffLogic_DefaultRandom::random(unsigned int r)
{
	return r;
};

ExponentialBackoffLogic_Base::ExponentialBackoffLogic_Base(double c, double timeout, double base, double maxSleepTime)
	: BackoffLogic(maxSleepTime),
	  _c(c),
	  _timeout(timeout),
	  _base(base),
	  _startTime(new boost::posix_time::ptime(boost::posix_time::second_clock::universal_time())),
	  _timeoutHappened(false),
	  _n(1),
	  _userSleepTime(-1),
	  _maxSleepTime(static_cast<unsigned int>(maxSleepTime))
{
	// EMPTY
}

ExponentialBackoffLogic_Base::~ExponentialBackoffLogic_Base()
{
	// EMPTY
}

bool ExponentialBackoffLogic_Base::timeout() const
{
	if (!_timeoutHappened)
		_timeoutHappened = (_elapsed() > _timeout);

	return _timeoutHappened;
}

void ExponentialBackoffLogic_Base::reset()
{
	reset( (time_t)NULL );
}

void ExponentialBackoffLogic_Base::reset(const time_t* t)
{
	_timeoutHappened = false;
	_n = 1;

	if (t)
		_startTime.reset(new boost::posix_time::ptime(boost::posix_time::from_time_t(*t)));
	else
		_startTime.reset(new boost::posix_time::ptime(boost::posix_time::second_clock::universal_time()));

	_userSleepTime = -1;
}

void ExponentialBackoffLogic_Base::sleep()
{
	unsigned int nextSleepTime = 0;

	// Check if the sleep time was set in a previous call.
	if (_userSleepTime >= 0) {
		// Specify it as actual sleep time
		nextSleepTime = _userSleepTime;
		// reset the exponential attempt counter to zero
		_n = 1;
		// reset the "sleep time was set" state
		_userSleepTime = -1;
	} else {
		// if not: use the exponential method, using the algorithm parameters
		// and the actual attempt counter, and set the actual sleep time
		nextSleepTime = static_cast<unsigned int>(_c * _random(
				static_cast<unsigned int>(std::pow(_base, static_cast<double>(_n))) - 1));

		++_n;
	}

	nextSleepTime = std::min(nextSleepTime, _maxSleepTime);
	// Check if it would be timeout. If so, set the "timeout" flag and return
	if (_elapsed() + nextSleepTime > _timeout) {
		_timeoutHappened = true;
	} else {
		// If not, then go to sleep.
		_sleep(nextSleepTime);
	}
}

void ExponentialBackoffLogic_Base::setSleepTime(unsigned int t)
{
	_userSleepTime = static_cast<int>(t);
}

void ExponentialBackoffLogic_Base::setTimeout(unsigned int t)
{
	_timeout = static_cast<double>(t);
}

void ExponentialBackoffLogic_Base::forceTimeout()
{
    _timeoutHappened = true;
}

long ExponentialBackoffLogic_Base::_elapsed() const
{
	boost::posix_time::ptime now(boost::posix_time::second_clock::universal_time());
	assert(_startTime.get());
	boost::posix_time::time_period elapsed(*_startTime, now);
	return elapsed.length().total_seconds();
}
