/*
 *  Copyright (c) Members of the EGEE Collaboration. 2004.
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  GLite Data SRM Util CPP: ExponentialBackoffLogics
 *
 *  Authors: Zsolt Molnar <zsolt.molnar@cern.ch - http://www.zsoltmolnar.hu>
 *  Version info: $Id: ExponentialBackoffLogic.h,v 1.2 2009/07/30 17:40:34 molnarzs Exp $
 *  Release: $Name:  $
 *
 */
#ifndef GLITE_DATA_SRM_UTIL_CPP_EXPONENTIAL_BACKOFF_LOGIC_H_
#define GLITE_DATA_SRM_UTIL_CPP_EXPONENTIAL_BACKOFF_LOGIC_H_

#include "BackoffLogic.h"

#include <boost/date_time/posix_time/posix_time.hpp>
#include <boost/scoped_ptr.hpp>
#include <cmath>

using namespace glite::data::srm::util::details;
class TestExponentialBackoffLogic;

/** Default process suspending policy (based on system sleep) */
struct ExponentialBackoffLogic_DefaultSleep
{
	static void sleep(const unsigned int seconds);
};

/** Default randomness policy (no randomness, simply return the parameter) */
struct ExponentialBackoffLogic_DefaultRandom
{
	/** Generate a random number between 0 and r (OK, now it is always r) */
	static unsigned int random(unsigned int r);
};

/** Default values for the exponential backoff algorithm:
 *
 *  c() * base()^n, at the n-th attempt
 */
struct ExponentialBackoffLogic_DefaultValues
{
	/** The global constant */
	static double c() { return 5;  /* sec */ }
	/** The global timeout */
	static double timeout() { return 300;  /* sec */}
	/** The base of the formula */
	static double base() { return 2; }
	/** Whatever the estimated weit time or the next exponential wait time,
	 * wait no more that this value */
	static double maxSleepTime() { return 30; /* sec */ }
};

/**
 * Implementation of the exponential backoff algorithm. The class properties are:
 *
 * - The backoff algorithm parameters can be given either in runtime or in compile time
 * - The algorithm may introduce randomness as well
 * - The sleep method suspends the process for the calculated amount of time.
 *   However, it is also parametrizable: it may be the system sleep (default
 *   implementation) or test sleep that may shorten test times by not waiting
 *   for the whole sleep timeout physically.
 * - The exponential timeout my be reset at any time, with the setSleepTime
 *   method. It has effect only for the very next sleep call. If a sleep call is not
 *   preceded by this method, then the sleep time is calculated by the exponential
 *   method - until an algorithm-resetting setSleepTime call (algorithm reset: n goes to one
 *   (n is the attempt counter)
 * - The total sleep times are accumulated during the lifetime of the object,
 *   until reaching AlgorithmConstants::timeout().
 *
 * Typical use of the class:
 *
 * ExponentialBackoffLogic logic; // default parameters: system sleep, no randomness
 *
 * while (we_need_to_backoff() && !logic.timeout()) {
 *     if (for_a_reason_wait_time_should_be_given())
 *         logic.setSleepTime(the_wait_time);
 *
 *     logic.sleep();
 * }
 */

/** The logic */
class ExponentialBackoffLogic_Base : public BackoffLogic
{
    friend class TestExponentialBackoffLogic;

public:
	ExponentialBackoffLogic_Base(double c, double timeout, double base, double maxSleepTime);
	virtual ~ExponentialBackoffLogic_Base();

	/** Check if the global timeout has been reached */
	virtual bool timeout() const;

	/** reset the global timeout counter and the exponential attempt counter to zero */
	virtual void reset();
	/** reset the exponential attempt counter to zero, and set an explicit start
	 * time value.  */
	void reset(const time_t* t);

	/** Send the process to sleep for the actual backoff period. If the sleep time
	 * would cause timeout, return immediately. In this case, the timeout
	 * will return true. If there is no specific wait time set by a previous
	 * setSleepTime call, then exponential backoff algorithm is used. In this case,
	 * the method advances the backoff times, according to the exponential formula:
	 *
	 * If the n-th attempt returns internal error, sleep k * c seconds, where
	 *
	 * - c is AlgorithmConstants::c()
	 * - k is a random number between 0 and pow(AlgorithmConstants::base(), n) - 1
	 *
	 * The random number is generated by ExponentialBackoffLogic_DefaultRandom::random()
	 * It means, that the "random" number may be a simple generic constant (no randomness).
	 *
	 * The function that send the process sleep is given by ExponentialBackoffLogic_DefaultSleep::sleep()
	 *
	 * That means that especially for testing purposes, we can specify different "clocks"
	 * or process suspending policies.
	 * */
	virtual void sleep();

	virtual void setSleepTime(unsigned int t);
	virtual void setTimeout(unsigned int t);
    virtual void forceTimeout();

protected:
	virtual unsigned int _random(unsigned int r) = 0;
	virtual void _sleep(const unsigned int seconds) = 0;

private:
	long _elapsed() const;

	const double _c;
	double _timeout;
	const double _base;
	boost::scoped_ptr<boost::posix_time::ptime> _startTime;
	// Mutable, because it caches the timeout state (avoid from re-calculating)
	mutable bool _timeoutHappened;
	unsigned int _n; // the exponential attempt counter
	int _userSleepTime; // The explicitly set sleep time for the next sleep call.
	                    // -1 means "not set"
	unsigned int _maxSleepTime;
};

/** Policy implementation */
template <
	typename AlgorithmConstants = ExponentialBackoffLogic_DefaultValues,
	typename SleepPolicy = ExponentialBackoffLogic_DefaultSleep,
	typename RandomPolicy = ExponentialBackoffLogic_DefaultRandom
>
class ExponentialBackoffLogic : public ExponentialBackoffLogic_Base
{
public:
	ExponentialBackoffLogic(const AlgorithmConstants& c = AlgorithmConstants())
		: ExponentialBackoffLogic_Base(c.c(), c.timeout(), c.base(), c.maxSleepTime()),
		  _constants(c)
	{
		// EMPTY
	}

	virtual ~ExponentialBackoffLogic()
	{
		//EMPTY
	}

	using ExponentialBackoffLogic_Base::timeout;
	using ExponentialBackoffLogic_Base::reset;
	using ExponentialBackoffLogic_Base::sleep;
	using ExponentialBackoffLogic_Base::setSleepTime;

	/** Clone the object */
	virtual IBackoffLogic* clone() const
	{
		return new ExponentialBackoffLogic(_constants);
	}

private:
	virtual unsigned int _random(unsigned int r)
	{
		return RandomPolicy::random(r);
	}

	virtual void _sleep(const unsigned int seconds)
	{
		SleepPolicy::sleep(seconds);
	}

	AlgorithmConstants _constants;
};

#endif //GLITE_DATA_SRM_UTIL_CPP_EXPONENTIAL_BACKOFF_LOGIC_H_
