/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004. 
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.  
 *
 *  Licensed under the Apache License, Version 2.0 (the "License"); 
 *  you may not use this file except in compliance with the License. 
 *  You may obtain a copy of the License at 
 *
 *      http://www.apache.org/licenses/LICENSE-2.0 
 *
 *  Unless required by applicable law or agreed to in writing, software 
 *  distributed under the License is distributed on an "AS IS" BASIS, 
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 *  See the License for the specific language governing permissions and 
 *  limitations under the License.
 *
 *  GLite Data SRM Util CPP: BringOnline Class
 *
 *  Authors: Paolo Badino <paolo.badino@cern.ch>, Paolo Tedesco <paolo.tedesco@cern.ch>
 *  Version info: $Id: BringOnline.cpp,v 1.10 2008/11/20 11:44:47 ptedesco Exp $
 *  Release: $Name:  $
 *
 */
#include "BringOnline.h"
#include "SrmVersion.h"
#include "GSoapContext.h"
#include "srm-utility.h"
#include "srm-common-bringonline.h"
#include "srm-common.h"
#include "context-caller.h"

#include <boost/lexical_cast.hpp>

using namespace glite::data::srm::util::srm1_1;

DEFINE_REQUEST_FACTORY(BringOnline, 1_1);
DEFINE_ASYNCH_REQUEST_FACTORY(BringOnline, 1_1);

/**
 * Execute the Request. In srm 1.1 BringOnline consist in a Get followed 
 * by a GetDone
 */
void BringOnline::execute() /* throw (LogicError, RemoteError, SrmBadResponse, SrmError)*/{
    CallInfo info;
    // CheckPreconditions
    this->checkToken(false);
    if(true == files.empty()){
        throw InvalidArgumentException("empty files array");
    }
    if(false == ctx().ignoreMissingFeatures){
        // Check that advanced features (not available in srm1.1) are 
        // not required
        // TODO: Check features not supported in srm1.1
    }
    
    GSoapContext g(this->ctx(),false);
                                                                                                                                               
    ::srm::ArrayOfstring surlarray;
    ::srm::ArrayOfstring protoarray;
                                                                                                                                                       
    // Set the Protocol array
    protoarray.__size = 1;
    protoarray.__ptr = (char**)soap_malloc(g.soap().soap,sizeof(char *));
    protoarray.__ptr[0] = &(*(this->transferParameters.protocol.begin()));
 
    // Set the SURL array
    surlarray.__size = files.size();
    surlarray.__ptr = (char**)soap_malloc(g.soap().soap,sizeof(char *) * surlarray.__size);
    std::vector<BringOnlineFileRequest>::iterator it;
    char ** ptr = &(surlarray.__ptr[0]);
    for(it = files.begin(); it != files.end(); ++it, ++ptr){
        info.add("surl",(*it).surl);
        (*ptr) = &(*((*it).surl.begin()));
    }
                                                                                                                                                       
    // Invoke SRM Get
    struct ::srm::srm__getResponse outg;
    this->ctx().beforeCall("srm__get");
    if(SOAP_OK != g.soap().srm__get(&surlarray,&protoarray,outg)){
        this->ctx().onFailure("srm__get",g.ip().c_str(),&info);
        // Failed to invoke remote Get: throw the correct exception
        g.handleError("SrmGet");
    }
    CallOnSuccess call(this->ctx(),"srm__get",g.ip().c_str(),&info);
     
    // Update Request Status
    update_request_status<BringOnline,BringOnlineFileRequest>(*this,outg._Result);

    // Check if the request is succeeded
    if(is_request_succeeded<BringOnline>(*this)){
        send_done_request<BringOnline,BringOnlineFileRequest>(g,*this);
    }
    info.token(m_token);
}

/**
 * update
 * Update the status of the Request.
 */
void BringOnline::update() /*throw (LogicError, RemoteError, SrmBadResponse, SrmError)*/{
    // CheckPreconditions
    this->checkToken(true);
    
    // Create gsoap context
    GSoapContext g(this->ctx(),false);
    
    // Update Request Status
    send_update_request<BringOnline,BringOnlineFileRequest>(g,*this);
    
    // Check if the request is succeeded
    if(is_request_succeeded<BringOnline>(*this)){
        // since for 1.1 BringOnline is simulated, release the resource immediately
        send_done_request<BringOnline,BringOnlineFileRequest>(g,*this);
        // For each file, adjust the status
        std::vector<BringOnlineFileRequest>::iterator it;
        for(it = this->files.begin(); it != this->files.end(); ++it){
            BringOnlineFileRequest& file = (*it);
            if(true == is_file_released(file)){
               // Set file status SUCCESS
               file.status.code = SrmStatus::SRM_SUCCESS;
            }
        }
    }
}

/*
 * done
 *
 * Complete a Request and release the allocated resources.
 */
void BringOnline::done() /*throw (LogicError, RemoteError, SrmBadResponse, SrmError)*/{
    // CheckPreconditions
    this->checkToken(true);
    
    // Set gsoap context
    GSoapContext g(this->ctx(),false);

    // Send done request    
    send_done_request<BringOnline,BringOnlineFileRequest>(g,*this);
}

/*
 * abort
 *
 * Abort a Request and cleanup the allocated resources.
 */
void BringOnline::abort() /*throw (LogicError, RemoteError, SrmBadResponse, SrmError)*/{
    // CheckPreconditions
    this->checkToken(true);

    // Create gsoap context
    GSoapContext g(this->ctx(),false);
    
    // Send abort request    
    send_abort_request<BringOnline,BringOnlineFileRequest>(g,*this);
}
