/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004. 
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.  
 *
 *  Licensed under the Apache License, Version 2.0 (the "License"); 
 *  you may not use this file except in compliance with the License. 
 *  You may obtain a copy of the License at 
 *
 *      http://www.apache.org/licenses/LICENSE-2.0 
 *
 *  Unless required by applicable law or agreed to in writing, software 
 *  distributed under the License is distributed on an "AS IS" BASIS, 
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 *  See the License for the specific language governing permissions and 
 *  limitations under the License.
 *
 *  GLite Data SRM Util CPP: GSoapContext Class
 *
 *  Authors: Paolo Badino <paolo.badino@cern.ch>, Paolo Tedesco <paolo.tedesco@cern.ch>
 *  Version info: $Id: GSoapContext.cpp,v 1.15 2008/05/08 12:48:26 ptedesco Exp $
 *  Release: $Name:  $
 *
 */
#include "GSoapContext.h"
#include "glite/data/srm/util/gsoap-common.h"
#include "glite/data/srm/util/SrmExceptions.h"
#include "glite/data/srm/cpp/srm.nsmap"
#include "ErrorPatterns.h"
#include <glite/data/srm/util/gsoap-common.h>

#include <sstream>

/*
 * Include CGSI GSOAP plugin
 */
extern "C"{
#include "cgsi_plugin.h"
}                                                                                                                                                       
using namespace glite::data::srm::util;
using namespace glite::data::srm::util::srm1_1;

/*
 * GSoapContext
 *
 * Constructor
 */
GSoapContext::GSoapContext(Context& ctx, bool deleg_cred):m_ctx(ctx){
    
    m_service.endpoint = m_ctx.endpoint.c_str();

    init_soap_ctx< ::srm::ISRM >(m_service,m_ctx.endpoint,deleg_cred,m_ctx.httpTimeout);
}

/*
 * ~GSoapContext
 *
 * Destructor
 */
GSoapContext::~GSoapContext(){
    // Nothing to Do
}

/*
 * getError
 *
 * Get an SrmError that correspond to teh reason string
 */
SrmStatus GSoapContext::getError(const std::string& reason) /*throw ()*/{
    return ErrorPatterns::instance().findSrmError(reason);
}  
    
/*
 * handleError
 *
 * Translate the Error into the appropriate Exception
 */
void GSoapContext::handleError(const std::string& method) /*throw (LogicError, RemoteError, SecurityError, SrmError)*/{

    //DEBUG: soap_print_fault(m_service.soap, stderr);
    // Check the Soap Fault     
    if(0 != m_service.soap->fault) {
	    // Check if it's a Remote Error
        if(is_TIMEOUT_ERROR(m_service.soap)){
            // Timeout in contacting the SRM
            std::stringstream reason;
            reason << "service timeout during [" << method << "]";
            throw HttpTimeoutError(reason.str());
	    }
	    // Parse the fault code in order to understand the reason
        std::stringstream ss;
        ss << "[" << method << "] failed: " 
           << m_service.soap->fault->faultcode << " - "
           << ((0 != m_service.soap->fault->faultstring)?
                     m_service.soap->fault->faultstring:"(null)");
        if (m_service.soap->version == 2){
            // SOAP 1.2 is used
            if((0 != m_service.soap->fault->SOAP_ENV__Detail) &&
               (0 != m_service.soap->fault->SOAP_ENV__Detail->fault)){
              ss << ". Details: " << reinterpret_cast<const char *>(
                    m_service.soap->fault->SOAP_ENV__Detail->fault);
            }
        } else {
            // SOAP 1.1 is used
            if((0 != m_service.soap->fault->detail) &&
               (0 != m_service.soap->fault->detail->fault)){
              ss << ". Details: " << reinterpret_cast<const char *>(
                    m_service.soap->fault->detail->fault);
            }
        }   
        std::string reason = ss.str();
        // Check if it's an GSI authentication error
        if(ErrorPatterns::instance().isGsiAuthenticationError(reason)){
            throw SecurityError(reason);
        } else if(ErrorPatterns::instance().isTcpConnectionError(reason)){
            throw RemoteError(reason);
        }
        // Raise an srm_status error
        SrmStatus srm_status = getError(reason);
        throw SrmError(srm_status);
    } else {
        // No SOAP fault returned, Generic Failure
        std::stringstream reason;
        reason << "[" << method <<"] failed without a reason";
        throw SrmError(SrmStatus::SRM_FAILURE, reason.str());
    }
}

/**
 * get the ip address
 */
std::string GSoapContext::ip(){
    return get_ip(m_service.soap);
}

