/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004. 
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.  
 *
 *  Licensed under the Apache License, Version 2.0 (the "License"); 
 *  you may not use this file except in compliance with the License. 
 *  You may obtain a copy of the License at 
 *
 *      http://www.apache.org/licenses/LICENSE-2.0 
 *
 *  Unless required by applicable law or agreed to in writing, software 
 *  distributed under the License is distributed on an "AS IS" BASIS, 
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 *  See the License for the specific language governing permissions and 
 *  limitations under the License.
 *
 *  GLite Data SRM Util CPP: PrepareToGet Class
 *
 *  Authors: Paolo Badino <paolo.badino@cern.ch>, Paolo Tedesco <paolo.tedesco@cern.ch>
 *  Version info: $Id: PrepareToGet.cpp,v 1.9 2008/11/20 11:44:47 ptedesco Exp $
 *  Release: $Name:  $
 *
 */
#include "PrepareToGet.h"
#include "SrmVersion.h"
#include "GSoapContext.h"
#include "srm-utility.h"
#include "srm-common-get.h"
#include "srm-common.h"
#include "context-caller.h"

#include <boost/lexical_cast.hpp>

using namespace glite::data::srm::util::srm1_1;

DEFINE_REQUEST_FACTORY(PrepareToGet, 1_1);
DEFINE_ASYNCH_REQUEST_FACTORY(PrepareToGet, 1_1);

/**
 * Execute the Request
 */
void PrepareToGet::execute() /* throw (LogicError, RemoteError, SrmBadResponse, SrmError)*/{
    
    CallInfo info;
    // CheckPreconditions
    this->checkToken(false);
    if(true == files.empty()){
        throw InvalidArgumentException("empty files array");
    }
    if(false == ctx().ignoreMissingFeatures){
        // Check that advanced features (not available in srm1.1) are 
        // not required
        // TODO: Check features not supported in srm1.1
    }
    
    GSoapContext g(this->ctx(),false);
                                                                                                                                               
    ::srm::ArrayOfstring surlarray;
    ::srm::ArrayOfstring protoarray;
                                                                                                                                                       
    // Set the Protocol array
    protoarray.__size = 1;
    protoarray.__ptr = (char**)soap_malloc(g.soap().soap,sizeof(char *));
    protoarray.__ptr[0] = &(*(this->transferParameters.protocol.begin()));
 
    // Set the SURL array
    surlarray.__size = files.size();
    surlarray.__ptr = (char**)soap_malloc(g.soap().soap,sizeof(char *) * surlarray.__size);
    std::vector<GetFileRequest>::iterator it;
    char ** ptr = &(surlarray.__ptr[0]);
    for(it = files.begin(); it != files.end(); ++it, ++ptr){
        (*ptr) = &(*((*it).surl.begin()));
        info.add("surl",(*it).surl);
    }
                                                                                                                                                       
    // Invoke SRM Get
    struct ::srm::srm__getResponse outg;
    this->ctx().beforeCall("srm__get");
    if(SOAP_OK != g.soap().srm__get(&surlarray,&protoarray,outg)){
        this->ctx().onFailure("srm__get",g.ip().c_str(),&info);
        // Failed to invoke remote Get: throw the correct exception
        g.handleError("SrmGet");
    }
    CallOnSuccess call(this->ctx(),"srm__get",g.ip().c_str(),&info);
    
    // Update Request Status
    update_request_status<PrepareToGet,GetFileRequest>(*this,outg._Result);

    // Set the File Status running if the request is completed
    // In case of errors, ignore them: an update should be performed
    try{
        set_running<PrepareToGet,GetFileRequest>(g,*this);
    } catch(const RemoteError& exc){
        this->ctx().onWarning("RemoteError while setting files' status to running",exc.what(),g.ip().c_str());
    } catch(const SrmBadResponse& exc){
        this->ctx().onWarning("SrmBadResponse while setting files' status to running",exc.what(),g.ip().c_str());
    } catch(const SrmError& exc){
        this->ctx().onWarning("SrmError while setting files' status to running",exc.what(),g.ip().c_str());
    }
    info.token(m_token);
}

/**
 * update
 * Update the status of the Request.
 */
void PrepareToGet::update() /*throw (LogicError, RemoteError, SrmBadResponse, SrmError)*/{
    // CheckPreconditions
    this->checkToken(true);
    
    // Create gsoap context
    GSoapContext g(this->ctx(),false);
    
    // Update Request Status
    send_update_request<PrepareToGet,GetFileRequest>(g,*this);
    
    // Set the File Status running if the request is completed
    set_running<PrepareToGet,GetFileRequest>(g,*this);
}

/*
 * done
 *
 * Complete a Request and release the allocated resources.
 */
void PrepareToGet::done() /*throw (LogicError, RemoteError, SrmBadResponse, SrmError)*/{
    // CheckPreconditions
    this->checkToken(true);
    
    // Set gsoap context
    GSoapContext g(this->ctx(),false);

    // Send done request    
    send_done_request<PrepareToGet,GetFileRequest>(g,*this);
}

/*
 * abort
 *
 * Abort a Request and cleanup the allocated resources.
 */
void PrepareToGet::abort() /*throw (LogicError, RemoteError, SrmBadResponse, SrmError)*/{
    // CheckPreconditions
    this->checkToken(true);

    // Create gsoap context
    GSoapContext g(this->ctx(),false);
    
    // Send abort request    
    send_abort_request<PrepareToGet,GetFileRequest>(g,*this);
}
