/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004. 
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.  
 *
 *  Licensed under the Apache License, Version 2.0 (the "License"); 
 *  you may not use this file except in compliance with the License. 
 *  You may obtain a copy of the License at 
 *
 *      http://www.apache.org/licenses/LICENSE-2.0 
 *
 *  Unless required by applicable law or agreed to in writing, software 
 *  distributed under the License is distributed on an "AS IS" BASIS, 
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 *  See the License for the specific language governing permissions and 
 *  limitations under the License.
 *
 *  GLite Data SRM Util CPP: PrepareToPut Class
 *
 *  Authors: Paolo Badino <paolo.badino@cern.ch>, Paolo Tedesco <paolo.tedesco@cern.ch>
 *  Version info: $Id: PrepareToPut.cpp,v 1.9 2008/11/20 11:44:47 ptedesco Exp $
 *  Release: $Name:  $
 *
 */
#include "PrepareToPut.h"
#include "SrmVersion.h"
#include "GSoapContext.h"
#include "srm-utility.h"
#include "srm-common-put.h"
#include "srm-common.h"
#include "context-caller.h"

#include <boost/lexical_cast.hpp>

using namespace glite::data::srm::util::srm1_1;

DEFINE_REQUEST_FACTORY(PrepareToPut, 1_1);
DEFINE_ASYNCH_REQUEST_FACTORY(PrepareToPut, 1_1);

/**
 * Execute the Request
 */
void PrepareToPut::execute() /* throw (LogicError, RemoteError, SrmBadResponse, SrmError)*/{
    CallInfo info;
    // CheckPreconditions
    this->checkToken(false);
    if(true == files.empty()){
        throw InvalidArgumentException("empty files array");
    }
    if(false == ctx().ignoreMissingFeatures){
        // Check that advanced features (not available in srm1.1) are 
        // not required
        // TODO: Check features not supported in srm1.1
    }
    
    GSoapContext g(this->ctx(),false);
                                             
    ::srm::ArrayOflong     sizearray;
    ::srm::ArrayOfboolean  permarray;
    ::srm::ArrayOfstring   surlarray;
    ::srm::ArrayOfstring   protoarray;

    // Set the Protocol array
    protoarray.__size = 1;
    protoarray.__ptr = (char**)soap_malloc(g.soap().soap,sizeof(char *));
    protoarray.__ptr[0] = &(*(this->transferParameters.protocol.begin()));
    
    // Set the Source and SURL array
    surlarray.__size = files.size();
    sizearray.__size = surlarray.__size;
    permarray.__size = surlarray.__size;
    surlarray.__ptr  = (char**)soap_malloc(g.soap().soap ,sizeof(char *) * surlarray.__size);
    sizearray.__ptr  = (LONG64*)soap_malloc(g.soap().soap,sizeof(LONG64) * sizearray.__size);
    permarray.__ptr  = (bool*)soap_malloc(g.soap().soap  ,sizeof(bool)   * permarray.__size);
    std::vector<PutFileRequest>::iterator it;
    char ** ptr    = &(surlarray.__ptr[0]);
    LONG64 * s_ptr = &(sizearray.__ptr[0]);
    int i = 0;
    for(it = files.begin(); it != files.end(); ++it, ++ptr, ++s_ptr, ++i){
        (*ptr)   = &(*((*it).surl.begin()));
        info.add("surl",(*it).surl);
        *s_ptr   = (*it).size;
        if(this->desiredFileStorageType == FileStorageType::VOLATILE){
            permarray.__ptr[i] = false;
        } else {
            permarray.__ptr[i] = true;
        }
    }    
    
    // Invoke SRM Get
    struct ::srm::srm__putResponse  outp;
    this->ctx().beforeCall("srm__put");
    if(SOAP_OK != g.soap().srm__put(&surlarray, &surlarray, &sizearray,&permarray, &protoarray, outp)){
        this->ctx().onFailure("srm__put",g.ip().c_str(),&info);
        // Failed to invoke remote Put: throw the correct exception
        g.handleError("SrmPut");
    }
    CallOnSuccess call(this->ctx(),"srm__put",g.ip().c_str(),&info);
    
    // Update Request Status
    update_request_status<PrepareToPut,PutFileRequest>(*this,outp._Result);

    // Set the File Status running if the request is completed
    // In case of errors, ignore them: an update should be performed
    try{
        set_running<PrepareToPut,PutFileRequest>(g,*this);
    } catch(const RemoteError& exc){
        this->ctx().onWarning("RemoteError while setting files' status to running",exc.what(),g.ip().c_str());
    } catch(const SrmBadResponse& exc){
        this->ctx().onWarning("SrmBadResponse while setting files' status to running",exc.what(),g.ip().c_str());
    } catch(const SrmError& exc){
        this->ctx().onWarning("SrmError while setting files' status to running",exc.what(),g.ip().c_str());
    }
    info.token(m_token);
}

/*
 * update
 *
 * Update the status of the Request.
 */
void PrepareToPut::update() /*throw (LogicError, RemoteError, SrmBadResponse, SrmError)*/{
    // CheckPreconditions
    this->checkToken(true);

    // Create gsoap context
    GSoapContext g(this->ctx(),false);

    // Update Request Status
    send_update_request<PrepareToPut,PutFileRequest>(g,*this);

    // Set the File Status running if the request is completed
    set_running<PrepareToPut,PutFileRequest>(g,*this);
}

/*
 * done
 *
 * Complete a Request and release the allocated resources.
 */
void PrepareToPut::done() /*throw (LogicError, RemoteError, SrmBadResponse, SrmError)*/{
    // CheckPreconditions
    this->checkToken(true);
    
    // Create gsoap context
    GSoapContext g(this->ctx(),false);
    
    // Send done request    
    send_done_request<PrepareToPut,PutFileRequest>(g,*this);
}

/*
 * abort
 *
 * Abort a Request and cleanup the allocated resources.
 */
void PrepareToPut::abort() /*throw (LogicError, RemoteError, SrmBadResponse, SrmError)*/{
    // CheckPreconditions
    this->checkToken(true);

    // Create gsoap context
    GSoapContext g(this->ctx(),false);
        
    // Send abort request    
    send_abort_request<PrepareToPut,PutFileRequest>(g,*this);
}
