/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004. 
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.  
 *
 *  Licensed under the Apache License, Version 2.0 (the "License"); 
 *  you may not use this file except in compliance with the License. 
 *  You may obtain a copy of the License at 
 *
 *      http://www.apache.org/licenses/LICENSE-2.0 
 *
 *  Unless required by applicable law or agreed to in writing, software 
 *  distributed under the License is distributed on an "AS IS" BASIS, 
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 *  See the License for the specific language governing permissions and 
 *  limitations under the License.
 *
 *  GLite Data SRM Util CPP: SrmRm Class
 *
 *  Authors: Paolo Badino <paolo.badino@cern.ch>, Paolo Tedesco <paolo.tedesco@cern.ch>
 *  Version info: $Id: SrmRm.cpp,v 1.9 2008/11/20 11:44:47 ptedesco Exp $
 *  Release: $Name:  $
 *
 */
#include "SrmRm.h"
#include "SrmVersion.h"
#include "GSoapContext.h"
#include "srm-utility.h"

using namespace glite::data::srm::util::srm1_1;

DEFINE_REQUEST_FACTORY(SrmRm, 1_1);

/**
 * Execute the Request
 */
glite::data::srm::util::SrmStatus SrmRm::execute() /* throw (LogicError, RemoteError, SecurityError, SrmError)*/{

    SrmStatus result;
    CallInfo info;

    if(false == ctx().ignoreMissingFeatures){
        // Check that advanced features (not available in srm1.1) are 
        // not required
        // TODO: Check features not supported in srm1.1
    }
    
    GSoapContext g(this->ctx(),false);
                           
    ::srm::ArrayOfstring surlarray;
                                                                                                                                                       
    // Set the SURL array
    surlarray.__size = files.size();
    surlarray.__ptr = (char**)soap_malloc(g.soap().soap,sizeof(char *) * surlarray.__size);
    std::vector<RmFileRequest>::iterator it;
    char ** ptr = &(surlarray.__ptr[0]);
    for(it = files.begin(); it != files.end(); ++it, ++ptr){
        (*ptr) = &(*((*it).surl.begin()));
        info.add("surl",(*it).surl);
    }
    
    // Invoke srm__advisoryDelete
    // ---------------------------------------------------------------------
    ::srm::srm__advisoryDeleteResponse out;
    this->ctx().beforeCall("srm__advisoryDelete");
    if(SOAP_OK != g.soap().srm__advisoryDelete(&surlarray, out)){
        this->ctx().onFailure("srm__advisoryDelete",g.ip().c_str(),&info);
        // Failed to invoke remote SrmRm: throw the correct exception
        try{
            g.handleError("SrmRm");
        } catch(const SrmError& exc){
            // In dCache, an error is returned when the file doesn't exist
            const std::string& reason_str = exc.status().message;
            // Request Failed: Set request status to failed and set individual 
            // file status
            result.code = SrmStatus::SRM_FAILURE;
            result.message = reason_str;
            // For Each file
            bool has_success = false;
            for(it = files.begin(); it != files.end(); ++it, ++ptr){
                // Check if there's a line containing either the SURL
                // or the Path name
                checkFileStatus((*it),reason_str);
                if((*it).status.code == SrmStatus::SRM_SUCCESS){
                    has_success = true;
                }
            }
            // Set to to partial success
            if(true == has_success){
                result.code = SrmStatus::SRM_PARTIAL_SUCCESS;
            }
        }
    } else {
        this->ctx().onSuccess("srm__advisoryDelete",g.ip().c_str(),&info);
    
        // Set the status for all files
        for(it = files.begin(); it != files.end(); ++it, ++ptr){
            (*it).status.code = SrmStatus::SRM_SUCCESS;
        }
        
        result.code = SrmStatus::SRM_SUCCESS;
    }
    return result;
}

/*
 * checkFileStatus
 *
 * Check the status of a single file
 */
void SrmRm::checkFileStatus(RmFileRequest& file,const std::string& reason) /*throw ()*/{
    // Get file path
    std::string file_path = get_file_path(file.surl);
    if(false == file_path.empty()){
        // Check if the message contais the file path
        std::string file_error = get_file_message(file_path,reason.c_str());
        if(false == file_error.empty()){
            // There is an error for that file
            file.status = GSoapContext::getError(file_error);
            // postcondition check
            switch(file.status.code){
                case SrmStatus::SRM_SUCCESS:
                case SrmStatus::SRM_INVALID_PATH:
                case SrmStatus::SRM_AUTHORIZATION_FAILURE:
                case SrmStatus::SRM_FILE_LOST:
                case SrmStatus::SRM_FILE_UNAVAILABLE:
                case SrmStatus::SRM_FAILURE:
                    // Ok
                break;
                default:
                    // Set it to generic failure
                    file.status.code = SrmStatus::SRM_FAILURE;
                break;
            }
        } else {
            // File is Success
            file.status.code = SrmStatus::SRM_SUCCESS;
        }
    } else {
        // Invalid surl specified ?
        file.status.code = SrmStatus::SRM_FAILURE;
        file.status.message = "Invalid SURL passed: no path contained";
    }
}
