/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004. 
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.  
 *
 *  Licensed under the Apache License, Version 2.0 (the "License"); 
 *  you may not use this file except in compliance with the License. 
 *  You may obtain a copy of the License at 
 *
 *      http://www.apache.org/licenses/LICENSE-2.0 
 *
 *  Unless required by applicable law or agreed to in writing, software 
 *  distributed under the License is distributed on an "AS IS" BASIS, 
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 *  See the License for the specific language governing permissions and 
 *  limitations under the License.
 *
 *  GLite Data SRM Util CPP: SrmStat Class
 *
 *  Authors: Paolo Badino <paolo.badino@cern.ch>, Paolo Tedesco <paolo.tedesco@cern.ch>
 *  Version info: $Id: SrmStat.cpp,v 1.10 2008/11/20 11:44:47 ptedesco Exp $
 *  Release: $Name:  $
 *
 */
#include "SrmStat.h"
#include "srm-common-stat.h"
#include "SrmVersion.h"
#include "GSoapContext.h"
#include <sstream>

using namespace glite::data::srm::util::srm1_1;

DEFINE_REQUEST_FACTORY(SrmStat, 1_1);

/**
 * Execute the Request
 */
glite::data::srm::util::SrmStatus SrmStat::execute() /* throw (LogicError, RemoteError, SecurityError, SrmError)*/{

    SrmStatus result;
    CallInfo info;
    
    if(false == ctx().ignoreMissingFeatures){
        // Check that advanced features (not available in srm1.1) are 
        // not required
        // TODO: Check features not supported in srm1.1
    }
    
    GSoapContext g(this->ctx(),false);
                           
    ::srm::ArrayOfstring surlarray;
                                                                                                                                                       
    // Set the SURL array
    surlarray.__size = files.size();
    surlarray.__ptr = (char**)soap_malloc(g.soap().soap,sizeof(char *) * surlarray.__size);
    std::vector<MetaDataPathDetail>::iterator it;
    char ** ptr = &(surlarray.__ptr[0]);
    for(it = files.begin(); it != files.end(); ++it, ++ptr){
        (*ptr) = &(*((*it).surl.begin()));
        info.add("surl",(*it).surl);
    }
    
    // Invoke srm__getFileMetaData
    // ---------------------------------------------------------------------
    ::srm::srm__getFileMetaDataResponse out;
    this->ctx().beforeCall("srm__getFileMetaData");
    if(SOAP_OK != g.soap().srm__getFileMetaData(&surlarray, out)){
        this->ctx().onFailure("srm__getFileMetaData",g.ip().c_str(),&info);
        // Failed to invoke remote SrmStat: throw the correct exception
        try{
            g.handleError("SrmStat");
        } catch(const SrmError& exc) {
            // Get Error
            const std::string& reason_str = exc.status().message;
            // Request Failed: Set request status to failed and set individual 
            // file status
            result.code = SrmStatus::SRM_FAILURE;
            result.message = reason_str;
            // For each file
            bool has_success = false;
            for(it = files.begin(); it != files.end(); ++it){
                // find the correct code
                checkFileStatus((*it),reason_str);
                if((*it).status.code == SrmStatus::SRM_SUCCESS){
                    has_success = true;
                }
            }
            // Set to to partial success
            if(true == has_success){
                result.code = SrmStatus::SRM_PARTIAL_SUCCESS;
            }
        }
    } else {
        this->ctx().onSuccess("srm__getFileMetaData",g.ip().c_str(),&info);
        // Parse the FileMetadata Response
        if((0 == out._Result) || 
           (out._Result->__size <= 0) ||
           (0 == out._Result->__ptr)){
            // Error: No result returned by the SRM
            throw SrmBadResponse("no Metadata returned by the SRM");
        }
        if(out._Result->__size != (int) files.size()){
            // Some file failed: Assume they don't exist
        }
        
        unsigned int nfiles = files.size();
        bool assume_same_ordering = false;
        unsigned int i = 0;
        for(it = files.begin(); it != files.end(); ++it, ++i){
            MetaDataPathDetail& file = (*it);
            // for each metadata
            unsigned int j = 0;
            unsigned int rfiles = 0;
            for(; j < (unsigned int)out._Result->__size; ++j){
                if(0 != out._Result->__ptr[j]){
                    // Reorder the file id based on the SURL
                    if(true == match_file<MetaDataPathDetail>(file,out._Result->__ptr[j])){
                        // Match Found
                        break;
                    }
                    ++rfiles;
                }
            } // end foreach returned metadata
            if(j != nfiles){
                // Set File Metadata
                update_file_metadata(file,out._Result->__ptr[j]);
            } else {
                if(rfiles == nfiles){
                    // No matching found, probably the SRM is not returning the
                    // Source SURL => assume the ordering is the same
                    assume_same_ordering = true;
                    break;
                } else {
                    // assume the file is missing
                    file.status.code = SrmStatus::SRM_INVALID_PATH;
                    file.status.message = "No status returned by the SRM for that file. Probably the file does not exist";
                }
            }
        }
        if(true == assume_same_ordering){
            if(0 != i){
                // Error : Inconsistency in the SRM Response!
                std::stringstream reason;
                reason << "The SRM didn't return enough "
                       << "information for assigning the metadata for all files. "
                       << "Only [" << i << "] files were correct";
                throw SrmBadResponse(reason.str().c_str());
            }
            unsigned int j = 0;
            for(; j < nfiles; ++j) {
                // Set File Metadata
                update_file_metadata(files[j],out._Result->__ptr[j]);
            }
        }
    
        // Check if request has failures
        bool has_failure = false;
        bool has_success = false;
        for(it = files.begin(); it != files.end(); ++it){
            if((*it).status.code == SrmStatus::SRM_SUCCESS){
                has_success = true;
            } else {
                has_failure = true;
            } 
            if(has_failure && has_success){
                break;
            }
        }
        // postconditions
        if(has_success && !has_failure){
            result.code = SrmStatus::SRM_SUCCESS;
        } else if(has_success && has_failure){
            result.code = SrmStatus::SRM_PARTIAL_SUCCESS;
            result.message = "Stat failed for some files";
        } else {
            result.code = SrmStatus::SRM_FAILURE;
            result.message = "Stat failed for all files";
        }
    }
    return result;
}

/*
 * checkFileStatus
 *
 * Check the status of a single file
 */
void SrmStat::checkFileStatus(MetaDataPathDetail& file,const std::string& reason) /*throw ()*/{
    // Get file path
    std::string file_path = get_file_path(file.surl);
    if(false == file_path.empty()){
        // Check if the message contais the file path
        std::string file_error = get_file_message(file_path,reason.c_str());
        if(true == file_error.empty()){
            // assume all the files failed with the same error
            file_error = reason;
        }
        // There is an error for that file
        file.status = GSoapContext::getError(file_error);
        // postcondition check
        switch(file.status.code){
            case SrmStatus::SRM_FILE_IN_CACHE:
            case SrmStatus::SRM_AUTHORIZATION_FAILURE:
            case SrmStatus::SRM_INVALID_PATH:
            case SrmStatus::SRM_FAILURE:
            case SrmStatus::SRM_SUCCESS:
               // Ok
            break;
            default:
                // Set it to generic failure
                file.status.code = SrmStatus::SRM_FAILURE;
            break;
        }
    } else {
        // Invalid surl specified ?
        file.status.code = SrmStatus::SRM_FAILURE;
        file.status.message = "Invalid SURL passed: no path contained";
    }
}
