/*
 *  Copyright (c) Members of the EGEE Collaboration. 2004. 
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.  
 *
 *  Licensed under the Apache License, Version 2.0 (the "License"); 
 *  you may not use this file except in compliance with the License. 
 *  You may obtain a copy of the License at 
 *
 *      http://www.apache.org/licenses/LICENSE-2.0 
 *
 *  Unless required by applicable law or agreed to in writing, software 
 *  distributed under the License is distributed on an "AS IS" BASIS, 
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 *  See the License for the specific language governing permissions and 
 *  limitations under the License.
 *
 *  GLite Data SRM Util CPP: srm common functions (abortfiles specialization)
 *
 *  Authors: Paolo Badino <paolo.badino@cern.ch>
 *  Version info: $Id: srm-common-abortfiles.h,v 1.3 2008/05/06 07:50:04 ptedesco Exp $
 *  Release: $Name:  $
 *
 */
#ifndef GLITE_DATA_SRM_UTIL_CPP_1_1_SRM_COMMON_ABORTFILES_H_
#define GLITE_DATA_SRM_UTIL_CPP_1_1_SRM_COMMON_ABORTFILES_H_

#include "AbortFiles.h"
#include "srm-common-defs.h"
#include "srm-utility.h"

#include <boost/lexical_cast.hpp>

namespace glite {
namespace data  {
namespace srm   {
namespace util  {
namespace srm1_1{

/**
 * Return true if the FileStatus structure is related to the given file
 */
template<>
inline bool match_file(const AbortFileRequest& file, const ::srm::srm2__RequestFileStatus * file_stat) {
    if(0 != file_stat){
        // Check SURL
        const char * name = file_stat->SURL;
        if((0 != name) && (strlen(name) > 0)){
            if(0 == file.surl.compare(name)){
                return true;
            }
        }
        // Check source file name
        name = file_stat->sourceFilename;
        if((0 != name) && (strlen(name) > 0)){
            if(0 == file.surl.compare(name)){
                return true;
            }
        }
        // Check dest file name
        name = file_stat->destFilename;
        if((0 != name) && (strlen(name) > 0)){
            if(0 == file.surl.compare(name)){
                return true;
            }
        }
        // Check TURL (abused in srmCopy)
        name   = file_stat->TURL;
        if((0 != name) && (strlen(name) > 0)){
            if(0 == file.surl.compare(name)){
                return true;
            }
        }
    }
    return false;
}

/**
 * Translate File State
 */    
template<>
inline int update_file_request(AbortFiles& request, AbortFileRequest& file, const ::srm::srm2__RequestFileStatus * file_stat, const ::srm::srm2__RequestStatus * srm_stat) /*throw (LogicError, SrmBadResponse)*/{
    if(0 == file_stat){
        throw NullPointerException("null file stat");
    }
    if(0 == srm_stat){
        throw NullPointerException("null srm stat");
    }

    // Return the file Id
    return file_stat->fileId;
}

/**
 * Overwrite template specialization
 */    
template<>
inline void set_file_failed(GSoapContext& g, int request_id, int file_id, AbortFiles& request, AbortFileRequest * file) /*throw (RemoteError, SrmBadResponse, SrmError)*/{
    if(file->status.code == SrmStatus::SRM_INVALID_PATH){
        // File doesn't exist in the request (skip it)
        return;
    }
    const char * next_status = SRM_1_1_STATUS_DONE;
    try{
        set_file_status(g,request_id,file_id,next_status);
    } catch(const SrmError& exc){
        if(exc.status().code == SrmStatus::SRM_INVALID_PATH){
            file->status = exc.status();
        } else {
            // Ignore SrmError: this method should fail only if the File 
            // status is already failed. Other kind of exception should pass 
            // through
            std::stringstream reason;
            reason << "SrmError when aborting file [" << request_id << ":" 
                   << file_id << "]";
            g.ctx().onWarning(reason.str().c_str(),exc.what(),g.ip().c_str());
        }
    }
    // Update File State
    if((0 != file) && (file->status.code != SrmStatus::SRM_INVALID_PATH)){
        file->status.code = SrmStatus::SRM_SUCCESS;
    }
}

/**
 * Update the state of a Request instance. AbortFiles specialization
 */
void update_request_status(AbortFiles& request, const ::srm::srm2__RequestStatus * srm_stat) /*throw (LogicError, RemoteError, SrmBadResponse, SrmError)*/{
   // Check Preconditions
    if(0 == srm_stat){
        // Error: No status returned by the SRM
        throw SrmBadResponse("no RequestStatus returned by the SRM");
    }
    // Check that the request is the same
    std::string id_str = boost::lexical_cast<std::string>(srm_stat->requestId);
    if(request.token != id_str){
        // Error: the status of a different request has been returned
        throw SrmBadResponse("invalid RequestId returned by the SRM");
    }
    
    // Check that the number of file states
    if((0 == srm_stat->fileStatuses) || 
       (0 == srm_stat->fileStatuses->__size) || 
       (0 == srm_stat->fileStatuses->__ptr)){
        // No file uids returned
        throw SrmBadResponse("Cannot retrieve file ids from request status");
    } else {
       unsigned long nfiles = request.files.size();
        if(request.fileIds.empty()){
            request.fileIds.resize(nfiles,-1);
        }
        // For each File
        unsigned int i = 0;
        std::vector<AbortFileRequest>::iterator it;
        for(it = request.files.begin(); it != request.files.end(); ++it, ++i){
            AbortFileRequest& file = (*it);
            // for each request stats
            int j = 0;
            for(; j < srm_stat->fileStatuses->__size; ++j){
                // Reorder the file id based on the Source/Dest SURLS
                if(true == match_file<AbortFileRequest>(file,srm_stat->fileStatuses->__ptr[j])){
                    // Match Found
                    break;
                }
            } // end foreach request stats
            if(j != srm_stat->fileStatuses->__size){
                request.fileIds[i] = update_file_request<AbortFiles,AbortFileRequest>(request,file,srm_stat->fileStatuses->__ptr[j], srm_stat);
            } else {
                // No matching found, assume invalid path
                request.files[i].status.code = SrmStatus::SRM_INVALID_PATH;
                request.files[i].status.message = "SURL is not part of the request";
            }
        } // End for Loop
    }
}

} // End namespace srm1_1
} // End namespace util
} // End namespace srm
} // End namespace data
} // End namespace glite

#endif //GLITE_DATA_SRM_UTIL_CPP_1_1_SRM_COMMON_ABORTFILES_H_
