/*
 *  Copyright (c) Members of the EGEE Collaboration. 2004. 
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.  
 *
 *  Licensed under the Apache License, Version 2.0 (the "License"); 
 *  you may not use this file except in compliance with the License. 
 *  You may obtain a copy of the License at 
 *
 *      http://www.apache.org/licenses/LICENSE-2.0 
 *
 *  Unless required by applicable law or agreed to in writing, software 
 *  distributed under the License is distributed on an "AS IS" BASIS, 
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 *  See the License for the specific language governing permissions and 
 *  limitations under the License.
 *
 *  GLite Data SRM Util CPP: srm common functions (put specialization)
 *
 *  Authors: Paolo Badino <paolo.badino@cern.ch>
 *  Version info: $Id: srm-common-copy.h,v 1.9 2008/05/06 07:50:04 ptedesco Exp $
 *  Release: $Name:  $
 *
 */
#ifndef GLITE_DATA_SRM_UTIL_CPP_1_1_SRM_COMMON_COPY_H_
#define GLITE_DATA_SRM_UTIL_CPP_1_1_SRM_COMMON_COPY_H_

#include "SrmCopy.h"
#include "srm-common-defs.h"
#include "srm-utility.h"

#include <boost/lexical_cast.hpp>

namespace glite {
namespace data  {
namespace srm   {
namespace util  {
namespace srm1_1{
    
/*
 * Override Template Specialization
 */
template<>
inline void translate_request_status(SrmCopy& request, char * state, const char * errorMessage) /* throw (LogicError,SrmBadResponse)*/{
    request.status.code  = translate_copyrequest_status(state,errorMessage);
}

/**
 * Return true if the FileStatus structure is related to the given file
 */
template<>
inline bool match_file(const CopyFileRequest& file, const ::srm::srm2__RequestFileStatus * file_stat) {
    bool result = false;
    if(0 != file_stat){
        const char * req_source   = file_stat->sourceFilename;
        if((0 != req_source) && (strlen(req_source) > 0)){
            req_source = file_stat->SURL;
        }
        const char * req_dest   = file_stat->destFilename;
        if((0 != req_dest) && (strlen(req_dest) > 0)){
            req_dest   = file_stat->TURL;
        }            
        if((0 != req_dest) && (strlen(req_dest) > 0)){
            // Compare the destination File name
            if(0 == file.targetSurl.compare(req_dest)){
                // Same destination: compare also the source 
                if((0 != req_source) && (strlen(req_source) > 0)){
                    if(0 == file.sourceSurl.compare(req_source)){
                        // Match Found
                        result = true;
                    }
                } else {
                    // No source specified: assume it's correct
                    result = true;
                }
            }
        } else {
            // No dest specified: compare only the SURL (assume the user didn't copy
            // the same file twice)
            if((0 != req_source) && (strlen(req_source) > 0)){
                // Compare the Source SURL
                if(0 == file.sourceSurl.compare(req_source)){
                    // Match Found
                    result = true;
                }
            }
        }
    }
    return result;
}

/**
 * Return true if the Request Status code is success after the execute command
 */
template<>
inline bool is_request_succeeded(const SrmCopy& request) {
    return (request.status.code == SrmStatus::SRM_SUCCESS)?true:false;
}
/**
 * Return true if the File Status code is pending
 */
template<>
inline bool is_file_pending(const CopyFileRequest& file) {
    return ((file.status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
            (file.status.code == SrmStatus::SRM_REQUEST_INPROGRESS))?true:false;
}

/**
 * Return true if the File Status code is success after the execute command
 */
template<>
inline bool is_file_succeeded(const CopyFileRequest& file) {
    return (file.status.code == SrmStatus::SRM_SUCCESS)?true:false;
}

/**
 * Return true if the File Status code is released after the done command
 */
template<>
inline bool is_file_released(const CopyFileRequest& file) {
    return (file.status.code == SrmStatus::SRM_RELEASED)?true:false;
}

/*
 * Override Template Specialization
 */
template<>
inline void file_from_request_status(SrmCopy& request, CopyFileRequest& file) {
    file.status.code = file_from_copyrequest_status(request.status.code);
}

/**
 * Translate File State
 */    
template<>
inline int update_file_request(SrmCopy& request, CopyFileRequest& file, const ::srm::srm2__RequestFileStatus * file_stat, const ::srm::srm2__RequestStatus * srm_stat) /*throw (LogicError, SrmBadResponse)*/{
    return update_copyfilerequest(file, file_stat, srm_stat);
}
/**
 * Overwrite template specialization
 */    
template<>
inline void set_file_running(GSoapContext& g, int request_id, int file_id, SrmCopy& request, CopyFileRequest * file) /*throw (RemoteError, SrmBadResponse, SrmError)*/{
    // Nothing to Do for Copy
}

/**
 * Overwrite template specialization
 */    
template<>
inline void set_file_done(GSoapContext& g, int request_id, int file_id, SrmCopy& request, CopyFileRequest * file) /*throw (RemoteError, SrmBadResponse, SrmError)*/{
    //Nothing to do for Copy 
}

/**
 * Overwrite template specialization
 */    
template<>
inline void set_file_failed(GSoapContext& g, int request_id, int file_id, SrmCopy& request, CopyFileRequest * file) /*throw (RemoteError, SrmBadResponse, SrmError)*/{
    const char * next_status = SRM_1_1_STATUS_DONE;
    if(0 != file) {
        switch(file->status.code)
        {
            case SrmStatus::SRM_UNDEF:
            case SrmStatus::SRM_REQUEST_QUEUED:
            case SrmStatus::SRM_REQUEST_INPROGRESS:
                next_status = SRM_1_1_STATUS_DONE;
            break;
            default:
                // Already failed or released: nothing to do
                next_status = 0;
        }
    }
    try{
        set_file_status(g,request_id,file_id,next_status);
    } catch(const SrmError& exc){
        // Ignore SrmError: this method should fail only if the File 
        // status is already failed. Other kind of exception should pass 
        // through
        std::stringstream reason;
        reason << "SrmError when aborting file [" << request_id << ":" 
               << file_id << "]";
        g.ctx().onWarning(reason.str().c_str(),exc.what(),g.ip().c_str());
    }
    // Update File State
    if((0 != next_status) && (0 != file)){
        file->status.code = SrmStatus::SRM_ABORTED;
    }
}

} // End namespace srm1_1
} // End namespace util
} // End namespace srm
} // End namespace data
} // End namespace glite

#endif //GLITE_DATA_SRM_UTIL_CPP_1_1_SRM_COMMON_COPY_H_
