/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004. 
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.  
 *
 *  Licensed under the Apache License, Version 2.0 (the "License"); 
 *  you may not use this file except in compliance with the License. 
 *  You may obtain a copy of the License at 
 *
 *      http://www.apache.org/licenses/LICENSE-2.0 
 *
 *  Unless required by applicable law or agreed to in writing, software 
 *  distributed under the License is distributed on an "AS IS" BASIS, 
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 *  See the License for the specific language governing permissions and 
 *  limitations under the License.
 *
 *  GLite Data SRM Util CPP: srm common functions (get specialization)
 *
 *  Authors: Paolo Badino <paolo.badino@cern.ch>, Paolo Tedesco <paolo.tedesco@cern.ch>
 *  Version info: $Id: srm-common-get.h,v 1.11 2008/11/20 11:44:47 ptedesco Exp $
 *  Release: $Name:  $
 *
 */
#ifndef GLITE_DATA_SRM_UTIL_CPP_1_1_SRM_COMMON_GET_H_
#define GLITE_DATA_SRM_UTIL_CPP_1_1_SRM_COMMON_GET_H_

#include "PrepareToGet.h"
#include "srm-common-defs.h"
#include "srm-utility.h"

#include <boost/lexical_cast.hpp>

namespace glite {
namespace data  {
namespace srm   {
namespace util  {
namespace srm1_1{
    
/*
 * Override Template Specialization
 */
template<>
inline void translate_request_status(PrepareToGet& request, char * state, const char * errorMessage) /* throw (LogicError,SrmBadResponse)*/{
    request.status.code  = translate_getrequest_status(state,errorMessage);
}

/**
 * Return true if the FileMetadata structure is related to the given file
 */
template<>
inline bool match_file(const GetFileRequest& file, const ::srm::srm2__FileMetaData * file_metadata){
    return match_file<FileRequest>(file,file_metadata);
}

/**
 * Return true if the FileStatus structure is related to the given file
 */
template<>
inline bool match_file(const GetFileRequest& file, const ::srm::srm2__RequestFileStatus * file_stat) {
    return match_file<GetFileRequest>(file,static_cast<const ::srm::srm2__FileMetaData *>(file_stat));
}

/**
 * Return true if the Request Status code is success after the execute command
 */
template<>
inline bool is_request_succeeded(const PrepareToGet& request) {
    return (request.status.code == SrmStatus::SRM_REQUEST_INPROGRESS)?true:false;
}

/**
 * Return true if the File Status code is pending
 */
template<>
inline bool is_file_pending(const GetFileRequest& file) {
    return ((file.status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
            (file.status.code == SrmStatus::SRM_REQUEST_INPROGRESS))?true:false;
}

/**
 * Return true if the File Status code is success after the execute command
 */
template<>
inline bool is_file_succeeded(const GetFileRequest& file) {
    return (file.status.code == SrmStatus::SRM_FILE_PINNED)?true:false;
}

/**
 * Return true if the File Status code is released after the done command
 */
template<>
inline bool is_file_released(const GetFileRequest& file) {
    return (file.status.code == SrmStatus::SRM_RELEASED)?true:false;
}

/*
 * Override Template Specialization
 */
template<>
inline void file_from_request_status(PrepareToGet& request, GetFileRequest& file) {
    file.status.code = file_from_getrequest_status(request.status.code);
}

/**
 * Translate File State
 */    
template<>
inline int update_file_request(PrepareToGet& request, GetFileRequest& file, const ::srm::srm2__RequestFileStatus * file_stat, const ::srm::srm2__RequestStatus * srm_stat) /*throw (LogicError, SrmBadResponse)*/{
    return update_getfilerequest(file, file_stat, srm_stat);
}
/**
 * Overwrite template specialization
 */    
template<>
inline void set_file_running(GSoapContext& g, int request_id, int file_id, PrepareToGet& request, GetFileRequest * file) /*throw (RemoteError, SrmBadResponse, SrmError)*/{
    const char * next_status = SRM_1_1_STATUS_RUNNING;
    if(0 != file) {
        switch(file->status.code)
        {
            case SrmStatus::SRM_UNDEF:
            case SrmStatus::SRM_REQUEST_INPROGRESS:
                next_status = SRM_1_1_STATUS_RUNNING;
            break;
            default:
                // File is not Pinned, don't set it to running
                next_status = 0;
        }
    }
    set_file_status(g,request_id,file_id,next_status);
    if((0 != next_status) && (0 != file)){
        // Update file status
        file->status.code = SrmStatus::SRM_FILE_PINNED;
    }
}

/**
 * Overwrite template specialization
 */    
template<>
inline void set_file_done(GSoapContext& g, int request_id, int file_id, PrepareToGet& request,GetFileRequest * file) /*throw (RemoteError, SrmBadResponse, SrmError)*/{
    const char * next_status = SRM_1_1_STATUS_DONE;
    if(0 != file) {
        switch(file->status.code)
        {
            case SrmStatus::SRM_UNDEF:
            case SrmStatus::SRM_REQUEST_QUEUED:
            case SrmStatus::SRM_REQUEST_INPROGRESS:
            case SrmStatus::SRM_FILE_PINNED:
                next_status = SRM_1_1_STATUS_DONE;
            break;
            case SrmStatus::SRM_FILE_BUSY:
            case SrmStatus::SRM_FILE_IN_CACHE:
                next_status = SRM_1_1_STATUS_FAILED;
            break;
            default:
                // Already failed or released: nothing to do
                next_status = 0;
        }
    }
    try{
        set_file_status(g,request_id,file_id,next_status);
    } catch(const SrmError& exc){
        // Ignore SrmError: this method should fail only if the File 
        // status is already failed or Done. Other kind of exception should pass 
        // through
        std::stringstream reason;
        reason << "SrmError when releasing file [" << request_id << ":" << file_id << "]";
        g.ctx().onWarning(reason.str().c_str(),exc.what(),g.ip().c_str());
        if(0 != file) {
            file->status.code = SrmStatus::SRM_FAILURE;
            file->status.message = exc.status().message; 
        }
    }
    // Update File State
    if((0 != next_status) && (0 != file)){
        file->status.code = SrmStatus::SRM_RELEASED;
    }
}

/**
 * Overwrite template specialization
 */    
template<>
inline void set_file_failed(GSoapContext& g, int request_id, int file_id, PrepareToGet& request, GetFileRequest * file) /*throw (RemoteError, SrmBadResponse, SrmError)*/{
    const char * next_status = SRM_1_1_STATUS_DONE;
    if(0 != file) {
        switch(file->status.code)
        {
            case SrmStatus::SRM_UNDEF:
            case SrmStatus::SRM_REQUEST_QUEUED:
            case SrmStatus::SRM_REQUEST_INPROGRESS:
            case SrmStatus::SRM_FILE_PINNED:
            case SrmStatus::SRM_FILE_BUSY:
                next_status = SRM_1_1_STATUS_DONE;
            break;
            default:
                // Already failed or released: nothing to do
                next_status = 0;
        }
    }
    try{
        set_file_status(g,request_id,file_id,next_status);
    } catch(const SrmError& exc){
        // Ignore SrmError: this method should fail only if the File 
        // status is already failed. Other kind of exception should pass 
        // through
        std::stringstream reason;
        reason << "SrmError when aborting file [" << request_id << ":" << file_id << "]";
        g.ctx().onWarning(reason.str().c_str(),exc.what(),g.ip().c_str());
    }
    // Update File State
    if((0 != next_status) && (0 != file)){
        file->status.code = SrmStatus::SRM_ABORTED;
    }
}

} // End namespace srm1_1
} // End namespace util
} // End namespace srm
} // End namespace data
} // End namespace glite

#endif //GLITE_DATA_SRM_UTIL_CPP_1_1_SRM_COMMON_GET_H_
