/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004. 
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.  
 *
 *  Licensed under the Apache License, Version 2.0 (the "License"); 
 *  you may not use this file except in compliance with the License. 
 *  You may obtain a copy of the License at 
 *
 *      http://www.apache.org/licenses/LICENSE-2.0 
 *
 *  Unless required by applicable law or agreed to in writing, software 
 *  distributed under the License is distributed on an "AS IS" BASIS, 
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 *  See the License for the specific language governing permissions and 
 *  limitations under the License.
 *
 *  GLite Data SRM Util CPP: srm common functions
 *
 *  Authors: Paolo Badino <paolo.badino@cern.ch>, Paolo Tedesco <paolo.tedesco@cern.ch>
 *  Version info: $Id: srm-common.h,v 1.11 2008/11/20 11:44:47 ptedesco Exp $
 *  Release: $Name:  $
 *
 */
#ifndef GLITE_DATA_SRM_UTIL_CPP_1_1_SRM_COMMON_H_
#define GLITE_DATA_SRM_UTIL_CPP_1_1_SRM_COMMON_H_

/*
 * Include template definitions
 */
#include "srm-common-defs.h"

namespace glite {
namespace data  {
namespace srm   {
namespace util  {
namespace srm1_1{

/**
 * Update the state of a Request instance
 */
template<typename R, typename F>
void update_request_status(R& request, const ::srm::srm2__RequestStatus * srm_stat) /*throw (LogicError, RemoteError, SrmBadResponse, SrmError)*/{
   // Check Preconditions
    if(0 == srm_stat){
        // Error: No status returned by the SRM
        throw SrmBadResponse("no RequestStatus returned by the SRM");
    }
    
    // Check Request Id
    if(true == request.token().empty()){
        // Assign the token to the request
        request.token(boost::lexical_cast<std::string>(srm_stat->requestId));
    } else {
        // Check that the request is the same
        std::string id_str = boost::lexical_cast<std::string>(srm_stat->requestId);
        if(request.token() != id_str){
            // Error: the status of a different request has been returned
            throw SrmBadResponse("invalid RequestId returned by the SRM");
        }
    }
    
    // Get Request State
    request.status.message = (0 != srm_stat->errorMessage)?srm_stat->errorMessage:"";
    translate_request_status<R>(request,srm_stat->state, request.status.message.c_str());
    
    // Check that the number of file states
    if((0 == srm_stat->fileStatuses) || 
       (0 == srm_stat->fileStatuses->__size) || 
       (0 == srm_stat->fileStatuses->__ptr)){
        // No file States return: assume the File State is the same as the 
        // Request
        typename std::vector<F>::iterator it;
        for(it = request.files.begin(); it != request.files.end(); ++it){
            file_from_request_status(request, (*it));
        }
        // Clear the File Ids vector 
        request.fileIds.clear();
    } else if(false == request.files.empty()){
        // Check that the size of the fileStatuses is the expected one
        if(srm_stat->fileStatuses->__size != (int) request.files.size()){
            // Invalid Response returned by the SRM
            std::stringstream reason;
            reason << "The SRM returned the status of [" 
                   << srm_stat->fileStatuses->__size
                   << "] file(s) while it was expected to return the one for ["
                   << request.files.size() << "] file(s)";
            throw SrmBadResponse(reason.str());
        }
        unsigned long nfiles = request.files.size();
        if(request.fileIds.empty()){
            request.fileIds.resize(nfiles,-1);
        }
        // For each File
        bool assume_same_ordering = false;
        unsigned int i = 0;
        typename std::vector<F>::iterator it;
        for(it = request.files.begin(); it != request.files.end(); ++it, ++i){
            F& file = (*it);
            // Check if the File Id is set
            if(-1 == request.fileIds[i]){
                // for each request stats
                unsigned int j = 0;
                for(; j < nfiles; ++j){
                    // Reorder the file id based on the Source/Dest SURLS
                    if(true == match_file<F>(file,srm_stat->fileStatuses->__ptr[j])){
                        // Match Found
                        break;
                    }
                } // end foreach request stats
                if(j != nfiles){
                    request.fileIds[i] = update_file_request<R,F>(request,file,srm_stat->fileStatuses->__ptr[j], srm_stat);
                } else {
                    // No matching found, probably the SRM is not returning the
                    // Source SURL => assume the ordering is the same
                    assume_same_ordering = true;
                    break;
                }
            } else {
                // Find the status with that file Id
                unsigned int j = 0;
                for(; j < nfiles; ++j){
                    // Reorder the file id based on the Source SURLS
                    int file_id = srm_stat->fileStatuses->__ptr[j]->fileId;
                    if(file_id == request.fileIds[i]){
                        break;
                    }
                }
                if(j != nfiles){
                    update_file_request<R,F>(request,file,srm_stat->fileStatuses->__ptr[j], srm_stat);
                } else {
                    // The request contains an invalid file id !
                    std::stringstream reason;
                    reason << "No File status returned by the SRM for the file Id ["
                           << request.fileIds[i] << "]";
                    throw SrmBadResponse(reason.str().c_str());
                }
            }
        } // End for Loop
        if(true == assume_same_ordering){
            if(0 != i){
                // Error : Inconsistency in the SRM Response!
                std::stringstream reason;
                reason << "The SRM didn't return enough "
                       << "information for assigning the file status for all files. "
                       << "Only [" << i << "] files were correct";
                throw SrmBadResponse(reason.str().c_str());
            }
            unsigned int j = 0;
            for(; j < nfiles; ++j) {
                // Check if the File Id is set
                int fid = update_file_request<R,F>(request,request.files[j],srm_stat->fileStatuses->__ptr[j],srm_stat);
                if(-1 == request.fileIds[j]){
                    // Set the File Id
                    request.fileIds[j] = fid;
                }
            }
        }
    } else {
        
        // Empty File List: this should be filled with the result returned by 
        // the SRM
        unsigned long nfiles = srm_stat->fileStatuses->__size;
        request.files.resize(nfiles);
        request.fileIds.resize(nfiles);
        unsigned int i = 0;
        for(; i < nfiles; ++i) {
            // Set the File Id
            request.fileIds[i] = update_file_request<R,F>(request,request.files[i],srm_stat->fileStatuses->__ptr[i], srm_stat);
        }
    }
    
    // Check Request Status
    if(request.status.code == SrmStatus::SRM_REQUEST_INPROGRESS){
        bool has_pending = false;
        bool has_success = false;
        // Check if all files are Succeeded
        typename std::vector<F>::iterator it;
        for(it = request.files.begin(); it != request.files.end(); ++it){
            F& file = (*it);
            if(true == is_file_succeeded<F>(file)){
                has_success = true;
            } else if(true == is_file_pending<F>(file)){
                has_pending = true;
                break;
            }
        }
        // no pending files => set request success (or failed)
        if(false == has_pending) {
            request.status.code = 
                (true == has_success)?SrmStatus::SRM_SUCCESS:SrmStatus::SRM_FAILURE;
        }
    } 
    if((request.status.code == SrmStatus::SRM_SUCCESS) ||
       (request.status.code == SrmStatus::SRM_FAILURE)) {
        bool has_success = false;
        bool is_partial = false;
        // Check if all files are Succeeded
        typename std::vector<F>::iterator it;
        for(it = request.files.begin(); it != request.files.end(); ++it){
            F& file = (*it);
            if(true == is_file_succeeded<F>(file)){
                has_success = true;
            } else if(true == is_file_released<F>(file)){
                // Ignore
            } else {
                is_partial = true;
            }
        }
        // There are some failures and some successes => SRM_PARTIAL_SUCCESS
        if((true == has_success) && (true == is_partial)){
            request.status.code = SrmStatus::SRM_PARTIAL_SUCCESS;
        }
    }
}

/**
 * Update the state of a Request instance
 */
template<typename R, typename F>
void send_update_request(GSoapContext& g, R& request) /*throw (LogicError, RemoteError, SrmBadResponse, SrmError)*/{
    int request_id = -1;
    try{
        request_id = boost::lexical_cast<int>(request.token());
    } catch(const boost::bad_lexical_cast&){
        throw LogicError("invalid request id: not a number");
    }

    // Invoke SRM RequestStatus
    struct ::srm::srm__getRequestStatusResponse out;
    g.ctx().beforeCall("srm__getRequestStatus");
    CallInfo info(request.token());
    if(SOAP_OK != g.soap().srm__getRequestStatus(request_id, out)){
        g.ctx().onFailure("srm__getRequestStatus",g.ip().c_str(),&info);
        // Failed to invoke remote GetRequestStatus: throw the correct exception
        g.handleError("SrmGetRequestStatus");
    }
    g.ctx().onSuccess("srm__getRequestStatus",g.ip().c_str(),&info);
    
    // Update Request
    update_request_status<R,F>(request,out._Result);
}

/**
 * Update the state of a Request instance
 */
template<typename R, typename F>
void set_running(GSoapContext& g, R& request) /*throw (LogicError, RemoteError, SrmBadResponse, SrmError)*/{
    int request_id = -1;
    try{
        request_id = boost::lexical_cast<int>(request.token());
    } catch(const boost::bad_lexical_cast&){
        throw LogicError("invalid request id: not a number");
    }

    // Check if the Request is succeeded
    if(is_request_succeeded<R>(request)){
       // Set the File status To Running
       bool has_file_list = false;
        if((false == request.files.empty()) && (request.files.size() == request.fileIds.size())){
            has_file_list = true;
        }
        // For each file, check the status
        std::vector<int>::iterator it;
        for(it = request.fileIds.begin(); it != request.fileIds.end(); ++it){
            int file_id = (*it);
            F * file = 0;
            // Check if the File status is set
            if(true == has_file_list){
                file = &(request.files[std::distance(request.fileIds.begin(),it)]);
            }
            try{
                // Set the file status to Runnig: Ignore error since the file
                // status would be update by next update request
                set_file_running<R,F>(g, request_id, file_id,request,file);
            } catch(const RemoteError& exc){
                std::stringstream reason;
                reason << "RemoteError when setting file [" << request_id << ":" << file_id << "] to Running";
                g.ctx().onWarning(reason.str().c_str(),exc.what(),g.ip().c_str());
            } catch(const SrmBadResponse& exc){
                std::stringstream reason;
                reason << "SrmBadResponse when setting file [" << request_id << ":" << file_id << "] to Running";
                g.ctx().onWarning(reason.str().c_str(),exc.what(),g.ip().c_str());
            } catch(const SrmError& exc){
                std::stringstream reason;
                reason << "SrmError when setting file [" << request_id << ":" << file_id << "] to Running";
                g.ctx().onWarning(reason.str().c_str(),exc.what(),g.ip().c_str());
            }
        }
        // Send an update Request in order to update the request correctly
        send_update_request<R,F>(g,request);
    }
}

/**
 * Complete a Request instance
 */
template<typename R, typename F>
void send_done_request(GSoapContext& g, R& request) /*throw (LogicError, RemoteError, SrmBadResponse, SrmError)*/{
    int request_id = -1;
    try{
        request_id = boost::lexical_cast<int>(request.token());
    } catch(const boost::bad_lexical_cast&){
        throw LogicError("invalid request id: not a number");
    }
    // Check that the fileIds vector is filled
    if(true == request.fileIds.empty()){
        // Invoke SRM RequestStatus
        struct ::srm::srm__getRequestStatusResponse outq;
        g.ctx().beforeCall("srm__getRequestStatus");
        CallInfo info(request.token());
        if(SOAP_OK != g.soap().srm__getRequestStatus(request_id, outq)){
            g.ctx().onFailure("srm__getRequestStatus",g.ip().c_str(),&info);
            // Failed to invoke remote GetRequestStatus: throw the correct exception
            g.handleError("SrmGetRequestStatus");
        }
        g.ctx().onSuccess("srm__getRequestStatus",g.ip().c_str(),&info);
        // Update Request Status
        update_request_status<R,F>(request,outq._Result);
    }
    
    bool has_file_list = false;
    if((false == request.files.empty()) && (request.files.size() == request.fileIds.size())){
        has_file_list = true;
    } 
    
    // For each file, check the status
    std::vector<int>::iterator it;
    for(it = request.fileIds.begin(); it != request.fileIds.end(); ++it){
        int file_id = (*it);
        F * file = 0;
        // Check if the File status is set
        if(true == has_file_list){
            file = &(request.files.at(std::distance(request.fileIds.begin(),it)));
        }
        set_file_done<R,F>(g, request_id, file_id,request,file);
    }
    // Everything went OK: Request status doesn't change (unless it's queued or inprogress)
    if((request.status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
       (request.status.code == SrmStatus::SRM_REQUEST_INPROGRESS)){
        request.status.code = SrmStatus::SRM_SUCCESS;
    }
    request.status.message.clear();
}    

/**
 * Abort a Request instance
 */
template<typename R, typename F>
void send_abort_request(GSoapContext& g, R& request) /*throw (LogicError, RemoteError, SrmBadResponse, SrmError)*/{
    int request_id = -1;
    try{
        request_id = boost::lexical_cast<int>(request.token());
    } catch(const boost::bad_lexical_cast&){
        throw LogicError("invalid request id: not a number");
    }
    // Check that the fileIds vector is filled
    if(true == request.fileIds.empty()){
        // Invoke SRM RequestStatus
        struct ::srm::srm__getRequestStatusResponse outq;
        g.ctx().beforeCall("srm__getRequestStatus");
        CallInfo info(request.token());
        if(SOAP_OK != g.soap().srm__getRequestStatus(request_id, outq)){
            g.ctx().onFailure("srm__getRequestStatus",g.ip().c_str(),&info);
            // Failed to invoke remote GetRequestStatus: throw the correct exception
            g.handleError("SrmGetRequestStatus");
        }
        g.ctx().onSuccess("srm__getRequestStatus",g.ip().c_str(),&info);
    
        // Update Request Status
        update_request_status<R,F>(request,outq._Result);
    }
    
    bool has_file_list = false;
    if((false == request.files.empty()) && (request.files.size() == request.fileIds.size())){
        has_file_list = true;
    }
    // For each file, check the status
    std::vector<int>::iterator it;
    for(it = request.fileIds.begin(); it != request.fileIds.end(); ++it){
        int file_id = (*it);
        F * file = 0;
        // Check if the File status is set
        if(true == has_file_list){
            file = &(request.files[std::distance(request.fileIds.begin(),it)]);
        }
        set_file_failed<R,F>(g, request_id, file_id,request,file);
    }
    // Everything went OK: Set the request status
    request.status.message.clear();
    request.status.code = SrmStatus::SRM_ABORTED;
}

        
} // End namespace srm1_1
} // End namespace util
} // End namespace srm
} // End namespace data
} // End namespace glite

#endif //GLITE_DATA_SRM_UTIL_CPP_1_1_SRM_COMMON_H_
