/*
 *  Copyright (c) Members of the EGEE Collaboration. 2004. 
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.  
 *
 *  Licensed under the Apache License, Version 2.0 (the "License"); 
 *  you may not use this file except in compliance with the License. 
 *  You may obtain a copy of the License at 
 *
 *      http://www.apache.org/licenses/LICENSE-2.0 
 *
 *  Unless required by applicable law or agreed to in writing, software 
 *  distributed under the License is distributed on an "AS IS" BASIS, 
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 *  See the License for the specific language governing permissions and 
 *  limitations under the License.
 *
 *  GLite Data SRM Util CPP: srm 1.1 utilities
 *
 *  Authors: Paolo Badino <paolo.badino@cern.ch>
 *  Version info: $Id: srm-utility.cpp,v 1.11 2007/05/08 09:28:19 ptedesco Exp $
 *  Release: $Name:  $
 *
 */
#include "GSoapContext.h"
#include "srm-utility.h"
#include "SrmVersion.h"
#include "glite/data/srm/util/SrmExceptions.h"
#include "glite/data/agents/url-utility.h"

#include <boost/tokenizer.hpp>
#include <string>
#include <sstream>
#include <algorithm>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>


namespace {
const char * const REQUEST_FILE_STATUS_PREFIX = "RequestFileStatus#";
const char * const SFN_TAG = "SFN";
}

namespace glite {
namespace data  {
namespace srm   {
namespace util  {
namespace srm1_1{
                                                                                                                                                         
const char * const SRM_1_1_STATUS_PENDING  = "Pending";
const char * const SRM_1_1_STATUS_READY    = "Ready";
const char * const SRM_1_1_STATUS_ACTIVE   = "Active";
const char * const SRM_1_1_STATUS_RUNNING  = "Running";
const char * const SRM_1_1_STATUS_DONE     = "Done";
const char * const SRM_1_1_STATUS_FAILED   = "Failed";
        
/*
 * get_file_message
 *
 * Extract the file message from the request message. In case no file message
 * is available, return an empty string
 */
std::string get_file_message(int file_id, const char * msg) /*throw ()*/{
    std::string file_msg;
    
    if((0 == msg) || (0 == strlen(msg))){
        // Return empty messsage
        return file_msg;
    }
    // Tokenize the error message
    std::string tmp_msg = msg;
    boost::char_separator<char> delim("\n");
    boost::tokenizer<boost::char_separator<char> > tok(tmp_msg,delim);
    boost::tokenizer<boost::char_separator<char> >::iterator it;
    for(it = tok.begin(); it != tok.end(); ++it){
        // Check if the error message starts with the pattern
        // RequestFileStatus#<file_id>
        std::stringstream s;
        s << REQUEST_FILE_STATUS_PREFIX << file_id;
        std::string prefix = s.str();
        if(((*it).length() >= prefix.length()) &&
           (true == std::equal(prefix.begin(),prefix.end(),(*it).begin()))){
            file_msg = (*it);
            break;
        }
    }
    return file_msg;
}

/*
 * get_file_message
 *
 * Extract the file message from the request message. In case no file message
 * is available, return an empty string
 */
std::string get_file_message(const std::string& filename, const char * msg) /*throw ()*/{
    std::string file_msg;
    
    if((0 == msg) || (0 == strlen(msg))){
        // Return empty messsage
        return file_msg;
    }
    // Tokenize the error message
    std::string tmp_msg = msg;
    boost::char_separator<char> delim("\n");
    boost::tokenizer<boost::char_separator<char> > tok(tmp_msg,delim);
    boost::tokenizer<boost::char_separator<char> >::iterator it;
    for(it = tok.begin(); it != tok.end(); ++it){
        // Check if the error message contains the filename
        if(std::string::npos != (*it).find(filename)){
            file_msg = (*it);
            break;
        }
    }
    return file_msg;
}

/**
 * Extract the file path from a SURL
 */
std::string get_file_path(const std::string& surl) /*throw ()*/{
    // Get the path from the SURL
    std::string file_path;
    try{
        std::string  hostname;
        unsigned int port = 0;
        std::string  protocol;
        std::string  path;
        std::string  query_str;
        // Parse surl
        glite::data::agents::parse_url(surl,hostname,port,protocol,path,query_str);
        // Check query string
        if(query_str.empty()){
            // SURL in compact form: take the path
            file_path = path;
        } else {
            // Parse Query string
            std::vector<std::pair<std::string,std::string> > values;
            glite::data::agents::parse_query_string(query_str,values);
            // Take the SFN
            std::vector<std::pair<std::string,std::string> >::iterator it;
            for(it = values.begin(); it != values.end(); ++it){
                if((*it).first == SFN_TAG){
                    file_path = (*it).second;
                    break;
                }
            }
        }
    } catch(const LogicError& exc){
        // Ignore error
    }
    return file_path;
}


// -----------------------------------------------------------------------------
// Get Helpers
// -----------------------------------------------------------------------------

/*
 * update_getfilerequest
 *
 * Update the GetFileRequest wiith the content of SRM response 
 */
int update_getfilerequest(GetFileRequest& file, const ::srm::srm2__RequestFileStatus * file_stat, const ::srm::srm2__RequestStatus * srm_stat) /*throw (LogicError, SrmBadResponse)*/ {
    if(0 == file_stat){
        throw NullPointerException("null file stat");
    }
    if(0 == srm_stat){
        throw NullPointerException("null srm stat");
    }

    // Set the Status Message
    file.status.message = get_file_message(file_stat->fileId, srm_stat->errorMessage);
    if((true == file.status.message.empty()) && 
       (0 == strcasecmp(file_stat->state, SRM_1_1_STATUS_FAILED)) &&
       (0 != srm_stat->errorMessage)){
        // DPM Workaround
        // TODO: find a better way to do it
        file.status.message = srm_stat->errorMessage;
    }
    // Set the Status Code
    file.status.code = translate_getfile_status(file_stat->state, file.status.message.c_str());
    
    // Set the Size
    file.size   = file_stat->size;
    
    // Update SURl (if empty)
    if((true == file.surl.empty()) && (0 != file_stat->SURL)){
        file.surl = file_stat->SURL;
    }
    // Update TURL
    if((true == file.turl.empty()) && (0 != file_stat->TURL)){
        file.turl = file_stat->TURL;
    }
    // Get the Estimated time 
    file.extimatedWaitTime = srm_stat->estTimeToStart;
    if(file.extimatedWaitTime <= 0){
        file.extimatedWaitTime = srm_stat->retryDeltaTime;
    }
    
    // Return the file Id
    return file_stat->fileId;
}

/*
 * translate_getrequest_status
 *
 * Translate the status of a get request
 */
SrmStatus::Code translate_getrequest_status(const char * state, const char * errorMessage) /* throw (LogicError,SrmBadResponse)*/ {
    SrmStatus::Code req_status = SrmStatus::SRM_UNDEF; 
    
    // Check Status
    if((0 == state) || (0 == strlen(state))){
        // Error: No Request State Set
        throw SrmBadResponse("no request status returned");
    }
                                                                                                                                                       
    // translate character state into SrmStatusCode
    if(0 == strcasecmp(state, SRM_1_1_STATUS_READY)){
        req_status = SrmStatus::SRM_REQUEST_INPROGRESS;
    } else if(0 == strcasecmp(state, SRM_1_1_STATUS_FAILED)){
        req_status = GSoapContext::getError(errorMessage).code;
    } else if(0 == strcasecmp(state, SRM_1_1_STATUS_DONE)){
        req_status = SrmStatus::SRM_SUCCESS;
    } else if(0 == strcasecmp(state, SRM_1_1_STATUS_ACTIVE)){
        req_status = SrmStatus::SRM_REQUEST_INPROGRESS;
    } else if(0 == strcasecmp(state, SRM_1_1_STATUS_PENDING)){
        req_status = SrmStatus::SRM_REQUEST_QUEUED;
    } else {
        // Error: Unknown Request State
        std::stringstream reason;
        reason << "invalid request status returned [" << state << "]";
        throw SrmBadResponse(reason.str());
    }
    
    // Enforce Error Codes
    switch(req_status){
        case SrmStatus::SRM_SUCCESS:
        case SrmStatus::SRM_REQUEST_QUEUED:
        case SrmStatus::SRM_REQUEST_INPROGRESS:
        case SrmStatus::SRM_PARTIAL_SUCCESS:
        case SrmStatus::SRM_AUTHENTICATION_FAILURE:
        case SrmStatus::SRM_AUTHORIZATION_FAILURE:
        case SrmStatus::SRM_INVALID_REQUEST:
        case SrmStatus::SRM_SPACE_LIFETIME_EXPIRED:
        case SrmStatus::SRM_EXCEED_ALLOCATION:
        case SrmStatus::SRM_NOT_SUPPORTED:
        case SrmStatus::SRM_ABORTED:
        case SrmStatus::SRM_REQUEST_TIMED_OUT:
        case SrmStatus::SRM_INTERNAL_ERROR:
        case SrmStatus::SRM_FAILURE:
            // Ok
        break;
        case SrmStatus::SRM_NO_USER_SPACE:
        case SrmStatus::SRM_NO_FREE_SPACE:
            // these are valid for the request, for probably in SRM 1.1 are
            // related only to a file: consider generic failure
        default:
            // Set to generic Failure
            req_status = SrmStatus::SRM_FAILURE;
    }
    
    return req_status;    
}

/*
 * translate_getfile_status
 *
 * Translate the file status of a get request
 */
SrmStatus::Code translate_getfile_status(const char * state, const char * errorMessage) /* throw (LogicError,SrmBadResponse)*/ {
    SrmStatus::Code file_status = SrmStatus::SRM_UNDEF; 

    // Check Status
    if((0 == state) || (0 == strlen(state))){
        // Error: No Request State Set
        throw SrmBadResponse("empty file status returned");
    }
                                                                                                                                                       
    // translate character state into SrmStatusCode
    if(0 == strcasecmp(state, SRM_1_1_STATUS_READY)){
        file_status = SrmStatus::SRM_REQUEST_INPROGRESS;
    } else if(0 == strcasecmp(state, SRM_1_1_STATUS_FAILED)){
        file_status = GSoapContext::getError(errorMessage).code;
    } else if(0 == strcasecmp(state, SRM_1_1_STATUS_DONE)){
        file_status = SrmStatus::SRM_RELEASED;
    } else if(0 == strcasecmp(state, SRM_1_1_STATUS_ACTIVE)){
        file_status = SrmStatus::SRM_FILE_PINNED;
    } else if(0 == strcasecmp(state, SRM_1_1_STATUS_RUNNING)){
        file_status = SrmStatus::SRM_FILE_PINNED;
    } else if(0 == strcasecmp(state, SRM_1_1_STATUS_PENDING)){
        file_status = SrmStatus::SRM_REQUEST_QUEUED;
    } else {
        // Error: Unknown Request State
        std::stringstream reason;
        reason << "invalid file status returned [" << state << "]";
        throw SrmBadResponse(reason.str());
    }
    
    // Enforce Error Codes
    switch(file_status){
        case SrmStatus::SRM_FILE_PINNED:
        case SrmStatus::SRM_REQUEST_QUEUED:
        case SrmStatus::SRM_REQUEST_INPROGRESS:
        case SrmStatus::SRM_ABORTED:
        case SrmStatus::SRM_RELEASED:
        case SrmStatus::SRM_FILE_LOST:
        case SrmStatus::SRM_FILE_BUSY:
        case SrmStatus::SRM_FILE_UNAVAILABLE:
        case SrmStatus::SRM_INVALID_PATH:
        case SrmStatus::SRM_AUTHORIZATION_FAILURE:
        case SrmStatus::SRM_FILE_LIFETIME_EXPIRED:
        case SrmStatus::SRM_FAILURE:
            // Ok
        break;
        default:
            // Set to generic Failure
            file_status = SrmStatus::SRM_FAILURE;
    }
    
    return file_status;
}
    
/*
 * file_from_getrequest_status
 *
 * Return the correct file status in case the response does not contain any file 
 */
SrmStatus::Code file_from_getrequest_status(SrmStatus::Code req_status) /*throw (SrmBadResponse)*/{
    SrmStatus::Code file_status = SrmStatus::SRM_UNDEF;
    switch(req_status){
        case SrmStatus::SRM_REQUEST_QUEUED:
            file_status = SrmStatus::SRM_REQUEST_QUEUED;
        break;
        case SrmStatus::SRM_REQUEST_INPROGRESS:
            file_status = SrmStatus::SRM_REQUEST_INPROGRESS;
        break;
        case SrmStatus::SRM_SUCCESS:
            file_status = SrmStatus::SRM_FILE_PINNED;
        break;
        case SrmStatus::SRM_PARTIAL_SUCCESS:
            // This should never happen
            throw SrmBadResponse("cannot have a partial success and no file status returned");
        break;
        case SrmStatus::SRM_FAILURE:
        case SrmStatus::SRM_AUTHENTICATION_FAILURE:
        case SrmStatus::SRM_AUTHORIZATION_FAILURE:
        case SrmStatus::SRM_ABORTED:
            file_status = req_status;
        break;
        default:
            file_status = SrmStatus::SRM_FAILURE;
    }
    return file_status;
}

// -----------------------------------------------------------------------------
// Put Helpers
// -----------------------------------------------------------------------------

/*
 * update_putfilerequest
 *
 * Update the PutFileRequest wiith the content of SRM response 
 */
int update_putfilerequest(PutFileRequest& file, const ::srm::srm2__RequestFileStatus * file_stat, const ::srm::srm2__RequestStatus * srm_stat) /*throw (LogicError, SrmBadResponse)*/ {
    if(0 == file_stat){
        throw NullPointerException("null file stat");
    }
    if(0 == srm_stat){
        throw NullPointerException("null srm stat");
    }

    // Set the Status Message
    file.status.message = get_file_message(file_stat->fileId, srm_stat->errorMessage);
    if((true == file.status.message.empty()) && 
       (0 == strcasecmp(file_stat->state, SRM_1_1_STATUS_FAILED)) &&
       (0 != srm_stat->errorMessage)){
        // DPM Workaround
        // TODO: find a better way to do it
        file.status.message = srm_stat->errorMessage;
    }
    
    // Set the Status Code
    file.status.code = translate_putfile_status(file_stat->state, file.status.message.c_str());
    
    // Set the Size
    if(file.size <= 0){
        file.size   = file_stat->size;
    }
    
    // Update SURl (if empty)
    if((true == file.surl.empty()) && (0 != file_stat->SURL)){
        file.surl = file_stat->SURL;
    }
    // Update TURL
    if((true == file.turl.empty()) && (0 != file_stat->TURL)){
        file.turl = file_stat->TURL;
    }
    // Get the Estimated time 
    file.extimatedWaitTime = srm_stat->estTimeToStart;
    if(file.extimatedWaitTime <= 0){
        file.extimatedWaitTime = srm_stat->retryDeltaTime;
    }
    
    // Return the file Id
    return file_stat->fileId;
}

/*
 * translate_putrequest_status
 *
 * Translate the status of a put request
 */
SrmStatus::Code translate_putrequest_status(const char * state, const char * errorMessage) /* throw (LogicError,SrmBadResponse)*/ {
    SrmStatus::Code req_status = SrmStatus::SRM_UNDEF; 
    
    // Check Status
    if((0 == state) || (0 == strlen(state))){
        // Error: No Request State Set
        throw SrmBadResponse("no request status returned");
    }
                                                                                                                                                       
    // translate character state into SrmStatusCode
    if(0 == strcasecmp(state, SRM_1_1_STATUS_READY)){
        req_status = SrmStatus::SRM_REQUEST_INPROGRESS;
    } else if(0 == strcasecmp(state, SRM_1_1_STATUS_FAILED)){
        req_status = GSoapContext::getError(errorMessage).code;
    } else if(0 == strcasecmp(state, SRM_1_1_STATUS_DONE)){
        req_status = SrmStatus::SRM_SUCCESS;
    } else if(0 == strcasecmp(state, SRM_1_1_STATUS_ACTIVE)){
        req_status = SrmStatus::SRM_SUCCESS;
    } else if(0 == strcasecmp(state, SRM_1_1_STATUS_PENDING)){
        req_status = SrmStatus::SRM_REQUEST_QUEUED;
    } else {
        // Error: Unknown Request State
        std::stringstream reason;
        reason << "invalid request status returned [" << state << "]";
        throw SrmBadResponse(reason.str());
    }
    
    // Enforce Error Codes
    switch(req_status){
        case SrmStatus::SRM_SUCCESS:
        case SrmStatus::SRM_REQUEST_QUEUED:
        case SrmStatus::SRM_REQUEST_INPROGRESS:
        case SrmStatus::SRM_PARTIAL_SUCCESS:
        case SrmStatus::SRM_AUTHENTICATION_FAILURE:
        case SrmStatus::SRM_AUTHORIZATION_FAILURE:
        case SrmStatus::SRM_INVALID_REQUEST:
        case SrmStatus::SRM_SPACE_LIFETIME_EXPIRED:
        case SrmStatus::SRM_EXCEED_ALLOCATION:
        case SrmStatus::SRM_REQUEST_TIMED_OUT:
        case SrmStatus::SRM_NOT_SUPPORTED:
        case SrmStatus::SRM_INTERNAL_ERROR:
        case SrmStatus::SRM_FAILURE:
            // Ok
        break;
        case SrmStatus::SRM_NO_USER_SPACE:
        case SrmStatus::SRM_NO_FREE_SPACE:
            // these are valid for the request, for probably in SRM 1.1 are
            // related only to a file: consider generic failure
        default:
            // Set to generic Failure
            req_status = SrmStatus::SRM_FAILURE;
    }
    
    return req_status;    
}

/*
 * translate_putfile_status
 *
 * Translate the file status of a put request
 */
SrmStatus::Code translate_putfile_status(const char * state, const char * errorMessage) /* throw (LogicError,SrmBadResponse)*/ {
    SrmStatus::Code file_status = SrmStatus::SRM_UNDEF; 

    // Check Status
    if((0 == state) || (0 == strlen(state))){
        // Error: No Request State Set
        throw SrmBadResponse("empty file status returned");
    }
                                                                                                                                                       
    // translate character state into SrmStatusCode
    if(0 == strcasecmp(state, SRM_1_1_STATUS_READY)){
        file_status = SrmStatus::SRM_REQUEST_INPROGRESS;
    } else if(0 == strcasecmp(state, SRM_1_1_STATUS_FAILED)){
        file_status = GSoapContext::getError(errorMessage).code;
    } else if(0 == strcasecmp(state, SRM_1_1_STATUS_DONE)){
        file_status = SrmStatus::SRM_SUCCESS;
    } else if(0 == strcasecmp(state, SRM_1_1_STATUS_ACTIVE)){
        file_status = SrmStatus::SRM_SPACE_AVAILABLE;
    } else if(0 == strcasecmp(state, SRM_1_1_STATUS_RUNNING)){
        file_status = SrmStatus::SRM_SPACE_AVAILABLE;
    } else if(0 == strcasecmp(state, SRM_1_1_STATUS_PENDING)){
        file_status = SrmStatus::SRM_REQUEST_QUEUED;
    } else {
        // Error: Unknown Request State
        std::stringstream reason;
        reason << "invalid file status returned [" << state << "]";
        throw SrmBadResponse(reason.str());
    }

    // Enforce Error Codes
    switch(file_status){
        case SrmStatus::SRM_SPACE_AVAILABLE:
        case SrmStatus::SRM_REQUEST_QUEUED:
        case SrmStatus::SRM_REQUEST_INPROGRESS:
        case SrmStatus::SRM_SUCCESS:
        case SrmStatus::SRM_FILE_BUSY:
        case SrmStatus::SRM_FILE_IN_CACHE:
        case SrmStatus::SRM_INVALID_PATH:
        case SrmStatus::SRM_DUPLICATION_ERROR:
        case SrmStatus::SRM_AUTHORIZATION_FAILURE:
        case SrmStatus::SRM_ABORTED:
        case SrmStatus::SRM_RELEASED:
        case SrmStatus::SRM_FAILURE:
		// added on 25/04/07
		case SrmStatus::SRM_NO_FREE_SPACE:
		case SrmStatus::SRM_NO_USER_SPACE:
		case SrmStatus::SRM_REQUEST_SUSPENDED:
            // Ok
        break;
        default:
            // Set to generic Failure
            file_status = SrmStatus::SRM_FAILURE;
    }
    return file_status;
}
    
/*
 * file_from_putrequest_status
 *
 * Return the correct file status in case the response does not contain any file 
 */
SrmStatus::Code file_from_putrequest_status(SrmStatus::Code req_status) /*throw (SrmBadResponse)*/{
    SrmStatus::Code file_status = SrmStatus::SRM_UNDEF;
    switch(req_status){
        case SrmStatus::SRM_REQUEST_QUEUED:
            file_status = SrmStatus::SRM_REQUEST_QUEUED;
        break;
        case SrmStatus::SRM_REQUEST_INPROGRESS:
            file_status = SrmStatus::SRM_REQUEST_INPROGRESS;
        break;
        case SrmStatus::SRM_SUCCESS:
            file_status = SrmStatus::SRM_SPACE_AVAILABLE;
        break;
        case SrmStatus::SRM_PARTIAL_SUCCESS:
            // This should never happen
            throw SrmBadResponse("cannot have a partial success and no file status returned");
        break;
        case SrmStatus::SRM_FAILURE:
        case SrmStatus::SRM_AUTHENTICATION_FAILURE:
        case SrmStatus::SRM_AUTHORIZATION_FAILURE:
        case SrmStatus::SRM_ABORTED:
            file_status = req_status;
        break;
        default:
            file_status = SrmStatus::SRM_FAILURE;
    }
    return file_status;
}

// -----------------------------------------------------------------------------
// Copy Helpers
// -----------------------------------------------------------------------------

/*
 * update_copyfilerequest
 *
 * Update the CopyFileRequest wiith the content of SRM response 
 */
int update_copyfilerequest(CopyFileRequest& file, const ::srm::srm2__RequestFileStatus * file_stat, const ::srm::srm2__RequestStatus * srm_stat) /*throw (LogicError, SrmBadResponse)*/ {
    if(0 == file_stat){
        throw NullPointerException("null file stat");
    }
    if(0 == srm_stat){
        throw NullPointerException("null srm stat");
    }

    // Set the Status Message
    file.status.message = get_file_message(file_stat->fileId, srm_stat->errorMessage);
    if((true == file.status.message.empty()) && 
       (0 == strcasecmp(file_stat->state, SRM_1_1_STATUS_FAILED)) &&
       (0 != srm_stat->errorMessage)){
        // DPM Workaround
        // TODO: find a better way to do it
        file.status.message = srm_stat->errorMessage;
    }
    
    // Set the Status Code
    file.status.code = translate_copyfile_status(file_stat->state, file.status.message.c_str());
    
    // Set the Size
    if(file.size <= 0){
        file.size   = file_stat->size;
    }

    // Update SURL (if empty)
    if(true == file.sourceSurl.empty()){
        if((0 != file_stat->sourceFilename) && (strlen(file_stat->sourceFilename) > 0)){
            file.sourceSurl = file_stat->sourceFilename;
        } else if((0 != file_stat->SURL) && (strlen(file_stat->SURL)> 0)){
            file.sourceSurl = file_stat->SURL;
        }
    }
    // Update TURL
    if(true == file.targetSurl.empty()){
        if((0 != file_stat->destFilename) && (strlen(file_stat->destFilename) > 0)){
            file.targetSurl = file_stat->destFilename;
        } else if((0 != file_stat->TURL) && (strlen(file_stat->TURL) > 0)){
            file.targetSurl = file_stat->TURL;
        }
    }
    // Get the Estimated time 
    file.extimatedWaitTime = srm_stat->estTimeToStart;
    if(file.extimatedWaitTime <= 0){
        file.extimatedWaitTime = srm_stat->retryDeltaTime;
    }
    
    // Return the file Id
    return file_stat->fileId;
}

/*
 * translate_copyrequest_status
 *
 * Translate the status of a copy request
 */
SrmStatus::Code translate_copyrequest_status(const char * state, const char * errorMessage) /* throw (LogicError,SrmBadResponse)*/ {
    SrmStatus::Code req_status = SrmStatus::SRM_UNDEF; 
    
    // Check Status
    if((0 == state) || (0 == strlen(state))){
        // Error: No Request State Set
        throw SrmBadResponse("no request status returned");
    }

    // translate character state into SrmStatusCode
    if(0 == strcasecmp(state, SRM_1_1_STATUS_READY)){
        req_status = SrmStatus::SRM_REQUEST_INPROGRESS;
    } else if(0 == strcasecmp(state, SRM_1_1_STATUS_FAILED)){
        req_status = GSoapContext::getError(errorMessage).code;
    } else if(0 == strcasecmp(state, SRM_1_1_STATUS_DONE)){
        req_status = SrmStatus::SRM_SUCCESS;
    } else if(0 == strcasecmp(state, SRM_1_1_STATUS_ACTIVE)){
        req_status = SrmStatus::SRM_REQUEST_INPROGRESS;
    } else if(0 == strcasecmp(state, SRM_1_1_STATUS_PENDING)){
        req_status = SrmStatus::SRM_REQUEST_QUEUED;
    } else {
        // Error: Unknown Request State
        std::stringstream reason;
        reason << "invalid request status returned [" << state << "]";
        throw SrmBadResponse(reason.str());
    }
    // Enforce Error Codes
    switch(req_status){
        case SrmStatus::SRM_SUCCESS:
        case SrmStatus::SRM_REQUEST_QUEUED:
        case SrmStatus::SRM_REQUEST_INPROGRESS:
        case SrmStatus::SRM_PARTIAL_SUCCESS:
        case SrmStatus::SRM_AUTHENTICATION_FAILURE:
        case SrmStatus::SRM_AUTHORIZATION_FAILURE:
        case SrmStatus::SRM_INVALID_REQUEST:
        case SrmStatus::SRM_TOO_MANY_RESULTS:
        case SrmStatus::SRM_REQUEST_TIMED_OUT:
        case SrmStatus::SRM_SPACE_LIFETIME_EXPIRED:
        case SrmStatus::SRM_ABORTED:
        case SrmStatus::SRM_EXCEED_ALLOCATION:
        case SrmStatus::SRM_NOT_SUPPORTED:
        case SrmStatus::SRM_INTERNAL_ERROR:
        case SrmStatus::SRM_FAILURE:
            // Ok
        break;
        case SrmStatus::SRM_NO_USER_SPACE:
        case SrmStatus::SRM_NO_FREE_SPACE:
            // these are valid for the request, for probably in SRM 1.1 are
            // related only to a file: consider generic failure
        default:
            // Set to generic Failure
            req_status = SrmStatus::SRM_FAILURE;
    }
    return req_status;    
}

/*
 * translate_copyfile_status
 *
 * Translate the file status of a copy request
 */
SrmStatus::Code translate_copyfile_status(const char * state, const char * errorMessage) /* throw (LogicError,SrmBadResponse)*/ {
    SrmStatus::Code file_status = SrmStatus::SRM_UNDEF; 

    // Check Status
    if((0 == state) || (0 == strlen(state))){
        // Error: No Request State Set
        throw SrmBadResponse("empty file status returned");
    }
                                       
    // translate character state into SrmStatusCode
    if(0 == strcasecmp(state, SRM_1_1_STATUS_READY)){
        file_status = SrmStatus::SRM_REQUEST_INPROGRESS;
    } else if(0 == strcasecmp(state, SRM_1_1_STATUS_FAILED)){
        file_status = GSoapContext::getError(errorMessage).code;
    } else if(0 == strcasecmp(state, SRM_1_1_STATUS_DONE)){
        file_status = SrmStatus::SRM_SUCCESS;
    } else if(0 == strcasecmp(state, SRM_1_1_STATUS_ACTIVE)){
        file_status = SrmStatus::SRM_REQUEST_INPROGRESS;
    } else if(0 == strcasecmp(state, SRM_1_1_STATUS_RUNNING)){
        file_status = SrmStatus::SRM_REQUEST_INPROGRESS;
    } else if(0 == strcasecmp(state, SRM_1_1_STATUS_PENDING)){
        file_status = SrmStatus::SRM_REQUEST_QUEUED;
    } else {
        // Error: Unknown Request State
        std::stringstream reason;
        reason << "invalid file status returned [" << state << "]";
        throw SrmBadResponse(reason.str());
    }
    
    // Enforce Error Codes
    switch(file_status){
        case SrmStatus::SRM_SUCCESS:
        case SrmStatus::SRM_REQUEST_QUEUED:
        case SrmStatus::SRM_REQUEST_INPROGRESS:
        case SrmStatus::SRM_FILE_LOST:
        case SrmStatus::SRM_FILE_BUSY:
        case SrmStatus::SRM_FILE_UNAVAILABLE:
        case SrmStatus::SRM_FILE_LIFETIME_EXPIRED:
        case SrmStatus::SRM_INVALID_PATH:
        case SrmStatus::SRM_DUPLICATION_ERROR:
        case SrmStatus::SRM_AUTHORIZATION_FAILURE:
        case SrmStatus::SRM_ABORTED:
        case SrmStatus::SRM_RELEASED:
        case SrmStatus::SRM_FAILURE:
            // Ok
        break;
        default:
            // Set to generic Failure
            file_status = SrmStatus::SRM_FAILURE;
    }
    return file_status;
}
    
/*
 * file_from_copyrequest_status
 *
 * Return the correct file status in case the response does not contain any file 
 */
SrmStatus::Code file_from_copyrequest_status(SrmStatus::Code req_status) /*throw (SrmBadResponse)*/{
    SrmStatus::Code file_status = SrmStatus::SRM_UNDEF;
    
    switch(req_status){
        case SrmStatus::SRM_REQUEST_QUEUED:
        case SrmStatus::SRM_REQUEST_INPROGRESS:
        case SrmStatus::SRM_SUCCESS:
            file_status = req_status;
        break;
        case SrmStatus::SRM_PARTIAL_SUCCESS:
            // This should never happen
            throw SrmBadResponse("cannot have a partial success and no file status returned");
        break;
        case SrmStatus::SRM_FAILURE:
        case SrmStatus::SRM_AUTHENTICATION_FAILURE:
        case SrmStatus::SRM_AUTHORIZATION_FAILURE:
        case SrmStatus::SRM_ABORTED:
            file_status = req_status;
        break;
        default:
            file_status = SrmStatus::SRM_FAILURE;
    }
    return file_status;
}

// -----------------------------------------------------------------------------
// BringOnline Helpers
// -----------------------------------------------------------------------------

/*
 * update_bringonlinefilerequest
 *
 * Update the GetFileRequest with the content of SRM response 
 */
int update_bringonlinefilerequest(BringOnlineFileRequest& file, const ::srm::srm2__RequestFileStatus * file_stat, const ::srm::srm2__RequestStatus * srm_stat) /*throw (LogicError, SrmBadResponse)*/ {
    if(0 == file_stat){
        throw NullPointerException("null file stat");
    }
    if(0 == srm_stat){
        throw NullPointerException("null srm stat");
    }

    // Set the Status Message
    file.status.message = get_file_message(file_stat->fileId, srm_stat->errorMessage);
    if((true == file.status.message.empty()) && 
       (0 == strcasecmp(file_stat->state, SRM_1_1_STATUS_FAILED)) &&
       (0 != srm_stat->errorMessage)){
        // DPM Workaround
        // TODO: find a better way to do it
        file.status.message = srm_stat->errorMessage;
    }
    
    // Set the Status Code
    file.status.code = translate_bringonlinefile_status(file_stat->state, file.status.message.c_str());
    
    // Set the Size
    file.size   = file_stat->size;
    
    // Update SURl (if empty)
    if((true == file.surl.empty()) && (0 != file_stat->SURL)){
        file.surl = file_stat->SURL;
    }
    // Get the Estimated time 
    file.extimatedWaitTime = srm_stat->estTimeToStart;
    if(file.extimatedWaitTime <= 0){
        file.extimatedWaitTime = srm_stat->retryDeltaTime;
    }
    
    // Return the file Id
    return file_stat->fileId;
}

/*
 * translate_bringonlinerequest_status
 *
 * Translate the status of a BringOnline request
 */
SrmStatus::Code translate_bringonlinerequest_status(const char * state, const char * errorMessage) /* throw (LogicError,SrmBadResponse)*/ {
    SrmStatus::Code req_status = SrmStatus::SRM_UNDEF; 
    
    // Check Status
    if((0 == state) || (0 == strlen(state))){
        // Error: No Request State Set
        throw SrmBadResponse("no request status returned");
    }
                                                                                                                                                       
    // translate character state into SrmStatusCode
    if(0 == strcasecmp(state, SRM_1_1_STATUS_READY)){
        req_status = SrmStatus::SRM_REQUEST_INPROGRESS;
    } else if(0 == strcasecmp(state, SRM_1_1_STATUS_FAILED)){
        req_status = GSoapContext::getError(errorMessage).code;
    } else if(0 == strcasecmp(state, SRM_1_1_STATUS_DONE)){
        req_status = SrmStatus::SRM_SUCCESS;
    } else if(0 == strcasecmp(state, SRM_1_1_STATUS_ACTIVE)){
        req_status = SrmStatus::SRM_REQUEST_INPROGRESS;
    } else if(0 == strcasecmp(state, SRM_1_1_STATUS_PENDING)){
        req_status = SrmStatus::SRM_REQUEST_QUEUED;
    } else {
        // Error: Unknown Request State
        std::stringstream reason;
        reason << "invalid request status returned [" << state << "]";
        throw SrmBadResponse(reason.str());
    }
    
    // Enforce Error Codes
    switch(req_status){
        case SrmStatus::SRM_SUCCESS:
        case SrmStatus::SRM_REQUEST_QUEUED:
        case SrmStatus::SRM_REQUEST_INPROGRESS:
        case SrmStatus::SRM_PARTIAL_SUCCESS:
        case SrmStatus::SRM_AUTHENTICATION_FAILURE:
        case SrmStatus::SRM_AUTHORIZATION_FAILURE:
        case SrmStatus::SRM_INVALID_REQUEST:
        case SrmStatus::SRM_NOT_SUPPORTED:
        case SrmStatus::SRM_SPACE_LIFETIME_EXPIRED:
        case SrmStatus::SRM_EXCEED_ALLOCATION:
        case SrmStatus::SRM_ABORTED:
        case SrmStatus::SRM_REQUEST_TIMED_OUT:
        case SrmStatus::SRM_INTERNAL_ERROR:
        case SrmStatus::SRM_FAILURE:
            // Ok
        break;
        case SrmStatus::SRM_NO_USER_SPACE:
        case SrmStatus::SRM_NO_FREE_SPACE:
            // these are valid for the request, for probably in SRM 1.1 are
            // related only to a file: consider generic failure
        default:
            // Set to generic Failure
            req_status = SrmStatus::SRM_FAILURE;
    }
    
    return req_status;    
}

/*
 * translate_getfile_status
 *
 * Translate the file status of a BringOnline request
 */
SrmStatus::Code translate_bringonlinefile_status(const char * state, const char * errorMessage) /* throw (LogicError,SrmBadResponse)*/ {
    SrmStatus::Code file_status = SrmStatus::SRM_UNDEF; 

    // Check Status
    if((0 == state) || (0 == strlen(state))){
        // Error: No Request State Set
        throw SrmBadResponse("empty file status returned");
    }
                                                                                                                                                       
    // translate character state into SrmStatusCode
    if(0 == strcasecmp(state, SRM_1_1_STATUS_READY)){
        file_status = SrmStatus::SRM_REQUEST_INPROGRESS;
    } else if(0 == strcasecmp(state, SRM_1_1_STATUS_FAILED)){
        file_status = GSoapContext::getError(errorMessage).code;
    } else if(0 == strcasecmp(state, SRM_1_1_STATUS_DONE)){
        file_status = SrmStatus::SRM_SUCCESS;
    } else if(0 == strcasecmp(state, SRM_1_1_STATUS_ACTIVE)){
        file_status = SrmStatus::SRM_REQUEST_INPROGRESS;
    } else if(0 == strcasecmp(state, SRM_1_1_STATUS_RUNNING)){
        file_status = SrmStatus::SRM_REQUEST_INPROGRESS;
    } else if(0 == strcasecmp(state, SRM_1_1_STATUS_PENDING)){
        file_status = SrmStatus::SRM_REQUEST_QUEUED;
    } else {
        // Error: Unknown Request State
        std::stringstream reason;
        reason << "invalid file status returned [" << state << "]";
        throw SrmBadResponse(reason.str());
    }
    
    // Enforce Error Codes
    switch(file_status){
        case SrmStatus::SRM_SUCCESS:
        case SrmStatus::SRM_REQUEST_QUEUED:
        case SrmStatus::SRM_REQUEST_INPROGRESS:
        case SrmStatus::SRM_AUTHORIZATION_FAILURE:
        case SrmStatus::SRM_ABORTED:
        case SrmStatus::SRM_RELEASED:
        case SrmStatus::SRM_FILE_LOST:
        case SrmStatus::SRM_FILE_BUSY:
        case SrmStatus::SRM_FILE_UNAVAILABLE:
        case SrmStatus::SRM_INVALID_PATH:
        case SrmStatus::SRM_FILE_LIFETIME_EXPIRED:
        case SrmStatus::SRM_FAILURE:
            // Ok
        break;
        default:
            // Set to generic Failure
            file_status = SrmStatus::SRM_FAILURE;
    }
    return file_status;
}
    
/*
 * file_from_bringonlinerequest_status
 *
 * Return the correct file status in case the response does not contain any file 
 */
SrmStatus::Code file_from_bringonlinerequest_status(SrmStatus::Code req_status) /*throw (SrmBadResponse)*/{
    SrmStatus::Code file_status = SrmStatus::SRM_UNDEF;
    switch(req_status){
        case SrmStatus::SRM_REQUEST_QUEUED:
            file_status = SrmStatus::SRM_REQUEST_QUEUED;
        break;
        case SrmStatus::SRM_REQUEST_INPROGRESS:
            file_status = SrmStatus::SRM_REQUEST_INPROGRESS;
        break;
        case SrmStatus::SRM_SUCCESS:
            file_status = SrmStatus::SRM_SUCCESS;
        break;
        case SrmStatus::SRM_PARTIAL_SUCCESS:
            // This should never happen
            throw SrmBadResponse("cannot have a partial success and no file status returned");
        break;
        case SrmStatus::SRM_FAILURE:
        case SrmStatus::SRM_AUTHENTICATION_FAILURE:
        case SrmStatus::SRM_AUTHORIZATION_FAILURE:
        case SrmStatus::SRM_ABORTED:
            file_status = req_status;
        break;
        default:
            file_status = SrmStatus::SRM_FAILURE;
    }
    return file_status;
}

// -----------------------------------------------------------------------------
// Stat helpers
// -----------------------------------------------------------------------------
    
/**
 * Update the MetadataPathDetail with the content of SRM response
 */
void update_file_metadata(MetaDataPathDetail& file, const ::srm::srm2__FileMetaData * file_metadata) /*throw (LogicError, SrmBadResponse)*/{
    if(0 == file_metadata){
        throw NullPointerException("null file metadata");
    }

    // Set the Status Code
    file.status.code = SrmStatus::SRM_SUCCESS;

    // SURL
    if((true == file.surl.empty()) && (0 != file_metadata->SURL)){
        file.surl = file_metadata->SURL;
    }
    // size
    if(file.size <= 0){
        file.size   = file_metadata->size;
    }
    // createdAtTime: Not Supported
    // lastModificationTime: Not Supported
    
    // fileStorageType & fileLocality
    if(true == file_metadata->isPermanent){
        if(true == file_metadata->isCached){
            file.fileLocality = FileLocality::ONLINE_AND_NEARLINE;
        } else {
            file.fileLocality = FileLocality::NEARLINE;
        }
        file.fileStorageType = FileStorageType::PERMANENT;
    } else {
        if(true == file_metadata->isCached){
            file.fileLocality = FileLocality::ONLINE;
        } else {
            file.fileLocality = FileLocality::UNAVAILABLE;
        }
        file.fileStorageType = FileStorageType::VOLATILE;
    }
    
    // spaceTokens: Not supported
    
    // fileType
    if(S_ISLNK(file_metadata->permMode)){
        file.fileType = FileType::LINK;
    } else if(S_ISDIR(file_metadata->permMode)){
        file.fileType = FileType::DIRECTORY;
    } else {
        // assume it's a file
        file.fileType = FileType::FILE;
    }
    
    // lifetimeAssigned: Not Supported
    // lifetimeLeft: Not Supported
    
    // ownerPermission
    if(0 != file_metadata->owner){
        file.ownerPermission.userID = file_metadata->owner;
    }
    file.ownerPermission.mode = (PermissionMode::Value)((file_metadata->permMode >> 6) & 07);
    // groupPermission
    if(0 != file_metadata->group){
        file.groupPermission.groupID = file_metadata->group;
    }
    file.groupPermission.mode = (PermissionMode::Value)((file_metadata->permMode >> 3) & 07);
    // otherPermission
    file.otherPermission = (PermissionMode::Value)(file_metadata->permMode & 07);
    
    // checkSumType
    if(0 != file_metadata->checksumType){
        file.checkSumType = file_metadata->checksumType;
    }
    // checkSumValue
    if(0 != file_metadata->checksumValue){
        file.checkSumValue = file_metadata->checksumValue;
    }
}

} // End namespace srm1_1
} // End namespace util
} // End namespace srm
} // End namespace data
} // End namespace glite
