/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004.
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  GLite Data SRM Util CPP: PrepareToGet Class
 *
 *  Authors: Paolo Badino <paolo.badino@cern.ch>,
 *           Paolo Tedesco <paolo.tedesco@cern.ch>
 *           Zsolt Molnar <zsolt.molnar@cern.ch - http://www.zsoltmolnar.hu>
 *  Version info: $Id: PrepareToGet.cpp,v 1.13 2009/07/30 17:40:34 molnarzs Exp $
 *  Release: $Name:  $
 *
 */
#include "PrepareToGet.h"
#include "SrmVersion.h"
#include "GSoapContext.h"
#include "srm-utility.h"
#include "srm-common-get.h"
#include "srm-common.h"
#include "context-caller.h"
#include "BackoffLogicExecutor.h"

using namespace glite::data::srm::util::srm2_2;
using namespace ::srm2;

DEFINE_REQUEST_FACTORY(PrepareToGet, 2_2);
DEFINE_ASYNCH_REQUEST_FACTORY(PrepareToGet, 2_2);

void PrepareToGet::execute()
{
	this->checkToken(false);
	this->status = BackoffLogicExecutor::executeAsync(*this);
}

/**
 * Execute the Request
 */
void PrepareToGet::executeOneCall() /* throw (LogicError, RemoteError, SrmError)*/{
    CallInfo info;
	this->m_token.clear();

	if(true == files.empty()){
        throw InvalidArgumentException("empty files array");
    }
    // Create GSoap Context
    GSoapContext g(this->ctx(),false);

    ::srm2__TFileStorageType       fileStorageType;
    ::srm2__TRetentionPolicyInfo   retentionPolicyInfo;
    ::srm2__TTransferParameters    transferParameters;
    ::srm2__TAccessPattern         accessPattern;
    ::srm2__TAccessLatency         accessLatency;
    ::srm2__TConnectionType        connectionType;
    ::srm2__ArrayOfString          protocols;
    ::srm2__ArrayOfTGetFileRequest arrayOfFileRequests;

    // Create Request
    ::srm2__srmPrepareToGetRequest req;
    req.authorizationID        = 0;
    req.userRequestDescription = 0;
    req.storageSystemInfo      = 0;
    if(this->desiredFileStorageType != FileStorageType::DEFAULT){
        fileStorageType = translate_FileStorageType(this->desiredFileStorageType);
        req.desiredFileStorageType = &(fileStorageType);
    } else {
        req.desiredFileStorageType = 0;
    }
    if(this->desiredTotalRequestTime != -1){
        req.desiredTotalRequestTime = &(this->desiredTotalRequestTime);
    } else {
        req.desiredTotalRequestTime = 0;
    }
    if(this->desiredPinLifeTime != -1){
        req.desiredPinLifeTime = &(this->desiredPinLifeTime);
    } else {
        req.desiredPinLifeTime = 0;
    }
    if(false == this->targetSpaceToken.empty()){
        req.targetSpaceToken = &(this->targetSpaceToken);
    } else {
        req.targetSpaceToken = 0;
    }
    // Retention Policy is a mandatory filed in RetentionPolicyInfo
    if(this->targetFileRetentionPolicyInfo.retentionPolicy != RetentionPolicy::DEFAULT){
        req.targetFileRetentionPolicyInfo = &(retentionPolicyInfo);
        // Translate RetentionPolicy
        req.targetFileRetentionPolicyInfo->retentionPolicy = translate_RetentionPolicy(this->targetFileRetentionPolicyInfo.retentionPolicy);
        if(this->targetFileRetentionPolicyInfo.accessLatency == AccessLatency::ONLINE){
            // Translate AccessLatency
            accessLatency = translate_AccessLatency(this->targetFileRetentionPolicyInfo.accessLatency);
            req.targetFileRetentionPolicyInfo->accessLatency = &(accessLatency);
        } else if(this->targetFileRetentionPolicyInfo.accessLatency != AccessLatency::DEFAULT){
            // AccessLatency should always be online !
            throw LogicError("AccessLatency in PrepareToGet can only be online");
        } else {
            req.targetFileRetentionPolicyInfo->accessLatency = 0;
        }
    } else {
        req.targetFileRetentionPolicyInfo = 0;
    }
    // Set Transfer Parameters
    req.transferParameters = &(transferParameters);
    if(this->transferParameters.accessPattern != AccessPattern::UNDEF){
        // Translate AccessPattern
        accessPattern = translate_AccessPattern(this->transferParameters.accessPattern);
        req.transferParameters->accessPattern = &(accessPattern);
    } else {
        req.transferParameters->accessPattern = 0;
    }
    if(this->transferParameters.connectionType != ConnectionType::UNDEF){
        // Translate ConnectionType
        connectionType = translate_ConnectionType(this->transferParameters.connectionType);
        req.transferParameters->connectionType = &(connectionType);
    } else {
        req.transferParameters->connectionType = 0;
    }
    req.transferParameters->arrayOfClientNetworks = 0;
    req.transferParameters->arrayOfTransferProtocols = &(protocols);
    req.transferParameters->arrayOfTransferProtocols->stringArray.push_back(this->transferParameters.protocol);

    req.arrayOfFileRequests = &(arrayOfFileRequests);
    // For each GetFileRequest
    ::srm2__TDirOption dir_option;
    dir_option.isSourceADirectory = false;
    dir_option.allLevelRecursive  = 0;
    dir_option.numOfLevels        = 0;
    std::vector<GetFileRequest>::iterator it;
    for(it = this->files.begin(); it != this->files.end(); ++it){
        info.add("surl",(*it).surl);
        // Add srm2__TGetFileRequest
        ::srm2__TGetFileRequest * file_request = ::soap_new_srm2__TGetFileRequest(g.soap().soap,-1);
        file_request->sourceSURL = (*it).surl;
        file_request->dirOption  = &(dir_option);
        req.arrayOfFileRequests->requestArray.push_back(file_request);
    }

    // Invoke SRM Get
    ::srm2__srmPrepareToGetResponse_ rsp;
    this->ctx().beforeCall("srm2__srmPrepareToGet");
    if(SOAP_OK != g.soap().srm2__srmPrepareToGet(&req,rsp)){
        this->ctx().onFailure("srm2__srmPrepareToGet",g.ip().c_str(),&info);
        // Failed to invoke remote procedure: throw the correct exception
        g.handleError("srm2__srmPrepareToGet");
    }
    CallOnSuccess call(this->ctx(),"srm2__srmPrepareToGet",g.ip().c_str(),&info);

    // Update the Request object
    SrmStatus result = update_request
        <PrepareToGet,GetFileRequest,::srm2__srmPrepareToGetResponse,::srm2__TGetRequestFileStatus>
        (*this,rsp.srmPrepareToGetResponse);

    // Update the request status
    this->status = result;
    info.token(m_token);

    if (rsp.srmPrepareToGetResponse->remainingTotalRequestTime)
    	m_eWT = *(rsp.srmPrepareToGetResponse->remainingTotalRequestTime);
}

/*
 * update
 *
 * Update the status of the Request.
 */
void PrepareToGet::update() /*throw (LogicError, RemoteError, SrmError)*/{
    // CheckPreconditions
    this->checkToken(true);
    CallInfo info(m_token);
    // Create GSoap Context
    GSoapContext g(this->ctx(),false);

    ::srm2__ArrayOfAnyURI sourceSURLs;

    // Create Request
    ::srm2__srmStatusOfGetRequestRequest req;
    req.authorizationID = 0;
    req.requestToken    = this->token();
    // #warning "probably the list should not be set."
    // TODO Always set the list ?
    // req.arrayOfSourceSURLs = &(sourceSURLs)
    if(false == this->files.empty()){
        // Fill the list of source Surls
        std::vector<GetFileRequest>::iterator it;
        for(it = this->files.begin(); it != this->files.end(); ++it){
            sourceSURLs.urlArray.push_back((*it).surl);
        }
        req.arrayOfSourceSURLs = &(sourceSURLs);
    } else {
        req.arrayOfSourceSURLs = 0;
    }

    // Invoke SRM StatusOfGetRequest
    ::srm2__srmStatusOfGetRequestResponse_ rsp;
    this->ctx().beforeCall("srm2__srmStatusOfGetRequest");
    if(SOAP_OK != g.soap().srm2__srmStatusOfGetRequest(&req,rsp)){
        this->ctx().onFailure("srm2__srmStatusOfGetRequest",g.ip().c_str(),&info);
        // Failed to invoke remote procedure: throw the correct exception
        g.handleError("srm2__srmStatusOfGetRequest");
    }
    this->ctx().onSuccess("srm2__srmStatusOfGetRequest",g.ip().c_str(),&info);

    // Update the Request object
    SrmStatus result = update_request
        <PrepareToGet,GetFileRequest,::srm2__srmStatusOfGetRequestResponse,::srm2__TGetRequestFileStatus>
        (*this,rsp.srmStatusOfGetRequestResponse);

    // Update the request status
    this->status = result;

    if (rsp.srmStatusOfGetRequestResponse->remainingTotalRequestTime)
    	m_eWT = *(rsp.srmStatusOfGetRequestResponse->remainingTotalRequestTime);
}

/*
 * done
 *
 * Complete a Request and release the allocated resources.
 */
void PrepareToGet::done() /*throw (LogicError, RemoteError, SrmError)*/{
    // CheckPreconditions
    this->checkToken(true);
    CallInfo info(m_token);
    // Create GSoap Context
    GSoapContext g(this->ctx(),false);

    ::srm2__ArrayOfAnyURI sourceSURLs;

    // Create Request
    ::srm2__srmReleaseFilesRequest req;
    std::string token   = this->token();
    req.requestToken    = &(token);
    req.authorizationID = 0;
    req.doRemove        = 0; // default is false
    req.arrayOfSURLs    = 0; // Release all SURLs
    // TODO Always set the list ?
    // req.arrayOfSURLs = &(sourceSURLs)
    if(false == this->files.empty()){
        // Fill the list of source Surls
        std::vector<GetFileRequest>::iterator it;
        for(it = this->files.begin(); it != this->files.end(); ++it){
            sourceSURLs.urlArray.push_back((*it).surl);
        }
        req.arrayOfSURLs = &(sourceSURLs);
    }

    // Invoke SRM ReleaseFiles
    ::srm2__srmReleaseFilesResponse_ rsp;
    this->ctx().beforeCall("srm2__srmReleaseFiles");
    if(SOAP_OK != g.soap().srm2__srmReleaseFiles(&req,rsp)){
        this->ctx().onFailure("srm2__srmReleaseFiles",g.ip().c_str(),&info);
        // Failed to invoke remote procedure: throw the correct exception
        g.handleError("srm2__srmReleaseFiles");
    }
    this->ctx().onSuccess("srm2__srmReleaseFiles",g.ip().c_str(),&info);

    // Update the Request object
    SrmStatus result = update_request
        <PrepareToGet,GetFileRequest,::srm2__srmReleaseFilesResponse,::srm2__TSURLReturnStatus>
        (*this,rsp.srmReleaseFilesResponse);

    // Update the request status
    this->status = result;
}

/*
 * abort
 *
 * Abort a Request and cleanup the allocated resources.
 */
void PrepareToGet::abort() /*throw (LogicError, RemoteError, SrmError)*/{
    // CheckPreconditions
    this->checkToken(true);
    CallInfo info(m_token);
    // Create GSoap Context
    GSoapContext g(this->ctx(),false);

    ::srm2__srmAbortRequestRequest req;
    req.requestToken    = this->token();
    req.authorizationID = 0;

    // Invoke SRM AbortRequest
    ::srm2__srmAbortRequestResponse_ rsp;
    this->ctx().beforeCall("srm2__srmAbortRequest");
    if(SOAP_OK != g.soap().srm2__srmAbortRequest(&req,rsp)){
        this->ctx().onFailure("srm2__srmAbortRequest",g.ip().c_str(),&info);
        // Failed to invoke remote procedure: throw the correct exception
        g.handleError("srm2__srmAbortRequest");
    }
    this->ctx().onSuccess("srm2__srmAbortRequest",g.ip().c_str(),&info);

    // Update the Request object
    SrmStatus result = update_request
        <PrepareToGet,GetFileRequest,::srm2__srmAbortRequestResponse>
        (*this,rsp.srmAbortRequestResponse);

    // Update the request status
    this->status = result;
}
