/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004.
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  GLite Data SRM Util CPP: PrepareToPut Class
 *
 *  Authors: Paolo Badino <paolo.badino@cern.ch>,
 *           Paolo Tedesco <paolo.tedesco@cern.ch>
 *           Zsolt Molnar <zsolt.molnar@cern.ch - http://www.zsoltmolnar.hu>
 *  Version info: $Id: PrepareToPut.cpp,v 1.9 2009/07/30 17:40:34 molnarzs Exp $
 *  Release: $Name:  $
 *
 */
#include "PrepareToPut.h"
#include "SrmVersion.h"
#include "GSoapContext.h"
#include "srm-utility.h"
#include "srm-common-put.h"
#include "srm-common.h"
#include "context-caller.h"
#include "BackoffLogicExecutor.h"

using namespace glite::data::srm::util::srm2_2;
using namespace ::srm2;

DEFINE_REQUEST_FACTORY(PrepareToPut, 2_2);
DEFINE_ASYNCH_REQUEST_FACTORY(PrepareToPut, 2_2);

/**
 * name of the TExtraInfo key to use to serialize copy pin lifetime
 */
const std::string PrepareToPut::CopyPinLifetimeKeyName = "CopyPinLifetime";

void PrepareToPut::execute()
{
	this->checkToken(false);
	this->status = BackoffLogicExecutor::executeAsync(*this);
}

/**
 * Execute the Request
 */
void PrepareToPut::executeOneCall() /* throw (LogicError, RemoteError, SrmError)*/{
    CallInfo info;
    this->m_token.clear();

    if(true == files.empty()){
        throw InvalidArgumentException("empty files array");
    }
    // Create GSoap Context
    GSoapContext g(this->ctx(),false);

    ::srm2__TOverwriteMode         overwriteOption;
    ::srm2__TFileStorageType       fileStorageType;
    ::srm2__TAccessPattern         accessPattern;
    ::srm2__TAccessLatency         accessLatency;
    ::srm2__TConnectionType        connectionType;
    ::srm2__TRetentionPolicyInfo   retentionPolicyInfo;
    ::srm2__ArrayOfString          protocols;
    ::srm2__TTransferParameters    transferParameters;
    ::srm2__ArrayOfTPutFileRequest arrayOfFileRequests;
    std::vector<ULONG64> expectedSizes;
    ::srm2__ArrayOfTExtraInfo      storageSystemInfo;

    // Create Request
    ::srm2__srmPrepareToPutRequest req;
    req.authorizationID        = 0;
    req.userRequestDescription = 0;
    req.storageSystemInfo      = 0;
    if(this->overwriteOption != OverwriteMode::DEFAULT){
        overwriteOption = translate_OverwriteMode(this->overwriteOption);
        req.overwriteOption = &(overwriteOption);
    } else {
        req.overwriteOption = 0;
    }
    if(this->desiredTotalRequestTime != -1){
        req.desiredTotalRequestTime = &(this->desiredTotalRequestTime);
    } else {
        req.desiredTotalRequestTime = 0;
    }
    if(this->desiredPinLifeTime != -1){
        req.desiredPinLifeTime = &(this->desiredPinLifeTime);
    } else {
        req.desiredPinLifeTime = 0;
    }
    if(this->desiredFileLifeTime != -1){
        req.desiredFileLifeTime = &(this->desiredFileLifeTime);
    } else {
        req.desiredFileLifeTime = 0;
    }
    if(this->desiredFileStorageType != FileStorageType::DEFAULT){
        fileStorageType = translate_FileStorageType(this->desiredFileStorageType);
        req.desiredFileStorageType = &(fileStorageType);
    } else {
        // For WLCG, default FileStorageType is PERMANENT
        // req.desiredFileStorageType = 0;
        fileStorageType = translate_FileStorageType(FileStorageType::PERMANENT);
        req.desiredFileStorageType = &(fileStorageType);
    }
    if(false == this->targetSpaceToken.empty()){
        req.targetSpaceToken = &(this->targetSpaceToken);
    } else {
        req.targetSpaceToken = 0;
    }
    // Retention Policy is a mandatory filed in RetentionPolicyInfo
    if(this->targetFileRetentionPolicyInfo.retentionPolicy != RetentionPolicy::DEFAULT){
        req.targetFileRetentionPolicyInfo = &(retentionPolicyInfo);
        // Translate RetentionPolicy
        req.targetFileRetentionPolicyInfo->retentionPolicy = translate_RetentionPolicy(this->targetFileRetentionPolicyInfo.retentionPolicy);
        if(this->targetFileRetentionPolicyInfo.accessLatency != AccessLatency::DEFAULT){
            // Translate AccessLatency
            accessLatency = translate_AccessLatency(this->targetFileRetentionPolicyInfo.accessLatency);
            req.targetFileRetentionPolicyInfo->accessLatency = &(accessLatency);
        } else {
            req.targetFileRetentionPolicyInfo->accessLatency = 0;
        }
    } else {
        req.targetFileRetentionPolicyInfo = 0;
    }
    // Set Transfer Parameters
    req.transferParameters = &(transferParameters);
    if(this->transferParameters.accessPattern != AccessPattern::UNDEF){
        // Translate AccessPattern
        accessPattern = translate_AccessPattern(this->transferParameters.accessPattern);
        req.transferParameters->accessPattern = &(accessPattern);
    } else {
        req.transferParameters->accessPattern = 0;
    }
    if(this->transferParameters.connectionType != ConnectionType::UNDEF){
        // Translate ConnectionType
        connectionType = translate_ConnectionType(this->transferParameters.connectionType);
        req.transferParameters->connectionType = &(connectionType);
    } else {
        req.transferParameters->connectionType = 0;
    }
    req.transferParameters->arrayOfClientNetworks = 0;
    req.transferParameters->arrayOfTransferProtocols = &(protocols);
    req.transferParameters->arrayOfTransferProtocols->stringArray.push_back(this->transferParameters.protocol);
    // copyPinLifetime
    if (-1 != this->copyPinLifetime){
        req.storageSystemInfo = &storageSystemInfo;
        ::srm2__TExtraInfo * copy_pin_lifetime = soap_new_srm2__TExtraInfo(g.soap().soap,-1);
        copy_pin_lifetime->key = CopyPinLifetimeKeyName;
        std::stringstream tmp;
        tmp << copyPinLifetime;
        copy_pin_lifetime->value = soap_new_std__string(g.soap().soap,-1);
        *(copy_pin_lifetime->value) = tmp.str().c_str();
        req.storageSystemInfo->extraInfoArray.push_back(copy_pin_lifetime);
    }


    req.arrayOfFileRequests = &(arrayOfFileRequests);
    // For each PutFileRequest
    expectedSizes.resize(this->files.size(),0);
    std::vector<PutFileRequest>::iterator it;
    std::vector<ULONG64>::iterator s_it;
    for(it = this->files.begin(), s_it = expectedSizes.begin(); it != this->files.end(); ++it, ++s_it){
        info.add("surl",(*it).surl);
        // Add srm2__TPutFileRequest
        ::srm2__TPutFileRequest * file_request = ::soap_new_srm2__TPutFileRequest(g.soap().soap,-1);
        file_request->targetSURL = &((*it).surl);
        (*s_it) = ((*it).size);
        file_request->expectedFileSize = &(*s_it);
        req.arrayOfFileRequests->requestArray.push_back(file_request);
    }

    // Invoke SRM Put
    ::srm2__srmPrepareToPutResponse_ rsp;
    this->ctx().beforeCall("srm2__srmPrepareToPut");
    if(SOAP_OK != g.soap().srm2__srmPrepareToPut(&req,rsp)){
        this->ctx().onFailure("srm2__srmPrepareToPut",g.ip().c_str(),&info);
        // Failed to invoke remote procedure: throw the correct exception
        g.handleError("srm2__srmPrepareToPut");
    }
    CallOnSuccess call(this->ctx(),"srm2__srmPrepareToPut",g.ip().c_str(),&info);

    // Update the Request object
    SrmStatus result = update_request
        <PrepareToPut,PutFileRequest,::srm2__srmPrepareToPutResponse,::srm2__TPutRequestFileStatus>
        (*this,rsp.srmPrepareToPutResponse);

    // Update the request status
    this->status = result;
    info.token(m_token);

    if (rsp.srmPrepareToPutResponse->remainingTotalRequestTime)
     	m_eWT = *(rsp.srmPrepareToPutResponse->remainingTotalRequestTime);
}

/*
 * update
 *
 * Update the status of the Request.
 */
void PrepareToPut::update() /*throw (LogicError, RemoteError, SrmError)*/{
    // CheckPreconditions
    this->checkToken(true);
    CallInfo info(m_token);
    // Create GSoap Context
    GSoapContext g(this->ctx(),false);

    ::srm2__ArrayOfAnyURI targetSURLs;

    // Create Request
    ::srm2__srmStatusOfPutRequestRequest req;
    req.authorizationID = 0;
    req.requestToken    = this->token();

    if(false == this->files.empty()){
        // Fill the list of source Surls
        std::vector<PutFileRequest>::iterator it;
        for(it = this->files.begin(); it != this->files.end(); ++it){
            targetSURLs.urlArray.push_back((*it).surl);
        }
        req.arrayOfTargetSURLs = &(targetSURLs);
    } else {
        req.arrayOfTargetSURLs = 0;
    }

    // Invoke SRM StatusOfPutRequest
    ::srm2__srmStatusOfPutRequestResponse_ rsp;
    this->ctx().beforeCall("srm2__srmStatusOfPutRequest");
    if(SOAP_OK != g.soap().srm2__srmStatusOfPutRequest(&req,rsp)){
        this->ctx().onFailure("srm2__srmStatusOfPutRequest",g.ip().c_str(),&info);
        // Failed to invoke remote procedure: throw the correct exception
        g.handleError("srm2__srmStatusOfPutRequest");
    }
    this->ctx().onSuccess("srm2__srmStatusOfPutRequest",g.ip().c_str(),&info);

    // Update the Request object
    SrmStatus result = update_request
        <PrepareToPut,PutFileRequest,::srm2__srmStatusOfPutRequestResponse,::srm2__TPutRequestFileStatus>
        (*this,rsp.srmStatusOfPutRequestResponse);

    // Update the request status
    this->status = result;

    if (rsp.srmStatusOfPutRequestResponse->remainingTotalRequestTime)
     	m_eWT = *(rsp.srmStatusOfPutRequestResponse->remainingTotalRequestTime);
}

/*
 * done
 *
 * Complete a Request and release the allocated resources.
 */
void PrepareToPut::done() /*throw (LogicError, RemoteError, SrmError)*/{
    // CheckPreconditions
    this->checkToken(true);
    CallInfo info(m_token);

    if(false == this->files.empty()){
        // Empty array: try to call update to get the SURL list
        // Ignore any failure
        try{
           this->update();
        } catch(...){
            // No files retrieved. Send a PutDone with an empty list
        }
    }
    // Create GSoap Context
    GSoapContext g(this->ctx(),false);

    ::srm2__ArrayOfAnyURI arrayOfSURLs;

    // Create Request
    ::srm2__srmPutDoneRequest req;
    req.requestToken    = this->token();
    req.authorizationID = 0;
    req.arrayOfSURLs    = &(arrayOfSURLs); // Release all SURLs
    if(false == this->files.empty()){
        // Fill the list
        std::vector<PutFileRequest>::iterator it;
        for(it = this->files.begin(); it != this->files.end(); ++it){
            // TODO Call PutDone only on succeded SURLs ?
            arrayOfSURLs.urlArray.push_back((*it).surl);
        }
    }

    // Invoke SRM PutDone
    ::srm2__srmPutDoneResponse_ rsp;
    this->ctx().beforeCall("srm2__srmPutDone");
    if(SOAP_OK != g.soap().srm2__srmPutDone(&req,rsp)){
        this->ctx().onFailure("srm2__srmPutDone",g.ip().c_str(),&info);
        // Failed to invoke remote procedure: throw the correct exception
        g.handleError("srm2__srmPutDone");
    }
    this->ctx().onSuccess("srm2__srmPutDone",g.ip().c_str(),&info);

    // Update the Request object
    SrmStatus result = update_request
        <PrepareToPut,PutFileRequest,::srm2__srmPutDoneResponse,::srm2__TSURLReturnStatus>
        (*this,rsp.srmPutDoneResponse);

    // Update the request status
    this->status = result;
}

/*
 * abort
 *
 * Abort a Request and cleanup the allocated resources.
 */
void PrepareToPut::abort() /*throw (LogicError, RemoteError, SrmError)*/{
    // CheckPreconditions
    this->checkToken(true);
    CallInfo info(m_token);

    // Create GSoap Context
    GSoapContext g(this->ctx(),false);

    ::srm2__srmAbortRequestRequest req;
    req.requestToken    = this->token();
    req.authorizationID = 0;

    // Invoke SRM AbortRequest
    ::srm2__srmAbortRequestResponse_ rsp;
    this->ctx().beforeCall("srm2__srmAbortRequest");
    if(SOAP_OK != g.soap().srm2__srmAbortRequest(&req,rsp)){
        this->ctx().onFailure("srm2__srmAbortRequest",g.ip().c_str(),&info);
        // Failed to invoke remote procedure: throw the correct exception
        g.handleError("srm2__srmAbortRequest");
    }
    this->ctx().onSuccess("srm2__srmAbortRequest",g.ip().c_str(),&info);

    // Update the Request object
    SrmStatus result = update_request
        <PrepareToPut,PutFileRequest,::srm2__srmAbortRequestResponse>
        (*this,rsp.srmAbortRequestResponse);

    // Update the request status
    this->status = result;
}
