/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004.
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  GLite Data SRM Util CPP: SrmCopy Class
 *
 *  Authors: Paolo Badino <paolo.badino@cern.ch>
 *           Paolo Tedesco <paolo.tedesco@cern.ch>
 *           Zsolt Molnar <zsolt.molnar@cern.ch - http://www.zsoltmolnar.hu>
 *  Version info: $Id: SrmCopy.cpp,v 1.12 2009/07/30 17:40:34 molnarzs Exp $
 *  Release: $Name:  $
 *
 */
#include "SrmCopy.h"
#include "SrmVersion.h"
#include "GSoapContext.h"
#include "srm-utility.h"
#include "srm-common-copy.h"
#include "srm-common.h"
#include "context-caller.h"
#include "BackoffLogicExecutor.h"

//using namespace glite::data::srm::util::srm2_2;
using namespace ::srm2;

namespace glite {
namespace data  {
namespace srm   {
namespace util  {
namespace srm2_2{

DEFINE_REQUEST_FACTORY(SrmCopy, 2_2);
DEFINE_ASYNCH_REQUEST_FACTORY(SrmCopy, 2_2);

/**
 * name of the TExtraInfo key to use to serialize source space token
 */
const std::string SrmCopy::SourceSpaceTokenKeyName = "SourceSpaceToken";

/**
 * name of the TExtraInfo key to use to serialize copy pin lifetime
 */
const std::string SrmCopy::CopyPinLifetimeKeyName = "CopyPinLifetime";

void SrmCopy::execute()
{
	this->checkToken(false);
	this->status = BackoffLogicExecutor::executeAsync(*this);
}

/**
 * Execute the Request
 */
void SrmCopy::executeOneCall() /* throw (LogicError, RemoteError, SrmError)*/{
    CallInfo info;
    this->m_token.clear();

    if(true == files.empty()){
        throw InvalidArgumentException("empty files array");
    }
    // Create GSoap Context
    GSoapContext g(this->ctx(),true);

    //srm2__TRetentionPolicyInfo *targetFileRetentionPolicyInfo;      /* optional element of type srm2:TRetentionPolicyInfo */

    ::srm2__TOverwriteMode          overwriteOption;
    ::srm2__TFileStorageType        fileStorageType;
    ::srm2__TAccessLatency          accessLatency;
    ::srm2__TRetentionPolicyInfo    retentionPolicyInfo;
    ::srm2__ArrayOfTCopyFileRequest arrayOfFileRequests;
    // sourceStorageSystemInfo and related keys
    ::srm2__ArrayOfTExtraInfo       sourceStorageSystemInfo;
    // targetStorageSystemInfo and related keys
    ::srm2__ArrayOfTExtraInfo       targetStorageSystemInfo;

    // Create Request
    ::srm2__srmCopyRequest req;
    req.authorizationID         = 0;
    req.userRequestDescription  = 0;
    req.sourceStorageSystemInfo = 0;
    req.targetStorageSystemInfo = 0;
    if(this->overwriteOption != OverwriteMode::DEFAULT){
        overwriteOption = translate_OverwriteMode(this->overwriteOption);
        req.overwriteOption = &(overwriteOption);
    } else {
        req.overwriteOption = 0;
    }
    if(this->desiredTotalRequestTime != -1){
        req.desiredTotalRequestTime = &(this->desiredTotalRequestTime);
    } else {
        req.desiredTotalRequestTime = 0;
    }
    if(this->desiredTargetSURLLifeTime != -1){
        req.desiredTargetSURLLifeTime = &(this->desiredTargetSURLLifeTime);
    } else {
        req.desiredTargetSURLLifeTime = 0;
    }
    if(this->targetFileStorageType != FileStorageType::DEFAULT){
        fileStorageType = translate_FileStorageType(this->targetFileStorageType);
        req.targetFileStorageType = &(fileStorageType);
    } else {
        // For WLCG, default FileStorageType is PERMANENT
        // req.desiredFileStorageType = 0;
        fileStorageType = translate_FileStorageType(FileStorageType::PERMANENT);
        req.targetFileStorageType = &(fileStorageType);
    }
    if(false == targetSpaceToken.empty()){
        req.targetSpaceToken = &(this->targetSpaceToken);
    } else {
        req.targetSpaceToken = 0;
    }
    // Retention Policy is a mandatory filed in RetentionPolicyInfo
    if(this->targetFileRetentionPolicyInfo.retentionPolicy != RetentionPolicy::DEFAULT){
        req.targetFileRetentionPolicyInfo = &(retentionPolicyInfo);
        // Translate RetentionPolicy
        req.targetFileRetentionPolicyInfo->retentionPolicy = translate_RetentionPolicy(this->targetFileRetentionPolicyInfo.retentionPolicy);
        if(this->targetFileRetentionPolicyInfo.accessLatency != AccessLatency::DEFAULT){
            // Translate AccessLatency
            accessLatency = translate_AccessLatency(this->targetFileRetentionPolicyInfo.accessLatency);
            req.targetFileRetentionPolicyInfo->accessLatency = &(accessLatency);
        } else {
            req.targetFileRetentionPolicyInfo->accessLatency = 0;
        }
    } else {
        req.targetFileRetentionPolicyInfo = 0;
    }

    // sourceSpaceToken
    if (!this->sourceSpaceToken.empty()){
        req.sourceStorageSystemInfo = &sourceStorageSystemInfo;
        ::srm2__TExtraInfo * source_space_token = soap_new_srm2__TExtraInfo(g.soap().soap,-1);
        source_space_token->key = SourceSpaceTokenKeyName;
        source_space_token->value = soap_new_std__string(g.soap().soap,-1);
        *(source_space_token->value) = sourceSpaceToken;
        req.sourceStorageSystemInfo->extraInfoArray.push_back(source_space_token);
    }
    // copyPinLifetime
    if (-1 != this->copyPinLifetime){
        req.targetStorageSystemInfo = &targetStorageSystemInfo;
        ::srm2__TExtraInfo * copy_pin_lifetime = soap_new_srm2__TExtraInfo(g.soap().soap,-1);
        copy_pin_lifetime->key = CopyPinLifetimeKeyName;
        std::stringstream tmp;
        tmp << copyPinLifetime;
        copy_pin_lifetime->value = soap_new_std__string(g.soap().soap,-1);
        *(copy_pin_lifetime->value) = tmp.str().c_str();
        req.targetStorageSystemInfo->extraInfoArray.push_back(copy_pin_lifetime);
    }
    req.arrayOfFileRequests = &(arrayOfFileRequests);
    // For each GetFileRequest
    ::srm2__TDirOption dir_option;
    dir_option.isSourceADirectory = false;
    dir_option.allLevelRecursive  = 0;
    dir_option.numOfLevels        = 0;
    std::vector<CopyFileRequest>::iterator it;
    for(it = this->files.begin(); it != this->files.end(); ++it){
        info.add("sourceSurl",(*it).sourceSurl);
        info.add("targetSurl",(*it).targetSurl);
        // Add srm2__TCopyFileRequest
        ::srm2__TCopyFileRequest * file_request = ::soap_new_srm2__TCopyFileRequest(g.soap().soap,-1);
        file_request->sourceSURL = (*it).sourceSurl;
        file_request->targetSURL = (*it).targetSurl;
        file_request->dirOption  = &(dir_option);
        req.arrayOfFileRequests->requestArray.push_back(file_request);
    }

    // Invoke SRM Copy
    ::srm2__srmCopyResponse_ rsp;
    this->ctx().beforeCall("srm2__srmCopy");
    if(SOAP_OK != g.soap().srm2__srmCopy(&req,rsp)){
        this->ctx().onFailure("srm2__srmCopy",g.ip().c_str(),&info);
        // Failed to invoke remote procedure: throw the correct exception
        g.handleError("srm2__srmCopy");
    }
    CallOnSuccess call(this->ctx(),"srm2__srmCopy",g.ip().c_str(),&info);

    // Update the Request object
    SrmStatus result = update_request
        <SrmCopy,CopyFileRequest,::srm2__srmCopyResponse,::srm2__TCopyRequestFileStatus>
        (*this,rsp.srmCopyResponse);

    // Update the request status
    this->status = result;
    info.token(m_token);

    if (rsp.srmCopyResponse->remainingTotalRequestTime)
    	m_eWT = *(rsp.srmCopyResponse->remainingTotalRequestTime);
}

/*
 * update
 *
 * Update the status of the Request.
 */
void SrmCopy::update() /*throw (LogicError, RemoteError, SrmError)*/{
    // CheckPreconditions
    this->checkToken(true);
    CallInfo info(m_token);

    // Create GSoap Context
    GSoapContext g(this->ctx(),false);

    ::srm2__ArrayOfAnyURI sourceSURLs;
    ::srm2__ArrayOfAnyURI targetSURLs;

    // Create Request
    ::srm2__srmStatusOfCopyRequestRequest req;
    req.authorizationID = 0;
    req.requestToken    = this->token();

    if(false == this->files.empty()){
        // Fill the list of source Surls
        std::vector<CopyFileRequest>::iterator it;
        for(it = this->files.begin(); it != this->files.end(); ++it){
            sourceSURLs.urlArray.push_back((*it).sourceSurl);
            targetSURLs.urlArray.push_back((*it).targetSurl);
        }
        req.arrayOfSourceSURLs = &(sourceSURLs);
        req.arrayOfTargetSURLs = &(targetSURLs);
    } else {
        req.arrayOfSourceSURLs = 0;
        req.arrayOfTargetSURLs = 0;
    }

    // Invoke SRM StatusOfCopyRequest
    ::srm2__srmStatusOfCopyRequestResponse_ rsp;
    this->ctx().beforeCall("srm2__srmStatusOfCopyRequest");
    if(SOAP_OK != g.soap().srm2__srmStatusOfCopyRequest(&req,rsp)){
        this->ctx().onFailure("srm2__srmStatusOfCopyRequest",g.ip().c_str(),&info);
        // Failed to invoke remote procedure: throw the correct exception
        g.handleError("srm2__srmStatusOfCopyRequest");
    }
    this->ctx().onSuccess("srm2__srmStatusOfCopyRequest",g.ip().c_str(),&info);

    // Update the Request object
    SrmStatus result = update_request
        <SrmCopy,CopyFileRequest,::srm2__srmStatusOfCopyRequestResponse,::srm2__TCopyRequestFileStatus>
        (*this,rsp.srmStatusOfCopyRequestResponse);

    // Update the request status
    this->status = result;

    if (rsp.srmStatusOfCopyRequestResponse->remainingTotalRequestTime)
    	m_eWT = *(rsp.srmStatusOfCopyRequestResponse->remainingTotalRequestTime);
}

/*
 * done
 *
 * Complete a Request and release the allocated resources.
 */
void SrmCopy::done() /*throw (LogicError, RemoteError, SrmError)*/{
    // Nothing To do
}

/*
 * abort
 *
 * Abort a Request and cleanup the allocated resources.
 */
void SrmCopy::abort() /*throw (LogicError, RemoteError, SrmError)*/{
    // CheckPreconditions
    this->checkToken(true);
    CallInfo info(m_token);

    // Create GSoap Context
    GSoapContext g(this->ctx(),false);

    ::srm2__srmAbortRequestRequest req;
    req.requestToken    = this->token();
    req.authorizationID = 0;

    // Invoke SRM AbortRequest
    ::srm2__srmAbortRequestResponse_ rsp;
    this->ctx().beforeCall("srm2__srmAbortRequest");
    if(SOAP_OK != g.soap().srm2__srmAbortRequest(&req,rsp)){
        this->ctx().onFailure("srm2__srmAbortRequest",g.ip().c_str(),&info);
        // Failed to invoke remote procedure: throw the correct exception
        g.handleError("srm2__srmAbortRequest");
    }
    this->ctx().onSuccess("srm2__srmAbortRequest",g.ip().c_str(),&info);

    // Update the Request object
    SrmStatus result = update_request
        <SrmCopy,CopyFileRequest,::srm2__srmAbortRequestResponse>
        (*this,rsp.srmAbortRequestResponse);

    // Update the request status
    this->status = result;
}

} // End namespace srm2_2
} // End namespace util
} // End namespace srm
} // End namespace data
} // End namespace glite
