/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004.
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  GLite Data SRM Util CPP: SrmLs Class
 *
 *  Authors: Paolo Tedesco <paolo.tedesco@cern.ch>,
 *           Zsolt Molnar <zsolt.molnar@cern.ch - http://www.zsoltmolnar.hu>
 *  Version info: $Id: SrmLs.cpp,v 1.6 2009/07/30 17:40:34 molnarzs Exp $
 *  Release: $Name:  $
 *
 */
#include "SrmLs.h"
#include "SrmVersion.h"
#include "GSoapContext.h"
#include "srm-utility.h"
#include "srm-common.h"
#include "srm-common-ls.h"
#include "context-caller.h"
#include "BackoffLogicExecutor.h"

using namespace glite::data::srm::util::srm2_2;
using namespace ::srm2;

DEFINE_REQUEST_FACTORY(SrmLs, 2_2);
DEFINE_ASYNCH_REQUEST_FACTORY(SrmLs, 2_2);

void SrmLs::execute()
{
	this->status = BackoffLogicExecutor::executeAsync(*this);
}

/**
 * _executeOneCall
 * Execute the Request
 */
void SrmLs::executeOneCall() /* throw (LogicError, RemoteError, SrmError)*/{
    SrmStatus result;
    CallInfo info;
    GSoapContext g(this->ctx(),false);

    ::srm2__TFileStorageType requestFileStorageType;

    // Create Request
    ::srm2__srmLsRequest req;
    req.authorizationID   = 0;

    // Set SURL array
    ::srm2__ArrayOfAnyURI arrayOfSURLs;
    std::vector<std::string>::iterator it;
    for(it = this->surls.begin(); it != this->surls.end(); ++it){
        info.add("surl",*it);
        arrayOfSURLs.urlArray.push_back(*it);
    }
    req.arrayOfSURLs = &arrayOfSURLs;

    ::srm2__ArrayOfTExtraInfo requestStorageSystemInfo;
    if (this->storageSystemInfo.empty()){
        req.storageSystemInfo = 0;
    } else {
        ExtraInfo::const_iterator eit = this->storageSystemInfo.begin();
        for (;eit != this->storageSystemInfo.end();++eit){
            srm2__TExtraInfo * extraInfo = new srm2__TExtraInfo();
            extraInfo->key = eit->first;
            if (!eit->second.empty()){
                extraInfo->value = new std::string(eit->second);
            } else {
                extraInfo->value = 0;
            }
            requestStorageSystemInfo.extraInfoArray.push_back(extraInfo);
        }
        req.storageSystemInfo = &requestStorageSystemInfo;
    }

    req.fullDetailedList  = &(this->fullDetailedList);
    req.allLevelRecursive = &(this->allLevelRecursive);
    req.numOfLevels       = &(this->numOfLevels);
    if (-1 == this->offset){
        req.offset = 0;
    } else {
        req.offset = &(this->offset);
    }
    if (-1 == this->count){
        req.count = 0;
    } else {
        req.count = &(this->count);
    }

    if(this->fileStorageType != FileStorageType::DEFAULT){
        requestFileStorageType = translate_FileStorageType(this->fileStorageType);
        req.fileStorageType = &(requestFileStorageType);
    } else {
        req.fileStorageType = 0;
    }

    // Invoke srm2__srmLs
    // ---------------------------------------------------------------------
    ::srm2__srmLsResponse_ rsp;
    this->ctx().beforeCall("srm2__srmLs");
    if(SOAP_OK != g.soap().srm2__srmLs(&req, rsp)){
        this->ctx().onFailure("srm2__srmLs",g.ip().c_str(),&info);
        // Failed to invoke remote call: throw the correct exception
        g.handleError("SrmLs");
    }
    CallOnSuccess call(this->ctx(),"srm2__srmLs",g.ip().c_str(),&info);

    if((0 == rsp.srmLsResponse) ||
       (0 == rsp.srmLsResponse->returnStatus)){
        // Error: No staqtus returned by the SRM
        throw SrmBadResponse("no ReturnStatus returned by the SRM");
    }

    // Get file details
    const std::vector< ::srm2__TMetaDataPathDetail* > * file_array = 0;
    if(0 != rsp.srmLsResponse->details){
        file_array = &(rsp.srmLsResponse->details->pathDetailArray);
    }

    // Get Result Code
    this->status = update_request
        <SrmLs,MetaDataPathDetail2,::srm2__srmLsResponse,::srm2__TMetaDataPathDetail>
        (*this,rsp.srmLsResponse,file_array);

    if (rsp.srmLsResponse->requestToken) {
    	m_token = *(rsp.srmLsResponse->requestToken);
    	info.token(m_token);
    }
}

/**
 * update
 * Update the status of the Request.
 */
void SrmLs::update() /*throw (LogicError, RemoteError, SrmError)*/{
    // CheckPreconditions
    this->checkToken(true);
    CallInfo info(m_token);
    // Create GSoap Context
    GSoapContext g(this->ctx(),false);

    // create request
    ::srm2__srmStatusOfLsRequestRequest req;
    req.authorizationID = 0;
    req.requestToken    = this->token();
    req.offset          = 0;
    req.count           = 0;

    // Invoke SRM StatusOfGetRequest
    ::srm2__srmStatusOfLsRequestResponse_ rsp;
    this->ctx().beforeCall("srm2__srmStatusOfLsRequest");
    if(SOAP_OK != g.soap().srm2__srmStatusOfLsRequest(&req,rsp)){
        this->ctx().onFailure("srm2__srmStatusOfLsRequest",g.ip().c_str(),&info);
        // Failed to invoke remote procedure: throw the correct exception
        g.handleError("srm2__srmStatusOfLsRequest");
    }
    this->ctx().onSuccess("srm2__srmStatusOfLsRequest",g.ip().c_str(),&info);

    // Update the Request object
    this->status = update_request
        <SrmLs,MetaDataPathDetail2,::srm2__srmStatusOfLsRequestResponse,::srm2__TMetaDataPathDetail>
        (*this,rsp.srmStatusOfLsRequestResponse);
}

/**
 * done
 * Complete a Request and release the allocated resources.
 */
void SrmLs::done() /*throw (LogicError, RemoteError, SrmError)*/{
    // do nothing
}

/**
 * abort
 * Abort a Request and cleanup the allocated resources.
 */
void SrmLs::abort() /*throw (LogicError, RemoteError, SrmError)*/{
    // CheckPreconditions
    this->checkToken(true);
    CallInfo info(m_token);
    // Create GSoap Context
    GSoapContext g(this->ctx(),false);

    ::srm2__srmAbortRequestRequest req;
    req.requestToken    = this->token();
    req.authorizationID = 0;

    // Invoke SRM AbortRequest
    ::srm2__srmAbortRequestResponse_ rsp;
    this->ctx().beforeCall("srm2__srmAbortRequest");
    if(SOAP_OK != g.soap().srm2__srmAbortRequest(&req,rsp)){
        this->ctx().onFailure("srm2__srmAbortRequest",g.ip().c_str(),&info);
        // Failed to invoke remote procedure: throw the correct exception
        g.handleError("srm2__srmAbortRequest");
    }
    this->ctx().onSuccess("srm2__srmAbortRequest",g.ip().c_str(),&info);

    // Update the Request object
    this->status = update_request
        <SrmLs,MetaDataPathDetail2,::srm2__srmAbortRequestResponse>
        (*this,rsp.srmAbortRequestResponse);
}
