/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004. 
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.  
 *
 *  Licensed under the Apache License, Version 2.0 (the "License"); 
 *  you may not use this file except in compliance with the License. 
 *  You may obtain a copy of the License at 
 *
 *      http://www.apache.org/licenses/LICENSE-2.0 
 *
 *  Unless required by applicable law or agreed to in writing, software 
 *  distributed under the License is distributed on an "AS IS" BASIS, 
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 *  See the License for the specific language governing permissions and 
 *  limitations under the License.
 *
 *  GLite Data SRM Util CPP: srm common functions (get space metadata specialization)
 *
 *  Authors: Paolo Tedesco <paolo.tedesco@cern.ch>
 *  Version info: $Id: srm-common-getspacemetadata.h,v 1.2 2008/03/14 15:19:23 ptedesco Exp $
 *  Release: $Name:  $
 *
 */
#ifndef GLITE_DATA_SRM_UTIL_CPP_2_2_SRM_COMMON_GETSPACEMETADATA_H_
#define GLITE_DATA_SRM_UTIL_CPP_2_2_SRM_COMMON_GETSPACEMETADATA_H_

#include "GetSpaceMetaData.h"
#include "srm-utility.h"
#include "srm-common-defs.h"

namespace glite {
namespace data  {
namespace srm   {
namespace util  {
namespace srm2_2{

/**
 * get the output details array from the response. 
 * Introduced to allow handling GetSpaceMetaData (which does not have an array
 * named "files")
 */
template<>
std::vector<MetaDataSpace>& get_array(GetSpaceMetaData& request){
    return request.spaceDetails;
}
    
/**
 * Check Preconditions
 */
template<>
inline void preconditions(
    const GetSpaceMetaData&                             request, 
    const ::srm2::srm2__srmGetSpaceMetaDataResponse *   response)  /*throw (LogicError, SrmBadResponse)*/{
    // Nothing To Check
}

/**
 * Check Postconditions
 */
template<>
inline void postconditions(
    GetSpaceMetaData&                                   request, 
    SrmStatus&                                          request_status, 
    const ::srm2::srm2__srmGetSpaceMetaDataResponse *   response)  /*throw (LogicError, SrmBadResponse)*/{
    // Verify Postconditions
    switch(request_status.code){
        case SrmStatus::SRM_SUCCESS:
        case SrmStatus::SRM_PARTIAL_SUCCESS:
        case SrmStatus::SRM_AUTHENTICATION_FAILURE:
        case SrmStatus::SRM_AUTHORIZATION_FAILURE:
        case SrmStatus::SRM_TOO_MANY_RESULTS:
        case SrmStatus::SRM_INVALID_REQUEST:
        case SrmStatus::SRM_INTERNAL_ERROR:
        case SrmStatus::SRM_FAILURE:
        case SrmStatus::SRM_NOT_SUPPORTED:
            // OK
        break;
        default:
            // Invalid result returned by the SRM: set it to generic 
            // failure
            failure_for_invalid_result(request_status);
        break;
    }
}

/**
 * Get File State From Request state
 */
template<>
inline void file_from_request_status(
    const GetSpaceMetaData&                             request, 
    const SrmStatus&                                    request_status, 
    MetaDataSpace&                                      file, 
    const ::srm2::srm2__srmGetSpaceMetaDataResponse *   response) {
#if 0    
    // Check the result and get the default file status
    switch(request_status.code){
        case SrmStatus::SRM_SUCCESS:
        case SrmStatus::SRM_FAILURE:
            file.status.code = request_status.code;
        break;
        case SrmStatus::SRM_AUTHENTICATION_FAILURE:
        case SrmStatus::SRM_AUTHORIZATION_FAILURE:
        case SrmStatus::SRM_INVALID_REQUEST:
        case SrmStatus::SRM_INTERNAL_ERROR:
        case SrmStatus::SRM_NOT_SUPPORTED:
        case SrmStatus::SRM_PARTIAL_SUCCESS:
            // Leave the file status undef
            file.status.code = SrmStatus::SRM_UNDEF;
        break;
        default:
            // This should never happen
        break;
    }
#endif    
}

/**
 * Return true if the File structure is related to the given file
 */
template<>
inline bool match_file(
    const MetaDataSpace&                    file, 
    const ::srm2::srm2__TMetaDataSpace *    file_stat) {
#if 0    
    bool result = false;
    if((file_stat != 0) && (file_stat->surl == file.surl)){
        result = true;
    }
    return result;
#else
    return true;    
#endif    
}

/**
 * Update Request Properties
 */    
template<>
inline void update_request_properties(
    GetSpaceMetaData&                                   request,
    const SrmStatus&                                    status, 
    const ::srm2::srm2__srmGetSpaceMetaDataResponse *   response) /*throw (LogicError, SrmBadResponse)*/{
    // Nothing to do
}

/**
 * Update File Properties
 * NOTE : the purpose of this function is to translate an ::srm2 structure into
 * an internal type
 */    
template<>
inline void update_file_properties(
    const GetSpaceMetaData&                             request, 
    const SrmStatus&                                    request_status, 
    MetaDataSpace&                                      space, 
    const ::srm2::srm2__TMetaDataSpace *                space_stat, 
    const ::srm2::srm2__srmGetSpaceMetaDataResponse *   /*response*/) /*throw (LogicError, SrmBadResponse)*/{
    if(0 == space_stat){
        // This should never happen: space_stat has been checked before
    } else {
        // Set the file Status
        if(0 != space_stat->status){
            space.status.code = translate_StatusCode(space_stat->status->statusCode);
            if(0 != space_stat->status->explanation){
                space.status.message = *(space_stat->status->explanation);
            }
            // Check that the status code  is valid
            switch(space.status.code){
                case SrmStatus::SRM_SUCCESS:
                case SrmStatus::SRM_AUTHORIZATION_FAILURE:
                case SrmStatus::SRM_INVALID_REQUEST:
                case SrmStatus::SRM_SPACE_LIFETIME_EXPIRED:
                case SrmStatus::SRM_EXCEED_ALLOCATION:
                case SrmStatus::SRM_FAILURE:
                    // Ok
                break;
                default:
                    // Invalid result returned by the SRM: set it to generic 
                    // failure
                    failure_for_invalid_result(space.status);
                break;
            }
        } else {
            // This should not happen: status is mandatory
            std::stringstream reason;
            reason << "The SRM returned no status for space metadata [" 
                    << space_stat->spaceToken
                   << "] in SrmGetSpaceMetaData. According to the standard, the status is mandatory";
            throw SrmBadResponse(reason.str());
        }
        space.spaceToken = space_stat->spaceToken;
        if (0 != space_stat->retentionPolicyInfo){
            space.retentionPolicyInfo.retentionPolicy = translate_RetentionPolicy(space_stat->retentionPolicyInfo->retentionPolicy);
            if (0 != space_stat->retentionPolicyInfo->accessLatency){
                space.retentionPolicyInfo.accessLatency = translate_AccessLatency(*space_stat->retentionPolicyInfo->accessLatency);
            }
        }
        if (0 != space_stat->owner){
            space.owner = (*space_stat->owner);
        }
        if (0 != space_stat->totalSize){
            space.totalSize = (*space_stat->totalSize);
        }
        if (0 != space_stat->guaranteedSize){
            space.guaranteedSize = (*space_stat->guaranteedSize);
        }
        if (0 != space_stat->unusedSize){
            space.unusedSize = (*space_stat->unusedSize);
        }
        if (0 != space_stat->lifetimeAssigned){
            space.lifetimeAssigned = (*space_stat->lifetimeAssigned);
        }
        if (0 != space_stat->lifetimeLeft){
            space.lifetimeLeft = (*space_stat->lifetimeLeft);
        }
        
    }
}

} // End namespace srm2_2
} // End namespace util
} // End namespace srm
} // End namespace data
} // End namespace glite

#endif // GLITE_DATA_SRM_UTIL_CPP_2_2_SRM_COMMON_GETSPACEMETADATA_H_

