/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004.
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  GLite Data SRM Util CPP: srm common functions
 *
 *  Authors: Paolo Badino <paolo.badino@cern.ch>
 *           Paolo Tedesco <paolo.tedesco@cern.ch>
 *  Version info: $Id: srm-common.h,v 1.14 2009/06/21 17:09:31 molnarzs Exp $
 *  Release: $Name:  $
 *
 */
#ifndef GLITE_DATA_SRM_UTIL_CPP_2_2_SRM_COMMON_H_
#define GLITE_DATA_SRM_UTIL_CPP_2_2_SRM_COMMON_H_

#include <sstream>

#include "srm-utility.h"
#include "srm-common-defs.h"

namespace glite {
namespace data  {
namespace srm   {
namespace util  {
namespace srm2_2{

//------------------------------------------------------------------------------
// NOTE ON TEMPLATE PARAMETERS:
//------------------------------------------------------------------------------
// R -> RequestType: the type of the request (class defined in the srm-util-cpp
// module)
// F -> FileType: the type of the file details in the RequestType class
// S -> SrmResponseType: the ::srm2 type of the response
// A -> ArrayType: the ::srm2 type of the file details array in the response
// These names are actually misleading, and inappropriate in case of (originally
// non foreseen) requests like GetSpaceMetaData (which does not handle file
// details).
// Newly created functions are using this convention, but this should be
// refactored to improve readability.
//------------------------------------------------------------------------------

/**
 * get the output details array from the response.
 * Introduced to allow handling GetSpaceMetaData (which does not have an array
 * named "files")
 */
template<typename RequestType,typename FileType>
std::vector<FileType>& get_array(RequestType& request){
    return request.files;
}

/**
 * Update the state of a Request instance
 */
template<typename R, typename F, typename S, typename A>
SrmStatus update_request(R& request, const S * srm_result, const std::vector<A*> * file_array) /*throw (LogicError, RemoteError, SrmBadResponse, SrmError)*/{

    // Check Preconditions
    if((0 == srm_result) ||
       (0 == srm_result->returnStatus)){
        // Error: No status returned by the SRM
        throw SrmBadResponse("no ReturnStatus returned by the SRM");
    }

    // Check Preconditions
    preconditions<R,S>(request,srm_result);

    // Get Result Code
    SrmStatus status;
    status.code = translate_StatusCode(srm_result->returnStatus->statusCode);
    if(0 != srm_result->returnStatus->explanation){
        status.message = *(srm_result->returnStatus->explanation);
    }

    // The exponential back-off logic is here.


    // Update Request Properties
    update_request_properties<R,S>(request,status,srm_result);
    std::vector<F>& files = get_array<R,F>(request);

    // Check that the number of file states
    if((0 == file_array) || (true == file_array->empty())){
        // Check that the request status is not partial Success
        if(status.code == SrmStatus::SRM_PARTIAL_SUCCESS){
            std::stringstream reason;
            reason << "Invalid response from the SRM. Request status is "\
                      "SRM_PARTIAL_SUCCESS but no object states are returned";
            throw SrmBadResponse(reason.str());
        }
        // No file States return: assume the File State is the same as the
        // Request
        typename std::vector<F>::iterator it;

        for(it = files.begin(); it != files.end(); ++it){
            file_from_request_status<R,F,S>(request, status, (*it),srm_result);
        }
    } else if(false == files.empty()){
        const std::vector<A*>& statusArray =  *(file_array);
        // Check that the size of the fileStatuses is the expected one
        if(statusArray.size() != files.size()){
            // Invalid Response returned by the SRM
            std::stringstream reason;
            reason << "The SRM returned the status of ["
                   << statusArray.size()
                   << "] file(s) while it was expected to return the status for ["
                   << files.size() << "] file(s)";
            throw SrmBadResponse(reason.str());
        }
        // Set the individual file status
        typename std::vector<F>::iterator it;
        typename std::vector<A*>::const_iterator s_it;
        unsigned int missing_count = 0;
        for(it = files.begin(); it != files.end(); ++it){
            // Find the related status
            for(s_it = statusArray.begin(); s_it != statusArray.end(); ++s_it){
                if(true == match_file<F,A>((*it),(*s_it))){
                    // Match found
                    update_file_properties<R,F,A,S>(request,status,(*it),(*s_it), srm_result);
                    break;
                }
            }
            if(s_it == statusArray.end()){
                ++missing_count;
            }
        } // end for each file
        if(0 != missing_count){
            if(missing_count == files.size()){
                // All missing: assume same order
                for(it = files.begin(), s_it = statusArray.begin();
                    it != files.end(); ++it, ++s_it){
                    // Update File property
                    update_file_properties<R,F,A,S>(request,status,(*it),(*s_it), srm_result);
                }
            } else {
                // No status returned by the SRM for that file
                std::stringstream reason;
                reason << "Inconsistent SRM response. No status returned for "\
                          "some of the files or an invalid file name has " \
                          "been returned";
                throw SrmBadResponse(reason.str());
            }
        }
    } else {
        // Empty File List: this should be filled with the result returned by
        // the SRM
        const std::vector<A*>& statusArray = *(file_array);
        // Resize the file array
        files.resize(statusArray.size());
        // For each file status
        typename std::vector<A*>::const_iterator s_it;
        typename std::vector<F>::iterator it = files.begin();
        for(s_it = statusArray.begin() ; s_it != statusArray.end(); ++s_it, ++it){
            // Set the individual file request
            update_file_properties<R,F,A,S>(request,status,(*it),*(s_it), srm_result);
        }
    }

    // Check PostConditions
    postconditions<R,S>(request,status,srm_result);

    return status;
}

/**
 * Update the state of a Request instance
 */
template<typename R, typename F, typename S, typename A>
SrmStatus update_request(R& request, const S * srm_result) /*throw (LogicError, RemoteError, SrmBadResponse, SrmError)*/{

    // For most of the method,  the file array is called in the same way
    const std::vector<A*> * file_array = 0;
    if((0 != srm_result) &&
       (0 != srm_result->arrayOfFileStatuses)){
           file_array = &(srm_result->arrayOfFileStatuses->statusArray);
    }
    return update_request<R,F,S,A>(request,srm_result,file_array);
}

/**
 * Update the state of a Request instance
 */
template<typename R, typename F, typename S>
SrmStatus update_request(R& request, const S * srm_result) /*throw (LogicError, RemoteError, SrmBadResponse, SrmError)*/{

    // Check Preconditions
    if((0 == srm_result) ||
       (0 == srm_result->returnStatus)){
        // Error: No status returned by the SRM
        throw SrmBadResponse("no ReturnStatus returned by the SRM");
    }

    // Check Preconditions
    preconditions<R,S>(request,srm_result);

    // Get Result Code
    SrmStatus status;
    status.code = translate_StatusCode(srm_result->returnStatus->statusCode);
    if(0 != srm_result->returnStatus->explanation){
        status.message = *(srm_result->returnStatus->explanation);
    }

    // Update Request Properties
    update_request_properties<R,S>(request,status,srm_result);

    // No files status is returned: update the files accordingly to the request
    // state
    typename std::vector<F>::iterator it;
    std::vector<F>& files = get_array<R,F>(request);
    for(it = files.begin(); it != files.end(); ++it){
        file_from_request_status<R,F,S>(request, status, (*it),srm_result);
    }

    // Check PostConditions
    postconditions<R,S>(request,status,srm_result);

    return status;
}

} // End namespace srm2_2
} // End namespace util
} // End namespace srm
} // End namespace data
} // End namespace glite

#endif //GLITE_DATA_SRM_UTIL_CPP_2_2_SRM_COMMON_H_
