/*
 *  Copyright (c) Members of the EGEE Collaboration. 2004. 
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.  
 *
 *  Licensed under the Apache License, Version 2.0 (the "License"); 
 *  you may not use this file except in compliance with the License. 
 *  You may obtain a copy of the License at 
 *
 *      http://www.apache.org/licenses/LICENSE-2.0 
 *
 *  Unless required by applicable law or agreed to in writing, software 
 *  distributed under the License is distributed on an "AS IS" BASIS, 
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 *  See the License for the specific language governing permissions and 
 *  limitations under the License.
 *
 *  GLite Data SRM Util CPP: srm 2.2 utilities
 *
 *  Authors: Paolo Badino <paolo.badino@cern.ch>
 *  Version info: $Id: srm-utility.cpp,v 1.7 2006/11/28 12:01:16 badino Exp $
 *  Release: $Name:  $
 *
 */
#include "srm-utility.h"
#include "glite/data/srm/util/SrmExceptions.h"

#include <string>
#include <sstream>
#include <algorithm>

using namespace ::srm2;

namespace glite {
namespace data  {
namespace srm   {
namespace util  {
namespace srm2_2{
       
/**
 * Set the status to generic failure 
 */
void failure_for_invalid_result(SrmStatus& status) /* throw ()*/{
    // Invalid result returned by the SRM: set it to generic 
    // failure
    std::stringstream reason;
    reason << "Invalid Status returned by the SRM ["
           << SrmStatus::toString(status.code) << "]. "
           << "Consider it as generic failure";
    if(false == status.message.empty()){
        reason << ". Original error message was: "
               << status.message;
    }
    status.code = SrmStatus::SRM_FAILURE;
    status.message = reason.str();
}
    
/*
 * check_abort_postconditions
 *
 * Check the validity of the request status after an abort
 */
void check_abort_postconditions(SrmStatus& status) /* throw ()*/{
    // Post-condition check
    switch(status.code){
        case SrmStatus::SRM_SUCCESS:
            // The Abort Request was SUCCESS, but the request status should 
            // now be ABORTED
            status.code = SrmStatus::SRM_ABORTED;
        break;
        case SrmStatus::SRM_ABORTED:
            // This should not be returned by teh SRM, but it's a valid value    
        case SrmStatus::SRM_AUTHENTICATION_FAILURE:
        case SrmStatus::SRM_AUTHORIZATION_FAILURE:
        case SrmStatus::SRM_INVALID_REQUEST:
        case SrmStatus::SRM_INTERNAL_ERROR:
        case SrmStatus::SRM_FAILURE:
        case SrmStatus::SRM_NOT_SUPPORTED:
            // Ok
        break;
        default:
            // Invalid result returned by the SRM: set it to generic 
            // failure
            failure_for_invalid_result(status);
        break;
    }
}

// -----------------------------------------------------------------------------
// Srm 2.2 Enum translation routines
// -----------------------------------------------------------------------------

/*
 * translate_StatusCode
 *
 * Translate SrmStatus Code
 */
SrmStatus::Code translate_StatusCode(::srm2__TStatusCode srm_code) /*throw (BadSrmResponse)*/{
    SrmStatus::Code code = SrmStatus::SRM_FAILURE;
    switch(srm_code){
    case ::srm2__TStatusCode__SRM_USCORESUCCESS:
        code = SrmStatus::SRM_SUCCESS;
	break;
	case ::srm2__TStatusCode__SRM_USCOREFAILURE:
	    code = SrmStatus::SRM_FAILURE;
	break;
	case ::srm2__TStatusCode__SRM_USCOREAUTHENTICATION_USCOREFAILURE:
	    code = SrmStatus::SRM_AUTHENTICATION_FAILURE;
	break;
	case ::srm2__TStatusCode__SRM_USCOREAUTHORIZATION_USCOREFAILURE:
	    code = SrmStatus::SRM_AUTHORIZATION_FAILURE;
	break;
	case ::srm2__TStatusCode__SRM_USCOREINVALID_USCOREREQUEST:
	    code = SrmStatus::SRM_INVALID_REQUEST;
	break;
	case ::srm2__TStatusCode__SRM_USCOREINVALID_USCOREPATH:
	    code = SrmStatus::SRM_INVALID_PATH;
	break;
	case ::srm2__TStatusCode__SRM_USCOREFILE_USCORELIFETIME_USCOREEXPIRED:
	    code = SrmStatus::SRM_FILE_LIFETIME_EXPIRED;
	break;
	case ::srm2__TStatusCode__SRM_USCORESPACE_USCORELIFETIME_USCOREEXPIRED:
	    code = SrmStatus::SRM_SPACE_LIFETIME_EXPIRED;
	break;
	case ::srm2__TStatusCode__SRM_USCOREEXCEED_USCOREALLOCATION:
	    code = SrmStatus::SRM_EXCEED_ALLOCATION;
	break;
	case ::srm2__TStatusCode__SRM_USCORENO_USCOREUSER_USCORESPACE:
	    code = SrmStatus::SRM_NO_USER_SPACE;
	break;
	case ::srm2__TStatusCode__SRM_USCORENO_USCOREFREE_USCORESPACE:
	    code = SrmStatus::SRM_NO_FREE_SPACE;
	break;
	case ::srm2__TStatusCode__SRM_USCOREDUPLICATION_USCOREERROR:
	    code = SrmStatus::SRM_DUPLICATION_ERROR;
	break;
	case ::srm2__TStatusCode__SRM_USCORENON_USCOREEMPTY_USCOREDIRECTORY:
	    code = SrmStatus::SRM_NON_EMPTY_DIRECTORY;
	break;
	case ::srm2__TStatusCode__SRM_USCORETOO_USCOREMANY_USCORERESULTS:
	    code = SrmStatus::SRM_TOO_MANY_RESULTS;
	break;
	case ::srm2__TStatusCode__SRM_USCOREINTERNAL_USCOREERROR:
	    code = SrmStatus::SRM_INTERNAL_ERROR;
	break;
	case ::srm2__TStatusCode__SRM_USCOREFATAL_USCOREINTERNAL_USCOREERROR:
	    code = SrmStatus::SRM_FATAL_INTERNAL_ERROR;
	break;
	case ::srm2__TStatusCode__SRM_USCORENOT_USCORESUPPORTED:
	    code = SrmStatus::SRM_NOT_SUPPORTED;
	break;
	case ::srm2__TStatusCode__SRM_USCOREREQUEST_USCOREQUEUED:
	    code = SrmStatus::SRM_REQUEST_QUEUED;
	break;
	case ::srm2__TStatusCode__SRM_USCOREREQUEST_USCOREINPROGRESS:
	    code = SrmStatus::SRM_REQUEST_INPROGRESS;
	break;
	case ::srm2__TStatusCode__SRM_USCOREREQUEST_USCORESUSPENDED:
	    code = SrmStatus::SRM_REQUEST_SUSPENDED;
	break;
	case ::srm2__TStatusCode__SRM_USCOREABORTED:
	    code = SrmStatus::SRM_ABORTED;
	break;
	case ::srm2__TStatusCode__SRM_USCORERELEASED:
	    code = SrmStatus::SRM_RELEASED;
	break;
	case ::srm2__TStatusCode__SRM_USCOREFILE_USCOREPINNED:
	    code = SrmStatus::SRM_FILE_PINNED;
	break;
	case ::srm2__TStatusCode__SRM_USCOREFILE_USCOREIN_USCORECACHE:
	    code = SrmStatus::SRM_FILE_IN_CACHE;
	break;
	case ::srm2__TStatusCode__SRM_USCORESPACE_USCOREAVAILABLE:
	    code = SrmStatus::SRM_SPACE_AVAILABLE;
	break;
	case ::srm2__TStatusCode__SRM_USCORELOWER_USCORESPACE_USCOREGRANTED:
	    code = SrmStatus::SRM_LOWER_SPACE_GRANTED;
	break;
	case ::srm2__TStatusCode__SRM_USCOREDONE:
	    code = SrmStatus::SRM_DONE;
	break;
	case ::srm2__TStatusCode__SRM_USCOREPARTIAL_USCORESUCCESS:
	    code = SrmStatus::SRM_PARTIAL_SUCCESS;
	break;
	case ::srm2__TStatusCode__SRM_USCOREREQUEST_USCORETIMED_USCOREOUT:
	    code = SrmStatus::SRM_REQUEST_TIMED_OUT;
	break;
	case ::srm2__TStatusCode__SRM_USCORELAST_USCORECOPY:
	    code = SrmStatus::SRM_LAST_COPY;
	break;
	case ::srm2__TStatusCode__SRM_USCOREFILE_USCOREBUSY:
	    code = SrmStatus::SRM_FILE_BUSY;
	break;
	case ::srm2__TStatusCode__SRM_USCOREFILE_USCORELOST:
	    code = SrmStatus::SRM_FILE_LOST;
	break;
	case ::srm2__TStatusCode__SRM_USCOREFILE_USCOREUNAVAILABLE:
	    code = SrmStatus::SRM_FILE_UNAVAILABLE;
	break;
	case ::srm2__TStatusCode__SRM_USCORECUSTOM_USCORESTATUS:
	    code = SrmStatus::SRM_CUSTOM_STATUS;
	break;
	default:
	    std::stringstream reason;
	    reason << "Invalid SrmStatus Code [" << srm_code << "] returned";
	    throw SrmBadResponse(reason.str());
    }    
    return code;
}

/*
 * translate_FileStorageType
 *
 * Translate FileStorageType
 */
::srm2__TFileStorageType translate_FileStorageType(FileStorageType::Value v) /* throw(LogicError) */{
    ::srm2__TFileStorageType result;
    switch(v){
        case FileStorageType::VOLATILE:
            result = ::srm2__TFileStorageType__VOLATILE;
        break;
        case FileStorageType::DURABLE:
            result = ::srm2__TFileStorageType__DURABLE;
        break;
        case FileStorageType::PERMANENT:
            result = ::srm2__TFileStorageType__PERMANENT;
        break;
        case FileStorageType::DEFAULT:
            // Note: This code should never be executed. However I still prefer 
            // to give a default
        
            // According to the standard, the default is Volatile. For 
            // WLCG I prefer to use Permanent
            result = ::srm2__TFileStorageType__PERMANENT;
        break;
        default:
            throw LogicError("Invalid FileStorageType");
        break;
    }
    return result;
}

/*
 * translate_FileStorageType
 *
 * Translate FileStorageType
 */
FileStorageType::Value translate_FileStorageType(::srm2__TFileStorageType v) /* throw(SrmBadResponse) */{
    FileStorageType::Value result = FileStorageType::DEFAULT;
    switch(v){
        case ::srm2__TFileStorageType__VOLATILE:
            result = FileStorageType::VOLATILE;
        break;
        case ::srm2__TFileStorageType__DURABLE:
            result = FileStorageType::DURABLE;
        break;
        case ::srm2__TFileStorageType__PERMANENT:
            result = FileStorageType::PERMANENT;
        break;
        default:{
            // Note: This code should never be executed.
            std::stringstream reason;
            reason << "Invalid FileStorageType [" << v << "] returned";
            throw SrmBadResponse(reason.str());
        } break;
    }
    return result;
}

/*
 * translate_RetentionPolicy
 *
 * Translate RetentionPolicy
 */
::srm2__TRetentionPolicy translate_RetentionPolicy(RetentionPolicy::Value v) /* throw(LogicError) */{
    ::srm2__TRetentionPolicy result;
    switch(v){
        case RetentionPolicy::REPLICA:
            result = ::srm2__TRetentionPolicy__REPLICA;
        break;
        case RetentionPolicy::OUTPUT:
            result = ::srm2__TRetentionPolicy__OUTPUT;
        break;
        case RetentionPolicy::CUSTODIAL:
            result = ::srm2__TRetentionPolicy__CUSTODIAL;
        break;
        case RetentionPolicy::DEFAULT:
            // Note: This code should never be executed. However I still prefer 
            // to give a default
        
            // The standard doesn't specify the default. For WLCG, I prefer 
            // to use Custodial
            result = ::srm2__TRetentionPolicy__CUSTODIAL;
        break;
        default:
            throw LogicError("Invalid RetentionPolicy");
        break;
    }
    return result;
}

/*
 * translate_RetentionPolicy
 *
 * Translate RetentionPolicy
 */
RetentionPolicy::Value translate_RetentionPolicy(::srm2__TRetentionPolicy v) /* throw(LogicError) */{
    RetentionPolicy::Value result = RetentionPolicy::DEFAULT;
    switch(v){
        case ::srm2__TRetentionPolicy__REPLICA:
            result = RetentionPolicy::REPLICA;
        break;
        case ::srm2__TRetentionPolicy__OUTPUT:
            result = RetentionPolicy::OUTPUT;
        break;
        case ::srm2__TRetentionPolicy__CUSTODIAL:
            result = RetentionPolicy::CUSTODIAL;
        break;
        default:{
            // Note: This code should never be executed.
            std::stringstream reason;
            reason << "Invalid RetentionPolicy [" << v << "] returned";
            throw SrmBadResponse(reason.str());
        } break;
    }
    return result;
}

/*
 * translate_AccessLatency
 *
 * Translate AccessLatency
 */
::srm2__TAccessLatency translate_AccessLatency(AccessLatency::Value v) /* throw(LogicError) */{
    ::srm2__TAccessLatency result;
    switch(v){
        case AccessLatency::ONLINE:
            result = ::srm2__TAccessLatency__ONLINE;
        break;
        case AccessLatency::NEARLINE:
            result = ::srm2__TAccessLatency__NEARLINE;
        break;
        case AccessLatency::DEFAULT:
            // Note: This code should never be executed. However I still prefer 
            // to give a default
        
            // For WLCG, I prefer to use Nearline
            result = ::srm2__TAccessLatency__NEARLINE;
        break;
        default:
            throw LogicError("Invalid AccessLatency");
        break;
    }
    return result;
}

/*
 * translate_AccessLatency
 *
 * Translate AccessLatency
 */
AccessLatency::Value translate_AccessLatency(::srm2__TAccessLatency v) /* throw(LogicError) */{
    AccessLatency::Value result = AccessLatency::DEFAULT;
    switch(v){
        case ::srm2__TAccessLatency__ONLINE:
            result = AccessLatency::ONLINE;
        break;
        case ::srm2__TAccessLatency__NEARLINE:
            result = AccessLatency::NEARLINE;
        break;
        default:{
            // Note: This code should never be executed.
            std::stringstream reason;
            reason << "Invalid AccessLatency [" << v << "] returned";
            throw SrmBadResponse(reason.str());
        } break;

    }
    return result;
}

/*
 * translate_AccessPattern
 *
 * Translate AccessPattern
 */
::srm2__TAccessPattern translate_AccessPattern(AccessPattern::Value v) /* throw(LogicError) */{
    ::srm2__TAccessPattern result;
    switch(v){
        case AccessPattern::TRANSFER_MODE:
            result = ::srm2__TAccessPattern__TRANSFER_USCOREMODE;
        break;
        case AccessPattern::PROCESSING_MODE:
            result = ::srm2__TAccessPattern__PROCESSING_USCOREMODE;
        break;
        case AccessPattern::UNDEF:
        default:
            throw LogicError("Invalid AccessPattern");
        break;
    }
    return result;
}

/*
 * translate_ConnectionType
 *
 * Translate ConnectionType
 */
::srm2__TConnectionType translate_ConnectionType(ConnectionType::Value v) /* throw(LogicError) */{
    ::srm2__TConnectionType result;
    switch(v){
        case ConnectionType::WAN:
            result = ::srm2__TConnectionType__WAN;
        break;
        case ConnectionType::LAN:
            result = ::srm2__TConnectionType__LAN;
        break;
        case ConnectionType::UNDEF:
        default:
            throw LogicError("Invalid ConnectionType");
        break;
    }
    return result;
}

/*
 * translate_OverwriteMode
 *
 * Translate OverwriteMode
 */
::srm2__TOverwriteMode translate_OverwriteMode(OverwriteMode::Value v) /* throw(LogicError) */{
    ::srm2__TOverwriteMode result;
    switch(v){
        case OverwriteMode::NEVER:
            result = ::srm2__TOverwriteMode__NEVER;
        break;
        case OverwriteMode::ALWAYS:
            result = ::srm2__TOverwriteMode__ALWAYS;
        break;
        case OverwriteMode::WHEN_FILES_ARE_DIFFERENT:
            result = ::srm2__TOverwriteMode__WHEN_USCOREFILES_USCOREARE_USCOREDIFFERENT;
        break;
        case OverwriteMode::DEFAULT:
            // Note: This code should never be executed. However I still prefer 
            // to give a default
        
            // For WLCG, I prefer to use Never
            result = ::srm2__TOverwriteMode__NEVER;
        break;
        default:
            throw LogicError("Invalid ConnectionType");
        break;
    }
    return result;
}

/*
 * translate_FileLocality
 *
 * Translate FileLocality
 */
FileLocality::Value translate_FileLocality(::srm2__TFileLocality v) /* throw(SrmBadResponse) */{
    FileLocality::Value result = FileLocality::UNDEF;
    switch(v){
        case ::srm2__TFileLocality__ONLINE:
            result = FileLocality::ONLINE;
        break;
        case ::srm2__TFileLocality__NEARLINE:
            result = FileLocality::NEARLINE;
        break;
        case ::srm2__TFileLocality__ONLINE_USCOREAND_USCORENEARLINE:
            result = FileLocality::ONLINE_AND_NEARLINE;
        break;
        case ::srm2__TFileLocality__LOST:
            result = FileLocality::LOST;
        break;
        case ::srm2__TFileLocality__NONE:
            result = FileLocality::NONE;
        break;
        case ::srm2__TFileLocality__UNAVAILABLE:
            result = FileLocality::UNAVAILABLE;
        break;
        default:{
            // Note: This code should never be executed.
            std::stringstream reason;
            reason << "Invalid FileLocality [" << v << "] returned";
            throw SrmBadResponse(reason.str());
        } break;
    }
    return result;
}

/*
 * translate_FileType
 *
 * Translate FileType
 */
FileType::Value translate_FileType(::srm2__TFileType v) /* throw(SrmBadResponse) */{
    FileType::Value result = FileType::UNDEF;
    switch(v){
        case ::srm2__TFileType__FILE_:
            result = FileType::FILE;
        break;
        case ::srm2__TFileType__DIRECTORY:
            result = FileType::DIRECTORY;
        break;
        case ::srm2__TFileType__LINK:
            result = FileType::LINK;
        break;
        default:{
            // Note: This code should never be executed.
            std::stringstream reason;
            reason << "Invalid FileType [" << v << "] returned";
            throw SrmBadResponse(reason.str());
        } break;
    }
    return result;
}

/*
 * translate_PermissionMode
 *
 * Translate PermissionMode
 */
PermissionMode::Value translate_PermissionMode(::srm2__TPermissionMode v) /* throw(SrmBadResponse) */{
    PermissionMode::Value result = PermissionMode::NONE;
    switch(v){
        case ::srm2__TPermissionMode__NONE:
            result = PermissionMode::NONE;
        break;
        case ::srm2__TPermissionMode__X:
            result = PermissionMode::X;
        break;
        case ::srm2__TPermissionMode__W:
            result = PermissionMode::W;
        break;
        case ::srm2__TPermissionMode__WX:
            result = PermissionMode::WX;
        break;
        case ::srm2__TPermissionMode__R:
            result = PermissionMode::R;
        break;
        case ::srm2__TPermissionMode__RX:
            result = PermissionMode::RX;
        break;
        case ::srm2__TPermissionMode__RW:
            result = PermissionMode::RW;
        break;
        case ::srm2__TPermissionMode__RWX:
            result = PermissionMode::RWX;
        break;
        default:{
            // Note: This code should never be executed.
            std::stringstream reason;
            reason << "Invalid PermissionMode [" << v << "] returned";
            throw SrmBadResponse(reason.str());
        } break;
    }
    return result;
}

} // End namespace srm2_2
} // End namespace util
} // End namespace srm
} // End namespace data
} // End namespace glite
