/*
 *  Copyright (c) Members of the EGEE Collaboration. 2004. 
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.  
 *
 *  Licensed under the Apache License, Version 2.0 (the "License"); 
 *  you may not use this file except in compliance with the License. 
 *  You may obtain a copy of the License at 
 *
 *      http://www.apache.org/licenses/LICENSE-2.0 
 *
 *  Unless required by applicable law or agreed to in writing, software 
 *  distributed under the License is distributed on an "AS IS" BASIS, 
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 *  See the License for the specific language governing permissions and 
 *  limitations under the License.
 *
 *  gLite SRM Util - Performs a request to get a file(s) from the SRM
 *
 *  Authors: Paolo Badino <Paolo.Badino@cern.ch>
 *  Version info: $Id: copy.cpp,v 1.1 2007/02/01 11:00:21 badino Exp $
 *  Release: $Name:  $
 *
 */

#include "util-main.h"
#include "copy-common.h"

using namespace glite::data::srm::util;

//
// Global variables
//
static const unsigned int DEFAULT_SLEEP_TIME = 10;

// Usage message
const char * const tool_usage = "SOURCE_SURL DEST_SURL [SOURCE_SURL DEST_SURL]";

// Help message
const char * const tool_help = 
	"\t-b\t\tRun the get in blocking mode\n";


// get specific command line options
const char * const tool_options = "b";

namespace {
// run teh get in blocking mode
bool g_blocking = false;
}

/*
 * tool_doit
 * 
 * Run the Get
 */
int tool_doit(Context::Ptr ctx, int argc, char *argv[]){
	
    if(argc <= 0){
        error("No Surl specified");
        print_usage();
        return EXIT_FAILURE;
    }
    if(0 != argc % 2){
        error("Surls are not specified as a pair");
        print_usage();
        return EXIT_FAILURE;
    }        
    
    // Create the Get Requestt Object
    boost::shared_ptr<SrmCopy> request(SrmCopy::create(ctx));
    
    // Set Request Parameters
    int num_pairs = argc/2;
    request->files.resize(num_pairs);
    for(int i = 0; i < num_pairs; ++i){
        request->files[i].sourceSurl = argv[2 * i];
        request->files[i].targetSurl = argv[(2 * i) + 1];
    }
    try{
        // Issue a SrmCopy Request
        request->execute();
    } catch(const SrmError& exc){
        error("SrmCopy Failed (SrmError): %s",exc.what());
        return EXIT_FAILURE;
    } catch(const RemoteError& exc){
        error("SrmCopy Failed (RemoteError): %s",exc.what());
        return EXIT_FAILURE;
    } catch(const SecurityError& exc){
        error("SrmCopy Failed (SecurityError): %s",exc.what());
        return EXIT_FAILURE;
    } catch(const SrmBadResponse& exc){
        error("SrmCopy Failed (SrmBadResponse): %s",exc.what());
        return EXIT_FAILURE;
    } catch(const LogicError& exc){
        error("SrmCopy Failed (LogicError): %s",exc.what());
        return EXIT_FAILURE;
    }
    
    // Print request
    print_request(*request);
    
    if(true == g_blocking){
        
        while((request->status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
              (request->status.code == SrmStatus::SRM_REQUEST_INPROGRESS) ||
              (request->status.code == SrmStatus::SRM_PARTIAL_SUCCESS)){
            // take the lower extimatedWaitTime 
            unsigned int sleep_time = DEFAULT_SLEEP_TIME;
            std::vector<CopyFileRequest>::const_iterator it;
            for(it = request->files.begin(); it != request->files.end(); ++it){
                if((((*it).status.code == SrmStatus::SRM_REQUEST_QUEUED) || 
                    ((*it).status.code == SrmStatus::SRM_REQUEST_INPROGRESS)) &&
                   ((*it).extimatedWaitTime > 0) && 
                   ((*it).extimatedWaitTime < sleep_time)){
                    sleep_time = (*it).extimatedWaitTime;
                   }                       
            }
            // Sleep for a while
            sleep(sleep_time);
            
            // print time
            time_t current;
            time(&current);
            char str[32];
            str[31] = 0;
            struct tm * tm = gmtime(&current);
            strftime(str, 31 , "%Y-%m-%d %H:%M:%S +00:00",tm);
            info("\n[%s] updating request status", str);
            
            // Pool the Srm until the Get is completed
            request->update();

            // Print request
            print_request(*request);
        }
        
        // Copy is done: complete the request
        try{   
            request->done();
        } catch(const SrmError& exc){
            error("SrmCopy Done Failed (SrmError): %s",exc.what());
            return EXIT_FAILURE;
        } catch(const RemoteError& exc){
            error("SrmCopy Done Failed (RemoteError): %s",exc.what());
            return EXIT_FAILURE;
        } catch(const SecurityError& exc){
            error("SrmCopy Done Failed (SecurityError): %s",exc.what());
            return EXIT_FAILURE;
        } catch(const SrmBadResponse& exc){
            error("SrmCopy Done Failed (SrmBadResponse): %s",exc.what());
            return EXIT_FAILURE;
        } catch(const LogicError& exc){
            error("SrmCopy Done Failed (LogicError): %s",exc.what());
            return EXIT_FAILURE;
        }    
    }
    
	return EXIT_SUCCESS;
}

/*
 * tool_parse_cmdline
 * 
 * Command line parsing
 */
int tool_parse_cmdline(int c, char *opt_arg) {
	switch(c)
	{
		case 'b':
			g_blocking = true;
			return 0;
		default:
			break;
	}
	return -1;
}
