/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004.
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  gLite SRM Util - srmLs common functions
 *
 *  Authors: Paolo Tedesco <paolo.tedesco@cern.ch>
 *  Version info: $Id: ls-common.cpp,v 1.4 2009/04/08 16:25:28 molnarzs Exp $
 *  Release: $Name:  $
 *
 */
#include "ls-common.h"
#include <sstream>

namespace {

static const std::string INDENT_STRING  = "|   ";
static const std::string INDENT_CHILD   = "|---";

/**
 * print_metadatapathdetail_list
 * @param request [IN] the SrmLs request
 * @param details [IN] the MetaDataPathDetail2 objects to print
 * @param indent [IN] a string for indentation
 */
void print_metadatapathdetail_list(
    const SrmLs&                            request,
    const std::vector<MetaDataPathDetail2>&  details,
    const std::string&                      indent){

    if (0 == details.size()){
        return;
    }

    char str[32];
    str[31] = 0;
    struct tm * tm = 0;
    // Display file Statuses
    std::vector<MetaDataPathDetail2>::const_iterator it;
    for(it = details.begin(); it != details.end(); ++it){
        const MetaDataPathDetail2& path = (*it);
        info("%s%s : %s (%ld)",indent.c_str(),FileType::toString(path.fileType),path.surl.c_str(),path.size);
        if (SrmStatus::SRM_SUCCESS != path.status.code){
            info("%s  Status           : %s",indent.c_str(),SrmStatus::toString(path.status.code));
            if(false == path.status.message.empty()){
                info("%s  Message          : %s",indent.c_str(),path.status.message.c_str());
            }
        }
        if (request.fullDetailedList){
            tm = gmtime(&path.createdAtTime);
            strftime(str, 31 , "%Y-%m-%d %H:%M:%S +00:00",tm);
            info("%s  Creation Time    : %s",indent.c_str(),(path.createdAtTime != (time_t)-1)?str:"");
            tm = gmtime(&path.lastModificationTime);
            strftime(str, 31 , "%Y-%m-%d %H:%M:%S +00:00",tm);
            info("%s  Last Modif. Time : %s",indent.c_str(),(path.lastModificationTime != (time_t)-1)?str:"");
            info("%s  Storage Type     : %s",indent.c_str(),FileStorageType::toString(path.fileStorageType));
            info("%s  Retention Policy Info :",indent.c_str());
            info("%s    Retention Policy : %s",indent.c_str(),RetentionPolicy::toString(path.retentionPolicyInfo.retentionPolicy));
            info("%s    Access Latency   : %s",indent.c_str(),AccessLatency::toString(path.retentionPolicyInfo.accessLatency));
            info("%s  Locality         : %s",indent.c_str(),FileLocality::toString(path.fileLocality));
            std::stringstream space_tokens;
            StringArray::const_iterator sit;
            for(sit = path.spaceTokens.begin();sit != path.spaceTokens.end();++sit){
                if(sit != path.spaceTokens.begin()){
                    space_tokens << "; ";
                }
                space_tokens << "[" << (*sit) << "]" ;
            }
            info("%s  Space Tokens     : %s",indent.c_str(),space_tokens.str().c_str());
            info("%s  Type             : %s",indent.c_str(),FileType::toString(path.fileType));
            if(path.lifetimeAssigned > 0){
                info("%s  Lifetime Assigned: %d",indent.c_str(),path.lifetimeAssigned);
            } else {
                info("%s  Lifetime Assigned: ",indent.c_str());
            }
            if(path.lifetimeLeft > 0){
                info("%s  Lifetime Left    : %d",indent.c_str(),path.lifetimeLeft);
            }else {
                info("%s  Lifetime Left    : ",indent.c_str());
            }
            info("%s  User  Permission : %-3s [%s]",indent.c_str(),PermissionMode::toString(path.ownerPermission.mode),path.ownerPermission.userID.c_str());
            info("%s  Group Permission : %-3s [%s]",indent.c_str(),PermissionMode::toString(path.groupPermission.mode),path.groupPermission.groupID.c_str());
            info("%s  Other Permission : %-3s",indent.c_str(),PermissionMode::toString(path.otherPermission));
            info("%s  CheckSum Type    : %s",indent.c_str(),path.checkSumType.c_str());
            info("%s  CheckSum Value   : %s",indent.c_str(),path.checkSumValue.c_str());
        }

        if (0 != path.arrayOfSubPaths.size()){
            std::string new_indent = (0 == indent.length()) ? INDENT_CHILD : INDENT_STRING + indent;
            print_metadatapathdetail_list(request,path.arrayOfSubPaths,new_indent);
        }

    }

}

} // namespace

void SrmLsTraits::PrintRequest(const SrmLs& request){

    // print the request token
    info(" ");
    info("Token  : %s", request.token().c_str());
    info("Status : %s", SrmStatus::toString(request.status.code));
    if(false == request.status.message.empty()){
        info("Message: %s", request.status.message.c_str());
    }

    info(" ");
    print_metadatapathdetail_list(request,request.details,"");
}
