/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004.
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  gLite SRM Util - Common functionality of srmLs-like requests
 *
 *  Authors: Zsolt Molnar <zsolt.molnar@cern.ch>
 *  Version info: $Id: ls_CommonOperations.h,v 1.1 2009/04/08 16:25:28 molnarzs Exp $
 *  Release: $Name:  $
 *
 */
#ifndef GLITE_DATA_SRM_UTIL_CPP_LS_COMMON_OPERATIONS_H_
#define GLITE_DATA_SRM_UTIL_CPP_LS_COMMON_OPERATIONS_H_

#include "glite/data/srm/util/Context.h"
#include "glite/data/srm/util/SrmExceptions.h"
#include "util-main.h"

using namespace glite::data::srm::util;

struct CommandLineParameters
{
	CommandLineParameters()
		: g_fullDetailedList(false),
		  g_allLevelRecursive(false),
		  g_numOfLevels(1),
		  g_offset(-1),
		  g_count(-1),
	      g_blocking(false)
	{}

	bool g_fullDetailedList;
	bool g_allLevelRecursive;
	int g_numOfLevels;
	int g_offset;
	int g_count;
	bool g_blocking;
};

namespace {
	static const unsigned int DEFAULT_SLEEP_TIME = 5;
}

extern CommandLineParameters g_parameters;

bool check_parameters_for_token(int argc, char **argv);
bool check_parameters_for_surl(int argc, char **argv);
int tool_parse_cmdline(int, char*);

template<class R>
struct SrmTraits
{
	typedef R RequestType;
};

template<class R> struct NoSetup
{
	void setup_request(R&, const CommandLineParameters&)
	{
		// EMPTY
	}
};

template<class R> struct Setup {};

template<class R> struct RequestWithToken
{
	RequestWithToken(int argc, char **argv)
		: _parametersOK(check_parameters_for_token(argc, argv))
	{
		if (_parametersOK)
			_token = argv[0];
	}

	R* createRequest(Context::Ptr ctx)
	{
		return R::create(ctx, _token.c_str());
	}

	bool checkParameters()
	{
		return _parametersOK;
	}

protected:
	bool _parametersOK;
	std::string _token;
};

template<class R> struct RequestWithoutToken
{
	RequestWithoutToken(int argc, char **argv)
		: _parametersOK(check_parameters_for_surl(argc, argv))
	{}

	R* createRequest(Context::Ptr ctx)
	{
		return R::create(ctx);
	}

	bool checkParameters()
	{
		return _parametersOK;
	}

private:
	bool _parametersOK;
};

template<class R>
struct SrmAbortTraits : public NoSetup<R>, public RequestWithToken<R>
{
	SrmAbortTraits(int argc, char **argv)
		: RequestWithToken<R>(argc, argv) {}

	static void Op(R& r)
	{
		r.abort();
	}
};

template<class R>
struct SrmExecuteTraits : public Setup<R>, public RequestWithoutToken<R>
{
	SrmExecuteTraits(int argc, char **argv)
		: RequestWithoutToken<R>(argc, argv),
		  Setup<R>(argc, argv)
    {}

	static void Op(R& r)
	{
		r.execute();
	}
};

template<class R>
struct SrmStatusTraits : public NoSetup<R>, public RequestWithToken<R>
{
	SrmStatusTraits(int argc, char **argv)
		: RequestWithToken<R>(argc, argv)
    {}

	static void Op(R& r)
	{
		r.update();
	}
};

template<class DoitTratis>
int tool_doit_general(Context::Ptr ctx, int argc, char **argv)
{
	DoitTratis doit(argc, argv);

	if (!doit.checkParameters())
		return EXIT_FAILURE;

    // Create the Get Request Object
    boost::shared_ptr<typename DoitTratis::RequestType> request(doit.createRequest(ctx));
    doit.setup_request(*request, g_parameters);
    // Set Request Parameters
    try{
    	DoitTratis::Op(*request);
    } catch(const SrmError& exc){
        error("Get status of %s Failed (SrmError): %s", DoitTratis::Label(), exc.what());
        return EXIT_FAILURE;
    } catch(const RemoteError& exc){
        error("Get status of %s Failed (RemoteError): %s", DoitTratis::Label(), exc.what());
        return EXIT_FAILURE;
    } catch(const SecurityError& exc){
        error("Get status of %s Failed (SecurityError): %s", DoitTratis::Label(), exc.what());
        return EXIT_FAILURE;
    } catch(const SrmBadResponse& exc){
        error("Get status of %s Failed (SrmBadResponse): %s", DoitTratis::Label(), exc.what());
        return EXIT_FAILURE;
    } catch(const LogicError& exc){
        error("Get status of %s Failed (LogicError): %s", DoitTratis::Label, exc.what());
        return EXIT_FAILURE;
    }
    // Print request
    doit.PrintRequest(*request);

    if (g_parameters.g_blocking)
    {
        while((request->status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
              (request->status.code == SrmStatus::SRM_REQUEST_INPROGRESS) ||
              (request->status.code == SrmStatus::SRM_PARTIAL_SUCCESS))
        {
            // Sleep for a while
            sleep(DEFAULT_SLEEP_TIME);
            // print time
            time_t current;
            time(&current);
            char str[32];
            str[31] = 0;
            struct tm * tm = gmtime(&current);
            strftime(str, 31 , "%Y-%m-%d %H:%M:%S +00:00",tm);
            info("\n[%s] updating request status", str);
            // Pool the Srm until the Get is completed
            request->update();
            // Print request
            doit.PrintRequest(*request);
        }
    }

	return EXIT_SUCCESS;
}

#endif // #ifndef GLITE_DATA_SRM_UTIL_CPP_LS_COMMON_OPERATIONS_H_
