/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004. 
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.  
 *
 *  Licensed under the Apache License, Version 2.0 (the "License"); 
 *  you may not use this file except in compliance with the License. 
 *  You may obtain a copy of the License at 
 *
 *      http://www.apache.org/licenses/LICENSE-2.0 
 *
 *  Unless required by applicable law or agreed to in writing, software 
 *  distributed under the License is distributed on an "AS IS" BASIS, 
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 *  See the License for the specific language governing permissions and 
 *  limitations under the License.
 *
 *  gLite SRM Util Main - Main function for the tools
 *
 *  Authors: Paolo Badino <Paolo.Badino@cern.ch>, Paolo Tedesco <paolo.tedesco@cern.ch>
 *  Version info: $Id: util-main.cpp,v 1.3 2008/11/20 11:44:47 ptedesco Exp $
 *  Release: $Name:  $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "util-main.h"
#include "glite/data/glite-util.h"
#include "glite/data/srm/util/Context.h"

#include <getopt.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <stdarg.h>

using namespace glite::data::srm::util;

// -----------------------------------------------------------------------------
// Global variables
// -----------------------------------------------------------------------------

static const char * const SRM_TYPE = "SRM";

// Command line options common to all the tools
static const char *common_options = "hqs:I:vV";

// Description of the common options
static const char *common_help =
    "\t-h\t\tPrint this help text and exit.\n"
    "\t-q\t\tQuiet operation.\n"
    "\t-s URL\t\tUse the SRM service at the specified URL.\n"
    "\t-I version\tThe SRM interface version to be used.\n"
    "\t-v\t\tBe more verbose.\n"
    "\t-V\t\tPrint the version number and exit.\n";

// Flag to turn on verbose output
int verbose_flag;

// Flag to request quiet operation
int quiet_flag;

// program name
static const char * prog_name = "";

/*
 * print program usage
 */
void print_usage() {
    printf("\nUsage: %s [options]", prog_name);
    if (tool_usage){
        printf(" %s\n", tool_usage);
    } else {
        printf("\n");
    }
    printf("Available options:\n%s", common_help);
    if (tool_help){
        printf("%s", tool_help);
    }
}

// -----------------------------------------------------------------------------
// Logging functions
// -----------------------------------------------------------------------------

/*
 * debug
 */
void debug(const char *fmt, ...){
    va_list ap;

    if (!verbose_flag){
        return;
    }

    va_start(ap, fmt);
    vfprintf(stdout,fmt, ap);
    va_end(ap);
    fflush(stdout);
    fprintf(stdout,"\n");
}


/*
 * info
 */
void info(const char *fmt, ...){
    va_list ap;

    va_start(ap, fmt);
    vfprintf(stdout,fmt, ap);
    va_end(ap);
    fflush(stdout);
    fprintf(stdout,"\n");
}

/*
 * error
 */
void error(const char *fmt, ...){
    va_list ap;

    va_start(ap, fmt);
    vfprintf(stderr,fmt, ap);
    va_end(ap);
    fflush(stderr);
    fprintf(stderr,"\n");
}

class ToolsContext : public Context{
protected:
    
    /**
     * before_call
     */
    virtual void before_call(
        const char * method){
        debug("# * Invoking %s",method);
    }
    
    /**
     * on_success
     */
    virtual void on_success(
        const char * method,
        const char * address,
        const CallInfo * info){
        debug("#   ---> %s success [%s]",method,address);
        if (info){
            if (!info->token().empty()){
                debug("#        request token = [%s]",info->token().c_str());
            }
            CallInfo::AdditionalInfos::const_iterator it;
            const CallInfo::AdditionalInfos& infos = info->additionalInfos();
            for (it = infos.begin();it != infos.end();++it){
                debug("#        %s : [%s]",it->key.c_str(),it->value.c_str());
            }
        }
    }
    
    /**
     * on_failure
     */
    virtual void on_failure(
        const char * method,
        const char * address,
        const CallInfo * info){
        debug("#   ---> %s failed [%s]",method,address);
        if (info){
            if (!info->token().empty()){
                debug("#        request token = [%s]",info->token().c_str());
            }
            CallInfo::AdditionalInfos::const_iterator it;
            const CallInfo::AdditionalInfos& infos = info->additionalInfos();
            for (it = infos.begin();it != infos.end();++it){
                debug("#        %s : [%s]",it->key.c_str(),it->value.c_str());
            }
        }
    }
    
    /**
     * on_warning
     */
    virtual void on_warning(
        const char * scope,
        const char * description,
        const char * address){
        debug("#   warning : %s - %s [%s]",scope,description,address);
    }

};

/**
 * main
 * Main routine for the gLite SRM tools
 */
int main(int argc, char *argv[]) {
    char *options;
    int optlen, ret, c;

    // Determine the program name we were called as
    prog_name = strrchr(argv[0], '/');
    if (!prog_name){
        prog_name = argv[0];
    } else {
        prog_name++;
    }
    
    // Build the option string for getopt()
    optlen = strlen(common_options) + 1;
    if (tool_options){
        optlen += strlen(tool_options);
    }
    options = new char[optlen];
    if (!options){
        error("Out of memory");
        exit(EXIT_FAILURE);
    }

    strcpy(options, common_options);
    if (tool_options){
        strcat(options, tool_options);
    }

    // Location of the srm service
    const char * service_location = 0;
    
    // the srm interface version
    const char * service_version = 0;
    
    // Parse the command line options
    while ((c = getopt(argc, argv, options)) != -1) switch (c)
    {
        case 'h':
            printf("\nGlite Data SRM Util CPP - Version %s by (C) EGEE\n", TOOLS_VERSION);
            printf("\nDisclaimer: This tool is intended to be used internally by the gLite File Transfer Service and therefore no user support will be provided\n");
            print_usage();
            exit(EXIT_SUCCESS);
        case 'q':
            quiet_flag++;
            verbose_flag = 0;
            break;
        case 's':
            service_location = optarg;
            break;
        case 'I':
            service_version = optarg;
            break;
        case 'v':
            verbose_flag++;
            quiet_flag = 0;
            break;
        case 'V':
            printf("%s\n", PACKAGE_VERSION);
            exit(EXIT_SUCCESS);
        case ':':
            error("Option argument is missing");
            print_usage();
            exit(EXIT_FAILURE);
        case '?':
            error("Unknown command line option");
            print_usage();
            exit(EXIT_FAILURE);
        default:
            ret = tool_parse_cmdline(c, optarg);
            if (ret){
                error("Illegal command line arguments");
                print_usage();
                exit(EXIT_FAILURE);
            }
            break;
    }

    // Drop the processed options from the argument list
    delete [] options;
    argc -= optind;
    argv += optind;

    char * service_name = 0;
    char * endpoint = 0;
    char * version = 0;
    char *error_msg = 0;
    Context::Ptr ctx(new ToolsContext());

    bool resolve_using_sd = true;
    // if an URL and the version have been passed, don't  resolve the name
    if((0 != service_location) && (strlen(service_location) > 0)){
       if(0 == strncmp(service_location,"http",4)){
           // location is an enpoit
           endpoint = strdup(service_location);
           // check version
           if((0 != service_version)  && (strlen(service_version) > 0 )){
               resolve_using_sd = false;
           } else {
               if(0 != strstr(endpoint,"srm/managerv1")){
                   service_version = "1.1";
                   resolve_using_sd = false;
               } else if(0 != strstr(endpoint,"srm/managerv2")){
                   service_version = "2.2";
                   resolve_using_sd = false;
               }
           }
       }
    }

    if(true == resolve_using_sd){
        // Get the SRM Service name
        service_name = glite_discover_service_by_version(SRM_TYPE, service_location, service_version, &error_msg);
        if(0 == service_name){
            error("Failed to determine the SRM endpoint: %s",error_msg);
            ret = EXIT_FAILURE;
            goto cleanup;
        }                                                                              
        // Get the SRM endpoit and version
        endpoint = glite_discover_endpoint(SRM_TYPE, service_name, &error_msg);
        if (0 == endpoint){
            error("Failed to determine the SRM endpoint: %s", error_msg);
            ret = EXIT_FAILURE;
            goto cleanup;
        }
        if(0 == service_version){
            version = glite_discover_version(SRM_TYPE, service_name, &error_msg);
            if (0 == version){
                error("Failed to determine the SRM endpoint: %s", error_msg);
                ret = EXIT_FAILURE;
                goto cleanup;
            }
            service_version = version;
        }
    }
    debug("# Using endpoint %s", endpoint);
    debug("# Using SRM version %s", service_version);

    // Create an Srm Context
    ctx->endpoint = endpoint;
    ctx->version  = service_version;
    
    // Perform the operation 
    ret = tool_doit(ctx, argc, argv);

cleanup:
    // Free memory
    if(0 != error_msg){
        free(error_msg);
    }
    if(0 != service_name){
        free(service_name);
    }
    if(0 != endpoint){
        free(endpoint);
    }
    if(0 != version){
        free(version);
    }
    return ret;
}
