/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004.
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  GLite Data SRM Util CPP: SRM 1.1 Test Case
 *
 *  Authors: Paolo Badino <paolo.badino@cern.ch>,
 *           Paolo Tedesco <paolo.tedesco@cern.ch>
 *           Zsolt Molnar <zsolt.molnar@cern.ch>
 *  Version info: $Id: Srm11_TestCase.cpp,v 1.20 2009/04/08 16:28:20 molnarzs Exp $
 *  Release: $Name:  $
 *
 */
#include "glite/data/srm/util/SrmExceptions.h"
#include "glite/data/srm/util/PrepareToPut.h"
#include "glite/data/srm/util/BringOnline.h"
#include "glite/data/srm/util/PrepareToGet.h"
#include "glite/data/srm/util/SrmCopy.h"
#include "glite/data/srm/util/SrmPing.h"
#include "glite/data/srm/util/SrmRm.h"
#include "glite/data/srm/util/SrmStat.h"
#include "glite/data/srm/util/SrmMkDir.h"
#include "glite/data/srm/util/SrmRmDir.h"
#include "glite/data/srm/util/AbortFiles.h"
#include "glite/data/srm/util/GetSpaceTokens.h"
#include "glite/data/srm/util/GetSpaceMetaData.h"
#include "glite/data/agents/UserProxyEnv.h"

#include <memory>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <boost/scoped_ptr.hpp>
#include <sstream>

#include "Srm11_TestCase.h"

/*
 * The SRM 1.1 endpoint
 */
extern std::string g_srm_11_endpoint;

/*
 * setUp
 *
 * set-up test case
 */
void Srm11_TestCase::setUp(){

    CPPUNIT_ASSERT_MESSAGE("Invalid Endpoint", g_srm_11_endpoint.empty() == false);

    // Create Context
    m_ctx.reset(new test_helpers::TestContext());

    // Create an Srm Context
    m_ctx->endpoint = g_srm_11_endpoint;
    m_ctx->version  = "1.1";
}

/*
 * tearDown
 *
 * tear-down test case
 */
void Srm11_TestCase::tearDown(){
    // Delete Context
    m_ctx.reset();
}

void Srm11_TestCase::_check_supported(const std::string& version)
{
    // Create Context
    Context::Ptr ctx(new test_helpers::TestContext());
    ctx->endpoint = g_srm_11_endpoint;
    ctx->version  = version;
    // Check supported methods
    CHECK_SUPPORTED(SrmPing,            ctx,true);
    CHECK_SUPPORTED(PrepareToPut,       ctx,true);
    CHECK_SUPPORTED(PrepareToGet,       ctx,true);
    CHECK_SUPPORTED(BringOnline,        ctx,true);
    CHECK_SUPPORTED(SrmStat,            ctx,true);
    CHECK_SUPPORTED(SrmCopy,            ctx,true);
    CHECK_SUPPORTED(SrmRm,              ctx,true);
    CHECK_SUPPORTED(GetSpaceTokens,     ctx,false);
    CHECK_SUPPORTED(GetSpaceMetaData,   ctx,false);
    CHECK_SUPPORTED(SrmMkDir,           ctx,false);
    CHECK_SUPPORTED(SrmRmDir,           ctx,true);
}

/*
 * Test Supported Method
 */
void Srm11_TestCase::testSupported(){
    // Do it with different values of the version (test tag stripping)
    _check_supported("1.1.0");
    _check_supported("1.1");
    _check_supported("1.1.x");
}

/*
 * Test Ping Method
 */
void Srm11_TestCase::testPing(){

    // Create Ping Request
    boost::shared_ptr<SrmPing> request(SrmPing::create(m_ctx));

    // Send Ping Request
    SrmStatus result = request->execute();

    // Check Return Value
    CHECKSRMSTATUS("Invalid Result Code",result.code,SrmStatus::SRM_SUCCESS);
    CPPUNIT_ASSERT_MESSAGE("Invalid Version Returned", request->versionInfo == "v1.1");

}

/*
 * Test Ping Method
 */
void Srm11_TestCase::testPing_SecurityError(){

    // Setup a wrong value for X509_USER_PROXY
    glite::data::agents::UserProxyEnv env("/tmp/x509_fake");

    // Create Ping Request
    boost::shared_ptr<SrmPing> request(SrmPing::create(m_ctx));

    // Send Ping Request
    try{
        SrmStatus result = request->execute();
        CPPUNIT_FAIL("This method should throw a SecurityError");
    } catch (const SecurityError& exc){
        // OK
    }

}

/*
 * Test Ping Method
 */
void Srm11_TestCase::testPing_RemoteError(){

    // Use an existing endpoint
    m_ctx->endpoint = "httpg://fake-srm-service-hopefully-it-will-not-exist.cern.cern.ch:8443/srm/managerv1";

    // Create Ping Request
    boost::shared_ptr<SrmPing> request(SrmPing::create(m_ctx));

    // Send Ping Request
    try{
        SrmStatus result = request->execute();
        CPPUNIT_FAIL("This method should throw a RemoteError: invalid host");
    } catch (const RemoteError& exc){
        // OK
    }

    // Use an existing endpoint
    m_ctx->endpoint = "httpg://localhost:88053/srm/managerv1";

    // Create Ping Request
    request.reset(SrmPing::create(m_ctx));

    // Send Ping Request
    try{
        SrmStatus result = request->execute();
        CPPUNIT_FAIL("This method should throw a RemoteError: invalid port");
    } catch (const RemoteError& exc){
        // OK
    }

}

/*
 * Test Put Method
 */
void Srm11_TestCase::testPut(){

    // Create Ping Request
    boost::shared_ptr<PrepareToPut> request(PrepareToPut::create(m_ctx));

    // Set Request Parameters
    request->files.resize(1);
    request->files[0].surl = "srm://localhost/srm/managerv1?SFN=/test/srm1/file001.txt";
    request->transferParameters.protocol = "file";

    // Send Ping Request
    request->execute();

    // Check Return Value
    bool is_request_failed = true;
    if((request->status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
       (request->status.code == SrmStatus::SRM_REQUEST_INPROGRESS) ||
       (request->status.code == SrmStatus::SRM_SUCCESS)){
           is_request_failed = false;
    }
    CPPUNIT_ASSERT_MESSAGE("Invalid Result Code on [execute]", is_request_failed == false);

    while((request->status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
          (request->status.code == SrmStatus::SRM_REQUEST_INPROGRESS)){
        // Sleep for a while
        //sleep(1);

        // Keep polling for the status
        request->update();

        // Check Return Value
        if((request->status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
           (request->status.code == SrmStatus::SRM_REQUEST_INPROGRESS) ||
           (request->status.code == SrmStatus::SRM_SUCCESS)){
               is_request_failed = false;
        }
        CPPUNIT_ASSERT_MESSAGE("Invalid Result Code on [update]", is_request_failed == false);
    }
    CHECKSRMSTATUS("Result Code should now be SUCCESS",request->status.code,SrmStatus::SRM_SUCCESS);

    if(request->status.code == SrmStatus::SRM_SUCCESS){
        // Check that the file properties are set correctly
        CHECKSRMSTATUS("Invalid File Status",request->files[0].status.code,SrmStatus::SRM_SPACE_AVAILABLE);
        CPPUNIT_ASSERT_MESSAGE("No turl retrieved", request->files[0].turl.empty() == false);
    }

    // Open the file
    if(0 == strncmp(request->files[0].turl.c_str(),"file://",7)){
        std::string path = request->files[0].turl.substr(7);
        int fd = open(path.c_str(),O_RDWR);
        CPPUNIT_ASSERT_MESSAGE("Cannot Open file", fd != -1);
        write(fd,SAMPLE_TEXT,strlen(SAMPLE_TEXT));
        close(fd);
    } else {
       CPPUNIT_FAIL("Invalid protocol for TURL: Not [file]");
    }

    // Complete the request
    request->done();
    // Check the Request status
    CHECKSRMSTATUS("Invalid Result Code on [DONE]",request->status.code,SrmStatus::SRM_SUCCESS);
    // Check the File status
    CHECKSRMSTATUS("Invalid File Status",request->files[0].status.code,SrmStatus::SRM_SUCCESS);

}

/*
 * Test Put Method, put should never return
 */
void Srm11_TestCase::testPut_forever_pending(){

    // Create Ping Request
    boost::shared_ptr<PrepareToPut> request(PrepareToPut::create(m_ctx));

    // Set Request Parameters
    request->files.resize(1);
    request->files[0].surl = "srm://localhost/srm/managerv1?SFN=/test/srm1/this-one-should-stay-pending_001.txt";
    request->transferParameters.protocol = "file";

    // Send Ping Request
    request->execute();

    // Check Return Value
    bool is_request_pending = false;
    if((request->status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
       (request->status.code == SrmStatus::SRM_REQUEST_INPROGRESS)){
           is_request_pending = true;
    }
    CPPUNIT_ASSERT_MESSAGE("Invalid Result Code on [execute]", is_request_pending == true);

    int count = 0;
    do{
        // Sleep for a while
        //sleep(1);

        // Keep polling for the status
        request->update();

        // Check Return Value
        if(!((request->status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
             (request->status.code == SrmStatus::SRM_REQUEST_INPROGRESS))){
            is_request_pending = false;
        }
        ++count;
    }while((count <= 3) && (true == is_request_pending));

    CPPUNIT_ASSERT_MESSAGE("Invalid Result Code on [update]", is_request_pending == true);

    // Abort the request
    request->abort();
    // Check the Request status
    CHECKSRMSTATUS("Invalid Result Code on [ABORT]",request->status.code,SrmStatus::SRM_ABORTED);
    // Check the File status
    CHECKSRMSTATUS("Invalid File Status",request->files[0].status.code,SrmStatus::SRM_ABORTED);

}

/**
 * testPutError
 * check that a put request fails with the expected result
 */
void Srm11_TestCase::putErrorUtil(
    const std::string& 		surl,
    const SrmStatus::Code& 	expected){

    // Create Ping Request
    boost::shared_ptr<PrepareToPut> request(PrepareToPut::create(m_ctx));

    // Set Request Parameters
    request->files.resize(1);
    request->files[0].surl = surl;
    request->transferParameters.protocol = "file";

    // Send Ping Request
    request->execute();

    // Check Return Value
    bool is_request_failed = true;
    if((request->status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
       (request->status.code == SrmStatus::SRM_REQUEST_INPROGRESS) ||
       (request->status.code == SrmStatus::SRM_SUCCESS)){
           is_request_failed = false;
    }
    CPPUNIT_ASSERT_MESSAGE("Invalid Result Code on [execute]", is_request_failed == false);

    while((request->status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
          (request->status.code == SrmStatus::SRM_REQUEST_INPROGRESS)){
        // Sleep for a while
        //sleep(1);

        // Keep polling for the status
        request->update();

        // Check Return Value
        if((request->status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
           (request->status.code == SrmStatus::SRM_REQUEST_INPROGRESS) ||
           (request->status.code == SrmStatus::SRM_SUCCESS)){
               is_request_failed = false;
        }
        CPPUNIT_ASSERT_MESSAGE("Invalid Result Code on [update]", is_request_failed == false);
    }
    CHECKSRMSTATUS("Result Code should now be FAILURE",request->status.code,SrmStatus::SRM_FAILURE);

    if(request->status.code == SrmStatus::SRM_FAILURE){
        // Check that the file properties are set correctly
        check_status("Invalid File Status", request->files[0].status.code,expected);
    }
    // Abort the request
    request->abort();
    // Check the Request status
    CHECKSRMSTATUS("Invalid Result Code on [ABORT]",request->status.code,SrmStatus::SRM_ABORTED);

}

/**
 * Test Put Method in case file exists
 */
void Srm11_TestCase::testPut_duplication_error(){
    putErrorUtil(
        "srm://localhost/srm/managerv1?SFN=/test/srm1/this-one-returns-Failed-file exists, cannot write-on-put.txt",
        SrmStatus::SRM_DUPLICATION_ERROR);
}

/**
 * Test Put Method for no free space
 */
void Srm11_TestCase::testPut_no_free_space(){
    putErrorUtil(
        "srm://localhost/srm/managerv1?SFN=/test/srm1/this-one-returns-Failed-no space left on device-on-put.txt",
        SrmStatus::SRM_NO_FREE_SPACE);
}

/*
 * Test Put Method, for a file busy
 */
void Srm11_TestCase::testPut_file_busy(){

    // Create Ping Request
    boost::shared_ptr<PrepareToPut> request(PrepareToPut::create(m_ctx));

    // Set Request Parameters
    request->files.resize(1);
    request->files[0].surl = "srm://localhost/srm/managerv1?SFN=/test/srm1/this-one-is-busy_001.txt";
    request->transferParameters.protocol = "file";

    // Send Ping Request
    request->execute();

    // Check Return Value
    bool is_request_pending = false;
    if((request->status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
       (request->status.code == SrmStatus::SRM_REQUEST_INPROGRESS)){
           is_request_pending = true;
    }
    CPPUNIT_ASSERT_MESSAGE("Invalid Result Code on [execute]", is_request_pending == true);

    do{
        // Sleep for a while
        //sleep(1);

        // Keep polling for the status
        request->update();

        // Check Return Value
        if(!((request->status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
             (request->status.code == SrmStatus::SRM_REQUEST_INPROGRESS))){
            is_request_pending = false;
        }
    } while(true == is_request_pending);

    CPPUNIT_ASSERT_MESSAGE("Request still pending [update]", is_request_pending == false);

    // Check result
    CHECKSRMSTATUS("Invalid Result Code [update]",request->status.code,SrmStatus::SRM_FAILURE);
    CHECKSRMSTATUS("Invalid File Status [update]",request->files[0].status.code,SrmStatus::SRM_FILE_BUSY);

    // Abort the request
    request->abort();
    // Check the Request status
    CHECKSRMSTATUS("Invalid Result Code on [ABORT]",request->status.code,SrmStatus::SRM_ABORTED);

}

/*
 * Test Get Method
 */
void Srm11_TestCase::testGet(){

    // Create Ping Request
    boost::shared_ptr<PrepareToGet> request(PrepareToGet::create(m_ctx));

    // Set Request Parameters
    request->files.resize(1);
    request->files[0].surl = "srm://localhost/srm/managerv1?SFN=/test/srm1/file001.txt";
    request->transferParameters.protocol = "file";

    // Send Ping Request
    request->execute();

    // Check Return Value
    bool is_request_failed = true;
    if((request->status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
        (request->status.code == SrmStatus::SRM_REQUEST_INPROGRESS) ||
        (request->status.code == SrmStatus::SRM_SUCCESS)){
        is_request_failed = false;
        }
        CPPUNIT_ASSERT_MESSAGE("Invalid Result Code on [execute]", is_request_failed == false);

        while((request->status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
               (request->status.code == SrmStatus::SRM_REQUEST_INPROGRESS)){
        // Sleep for a while
            //sleep(1);

        // Keep polling for the status
            request->update();

        // Check Return Value
            if((request->status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
                (request->status.code == SrmStatus::SRM_REQUEST_INPROGRESS) ||
                (request->status.code == SrmStatus::SRM_SUCCESS)){
                is_request_failed = false;
                }
                CPPUNIT_ASSERT_MESSAGE("Invalid Result Code on [update]", is_request_failed == false);
               }
               CHECKSRMSTATUS("Result Code should now be SUCCESS",request->status.code,SrmStatus::SRM_SUCCESS);

               if(request->status.code == SrmStatus::SRM_SUCCESS){
        // Check that the file properties are set correctly
                   CHECKSRMSTATUS("Invalid File Status",request->files[0].status.code,SrmStatus::SRM_FILE_PINNED);
                   CPPUNIT_ASSERT_MESSAGE("No turl retrieved", request->files[0].turl.empty() == false);
               }

    // Open the file
               if(0 == strncmp(request->files[0].turl.c_str(),"file://",7)){
                   std::string path = request->files[0].turl.substr(7);
                   int fd = open(path.c_str(),O_RDONLY);
                   CPPUNIT_ASSERT_MESSAGE("Cannot Open file", fd != -1);
                   char * buffer = new char[strlen(SAMPLE_TEXT) + 100];
                   memset(buffer,0,strlen(SAMPLE_TEXT) + 100);
                   int br = read(fd,buffer,strlen(SAMPLE_TEXT) + 100);
                   CPPUNIT_ASSERT_MESSAGE("Invalid Number of bytes read", (size_t)br == strlen(SAMPLE_TEXT));
                   CPPUNIT_ASSERT_MESSAGE("Invalid text read", 0 == strcmp(buffer,SAMPLE_TEXT));
                   delete [] buffer;
                   close(fd);
               } else {
                   CPPUNIT_FAIL("Invalid protocol for TURL: Not [file]");
               }

    // Complete the request
               request->done();
    // Check the Request status
               CHECKSRMSTATUS("Invalid Result Code on [Done]",request->status.code,SrmStatus::SRM_SUCCESS);
    // Check the File status
               CHECKSRMSTATUS("Invalid File Status on [Done]",request->files[0].status.code,SrmStatus::SRM_RELEASED);

}

/*
 * Test Get Method
 */
void Srm11_TestCase::testGet_forever_pending(){

    // Create Ping Request
    boost::shared_ptr<PrepareToGet> request(PrepareToGet::create(m_ctx));

    // Set Request Parameters
    request->files.resize(1);
    request->files[0].surl = "srm://localhost/srm/managerv1?SFN=/test/srm1/this-one-should-stay-pending_001.txt";
    request->transferParameters.protocol = "file";

    // Send Ping Request
    request->execute();

    // Check Return Value
    bool is_request_pending = false;
    if((request->status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
        (request->status.code == SrmStatus::SRM_REQUEST_INPROGRESS)){
        is_request_pending = true;
        }
        CPPUNIT_ASSERT_MESSAGE("Invalid Result Code on [execute]", is_request_pending == true);

        int count = 0;
        do{
        // Sleep for a while
            //sleep(1);

        // Keep polling for the status
            request->update();

        // Check Return Value
            if(!((request->status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
                  (request->status.code == SrmStatus::SRM_REQUEST_INPROGRESS))){
                is_request_pending = false;
                  }
                  ++count;
        } while((count <= 3) && (is_request_pending == true));

        CPPUNIT_ASSERT_MESSAGE("Invalid Result Code on [update]", is_request_pending == true);

    // Abort the request
        request->abort();
    // Check the Request status
        CHECKSRMSTATUS("Invalid Result Code on [Done]",request->status.code,SrmStatus::SRM_ABORTED);
    // Check the File status
        CHECKSRMSTATUS("Invalid File Status on [Done]",request->files[0].status.code,SrmStatus::SRM_ABORTED);

}

/*
 * Test BringOnline Method
 */
void Srm11_TestCase::testBringOnline(){

    // Create Ping Request
    boost::shared_ptr<BringOnline> request(BringOnline::create(m_ctx));

    // Set Request Parameters
    request->files.resize(1);
    request->files[0].surl = "srm://localhost/srm/managerv1?SFN=/test/srm1/file001.txt";
    request->transferParameters.protocol = "file";

    // Send Ping Request
    request->execute();

    // Check Return Value
    bool is_request_failed = true;
    if((request->status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
        (request->status.code == SrmStatus::SRM_REQUEST_INPROGRESS) ||
        (request->status.code == SrmStatus::SRM_SUCCESS)){
        is_request_failed = false;
        }
        CPPUNIT_ASSERT_MESSAGE("Invalid Result Code on [execute]", is_request_failed == false);

        while((request->status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
               (request->status.code == SrmStatus::SRM_REQUEST_INPROGRESS)){
        // Sleep for a while
            //sleep(1);

        // Keep polling for the status
            request->update();

        // Check Return Value
            if((request->status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
                (request->status.code == SrmStatus::SRM_REQUEST_INPROGRESS) ||
                (request->status.code == SrmStatus::SRM_SUCCESS)){
                is_request_failed = false;
                }
                CPPUNIT_ASSERT_MESSAGE("Invalid Result Code on [update]", is_request_failed == false);
               }
               CHECKSRMSTATUS("Result Code should now be SUCCESS",request->status.code,SrmStatus::SRM_SUCCESS);

               if(request->status.code == SrmStatus::SRM_SUCCESS){
        // Check that the file properties are set correctly
                   CHECKSRMSTATUS("Invalid File Status after success",request->files[0].status.code,SrmStatus::SRM_SUCCESS);
               }

    // Complete the request
               request->done();
    // Check the Request status
               CHECKSRMSTATUS("Invalid Result Code on [DONE]",request->status.code,SrmStatus::SRM_SUCCESS);
    // Check the File status
               CHECKSRMSTATUS("Invalid File Status after Done",request->files[0].status.code,SrmStatus::SRM_RELEASED);

}

/*
 * Test BringOnline Method
 */
void Srm11_TestCase::testBringOnline_forever_pending(){

    // Create Ping Request
    boost::shared_ptr<BringOnline> request(BringOnline::create(m_ctx));

    // Set Request Parameters
    request->files.resize(1);
    request->files[0].surl = "srm://localhost/srm/managerv1?SFN=/test/srm1/this-one-should-stay-pending.txt";
    request->transferParameters.protocol = "file";

    // Send Ping Request
    request->execute();

    // Check Return Value
    bool is_request_pending = false;
    if((request->status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
        (request->status.code == SrmStatus::SRM_REQUEST_INPROGRESS)){
        is_request_pending = true;
        }
        CPPUNIT_ASSERT_MESSAGE("Invalid Result Code on [execute]", is_request_pending == true);

        int count = 0;
        do{
        // Sleep for a while
            //sleep(1);

        // Keep polling for the status
            request->update();

        // Check Return Value
            if(!((request->status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
                  (request->status.code == SrmStatus::SRM_REQUEST_INPROGRESS))){
                is_request_pending = false;
                  }
                  ++count;
        } while ((count <= 3) && (is_request_pending == true));

        CPPUNIT_ASSERT_MESSAGE("Invalid Result Code on [update]", is_request_pending == true);

    // Abort the request
        request->abort();
    // Check the Request status
        CHECKSRMSTATUS("Invalid Result Code on [DONE]",request->status.code,SrmStatus::SRM_ABORTED);
    // Check the File status
        CHECKSRMSTATUS("Invalid File Status after Done",request->files[0].status.code,SrmStatus::SRM_ABORTED);

}

/*
 * Test Copy Method
 */
void Srm11_TestCase::testCopy(){

    // Create Ping Request
    boost::shared_ptr<SrmCopy> request(SrmCopy::create(m_ctx));

    // Set Request Parameters
    request->files.resize(1);
    request->files[0].sourceSurl = "srm://localhost/srm/managerv1?SFN=/test/srm1/file001.txt";
    request->files[0].targetSurl = "srm://localhost/srm/managerv1?SFN=/test/srm1/file002.txt";

    // Send Ping Request
    request->execute();

    // Check Return Value
    bool is_request_failed = true;
    if((request->status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
        (request->status.code == SrmStatus::SRM_REQUEST_INPROGRESS) ||
        (request->status.code == SrmStatus::SRM_SUCCESS)){
        is_request_failed = false;
        }
        CPPUNIT_ASSERT_MESSAGE("Invalid Result Code on [execute]", is_request_failed == false);

        while((request->status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
               (request->status.code == SrmStatus::SRM_REQUEST_INPROGRESS)){
        // Sleep for a while
            //sleep(1);

        // Keep polling for the status
            request->update();

        // Check Return Value
            if((request->status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
                (request->status.code == SrmStatus::SRM_REQUEST_INPROGRESS) ||
                (request->status.code == SrmStatus::SRM_SUCCESS)){
                is_request_failed = false;
                }
                CPPUNIT_ASSERT_MESSAGE("Invalid Result Code on [update]", is_request_failed == false);
               }
               CHECKSRMSTATUS("Result Code should now be SUCCESS",request->status.code,SrmStatus::SRM_SUCCESS);

               if(request->status.code == SrmStatus::SRM_SUCCESS){
        // Check that the file properties are set correctly
                   CHECKSRMSTATUS("Invalid File Status",request->files[0].status.code,SrmStatus::SRM_SUCCESS);
                   CPPUNIT_ASSERT_MESSAGE("No File size", (size_t)request->files[0].size == strlen(SAMPLE_TEXT));
               }

    // Complete the request
               request->done();
    // Check the Request status
               CHECKSRMSTATUS("Invalid Result Code on [DONE]",request->status.code,SrmStatus::SRM_SUCCESS);
    // Check the File status
               CHECKSRMSTATUS("Invalid File Status",request->files[0].status.code,SrmStatus::SRM_SUCCESS);

}

/*
 * Test Copy Method
 */
void Srm11_TestCase::testCopy_forever_pending(){

    // Create Ping Request
    boost::shared_ptr<SrmCopy> request(SrmCopy::create(m_ctx));

    // Set Request Parameters
    request->files.resize(1);
    request->files[0].sourceSurl = "srm://localhost/srm/managerv1?SFN=/test/srm1/this-one-should-stay-pending.txt";
    request->files[0].targetSurl = "srm://localhost/srm/managerv1?SFN=/test/srm1/file002.txt";

    // Send Ping Request
    request->execute();

    // Check Return Value
    bool is_request_pending = false;
    if((request->status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
        (request->status.code == SrmStatus::SRM_REQUEST_INPROGRESS)){
        is_request_pending = true;
        }
        CPPUNIT_ASSERT_MESSAGE("Invalid Result Code on [execute]", is_request_pending == true);

        int count = 0;
        do{
        // Sleep for a while
            //sleep(1);

        // Keep polling for the status
            request->update();

        // Check Return Value
            if(!((request->status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
                  (request->status.code == SrmStatus::SRM_REQUEST_INPROGRESS))){
                is_request_pending = false;
                  }
                  ++count;
        } while((count <= 3) && (is_request_pending == true));

        CPPUNIT_ASSERT_MESSAGE("Invalid Result Code on [update]", is_request_pending == true);

    // Abort the request
        request->abort();
    // Check the Request status
        CHECKSRMSTATUS("Invalid Result Code on [DONE]",request->status.code,SrmStatus::SRM_ABORTED);
    // Check the File status
        CHECKSRMSTATUS("Invalid File Status",request->files[0].status.code,SrmStatus::SRM_ABORTED);

}

/*
 * Test Rm Method
 */
void Srm11_TestCase::testRm(){

    // Create Request
    boost::shared_ptr<SrmRm> request(SrmRm::create(m_ctx));

    // Set Request Parameters
    request->files.resize(1);
    request->files[0].surl = "srm://localhost/srm/managerv1?SFN=/test/srm1/file002.txt";

    // Send Request
    SrmStatus result = request->execute();

    // Check Return Value
    CHECKSRMSTATUS("Invalid Result Code",result.code,SrmStatus::SRM_SUCCESS);

    // Check file Properties
    CHECKSRMSTATUS("Invalid Status",request->files[0].status.code,SrmStatus::SRM_SUCCESS);

}

/*
 * Test Rm Method
 */
void Srm11_TestCase::testRm_Again(){

    // Create Request
    boost::shared_ptr<SrmRm> request(SrmRm::create(m_ctx));

    // Set Request Parameters
    request->files.resize(1);
    request->files[0].surl = "srm://localhost/srm/managerv1?SFN=/test/srm1/file002.txt";

    // Send Request
    SrmStatus result = request->execute();

    // Check Return Value
    CHECKSRMSTATUS("Invalid Result Code",result.code,SrmStatus::SRM_FAILURE);

    // Check file Properties
    CHECKSRMSTATUS("Invalid Status",request->files[0].status.code,SrmStatus::SRM_INVALID_PATH);

}

/*
 * Test GetSpaceTokens Method
 */
void Srm11_TestCase::testGetSpaceTokens(){

    CHECK_SUPPORTED(GetSpaceTokens,m_ctx,false);

    // Create Request
    boost::shared_ptr<GetSpaceTokens> request(GetSpaceTokens::create(m_ctx));

    // Set Request Parameters
    request->description = "Test token description";

    // Send Request
    SrmStatus result = request->execute();

    // Check Return Value
    CHECKSRMSTATUS("Invalid Result Code",result.code,SrmStatus::SRM_NOT_SUPPORTED);

    // Check file Properties
    CPPUNIT_ASSERT_MESSAGE("Invalid token returned", request->tokens.empty() == true);

}

/**
 * test GetSpaceMetaData method
 */
void Srm11_TestCase::testGetSpaceMetaData(){
}

/*
 * Test Mkdir Method
 */
void Srm11_TestCase::testMkDir(){

    CHECK_SUPPORTED(SrmMkDir,m_ctx,false);

    // Create Request
    boost::shared_ptr<SrmMkDir> request(SrmMkDir::create(m_ctx));

    // Set Request Parameters
    request->directoryPath = "srm://localhost/srm/managerv1?SFN=/test/srm1/dir";

    // Send Request
    SrmStatus result = request->execute();

    // Check Return Value
    CHECKSRMSTATUS("Invalid Result Code",result.code,SrmStatus::SRM_SUCCESS);

}

/*
 * Test Rmdir Method
 */
void Srm11_TestCase::testRmDir_NoSuchDir(){

    // Create Request
    boost::shared_ptr<SrmRmDir> request(SrmRmDir::create(m_ctx));

    // Set Request Parameters
    request->directoryPath = "srm://localhost/srm/managerv1?SFN=/test/srm1/dir";

    // Send Request
    SrmStatus result = request->execute();

    // Check Return Value
    CHECKSRMSTATUS("Invalid Result Code",result.code,SrmStatus::SRM_INVALID_PATH);

}

/*
 * Test Rmdir Method
 */
void Srm11_TestCase::testRmDir_NonEmptyDir(){

    // Create Request
    boost::shared_ptr<SrmRmDir> request(SrmRmDir::create(m_ctx));

    // Set Request Parameters
    request->directoryPath = "srm://localhost/srm/managerv1?SFN=/test/srm1";

    // Send Request
    SrmStatus result = request->execute();

    // Check Return Value
    CHECKSRMSTATUS("Invalid Result Code",result.code,SrmStatus::SRM_NON_EMPTY_DIRECTORY);

}

/*
 * Test Rm Method
 */
void Srm11_TestCase::testRm_FirstFile(){

    // Create Request
    boost::shared_ptr<SrmRm> request(SrmRm::create(m_ctx));

    // Set Request Parameters
    request->files.resize(1);
    request->files[0].surl = "srm://localhost/srm/managerv1?SFN=/test/srm1/file001.txt";

    // Send Request
    SrmStatus result = request->execute();

    // Check Return Value
    CHECKSRMSTATUS("Invalid Result Code",result.code,SrmStatus::SRM_SUCCESS);

    // Check file Properties
    CHECKSRMSTATUS("Invalid Status",request->files[0].status.code,SrmStatus::SRM_SUCCESS);

}

/*
 * Test Rmdir Method
 */
void Srm11_TestCase::testRmDir(){

    // Create Request
    boost::shared_ptr<SrmRmDir> request(SrmRmDir::create(m_ctx));

    // Set Request Parameters
    request->directoryPath = "srm://localhost/srm/managerv1?SFN=/test/srm1";

    // Send Request
    SrmStatus result = request->execute();

    // Check Return Value: Now the Directory should be deleted
    CHECKSRMSTATUS("Invalid Result Code",result.code,SrmStatus::SRM_SUCCESS);

}

/*
 * Test AbortFiles Method
 */
void Srm11_TestCase::testAbortFiles(){

    // Create Ping Request
    boost::shared_ptr<PrepareToPut> request(PrepareToPut::create(m_ctx));

    // Set Request Parameters
    request->files.resize(2);
    request->files[0].surl = "srm://localhost/srm/managerv1?SFN=/test/srm1/this-one-should-stay-pending_001.txt";
    request->files[1].surl = "srm://localhost/srm/managerv1?SFN=/test/srm1/this-one-should-stay-pending_002.txt";
    request->transferParameters.protocol = "file";

    // Send Ping Request
    request->execute();

    // Check Return Value
    bool is_request_pending = false;
    if((request->status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
       (request->status.code == SrmStatus::SRM_REQUEST_INPROGRESS)){
           is_request_pending = true;
    }
    CPPUNIT_ASSERT_MESSAGE("Invalid Result Code on [execute]", is_request_pending == true);

    // Sleep for a while
    //sleep(1);

    // Keep polling for the status
    request->update();

    // Check Return Value
    if(!((request->status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
         (request->status.code == SrmStatus::SRM_REQUEST_INPROGRESS))){
        is_request_pending = false;
    }
    CPPUNIT_ASSERT_MESSAGE("Invalid Result Code on [update]", is_request_pending == true);

    // Abort one of the files
    boost::shared_ptr<AbortFiles> abort_request(AbortFiles::create(m_ctx));
    // Set Request Parameters
    abort_request->token = request->token();
    abort_request->files.resize(1);
    abort_request->files[0].surl = "srm://localhost/srm/managerv1?SFN=/test/srm1/this-one-should-stay-pending_002.txt";

    // Send AbortFiles Request
    SrmStatus abort_result = abort_request->execute();
    CHECKSRMSTATUS("Invalid Result Code on [ABORT_FILES]",abort_result.code,SrmStatus::SRM_SUCCESS);
    CHECKSRMSTATUS("Invalid File Result Code on [ABORT_FILES]",abort_request->files.at(0).status.code,SrmStatus::SRM_SUCCESS);

    // Check request status
    request->update();

    // Check Return Value
    if(!((request->status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
         (request->status.code == SrmStatus::SRM_REQUEST_INPROGRESS))){
        is_request_pending = false;
    }
    CPPUNIT_ASSERT_MESSAGE("Invalid Result Code on [update after abort]", is_request_pending == true);
    // Check Files' State
    bool is_file_pending = true;
    if(!((request->files.at(0).status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
         (request->files.at(0).status.code == SrmStatus::SRM_REQUEST_INPROGRESS))){
        is_file_pending= false;
    }
    CPPUNIT_ASSERT_MESSAGE("Invalid File #0 Code on [update after abort]", is_file_pending == true);
    if(!((request->files.at(1).status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
         (request->files.at(1).status.code == SrmStatus::SRM_REQUEST_INPROGRESS))){
        is_file_pending= false;
    }
    CPPUNIT_ASSERT_MESSAGE("Invalid File #1 Code on [update after abort]", is_file_pending == false);
    CPPUNIT_ASSERT_MESSAGE("Invalid File #1 Status [after abort files]", (request->files.at(1).status.code == SrmStatus::SRM_ABORTED) || (request->files.at(1).status.code == SrmStatus::SRM_FAILURE));

    // Abort the request
    request->abort();
    // Check the Request status
    CHECKSRMSTATUS("Invalid Result Code on [ABORT]",request->status.code,SrmStatus::SRM_ABORTED);
    // Check the File status
    CHECKSRMSTATUS("Invalid File Status #0",request->files[0].status.code,SrmStatus::SRM_ABORTED);
    CPPUNIT_ASSERT_MESSAGE("Invalid File Status #1", (request->files[1].status.code == SrmStatus::SRM_ABORTED) || (request->files.at(1).status.code == SrmStatus::SRM_FAILURE));

}

/*
 * Test AbortFiles Method
 */
void Srm11_TestCase::testAbortFiles_AlreadyAborted(){

    // Create Ping Request
    boost::shared_ptr<PrepareToPut> request(PrepareToPut::create(m_ctx));

    // Set Request Parameters
    request->files.resize(1);
    request->files[0].surl = "srm://localhost/srm/managerv1?SFN=/test/srm1/this-one-should-stay-pending_001.txt";
    request->transferParameters.protocol = "file";

    // Send Ping Request
    request->execute();

    // Check Return Value
    bool is_request_pending = false;
    if((request->status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
       (request->status.code == SrmStatus::SRM_REQUEST_INPROGRESS)){
           is_request_pending = true;
    }
    CPPUNIT_ASSERT_MESSAGE("Invalid Result Code on [execute]", is_request_pending == true);

    // Sleep for a while
    //sleep(1);

    // Keep polling for the status
    request->update();

    // Check Return Value
    if(!((request->status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
         (request->status.code == SrmStatus::SRM_REQUEST_INPROGRESS))){
        is_request_pending = false;
    }
    CPPUNIT_ASSERT_MESSAGE("Invalid Result Code on [update]", is_request_pending == true);

    // Abort the request
    request->abort();
    // Check the Request status
    CHECKSRMSTATUS("Invalid Result Code on [ABORT]",request->status.code,SrmStatus::SRM_ABORTED);
    // Check the File status
    CHECKSRMSTATUS("Invalid File Status #0",request->files[0].status.code,SrmStatus::SRM_ABORTED);

    // Abort one of the files
    boost::shared_ptr<AbortFiles> abort_request(AbortFiles::create(m_ctx));
    // Set Request Parameters
    abort_request->token = request->token();
    abort_request->files.resize(1);
    abort_request->files[0].surl = "srm://localhost/srm/managerv1?SFN=/test/srm1/this-one-should-stay-pending_001.txt";

    // Send AbortFiles Request
    SrmStatus abort_result = abort_request->execute();
    CHECKSRMSTATUS("Invalid Result Code on [ABORT_FILES]",abort_result.code,SrmStatus::SRM_SUCCESS);
    CHECKSRMSTATUS("Invalid File Result Code on [ABORT_FILES]",abort_request->files.at(0).status.code,SrmStatus::SRM_SUCCESS);

}

/*
 * Test AbortFiles Method
 */
void Srm11_TestCase::testAbortFiles_ErrorNoFile(){

    // Create Ping Request
    boost::shared_ptr<PrepareToPut> request(PrepareToPut::create(m_ctx));

    // Set Request Parameters
    request->files.resize(1);
    request->files[0].surl = "srm://localhost/srm/managerv1?SFN=/test/srm1/this-one-should-stay-pending_001.txt";
    request->transferParameters.protocol = "file";

    // Send Ping Request
    request->execute();

    // Check Return Value
    bool is_request_pending = false;
    if((request->status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
       (request->status.code == SrmStatus::SRM_REQUEST_INPROGRESS)){
           is_request_pending = true;
    }
    CPPUNIT_ASSERT_MESSAGE("Invalid Result Code on [execute]", is_request_pending == true);

    // Sleep for a while
    //sleep(1);

    // Keep polling for the status
    request->update();

    // Check Return Value
    if(!((request->status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
         (request->status.code == SrmStatus::SRM_REQUEST_INPROGRESS))){
        is_request_pending = false;
    }
    CPPUNIT_ASSERT_MESSAGE("Invalid Result Code on [update]", is_request_pending == true);

    // Abort the request
    request->abort();
    // Check the Request status
    CHECKSRMSTATUS("Invalid Result Code on [ABORT]",request->status.code,SrmStatus::SRM_ABORTED);
    // Check the File status
    CHECKSRMSTATUS("Invalid File Status #0",request->files[0].status.code,SrmStatus::SRM_ABORTED);

    // Abort one of the files
    boost::shared_ptr<AbortFiles> abort_request(AbortFiles::create(m_ctx));
    // Set Request Parameters
    abort_request->token = request->token();
    abort_request->files.resize(1);
    abort_request->files[0].surl = "srm://localhost/srm/managerv1?SFN=/test/srm1/this-one-should-stay-pending_666.txt";

    // Send AbortFiles Request
    SrmStatus abort_result = abort_request->execute();
    CHECKSRMSTATUS("Invalid Result Code on [ABORT_FILES]",abort_result.code,SrmStatus::SRM_FAILURE);
    CHECKSRMSTATUS("Invalid File Result Code on [ABORT_FILES]",abort_request->files.at(0).status.code,SrmStatus::SRM_INVALID_PATH);

}

void Srm11_TestCase::testStat_AfterCopy(){

    // Create Request
    boost::shared_ptr<SrmStat> request(SrmStat::create(m_ctx));

    // Set Request Parameters
    request->files.resize(1);
    request->files[0].surl = "srm://localhost/srm/managerv1?SFN=/test/srm1/file002.txt";

    // Send Request
    SrmStatus result = request->execute();

    // Check Return Value
    CHECKSRMSTATUS("Invalid Result Code",result.code,SrmStatus::SRM_SUCCESS);

    // Check file Properties
    CPPUNIT_ASSERT_MESSAGE("Invalid SURL",   request->files[0].surl == "srm://localhost/srm/managerv1?SFN=/test/srm1/file002.txt");
    CPPUNIT_ASSERT_MESSAGE("Invalid Status", (request->files[0].status.code == SrmStatus::SRM_SUCCESS) || (request->files[0].status.code == SrmStatus::SRM_UNDEF));
    CPPUNIT_ASSERT_MESSAGE("Null Size"   , request->files[0].size > 0);
    CPPUNIT_ASSERT_MESSAGE("Invalid Size", (size_t)request->files[0].size == strlen(SAMPLE_TEXT));
    CPPUNIT_ASSERT_MESSAGE("Invalid File Locality", request->files[0].fileLocality == FileLocality::ONLINE);
    CPPUNIT_ASSERT_MESSAGE("Invalid File Type", request->files[0].fileType == FileType::FILE);

}

