/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004.
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  GLite Data SRM Util CPP: SRM 2.2 Test Case
 *
 *  Authors: Paolo Badino <paolo.badino@cern.ch>,
 *           Paolo Tedesco <paolo.tedesco@cern.ch>
 *           Zsolt Molnar <zsolt.molnar@cern.ch>
 *  Version info: $Id: Srm22_SrmLs.cpp,v 1.2 2009/06/21 17:09:31 molnarzs Exp $
 *  Release: $Name:  $
 *
 */

#include "glite/data/srm/util/Context.h"
#include "glite/data/srm/util/SrmExceptions.h"
#include "glite/data/srm/util/SrmLs.h"
#include "glite/data/srm/util/AbortFiles.h"
#include "glite/data/agents/UserProxyEnv.h"

#include <memory>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <boost/scoped_ptr.hpp>

#include "Srm22_SrmLs.h"

using boost::scoped_ptr;

extern std::string g_srm_22_endpoint;

void Srm22_SrmLs::setUp()
{
    CPPUNIT_ASSERT_MESSAGE("Invalid Endpoint", g_srm_22_endpoint.empty() == false);
    m_ctx.reset(new test_helpers::TestContext());
    m_ctx->endpoint = g_srm_22_endpoint;
    m_ctx->version  = "2.2";

    TestBase::setUp();
    put_file(_TestFilePath_synch(), true);
    put_file(_TestFilePath_asynch(), true);
}

void Srm22_SrmLs::tearDown()
{
	remove_file(_TestFilePath_synch());
	remove_file(_TestFilePath_asynch());
	TestBase::tearDown();
}

void Srm22_SrmLs::_check_supported(const std::string& version)
{
    // Create Context
    Context::Ptr ctx(new test_helpers::TestContext());
    ctx->endpoint = g_srm_22_endpoint;
    ctx->version  = version;
    TestBase::check_supported(ctx, v2_2);
    // Check supported methods
    CHECK_SUPPORTED(SrmLs,ctx,true);
}

/**
 * Test Ls Method
 */
void Srm22_SrmLs::testLs_synch_request()
{
	std::string surl = _TestFilePath_synch();
    // Create Request
    boost::shared_ptr<SrmLs> request(SrmLs::create(m_ctx));

    // Set Request Parameters
    request->surls.resize(1);
    request->surls[0] = surl;
    request->numOfLevels = 0;

    // Send Request
    request->execute();

    // Check Return Value
    CHECKSRMSTATUS("Invalid Result Code",request->status.code,SrmStatus::SRM_SUCCESS);

    // Check file Properties
    CPPUNIT_ASSERT_MESSAGE("Invalid SURL",   request->details[0].surl == surl);
    CPPUNIT_ASSERT_MESSAGE("Invalid Status", (request->details[0].status.code == SrmStatus::SRM_SUCCESS) || (request->details[0].status.code == SrmStatus::SRM_UNDEF));
    CPPUNIT_ASSERT_MESSAGE("Null Size"   , request->details[0].size > 0);
    CPPUNIT_ASSERT_MESSAGE("Invalid Size", (size_t)request->details[0].size == strlen(SAMPLE_TEXT));
    CPPUNIT_ASSERT_MESSAGE("Invalid File Locality", request->details[0].fileLocality == FileLocality::ONLINE);
    CPPUNIT_ASSERT_MESSAGE("Invalid File Type", request->details[0].fileType == FileType::FILE);
    CPPUNIT_ASSERT_MESSAGE("Invalid Storage Type", request->details[0].fileStorageType == FileStorageType::VOLATILE);
}

void Srm22_SrmLs::testLs_02()
{
    std::string surl = testPath() + "locality-NEARLINE.txt";
    put_file(surl);
    // Create Request
    boost::shared_ptr<SrmLs> request(SrmLs::create(m_ctx));

    // Set Request Parameters
    request->surls.resize(1);
    request->surls[0] = surl;
    request->numOfLevels = 0;

    // Send Request
    request->execute();

    remove_file(surl);

    // Check Return Value
    CHECKSRMSTATUS("Invalid Result Code",request->status.code,SrmStatus::SRM_SUCCESS);

    // Check file Properties
    CPPUNIT_ASSERT_MESSAGE("Invalid SURL",   request->details[0].surl == surl);
    CPPUNIT_ASSERT_MESSAGE("Invalid Status", (request->details[0].status.code == SrmStatus::SRM_SUCCESS) || (request->details[0].status.code == SrmStatus::SRM_UNDEF));
    CPPUNIT_ASSERT_MESSAGE("Null Size"   , request->details[0].size > 0);
    CPPUNIT_ASSERT_MESSAGE("Invalid Size", (size_t)request->details[0].size == strlen(SAMPLE_TEXT));
    CPPUNIT_ASSERT_MESSAGE("Invalid File Locality", request->details[0].fileLocality == FileLocality::NEARLINE);
    CPPUNIT_ASSERT_MESSAGE("Invalid File Type", request->details[0].fileType == FileType::FILE);
    CPPUNIT_ASSERT_MESSAGE("Invalid Storage Type", request->details[0].fileStorageType == FileStorageType::VOLATILE);
}

void Srm22_SrmLs::testLs_03()
{
    std::string surl = testPath() + "this-one-returns-SRM_AUTHORIZATION_FAILURE-on-ls.dat";
    // Create Request
    boost::shared_ptr<SrmLs> request(SrmLs::create(m_ctx));
    // Set Request Parameters
    request->surls.resize(1);
    request->surls[0] = surl;
    request->numOfLevels = 0;
    // Send Request
    request->execute();
    // Check Return Value
    CHECKSRMSTATUS("Invalid Result Code",request->status.code,SrmStatus::SRM_FAILURE);
    // Check file Properties
    CPPUNIT_ASSERT_MESSAGE("Invalid SURL", request->details[0].surl == surl);
 //   CHECKSRMSTATUS("Invalid Status",request->details[0].status.code,SrmStatus::SRM_AUTHORIZATION_FAILURE);
}

void Srm22_SrmLs::testLs_04()
{
    std::string surl = testPath() + "test-cksm-ADLER32-141232401ab6571258631c5107aa9b65.dat";
    put_file(surl);
    // Create Request
    boost::shared_ptr<SrmLs> request(SrmLs::create(m_ctx));
    // Set Request Parameters
    request->surls.resize(1);
    request->surls[0] = surl;
    request->numOfLevels = 0;
    // Send Request
    request->execute();
    remove_file(surl);
    // Check Return Value
    CHECKSRMSTATUS("Invalid Result Code",request->status.code,SrmStatus::SRM_SUCCESS);
    // Check file Properties
    CPPUNIT_ASSERT_MESSAGE("Invalid SURL",   request->details[0].surl == surl);
    CHECKSRMSTATUS("Invalid Status",request->details[0].status.code,SrmStatus::SRM_SUCCESS);
    CPPUNIT_ASSERT_MESSAGE("Invalid checksum algorithm",request->details[0].checkSumType == "ADLER32");
    CPPUNIT_ASSERT_MESSAGE("Invalid checksum value",request->details[0].checkSumValue == "141232401ab6571258631c5107aa9b65");
}

void Srm22_SrmLs::testLs_asynch_request()
{
	std::string surl = _TestFilePath_asynch();

    boost::shared_ptr<SrmLs> request(SrmLs::create(m_ctx));
    request->surls.resize(1);
    request->surls[0] = surl;
    request->numOfLevels = 0;

    request->execute();
    CHECKSRMSTATUS("Invalid Result Code",request->status.code,SrmStatus::SRM_REQUEST_QUEUED);

    request->update();
    CHECKSRMSTATUS("Invalid Result Code",request->status.code,SrmStatus::SRM_SUCCESS);

    CPPUNIT_ASSERT_MESSAGE("Invalid SURL", request->details[0].surl == surl);
    CPPUNIT_ASSERT_MESSAGE("Invalid Status", (request->details[0].status.code == SrmStatus::SRM_SUCCESS) || (request->details[0].status.code == SrmStatus::SRM_UNDEF));
    CPPUNIT_ASSERT_MESSAGE("Null Size"   , request->details[0].size > 0);
    CPPUNIT_ASSERT_MESSAGE("Invalid Size", (size_t)request->details[0].size == strlen(SAMPLE_TEXT));
    CPPUNIT_ASSERT_MESSAGE("Invalid File Locality", request->details[0].fileLocality == FileLocality::ONLINE);
    CPPUNIT_ASSERT_MESSAGE("Invalid File Type", request->details[0].fileType == FileType::FILE);
    CPPUNIT_ASSERT_MESSAGE("Invalid Storage Type", request->details[0].fileStorageType == FileStorageType::VOLATILE);
}


void Srm22_SrmLs::testLs_ErrorNoFile()
{
	std::string surl = testPath() + "this-file-does-not-exist.txt";
    // Create Request
    boost::shared_ptr<SrmLs> request(SrmLs::create(m_ctx));

    // Set Request Parameters
    request->surls.resize(1);
    request->surls[0] = surl;
    request->numOfLevels = 0;

    // Send Request
    request->execute();
    // Check Return Value
    CHECKSRMSTATUS("Invalid Result Code",request->status.code,SrmStatus::SRM_FAILURE);
    // Check file Properties
    CHECKSRMSTATUS("Invalid Status",request->details[0].status.code, SrmStatus::SRM_INVALID_PATH);
}

void Srm22_SrmLs::testStat_ErrorNoFile_partial()
{
    // Create Request
	boost::shared_ptr<SrmLs> request(SrmLs::create(m_ctx));
    // Set Request Parameters
    request->surls.resize(2);
    request->surls[0] = _TestFilePath_synch();
    request->surls[1] = testPath() + "this-file-does-not-exist.txt";
    request->numOfLevels = 0;
    // Send Request
    request->execute();
    // Check Return Value
    CHECKSRMSTATUS("Invalid Result Code",request->status.code,SrmStatus::SRM_PARTIAL_SUCCESS);
    // Check file Properties
    CPPUNIT_ASSERT_MESSAGE("Invalid Status #0", (request->details[0].status.code == SrmStatus::SRM_SUCCESS) ||
    		                                    (request->details[0].status.code == SrmStatus::SRM_UNDEF));
    CHECKSRMSTATUS("Invalid Status #1",request->details[1].status.code,SrmStatus::SRM_INVALID_PATH);

}

const std::string& Srm22_SrmLs::_TestFilePath_synch()
{
	static const std::string s(testPath() + "file001.txt");
	return s;
}

const std::string& Srm22_SrmLs::_TestFilePath_asynch()
{
	static const std::string s(testPath() + "this-one-returns-SRM_REQUEST_QUEUED-on-ls.dat");
	return s;
}
