/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004.
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  GLite Data SRM Util CPP: SRM 2.2 Test Case
 *
 *  Authors: Paolo Badino <paolo.badino@cern.ch>,
 *           Paolo Tedesco <paolo.tedesco@cern.ch>
 *           Zsolt Molnar <zsolt.molnar@cern.ch>
 *  Version info: $Id: Srm22_TestCase.cpp,v 1.21 2009/06/21 17:09:31 molnarzs Exp $
 *  Release: $Name:  $
 *
 */
#include "glite/data/srm/util/SrmExceptions.h"

#include "glite/data/srm/util/PrepareToPut.h"
#include "glite/data/srm/util/BringOnline.h"
#include "glite/data/srm/util/PrepareToGet.h"
#include "glite/data/srm/util/SrmCopy.h"
#include "glite/data/srm/util/SrmStat.h"

#include "glite/data/srm/util/SrmPing.h"
#include "glite/data/srm/util/SrmRm.h"
#include "glite/data/srm/util/SrmMkDir.h"
#include "glite/data/srm/util/SrmRmDir.h"
#include "glite/data/srm/util/GetSpaceTokens.h"
#include "glite/data/srm/util/GetSpaceMetaData.h"
#include "glite/data/srm/util/AbortFiles.h"
#include "glite/data/agents/UserProxyEnv.h"

#include <memory>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <boost/scoped_ptr.hpp>
#include <cppunit/TestCase.h>
#include <cppunit/extensions/HelperMacros.h>

#include "Srm22_TestCase.h"

using namespace glite::data::srm::util;

using boost::scoped_ptr;

/*
 * The SRM 2.2 endpoint
 */
extern std::string g_srm_22_endpoint;

/*
 * setUp
 *
 * set-up test case
 */
void Srm22_TestCase::setUp(){

    CPPUNIT_ASSERT_MESSAGE("Invalid Endpoint", g_srm_22_endpoint.empty() == false);

    // Create Context
    m_ctx.reset(new test_helpers::TestContext());

    // Create an Srm Context
    m_ctx->endpoint = g_srm_22_endpoint;
    m_ctx->version  = "2.2";
}

/*
 * tearDown
 *
 * tear-down test case
 */
void Srm22_TestCase::tearDown(){
    // Delete Context
    m_ctx.reset();
}

void Srm22_TestCase::_check_supported(const std::string& version)
{
    // Create Context
    Context::Ptr ctx(new test_helpers::TestContext());
    ctx->endpoint = g_srm_22_endpoint;
    ctx->version  = version;
    // Check supported methods
    CHECK_SUPPORTED(SrmPing,            ctx,true);
    CHECK_SUPPORTED(PrepareToPut,       ctx,true);
    CHECK_SUPPORTED(PrepareToGet,       ctx,true);
    CHECK_SUPPORTED(BringOnline,        ctx,true);
    CHECK_SUPPORTED(SrmCopy,            ctx,true);
    CHECK_SUPPORTED(SrmRm,              ctx,true);
    CHECK_SUPPORTED(SrmStat,            ctx,true);
    CHECK_SUPPORTED(GetSpaceTokens,     ctx,true);
    CHECK_SUPPORTED(GetSpaceMetaData,   ctx,true);
    CHECK_SUPPORTED(SrmMkDir,           ctx,true);
    CHECK_SUPPORTED(SrmRmDir,           ctx,true);
}

/*
 * Test Ping Method
 */
void Srm22_TestCase::testPing(){

    // Create Ping Request
    boost::shared_ptr<SrmPing> request(SrmPing::create(m_ctx));

    // Send Ping Request
    SrmStatus result = request->execute();

    // Check Return Value
    CHECKSRMSTATUS("Invalid Result Code",result.code,SrmStatus::SRM_SUCCESS);
    CPPUNIT_ASSERT_MESSAGE("Invalid Version Returned", request->versionInfo == "v2.2");

}

/*
 * Test Ping Method
 */
void Srm22_TestCase::testPing_SecurityError(){

    // Setup a wrong value for X509_USER_PROXY
    glite::data::agents::UserProxyEnv env("/tmp/x509_fake");

    // Create Ping Request
    boost::shared_ptr<SrmPing> request(SrmPing::create(m_ctx));

    // Send Ping Request
    try{
        SrmStatus result = request->execute();
        CPPUNIT_FAIL("This method should throw a SecurityError");
    } catch (const SecurityError& exc){
        // OK
    }

}

/*
 * Test Ping Method
 */
void Srm22_TestCase::testPing_RemoteError(){

    // Use an existing endpoint
    m_ctx->endpoint = "httpg://fake-srm-service-hopefully-it-will-not-exist.cern.cern.ch:8443/srm/managerv2";

    // Create Ping Request
    boost::shared_ptr<SrmPing> request(SrmPing::create(m_ctx));

    // Send Ping Request
    try{
        SrmStatus result = request->execute();
        CPPUNIT_FAIL("This method should throw a RemoteError: invalid host");
    } catch (const RemoteError& exc){
        // OK
    }

    // Use an existing endpoint
    m_ctx->endpoint = "httpg://localhost:88054/srm/managerv2";

    // Create Ping Request
    request.reset(SrmPing::create(m_ctx));

    // Send Ping Request
    try{
        SrmStatus result = request->execute();
        CPPUNIT_FAIL("This method should throw a RemoteError: invalid port");
    } catch (const RemoteError& exc){
        // OK
    }

}

/*
 * Test Mkdir Method
 */
void Srm22_TestCase::testMkDir(){

    CHECK_SUPPORTED(SrmMkDir,m_ctx,true);

    // Create Request
    boost::shared_ptr<SrmMkDir> request(SrmMkDir::create(m_ctx));

    // Set Request Parameters
    request->directoryPath = "srm://localhost/srm/managerv2?SFN=/test";

    // Send Request
    SrmStatus result = request->execute();

    // Set Request Parameters
    request->directoryPath = "srm://localhost/srm/managerv2?SFN=/test/srm2";

    // Send Request
    result = request->execute();

    // Check Return Value
    CHECKSRMSTATUS("Invalid Result Code [2]",result.code,SrmStatus::SRM_SUCCESS);

}

/*
 * Test Mkdir Method
 */
void Srm22_TestCase::testMkDir_Again(){

    // Create Request
    boost::shared_ptr<SrmMkDir> request(SrmMkDir::create(m_ctx));

    // Set Request Parameters
    request->directoryPath = "srm://localhost/srm/managerv2?SFN=/test/srm2";

    // Send Request
    SrmStatus result = request->execute();

    // Check Return Value
    CHECKSRMSTATUS("Invalid Result Code",result.code,SrmStatus::SRM_DUPLICATION_ERROR);

}
/*
 * Test Mkdir Method
 */
void Srm22_TestCase::testMkDir_NoParentDir(){

    // Create Request
    boost::shared_ptr<SrmMkDir> request(SrmMkDir::create(m_ctx));

    // Set Request Parameters
    request->directoryPath = "srm://localhost/srm/managerv2?SFN=/test/srm2/this-dir-does-not-exist/foo";

    // Send Request
    SrmStatus result = request->execute();

    // Check Return Value
    CHECKSRMSTATUS("Invalid Result Code",result.code,SrmStatus::SRM_INVALID_PATH);

}

/*
 * Test Put Method
 */
void Srm22_TestCase::testPut(){
    put_file("srm://localhost/srm/managerv2?SFN=/test/srm2/file001.txt");
}

/*
 * Test Put Method, put should never return
 */
void Srm22_TestCase::testPut_forever_pending(){

    // Create Ping Request
    boost::shared_ptr<PrepareToPut> request(PrepareToPut::create(m_ctx));

    // Set Request Parameters
    request->files.resize(1);
    request->files[0].surl = "srm://localhost/srm/managerv2?SFN=/test/srm2/this-one-should-stay-pending_001.txt";
    request->transferParameters.protocol = "file";

    // Send Ping Request
    request->execute();

    // Check Return Value
    bool is_request_pending = false;
    if((request->status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
       (request->status.code == SrmStatus::SRM_REQUEST_INPROGRESS)){
           is_request_pending = true;
    }
    CPPUNIT_ASSERT_MESSAGE("Invalid Result Code on [execute]", is_request_pending == true);

    int count = 0;
    do{
        // Sleep for a while
        //sleep(1);

        // Keep polling for the status
        request->update();

        // Check Return Value
        if(!((request->status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
             (request->status.code == SrmStatus::SRM_REQUEST_INPROGRESS))){
            is_request_pending = false;
        }
        ++count;
    }while((count <= 3) && (true == is_request_pending));

    CPPUNIT_ASSERT_MESSAGE("Invalid Result Code on [update]", is_request_pending == true);

    // Abort the request
    request->abort();
    // Check the Request status
    CHECKSRMSTATUS("Invalid Result Code on [ABORT]",request->status.code,SrmStatus::SRM_ABORTED);
    // Check the File status
    CHECKSRMSTATUS("Invalid File Status",request->files[0].status.code,SrmStatus::SRM_ABORTED);

}

/*
 * Test Put Method, for a file busy
 */
void Srm22_TestCase::testPut_file_busy(){

    // Create Ping Request
    boost::shared_ptr<PrepareToPut> request(PrepareToPut::create(m_ctx));

    // Set Request Parameters
    request->files.resize(1);
    request->files[0].surl = "srm://localhost/srm/managerv2?SFN=/test/srm2/this-one-is-busy_001.txt";
    request->transferParameters.protocol = "file";

    // Send Ping Request
    request->execute();

    // Check Return Value
    bool is_request_pending = false;
    if((request->status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
       (request->status.code == SrmStatus::SRM_REQUEST_INPROGRESS)){
           is_request_pending = true;
    }
    CPPUNIT_ASSERT_MESSAGE("Invalid Result Code on [execute]", is_request_pending == true);

    do{
        // Sleep for a while
        //sleep(1);

        // Keep polling for the status
        request->update();

        // Check Return Value
        if(!((request->status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
             (request->status.code == SrmStatus::SRM_REQUEST_INPROGRESS))){
            is_request_pending = false;
        }
    } while(true == is_request_pending);

    CPPUNIT_ASSERT_MESSAGE("Request still pending [update]", is_request_pending == false);

    // Check result
    CHECKSRMSTATUS("Invalid Result Code [update]",request->status.code,SrmStatus::SRM_FAILURE);
    CHECKSRMSTATUS("Invalid File Status [update]",request->files[0].status.code,SrmStatus::SRM_FILE_BUSY);

    // Abort the request
    request->abort();
    // Check the Request status
    CHECKSRMSTATUS("Invalid Result Code on [ABORT]",request->status.code,SrmStatus::SRM_ABORTED);

}

/*
 * Test Put Method in case SRM_NO_FREE_SPACE is returned
 */
void Srm22_TestCase::testPut_no_free_space(){

    // Create Ping Request
    boost::shared_ptr<PrepareToPut> request(PrepareToPut::create(m_ctx));

    // Set Request Parameters
    request->files.resize(1);
    request->files[0].surl = "srm://localhost/srm/managerv2?SFN=/test/srm2/this-one-returns-SRM_NO_FREE_SPACE-on-put.txt";
    request->transferParameters.protocol = "file";

    // Send Ping Request
    request->execute();

    // Check Return Value
    bool is_request_pending = false;
    if((request->status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
       (request->status.code == SrmStatus::SRM_REQUEST_INPROGRESS)){
           is_request_pending = true;
    }
    CPPUNIT_ASSERT_MESSAGE("Invalid Result Code on [execute]", is_request_pending == true);

    do{
        // Sleep for a while
        //sleep(1);

        // Keep polling for the status
        request->update();

        // Check Return Value
        if(!((request->status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
             (request->status.code == SrmStatus::SRM_REQUEST_INPROGRESS))){
            is_request_pending = false;
        }
    } while(true == is_request_pending);

    CPPUNIT_ASSERT_MESSAGE("Request still pending [update]", is_request_pending == false);

    // Check result
    CHECKSRMSTATUS("Invalid Result Code [update]",request->status.code,SrmStatus::SRM_FAILURE);
    CHECKSRMSTATUS("Invalid File Status [update]",request->files[0].status.code,SrmStatus::SRM_NO_FREE_SPACE);

    // Abort the request
    request->abort();
    // Check the Request status
    CHECKSRMSTATUS("Invalid Result Code on [ABORT]",request->status.code,SrmStatus::SRM_ABORTED);

}

/*
 * Test Get Method
 */
void Srm22_TestCase::testGet(){

    // Create Ping Request
    boost::shared_ptr<PrepareToGet> request(PrepareToGet::create(m_ctx));

    // Set Request Parameters
    request->files.resize(1);
    request->files[0].surl = "srm://localhost/srm/managerv2?SFN=/test/srm2/file001.txt";
    request->transferParameters.protocol = "file";

    // Send Ping Request
    request->execute();

    // Check Return Value
    bool is_request_failed = true;
    if((request->status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
       (request->status.code == SrmStatus::SRM_REQUEST_INPROGRESS) ||
       (request->status.code == SrmStatus::SRM_SUCCESS)){
           is_request_failed = false;
    }
    CPPUNIT_ASSERT_MESSAGE("Invalid Result Code on [execute]", is_request_failed == false);

    while((request->status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
          (request->status.code == SrmStatus::SRM_REQUEST_INPROGRESS)){
        // Sleep for a while
        //sleep(1);

        // Keep polling for the status
        request->update();

        // Check Return Value
        if((request->status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
           (request->status.code == SrmStatus::SRM_REQUEST_INPROGRESS) ||
           (request->status.code == SrmStatus::SRM_SUCCESS)){
               is_request_failed = false;
        }
        CPPUNIT_ASSERT_MESSAGE("Invalid Result Code on [update]", is_request_failed == false);
    }
    CHECKSRMSTATUS("Result Code should now be SUCCESS",request->status.code,SrmStatus::SRM_SUCCESS);

    if(request->status.code == SrmStatus::SRM_SUCCESS){
        // Check that the file properties are set correctly
        CHECKSRMSTATUS("Invalid File Status",request->files[0].status.code,SrmStatus::SRM_FILE_PINNED);
        CPPUNIT_ASSERT_MESSAGE("No turl retrieved", request->files[0].turl.empty() == false);
    }

    // Open the file
    if(0 == strncmp(request->files[0].turl.c_str(),"file://",7)){
        std::string path = request->files[0].turl.substr(7);
        int fd = open(path.c_str(),O_RDONLY);
        CPPUNIT_ASSERT_MESSAGE("Cannot Open file", fd != -1);
        char * buffer = new char[strlen(SAMPLE_TEXT) + 100];
        memset(buffer,0,strlen(SAMPLE_TEXT) + 100);
        int br = read(fd,buffer,strlen(SAMPLE_TEXT) + 100);
        CPPUNIT_ASSERT_MESSAGE("Invalid Number of bytes read", (size_t)br == strlen(SAMPLE_TEXT));
        CPPUNIT_ASSERT_MESSAGE("Invalid text read", 0 == strcmp(buffer,SAMPLE_TEXT));
        delete [] buffer;
        close(fd);
    } else {
       CPPUNIT_FAIL("Invalid protocol for TURL: Not [file]");
    }

    // Complete the request
    request->done();
    // Check the Request status
    CHECKSRMSTATUS("Invalid Result Code on [Done]",request->status.code,SrmStatus::SRM_SUCCESS);
    // Check the File status
    CHECKSRMSTATUS("Invalid File Status on [Done]",request->files[0].status.code,SrmStatus::SRM_RELEASED);

}

/*
 * Test Get Method
 */
void Srm22_TestCase::testGet_forever_pending(){

    // Create Ping Request
    boost::shared_ptr<PrepareToGet> request(PrepareToGet::create(m_ctx));

    // Set Request Parameters
    request->files.resize(1);
    request->files[0].surl = "srm://localhost/srm/managerv2?SFN=/test/srm2/this-one-should-stay-pending_001.txt";
    request->transferParameters.protocol = "file";

    // Send Ping Request
    request->execute();

    // Check Return Value
    bool is_request_pending = false;
    if((request->status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
       (request->status.code == SrmStatus::SRM_REQUEST_INPROGRESS)){
           is_request_pending = true;
    }
    CPPUNIT_ASSERT_MESSAGE("Invalid Result Code on [execute]", is_request_pending == true);

    int count = 0;
    do{
        // Sleep for a while
        //sleep(1);

        // Keep polling for the status
        request->update();

        // Check Return Value
        if(!((request->status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
             (request->status.code == SrmStatus::SRM_REQUEST_INPROGRESS))){

        	std::cout << "*** " << request->status.code << std::endl;

            is_request_pending = false;
        }
        ++count;
    } while((count <= 3) && (is_request_pending == true));

    CPPUNIT_ASSERT_MESSAGE("Invalid Result Code on [update]", is_request_pending == true);

    // Abort the request
    request->abort();
    // Check the Request status
    CHECKSRMSTATUS("Invalid Result Code on [Done]",request->status.code,SrmStatus::SRM_ABORTED);
    // Check the File status
    CHECKSRMSTATUS("Invalid File Status on [Done]",request->files[0].status.code,SrmStatus::SRM_ABORTED);

}

/*
 * Test BringOnline Method
 */
void Srm22_TestCase::testBringOnline(){

    // Create Ping Request
    boost::shared_ptr<BringOnline> request(BringOnline::create(m_ctx));

    // Set Request Parameters
    request->files.resize(1);
    request->files[0].surl = "srm://localhost/srm/managerv2?SFN=/test/srm2/file001.txt";
    request->transferParameters.protocol = "file";

    // Send Ping Request
    request->execute();

    // Check Return Value
    bool is_request_failed = true;
    if((request->status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
       (request->status.code == SrmStatus::SRM_REQUEST_INPROGRESS) ||
       (request->status.code == SrmStatus::SRM_SUCCESS)){
           is_request_failed = false;
    }
    CPPUNIT_ASSERT_MESSAGE("Invalid Result Code on [execute]", is_request_failed == false);

    while((request->status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
          (request->status.code == SrmStatus::SRM_REQUEST_INPROGRESS)){
        // Sleep for a while
        //sleep(1);

        // Keep polling for the status
        request->update();

        // Check Return Value
        if((request->status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
           (request->status.code == SrmStatus::SRM_REQUEST_INPROGRESS) ||
           (request->status.code == SrmStatus::SRM_SUCCESS)){
               is_request_failed = false;
        }
        CPPUNIT_ASSERT_MESSAGE("Invalid Result Code on [update]", is_request_failed == false);
    }
    CHECKSRMSTATUS("Result Code should now be SUCCESS",request->status.code,SrmStatus::SRM_SUCCESS);

    if(request->status.code == SrmStatus::SRM_SUCCESS){
        // Check that the file properties are set correctly
        CHECKSRMSTATUS("Invalid File Status",request->files[0].status.code,SrmStatus::SRM_SUCCESS);
    }

    // Complete the request
    request->done();
    // Check the Request status
    CHECKSRMSTATUS("Invalid Result Code on [Done]",request->status.code,SrmStatus::SRM_SUCCESS);
    // Check the File status
    CHECKSRMSTATUS("Invalid File Status on [Done]",request->files[0].status.code,SrmStatus::SRM_RELEASED);

}

/*
 * Test BringOnline Method
 */
void Srm22_TestCase::testBringOnline_forever_pending(){

    // Create Ping Request
    boost::shared_ptr<BringOnline> request(BringOnline::create(m_ctx));

    // Set Request Parameters
    request->files.resize(1);
    request->files[0].surl = "srm://localhost/srm/managerv2?SFN=/test/srm2/this-one-should-stay-pending.txt";
    request->transferParameters.protocol = "file";

    // Send Ping Request
    request->execute();

    // Check Return Value
    bool is_request_pending = false;
    if((request->status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
       (request->status.code == SrmStatus::SRM_REQUEST_INPROGRESS)){
           is_request_pending = true;
    }
    CPPUNIT_ASSERT_MESSAGE("Invalid Result Code on [execute]", is_request_pending == true);

    int count = 0;
    do{
        // Sleep for a while
        //sleep(1);

        // Keep polling for the status
        request->update();

        // Check Return Value
        if(!((request->status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
             (request->status.code == SrmStatus::SRM_REQUEST_INPROGRESS))){
            is_request_pending = false;
        }
        ++count;
    } while ((count <= 3) && (is_request_pending == true));

    CPPUNIT_ASSERT_MESSAGE("Invalid Result Code on [update]", is_request_pending == true);

    // Abort the request
    request->abort();
    // Check the Request status
    CHECKSRMSTATUS("Invalid Result Code on [DONE]",request->status.code,SrmStatus::SRM_ABORTED);
    // Check the File status
    CHECKSRMSTATUS("Invalid File Status after Done",request->files[0].status.code,SrmStatus::SRM_ABORTED);

}

/*
 * Test Copy Method
 */
void Srm22_TestCase::testCopy(){

    // Create Ping Request
    boost::shared_ptr<SrmCopy> request(SrmCopy::create(m_ctx));

    // Set Request Parameters
    request->files.resize(1);
    request->files[0].sourceSurl = "srm://localhost/srm/managerv2?SFN=/test/srm2/file001.txt";
    request->files[0].targetSurl = "srm://localhost/srm/managerv2?SFN=/test/srm2/file002.txt";

    // Send Ping Request
    request->execute();

    // Check Return Value
    bool is_request_failed = true;
    if((request->status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
       (request->status.code == SrmStatus::SRM_REQUEST_INPROGRESS) ||
       (request->status.code == SrmStatus::SRM_SUCCESS)){
           is_request_failed = false;
    }
    CPPUNIT_ASSERT_MESSAGE("Invalid Result Code on [execute]", is_request_failed == false);

    while((request->status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
          (request->status.code == SrmStatus::SRM_REQUEST_INPROGRESS)){
        // Sleep for a while
        //sleep(1);

        // Keep polling for the status
        request->update();

        // Check Return Value
        if((request->status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
           (request->status.code == SrmStatus::SRM_REQUEST_INPROGRESS) ||
           (request->status.code == SrmStatus::SRM_SUCCESS)){
               is_request_failed = false;
        }
        CPPUNIT_ASSERT_MESSAGE("Invalid Result Code on [update]", is_request_failed == false);
    }
    CHECKSRMSTATUS("Result Code should now be SUCCESS",request->status.code,SrmStatus::SRM_SUCCESS);

    if(request->status.code == SrmStatus::SRM_SUCCESS){
        // Check that the file properties are set correctly
        CHECKSRMSTATUS("Invalid File Status",request->files[0].status.code,SrmStatus::SRM_SUCCESS);
        CPPUNIT_ASSERT_MESSAGE("Invalid File size", (size_t)request->files[0].size == strlen(SAMPLE_TEXT));
    }

    // Complete the request
    request->done();
    // Check the Request status
    CHECKSRMSTATUS("Invalid Result Code on [DONE]",request->status.code,SrmStatus::SRM_SUCCESS);
    // Check the File status
    CHECKSRMSTATUS("Invalid File Status",request->files[0].status.code,SrmStatus::SRM_SUCCESS);

}

/*
 * Test Copy Method
 */
void Srm22_TestCase::testCopy_forever_pending(){

    // Create Request
    boost::shared_ptr<SrmCopy> request(SrmCopy::create(m_ctx));

    // Set Request Parameters
    request->files.resize(1);
    request->files[0].sourceSurl = "srm://localhost/srm/managerv2?SFN=/test/srm2/this-one-should-stay-pending.txt";
    request->files[0].targetSurl = "srm://localhost/srm/managerv2?SFN=/test/srm2/file002.txt";

    // Send Ping Request
    request->execute();

    // Check Return Value
    bool is_request_pending = false;
    if((request->status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
       (request->status.code == SrmStatus::SRM_REQUEST_INPROGRESS)){
           is_request_pending = true;
    }
    CPPUNIT_ASSERT_MESSAGE("Invalid Result Code on [execute]", is_request_pending == true);

    int count = 0;
    do{
        // Sleep for a while
        //sleep(1);

        // Keep polling for the status
        request->update();

        // Check Return Value
        if(!((request->status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
             (request->status.code == SrmStatus::SRM_REQUEST_INPROGRESS))){
               is_request_pending = false;
        }
        ++count;
    } while((count <= 3) && (is_request_pending == true));

    CPPUNIT_ASSERT_MESSAGE("Invalid Result Code on [update]", is_request_pending == true);

    // Abort the request
    request->abort();
    // Check the Request status
    CHECKSRMSTATUS("Invalid Result Code on [DONE]",request->status.code,SrmStatus::SRM_ABORTED);
    // Check the File status
    CHECKSRMSTATUS("Invalid File Status",request->files[0].status.code,SrmStatus::SRM_ABORTED);

}

/**
 * SrmCopy: test setting the source space token parameter
 */
void Srm22_TestCase::testCopy_sourceSpaceToken(){
    static const std::string SOURCE_SPACE_TOKEN = "test-source-space-token";
    // Create Ping Request
    boost::shared_ptr<SrmCopy> request(SrmCopy::create(m_ctx));
    // Set Request Parameters
    request->files.resize(1);
    request->files[0].sourceSurl = "srm://localhost/srm/managerv2?SFN=/test/srm2/file001.txt";
    request->files[0].targetSurl = "srm://localhost/srm/managerv2?SFN=/test/srm2/testCopy_sourceSpaceToken.txt";
    // set source space token
    request->sourceSpaceToken = SOURCE_SPACE_TOKEN;
    // Send Ping Request
    request->execute();
    // Check Return Value
    bool is_request_failed = true;
    if((request->status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
       (request->status.code == SrmStatus::SRM_REQUEST_INPROGRESS) ||
       (request->status.code == SrmStatus::SRM_SUCCESS)){
           is_request_failed = false;
    }
    CPPUNIT_ASSERT_MESSAGE("Invalid Result Code on [execute]", is_request_failed == false);
    // HACK: check the stub server has returned the string in the message
    if (std::string::npos == request->status.message.find(SOURCE_SPACE_TOKEN)) {
        FAIL_TEST("request return message '" << request->status.message << "' does not contain source space token '" << SOURCE_SPACE_TOKEN << "'");
    }
    request->abort();
}

/**
 * SrmCopy: test setting the copy pin lifetime parameter
 */
void Srm22_TestCase::testCopy_copyPinLifeTime(){
    static const int COPY_PIN_LIFETIME = 1022008;
    // Create Ping Request
    boost::shared_ptr<SrmCopy> request(SrmCopy::create(m_ctx));
    // Set Request Parameters
    request->files.resize(1);
    request->files[0].sourceSurl = "srm://localhost/srm/managerv2?SFN=/test/srm2/file001.txt";
    request->files[0].targetSurl = "srm://localhost/srm/managerv2?SFN=/test/srm2/testCopy_copyPinLifeTime.txt";
    // set source space token
    request->copyPinLifetime = COPY_PIN_LIFETIME;
    // Send Ping Request
    request->execute();
    // Check Return Value
    bool is_request_failed = true;
    if((request->status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
        (request->status.code == SrmStatus::SRM_REQUEST_INPROGRESS) ||
        (request->status.code == SrmStatus::SRM_SUCCESS)){
        is_request_failed = false;
    }
    CPPUNIT_ASSERT_MESSAGE("Invalid Result Code on [execute]", is_request_failed == false);
    // HACK: check the stub server has returned the string in the message
    std::stringstream tmp;
    tmp << COPY_PIN_LIFETIME;
    if (std::string::npos == request->status.message.find(tmp.str())) {
        FAIL_TEST("request return message '" << request->status.message << "' does not contain copy pin lifetime '" << COPY_PIN_LIFETIME << "'");
    }
    request->abort();
}

/**
 * SrmPrepareToPut: test setting the copy pin lifetime parameter
 */
void Srm22_TestCase::testPut_copyPinLifeTime(){
    static const int COPY_PIN_LIFETIME = 12345;
    // Create Ping Request
    boost::shared_ptr<PrepareToPut> request(PrepareToPut::create(m_ctx));
    // Set Request Parameters
    request->files.resize(1);
    request->files[0].surl = "srm://localhost/srm/managerv2?SFN=/test/srm2/file001.txt";
    request->transferParameters.protocol = "file";
    request->copyPinLifetime = COPY_PIN_LIFETIME;
    // Send Ping Request
    request->execute();

    // Check Return Value
    bool is_request_failed = true;
    if((request->status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
       (request->status.code == SrmStatus::SRM_REQUEST_INPROGRESS) ||
       (request->status.code == SrmStatus::SRM_SUCCESS)){
           is_request_failed = false;
    }
    CPPUNIT_ASSERT_MESSAGE("Invalid Result Code on [execute]", is_request_failed == false);
    // HACK: check the stub server has returned the string in the message
    std::stringstream tmp;
    tmp << COPY_PIN_LIFETIME;
    if (std::string::npos == request->status.message.find(tmp.str())) {
        FAIL_TEST("request return message '" << request->status.message << "' does not contain copy pin lifetime '" << COPY_PIN_LIFETIME << "'");
    }
    request->abort();
}

/*
 * Test Rm Method
 */
void Srm22_TestCase::testRm(){
    remove_file("srm://localhost/srm/managerv2?SFN=/test/srm2/file002.txt");
}

/*
 * Test Rm Method
 */
void Srm22_TestCase::testRm_Again(){

    // Create Request
    boost::shared_ptr<SrmRm> request(SrmRm::create(m_ctx));

    // Set Request Parameters
    request->files.resize(1);
    request->files[0].surl = "srm://localhost/srm/managerv2?SFN=/test/srm2/file002.txt";

    // Send Request
    SrmStatus result = request->execute();

    // Check Return Value
    CHECKSRMSTATUS("Invalid Result Code",result.code,SrmStatus::SRM_FAILURE);

    // Check file Properties
    CHECKSRMSTATUS("Invalid Status",request->files[0].status.code,SrmStatus::SRM_INVALID_PATH);

}

/*
 * Test Rmdir Method
 */
void Srm22_TestCase::testRmDir_NoSuchDir(){

    // Create Request
    boost::shared_ptr<SrmRmDir> request(SrmRmDir::create(m_ctx));

    // Set Request Parameters
    request->directoryPath = "srm://localhost/srm/managerv2?SFN=/test/srm2/dir";

    // Send Request
    SrmStatus result = request->execute();

    // Check Return Value
    CHECKSRMSTATUS("Invalid Result Code",result.code,SrmStatus::SRM_INVALID_PATH);

}

/*
 * Test Rmdir Method
 */
void Srm22_TestCase::testRmDir_NonEmptyDir(){

    // Create Request
    boost::shared_ptr<SrmRmDir> request(SrmRmDir::create(m_ctx));

    // Set Request Parameters
    request->directoryPath = "srm://localhost/srm/managerv2?SFN=/test/srm2";

    // Send Request
    SrmStatus result = request->execute();

    // Check Return Value
    CHECKSRMSTATUS("Invalid Result Code",result.code,SrmStatus::SRM_NON_EMPTY_DIRECTORY);

}

/*
 * Test Rm Method
 */
void Srm22_TestCase::testRm_FirstFile(){

    // Create Request
    boost::shared_ptr<SrmRm> request(SrmRm::create(m_ctx));

    // Set Request Parameters
    request->files.resize(1);
    request->files[0].surl = "srm://localhost/srm/managerv2?SFN=/test/srm2/file001.txt";

    // Send Request
    SrmStatus result = request->execute();

    // Check Return Value
    CHECKSRMSTATUS("Invalid Result Code",result.code,SrmStatus::SRM_SUCCESS);

    // Check file Properties
    CHECKSRMSTATUS("Invalid Status",request->files[0].status.code,SrmStatus::SRM_SUCCESS);

}

/*
 * Test Rmdir Method
 */
void Srm22_TestCase::testRmDir(){

    // Create Request
    boost::shared_ptr<SrmRmDir> request(SrmRmDir::create(m_ctx));

    // Set Request Parameters
    request->directoryPath = "srm://localhost/srm/managerv2?SFN=/test/srm2";

    // Send Request
    SrmStatus result = request->execute();

    // Check Return Value: Now the Directory should be deleted
    CHECKSRMSTATUS("Invalid Result Code",result.code,SrmStatus::SRM_SUCCESS);

}

/*
 * Test GetSpaceTokens Method
 */
void Srm22_TestCase::testGetSpaceTokens_01(){
    CHECK_SUPPORTED(GetSpaceTokens,m_ctx,true);
    // Create Request
    boost::shared_ptr<GetSpaceTokens> request(GetSpaceTokens::create(m_ctx));
    // Set Request Parameters
    request->description = "Description_for:TEST_TOKEN";
    // Send Request
    SrmStatus result = request->execute();
    // Check Return Value
    CHECKSRMSTATUS("Invalid Result Code",result.code,SrmStatus::SRM_SUCCESS);
    // Check file Properties
    CPPUNIT_ASSERT_MESSAGE("Invalid number of tokens returned", request->tokens.size() == 1);
    CPPUNIT_ASSERT_MESSAGE("Invalid token returned", request->tokens.at(0) == "TEST_TOKEN");
}

void Srm22_TestCase::testGetSpaceTokens_02(){
    CHECK_SUPPORTED(GetSpaceTokens,m_ctx,true);
    // Create Request
    boost::shared_ptr<GetSpaceTokens> request(GetSpaceTokens::create(m_ctx));
    // Set Request Parameters
    request->description = "Description_for:TOKEN0;TOKEN1;TOKEN2";
    // Send Request
    SrmStatus result = request->execute();
    // Check Return Value
    CHECKSRMSTATUS("Invalid Result Code",result.code,SrmStatus::SRM_SUCCESS);
    // Check file Properties
    CHECK_VALUE(request->tokens.size(),3);
    CPPUNIT_ASSERT_MESSAGE("Invalid token returned (0)", request->tokens.at(0) == "TOKEN0");
    CPPUNIT_ASSERT_MESSAGE("Invalid token returned (1)", request->tokens.at(1) == "TOKEN1");
    CPPUNIT_ASSERT_MESSAGE("Invalid token returned (2)", request->tokens.at(2) == "TOKEN2");
}

/**
 * Test GetSpaceMetaData method
 */
void Srm22_TestCase::testGetSpaceMetaData(){
    CHECK_SUPPORTED(GetSpaceMetaData,m_ctx,true);
    static const std::string token_name = "rp=CUSTODIAL:al=ONLINE:o=dteam:ts=2048:gs=1024:us=512:la=256:ll=128";
    // Create Request
    boost::shared_ptr<GetSpaceMetaData> request(GetSpaceMetaData::create(m_ctx));
    // Set tokens
    request->tokens.resize(1);
    request->tokens[0] = token_name;
    // Send Request
    SrmStatus result = request->execute();
    // Check Return Value
    CHECKSRMSTATUS("Invalid Result Code",result.code,SrmStatus::SRM_SUCCESS);

    // Check file Properties
    CPPUNIT_ASSERT_MESSAGE("Invalid number of spaces returned", request->spaceDetails.size() == 1);

    const MetaDataSpace& space = request->spaceDetails.at(0);
    CHECKSRMSTATUS("Invalid space-level code",space.status.code,SrmStatus::SRM_SUCCESS);
    CHECK_VALUE(space.spaceToken,token_name);
    CHECK_VALUE(space.retentionPolicyInfo.retentionPolicy,RetentionPolicy::CUSTODIAL);
    CHECK_VALUE(space.retentionPolicyInfo.accessLatency,AccessLatency::ONLINE);
    CHECK_VALUE(space.owner,std::string("dteam"));
    CHECK_VALUE(space.totalSize,(longlong)2048);
    CHECK_VALUE(space.guaranteedSize,(longlong)1024);
    CHECK_VALUE(space.unusedSize,(longlong)512);
    CHECK_VALUE(space.lifetimeAssigned,256);
    CHECK_VALUE(space.lifetimeLeft,128);
}

#define CHECK_GETSPACEMETADATA_REQUESTFAILURE(code_string,expected){\
    std::string token_name = "this-request-returns-" + std::string(#code_string);\
    boost::shared_ptr<GetSpaceMetaData> request(GetSpaceMetaData::create(m_ctx));\
    request->tokens.resize(1);\
    request->tokens[0] = token_name;\
    SrmStatus result = request->execute();\
    CHECKSRMSTATUS("Invalid Result Code",result.code,SrmStatus::expected);\
    CPPUNIT_ASSERT_MESSAGE("Invalid number of spaces returned", request->spaceDetails.size() == 0);\
}
#define CHECK_GETSPACEMETADATA_REQUESTFAILURE_VALID(a) CHECK_GETSPACEMETADATA_REQUESTFAILURE(a,a)
/**
 * Test request failure (check valid return codes)
 */
void Srm22_TestCase::testGetSpaceMetaData_RequestFailure(){
    CHECK_SUPPORTED(GetSpaceMetaData,m_ctx,true);
    //CHECK_GETSPACEMETADATA_REQUESTFAILURE_VALID(SRM_SUCCESS);
    //CHECK_GETSPACEMETADATA_REQUESTFAILURE_VALID(SRM_PARTIAL_SUCCESS);
    CHECK_GETSPACEMETADATA_REQUESTFAILURE_VALID(SRM_AUTHENTICATION_FAILURE);
    CHECK_GETSPACEMETADATA_REQUESTFAILURE_VALID(SRM_AUTHORIZATION_FAILURE);
    CHECK_GETSPACEMETADATA_REQUESTFAILURE_VALID(SRM_TOO_MANY_RESULTS);
    CHECK_GETSPACEMETADATA_REQUESTFAILURE_VALID(SRM_INVALID_REQUEST);
    // In case of internal error, the request retries. Therefore, the client never
    // sees this eror,timeout should occur instead.
    //CHECK_GETSPACEMETADATA_REQUESTFAILURE_VALID(SRM_INTERNAL_ERROR);
    CHECK_GETSPACEMETADATA_REQUESTFAILURE_VALID(SRM_FAILURE);
    CHECK_GETSPACEMETADATA_REQUESTFAILURE_VALID(SRM_NOT_SUPPORTED);
}

#define CHECK_GETSPACEMETADATA_REQUESTFAILURE_INVALID(a) CHECK_GETSPACEMETADATA_REQUESTFAILURE(a,SRM_FAILURE)
/**
 * Test request failure (check invalid return codes)
 */
void Srm22_TestCase::testGetSpaceMetaData_InvalidRequestFailure(){
    CHECK_GETSPACEMETADATA_REQUESTFAILURE_INVALID(SRM_INVALID_PATH);
    CHECK_GETSPACEMETADATA_REQUESTFAILURE_INVALID(SRM_FILE_LIFETIME_EXPIRED);
    CHECK_GETSPACEMETADATA_REQUESTFAILURE_INVALID(SRM_SPACE_LIFETIME_EXPIRED);
    CHECK_GETSPACEMETADATA_REQUESTFAILURE_INVALID(SRM_EXCEED_ALLOCATION);
    CHECK_GETSPACEMETADATA_REQUESTFAILURE_INVALID(SRM_NO_USER_SPACE);
    CHECK_GETSPACEMETADATA_REQUESTFAILURE_INVALID(SRM_NO_FREE_SPACE);
    CHECK_GETSPACEMETADATA_REQUESTFAILURE_INVALID(SRM_DUPLICATION_ERROR);
    CHECK_GETSPACEMETADATA_REQUESTFAILURE_INVALID(SRM_NON_EMPTY_DIRECTORY);
    CHECK_GETSPACEMETADATA_REQUESTFAILURE_INVALID(SRM_FATAL_INTERNAL_ERROR);
    CHECK_GETSPACEMETADATA_REQUESTFAILURE_INVALID(SRM_REQUEST_QUEUED);
    CHECK_GETSPACEMETADATA_REQUESTFAILURE_INVALID(SRM_REQUEST_INPROGRESS);
    CHECK_GETSPACEMETADATA_REQUESTFAILURE_INVALID(SRM_REQUEST_SUSPENDED);
    CHECK_GETSPACEMETADATA_REQUESTFAILURE_INVALID(SRM_ABORTED);
    CHECK_GETSPACEMETADATA_REQUESTFAILURE_INVALID(SRM_RELEASED);
    CHECK_GETSPACEMETADATA_REQUESTFAILURE_INVALID(SRM_FILE_PINNED);
    CHECK_GETSPACEMETADATA_REQUESTFAILURE_INVALID(SRM_FILE_IN_CACHE);
    CHECK_GETSPACEMETADATA_REQUESTFAILURE_INVALID(SRM_SPACE_AVAILABLE);
    CHECK_GETSPACEMETADATA_REQUESTFAILURE_INVALID(SRM_LOWER_SPACE_GRANTED);
    CHECK_GETSPACEMETADATA_REQUESTFAILURE_INVALID(SRM_DONE);
    CHECK_GETSPACEMETADATA_REQUESTFAILURE_INVALID(SRM_REQUEST_TIMED_OUT);
    CHECK_GETSPACEMETADATA_REQUESTFAILURE_INVALID(SRM_LAST_COPY);
    CHECK_GETSPACEMETADATA_REQUESTFAILURE_INVALID(SRM_FILE_BUSY);
    CHECK_GETSPACEMETADATA_REQUESTFAILURE_INVALID(SRM_FILE_LOST);
    CHECK_GETSPACEMETADATA_REQUESTFAILURE_INVALID(SRM_FILE_UNAVAILABLE);
}

#define CHECK_GETSPACEMETADATA_TOKENFAILURE(code_string,expected){\
    std::string token_name = "this-token-returns-" + std::string(#code_string);\
    boost::shared_ptr<GetSpaceMetaData> request(GetSpaceMetaData::create(m_ctx));\
    request->tokens.resize(1);\
    request->tokens[0] = token_name;\
    SrmStatus result = request->execute();\
    CHECKSRMSTATUS("Invalid Result Code",result.code,SrmStatus::SRM_FAILURE);\
    CPPUNIT_ASSERT_MESSAGE("Invalid number of spaces returned", request->spaceDetails.size() == 1);\
    const MetaDataSpace& space = request->spaceDetails.at(0);\
    CHECKSRMSTATUS("Invalid space-level code",space.status.code,SrmStatus::expected);\
    CHECK_VALUE(space.spaceToken,token_name);\
}
#define CHECK_GETSPACEMETADATA_TOKENFAILURE_VALID(a) CHECK_GETSPACEMETADATA_TOKENFAILURE(a,a)

/**
 * Test request failure (check valid token level return codes)
 */
void Srm22_TestCase::testGetSpaceMetaData_TokenFailure(){
    CHECK_GETSPACEMETADATA_TOKENFAILURE_VALID(SRM_AUTHORIZATION_FAILURE);
    CHECK_GETSPACEMETADATA_TOKENFAILURE_VALID(SRM_INVALID_REQUEST);
    CHECK_GETSPACEMETADATA_TOKENFAILURE_VALID(SRM_SPACE_LIFETIME_EXPIRED);
    CHECK_GETSPACEMETADATA_TOKENFAILURE_VALID(SRM_EXCEED_ALLOCATION);
}

#define CHECK_GETSPACEMETADATA_TOKENFAILURE_INVALID(a) CHECK_GETSPACEMETADATA_TOKENFAILURE(a,SRM_FAILURE)
/**
 * Test request failure (check invalid token level return codes)
 */
void Srm22_TestCase::testGetSpaceMetaData_InvalidTokenFailure(){
    CHECK_GETSPACEMETADATA_TOKENFAILURE_INVALID(SRM_AUTHENTICATION_FAILURE);
    CHECK_GETSPACEMETADATA_TOKENFAILURE_INVALID(SRM_INVALID_PATH);
    CHECK_GETSPACEMETADATA_TOKENFAILURE_INVALID(SRM_FILE_LIFETIME_EXPIRED);
    CHECK_GETSPACEMETADATA_TOKENFAILURE_INVALID(SRM_NO_USER_SPACE);
    CHECK_GETSPACEMETADATA_TOKENFAILURE_INVALID(SRM_NO_FREE_SPACE);
    CHECK_GETSPACEMETADATA_TOKENFAILURE_INVALID(SRM_DUPLICATION_ERROR);
    CHECK_GETSPACEMETADATA_TOKENFAILURE_INVALID(SRM_NON_EMPTY_DIRECTORY);
    CHECK_GETSPACEMETADATA_TOKENFAILURE_INVALID(SRM_TOO_MANY_RESULTS);
    CHECK_GETSPACEMETADATA_TOKENFAILURE_INVALID(SRM_INTERNAL_ERROR);
    CHECK_GETSPACEMETADATA_TOKENFAILURE_INVALID(SRM_FATAL_INTERNAL_ERROR);
    CHECK_GETSPACEMETADATA_TOKENFAILURE_INVALID(SRM_NOT_SUPPORTED);
    CHECK_GETSPACEMETADATA_TOKENFAILURE_INVALID(SRM_REQUEST_QUEUED);
    CHECK_GETSPACEMETADATA_TOKENFAILURE_INVALID(SRM_REQUEST_INPROGRESS);
    CHECK_GETSPACEMETADATA_TOKENFAILURE_INVALID(SRM_REQUEST_SUSPENDED);
    CHECK_GETSPACEMETADATA_TOKENFAILURE_INVALID(SRM_ABORTED);
    CHECK_GETSPACEMETADATA_TOKENFAILURE_INVALID(SRM_RELEASED);
    CHECK_GETSPACEMETADATA_TOKENFAILURE_INVALID(SRM_FILE_PINNED);
    CHECK_GETSPACEMETADATA_TOKENFAILURE_INVALID(SRM_FILE_IN_CACHE);
    CHECK_GETSPACEMETADATA_TOKENFAILURE_INVALID(SRM_SPACE_AVAILABLE);
    CHECK_GETSPACEMETADATA_TOKENFAILURE_INVALID(SRM_LOWER_SPACE_GRANTED);
    CHECK_GETSPACEMETADATA_TOKENFAILURE_INVALID(SRM_DONE);
    CHECK_GETSPACEMETADATA_TOKENFAILURE_INVALID(SRM_PARTIAL_SUCCESS);
    CHECK_GETSPACEMETADATA_TOKENFAILURE_INVALID(SRM_REQUEST_TIMED_OUT);
    CHECK_GETSPACEMETADATA_TOKENFAILURE_INVALID(SRM_LAST_COPY);
    CHECK_GETSPACEMETADATA_TOKENFAILURE_INVALID(SRM_FILE_BUSY);
    CHECK_GETSPACEMETADATA_TOKENFAILURE_INVALID(SRM_FILE_LOST);
    CHECK_GETSPACEMETADATA_TOKENFAILURE_INVALID(SRM_FILE_UNAVAILABLE);
}


/*
 * Test GetSpaceTokens Method
 */
void Srm22_TestCase::testGetSpaceTokens_InvalidDescr(){

    // Create Request
    boost::shared_ptr<GetSpaceTokens> request(GetSpaceTokens::create(m_ctx));

    // Set Request Parameters
    request->description = "this is an invalid description";

    // Send Request
    SrmStatus result = request->execute();

    // Check Return Value
    CHECKSRMSTATUS("Invalid Result Code",result.code,SrmStatus::SRM_INVALID_REQUEST);

    // Check file Properties
    CPPUNIT_ASSERT_MESSAGE("Invalid number of tokens returned", request->tokens.empty() == true);

}

/*
 * Test AbortFiles Method
 */
void Srm22_TestCase::testAbortFiles(){

    // Create Ping Request
    boost::shared_ptr<PrepareToPut> request(PrepareToPut::create(m_ctx));

    // Set Request Parameters
    request->files.resize(2);
    request->files[0].surl = "srm://localhost/srm/managerv2?SFN=/test/srm2/this-one-should-stay-pending_001.txt";
    request->files[1].surl = "srm://localhost/srm/managerv2?SFN=/test/srm2/this-one-should-stay-pending_002.txt";
    request->transferParameters.protocol = "file";

    // Send Ping Request
    request->execute();

    // Check Return Value
    bool is_request_pending = false;
    if((request->status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
       (request->status.code == SrmStatus::SRM_REQUEST_INPROGRESS)){
           is_request_pending = true;
    }
    CPPUNIT_ASSERT_MESSAGE("Invalid Result Code on [execute]", is_request_pending == true);

    // Sleep for a while
    //sleep(1);

    // Keep polling for the status
    request->update();

    // Check Return Value
    if(!((request->status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
         (request->status.code == SrmStatus::SRM_REQUEST_INPROGRESS))){
        is_request_pending = false;
    }
    CPPUNIT_ASSERT_MESSAGE("Invalid Result Code on [update]", is_request_pending == true);

    // Abort one of the files
    boost::shared_ptr<AbortFiles> abort_request(AbortFiles::create(m_ctx));
    // Set Request Parameters
    abort_request->token = request->token();
    abort_request->files.resize(1);
    abort_request->files[0].surl = "srm://localhost/srm/managerv2?SFN=/test/srm2/this-one-should-stay-pending_002.txt";

    // Send AbortFiles Request
    SrmStatus abort_result = abort_request->execute();
    CHECKSRMSTATUS("Invalid Result Code on [ABORT_FILES]",abort_result.code,SrmStatus::SRM_SUCCESS);
    CHECKSRMSTATUS("Invalid File Result Code on [ABORT_FILES]",abort_request->files.at(0).status.code,SrmStatus::SRM_SUCCESS);

    // Check request status
    request->update();

    // Check Return Value
    if(!((request->status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
         (request->status.code == SrmStatus::SRM_REQUEST_INPROGRESS))){
        is_request_pending = false;
    }
    CPPUNIT_ASSERT_MESSAGE("Invalid Result Code on [update after abort]", is_request_pending == true);
    // Check Files' State
    bool is_file_pending = true;
    if(!((request->files.at(0).status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
         (request->files.at(0).status.code == SrmStatus::SRM_REQUEST_INPROGRESS))){
        is_file_pending= false;
    }
    CPPUNIT_ASSERT_MESSAGE("Invalid File #0 Code on [update after abort]", is_file_pending == true);
    if(!((request->files.at(1).status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
         (request->files.at(1).status.code == SrmStatus::SRM_REQUEST_INPROGRESS))){
        is_file_pending= false;
    }
    CPPUNIT_ASSERT_MESSAGE("Invalid File #1 Code on [update after abort]", is_file_pending == false);
    CHECKSRMSTATUS("Invalid File #1 Status [after abort files]",request->files.at(1).status.code,SrmStatus::SRM_ABORTED);

    // Abort the request
    request->abort();
    // Check the Request status
    CHECKSRMSTATUS("Invalid Result Code on [ABORT]",request->status.code,SrmStatus::SRM_ABORTED);
    // Check the File status
    CHECKSRMSTATUS("Invalid File Status #0",request->files[0].status.code,SrmStatus::SRM_ABORTED);
    CHECKSRMSTATUS("Invalid File Status #1",request->files[1].status.code,SrmStatus::SRM_ABORTED);

}

/*
 * Test AbortFiles Method
 */
void Srm22_TestCase::testAbortFiles_AlreadyAborted(){

    // Create Ping Request
    boost::shared_ptr<PrepareToPut> request(PrepareToPut::create(m_ctx));

    // Set Request Parameters
    request->files.resize(1);
    request->files[0].surl = "srm://localhost/srm/managerv2?SFN=/test/srm2/this-one-should-stay-pending_001.txt";
    request->transferParameters.protocol = "file";

    // Send Ping Request
    request->execute();

    // Check Return Value
    bool is_request_pending = false;
    if((request->status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
       (request->status.code == SrmStatus::SRM_REQUEST_INPROGRESS)){
           is_request_pending = true;
    }
    CPPUNIT_ASSERT_MESSAGE("Invalid Result Code on [execute]", is_request_pending == true);

    // Sleep for a while
    //sleep(1);

    // Keep polling for the status
    request->update();

    // Check Return Value
    if(!((request->status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
         (request->status.code == SrmStatus::SRM_REQUEST_INPROGRESS))){
        is_request_pending = false;
    }
    CPPUNIT_ASSERT_MESSAGE("Invalid Result Code on [update]", is_request_pending == true);

    // Abort the request
    request->abort();
    // Check the Request status
    CHECKSRMSTATUS("Invalid Result Code on [ABORT]",request->status.code,SrmStatus::SRM_ABORTED);
    // Check the File status
    CHECKSRMSTATUS("Invalid File Status #0",request->files[0].status.code,SrmStatus::SRM_ABORTED);

    // Abort one of the files
    boost::shared_ptr<AbortFiles> abort_request(AbortFiles::create(m_ctx));
    // Set Request Parameters
    abort_request->token = request->token();
    abort_request->files.resize(1);
    abort_request->files[0].surl = "srm://localhost/srm/managerv2?SFN=/test/srm2/this-one-should-stay-pending_001.txt";

    // Send AbortFiles Request
    SrmStatus abort_result = abort_request->execute();
    CHECKSRMSTATUS("Invalid Result Code on [ABORT_FILES]",abort_result.code,SrmStatus::SRM_SUCCESS);
    CHECKSRMSTATUS("Invalid File Result Code on [ABORT_FILES]",abort_request->files.at(0).status.code,SrmStatus::SRM_SUCCESS);

}

/*
 * Test AbortFiles Method
 */
void Srm22_TestCase::testAbortFiles_ErrorNoFile(){

    // Create Ping Request
    boost::shared_ptr<PrepareToPut> request(PrepareToPut::create(m_ctx));

    // Set Request Parameters
    request->files.resize(1);
    request->files[0].surl = "srm://localhost/srm/managerv2?SFN=/test/srm1/this-one-should-stay-pending_001.txt";
    request->transferParameters.protocol = "file";

    // Send Ping Request
    request->execute();

    // Check Return Value
    bool is_request_pending = false;
    if((request->status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
       (request->status.code == SrmStatus::SRM_REQUEST_INPROGRESS)){
           is_request_pending = true;
    }
    CPPUNIT_ASSERT_MESSAGE("Invalid Result Code on [execute]", is_request_pending == true);

    // Sleep for a while
    //sleep(1);

    // Keep polling for the status
    request->update();

    // Check Return Value
    if(!((request->status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
         (request->status.code == SrmStatus::SRM_REQUEST_INPROGRESS))){
        is_request_pending = false;
    }
    CPPUNIT_ASSERT_MESSAGE("Invalid Result Code on [update]", is_request_pending == true);

    // Abort the request
    request->abort();
    // Check the Request status
    CHECKSRMSTATUS("Invalid Result Code on [ABORT]",request->status.code,SrmStatus::SRM_ABORTED);
    // Check the File status
    CHECKSRMSTATUS("Invalid File Status #0",request->files[0].status.code,SrmStatus::SRM_ABORTED);

    // Abort one of the files
    boost::shared_ptr<AbortFiles> abort_request(AbortFiles::create(m_ctx));
    // Set Request Parameters
    abort_request->token = request->token();
    abort_request->files.resize(1);
    abort_request->files[0].surl = "srm://localhost/srm/managerv2?SFN=/test/srm2/this-one-should-stay-pending_666.txt";

    // Send AbortFiles Request
    SrmStatus abort_result = abort_request->execute();
    CHECKSRMSTATUS("Invalid Result Code on [ABORT_FILES]",abort_result.code,SrmStatus::SRM_FAILURE);
    CHECKSRMSTATUS("Invalid File Result Code on [ABORT_FILES]",abort_request->files.at(0).status.code,SrmStatus::SRM_INVALID_PATH);

}


/*
 * Test Stat Method
 */
void Srm22_TestCase::testStat_AfterCopy(){

    // Create Request
    boost::shared_ptr<SrmStat> request(SrmStat::create(m_ctx));

    // Set Request Parameters
    request->files.resize(1);
    request->files[0].surl = "srm://localhost/srm/managerv2?SFN=/test/srm2/file002.txt";

    // Send Request
    SrmStatus result = request->execute();

    // Check Return Value
    CHECKSRMSTATUS("Invalid Result Code",result.code,SrmStatus::SRM_SUCCESS);

    // Check file Properties
    CPPUNIT_ASSERT_MESSAGE("Invalid SURL",   request->files[0].surl == "srm://localhost/srm/managerv2?SFN=/test/srm2/file002.txt");
    CPPUNIT_ASSERT_MESSAGE("Invalid Status", (request->files[0].status.code == SrmStatus::SRM_SUCCESS) || (request->files[0].status.code == SrmStatus::SRM_UNDEF));
    CPPUNIT_ASSERT_MESSAGE("Null Size"   , request->files[0].size > 0);
    CPPUNIT_ASSERT_MESSAGE("Invalid Size", (size_t)request->files[0].size == strlen(SAMPLE_TEXT));
    CPPUNIT_ASSERT_MESSAGE("Invalid File Locality", request->files[0].fileLocality == FileLocality::ONLINE);
    CPPUNIT_ASSERT_MESSAGE("Invalid File Type", request->files[0].fileType == (int)FileType::FILE);

}
