/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004.
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  GLite Data SRM Util CPP: Common base of the SRM Util CPOP tests.
 *
 *  Authors: Zsolt Molnar <zsolt.molnar@cern.ch>
 *  Version info: $Id: TestBase.cpp,v 1.3 2009/07/30 17:40:34 molnarzs Exp $
 *  Release: $Name:  $
 *
 */

#include "glite/data/srm/util/SrmExceptions.h"
#include "glite/data/srm/util/AbortFiles.h"
#include "glite/data/agents/UserProxyEnv.h"
#include "glite/data/srm/util/PrepareToPut.h"
#include "glite/data/srm/util/SrmRm.h"
#include "glite/data/srm/util/SrmMkDir.h"
#include "glite/data/srm/util/SrmRmDir.h"

#include <boost/scoped_ptr.hpp>

#include "TestBase.h"

const char SAMPLE_TEXT[] ="Hello World! This is a test\nWrite some data for SRM"
	"Util tests\n";

void TestBase::setUp()
{
    make_dir(_testBasePath(), true);
    make_dir(testPath(), true);
    _logic.reset(dynamic_cast<glite::data::srm::util::details::IBackoffLogic*>(new ExponentialBackoffLogic<AlgorithmConstants>()));
}

void TestBase::tearDown()
{
	remove_dir(testPath(), true);
	remove_dir(_testBasePath());
	m_ctx.reset();
}

void TestBase::put_file(const std::string& surl, bool overWrite)
{
    // Create Ping Request
    boost::shared_ptr<PrepareToPut> request(PrepareToPut::create(m_ctx));

    // Set Request Parameters
    request->files.resize(1);

    if (overWrite)
    	request->overwriteOption = OverwriteMode::ALWAYS;

    request->files[0].surl = surl;
    request->transferParameters.protocol = "file";

    // Send Ping Request
    request->execute();

    // Check Return Value
    bool is_request_failed = true;
    if((request->status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
       (request->status.code == SrmStatus::SRM_REQUEST_INPROGRESS) ||
       (request->status.code == SrmStatus::SRM_SUCCESS)){
           is_request_failed = false;
    }
    CPPUNIT_ASSERT_MESSAGE("Invalid Result Code on [execute]", is_request_failed == false);

    while((request->status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
          (request->status.code == SrmStatus::SRM_REQUEST_INPROGRESS)){
        // Sleep for a while
        //sleep(1);

        // Keep polling for the status
        request->update();

        // Check Return Value
        if((request->status.code == SrmStatus::SRM_REQUEST_QUEUED) ||
           (request->status.code == SrmStatus::SRM_REQUEST_INPROGRESS) ||
           (request->status.code == SrmStatus::SRM_SUCCESS)){
               is_request_failed = false;
        }
        CPPUNIT_ASSERT_MESSAGE("Invalid Result Code on [update]", is_request_failed == false);
    }
    CHECKSRMSTATUS("Result Code should now be SUCCESS",request->status.code,SrmStatus::SRM_SUCCESS);

    if(request->status.code == SrmStatus::SRM_SUCCESS){
        // Check that the file properties are set correctly
        CHECKSRMSTATUS("Invalid File Status on [update]",request->files[0].status.code,SrmStatus::SRM_SPACE_AVAILABLE);
        CPPUNIT_ASSERT_MESSAGE("No turl retrieved", request->files[0].turl.empty() == false);
    }

    // Open the file
    if(0 == strncmp(request->files[0].turl.c_str(),"file://",7)){
        std::string path = request->files[0].turl.substr(7);
        int fd = open(path.c_str(),O_RDWR);
        CPPUNIT_ASSERT_MESSAGE("Cannot Open file", fd != -1);
        write(fd,SAMPLE_TEXT,strlen(SAMPLE_TEXT));
        close(fd);
    } else {
       CPPUNIT_FAIL("Invalid protocol for TURL: Not [file]");
    }

    // Complete the request
    request->done();
    // Check the Request status
    CHECKSRMSTATUS("Invalid Result Code on [done]",request->status.code,SrmStatus::SRM_SUCCESS);
    // Check the File status
    CHECKSRMSTATUS("Invalid File Status on [done]",request->files[0].status.code,SrmStatus::SRM_SUCCESS);
}

void TestBase::remove_file(const std::string& surl)
{
    // Create Request
    boost::shared_ptr<SrmRm> request(SrmRm::create(m_ctx));
    // Set Request Parameters
    request->files.resize(1);
    request->files[0].surl = surl;
    // Send Request
    SrmStatus result = request->execute();

   	CHECKSRMSTATUS(std::string("Invalid Result Code (") + surl + ")" ,
   			       result.code,SrmStatus::SRM_SUCCESS);
   	CHECKSRMSTATUS("Invalid Status",request->files[0].status.code,SrmStatus::SRM_SUCCESS);
}

void TestBase::check_supported(Context::Ptr ctx, SrmVersion v)
{
    // Check supported methods
    CHECK_SUPPORTED(PrepareToPut, ctx, true);
    CHECK_SUPPORTED(SrmRm, ctx, true);
    CHECK_SUPPORTED(SrmMkDir, ctx, v != v1_1);
    CHECK_SUPPORTED(SrmRmDir, ctx, true);
}

void TestBase::make_dir(const std::string& surl, bool ignoreIfExists)
{
    boost::shared_ptr<SrmMkDir> request(SrmMkDir::create(m_ctx));
    request->directoryPath = surl;
    SrmStatus result = request->execute();

    if (!(ignoreIfExists && result.code == SrmStatus::SRM_DUPLICATION_ERROR))
    	CHECKSRMSTATUS(std::string("Invalid Result Code in SrmMkDir") + surl + ")",
    			       result.code, SrmStatus::SRM_SUCCESS);
}

void TestBase::remove_dir(const std::string& surl, bool recursive, bool ignoreIfNotExists)
{
	boost::shared_ptr<SrmRmDir> request(SrmRmDir::create(m_ctx));
	request->directoryPath = surl;
	request->recursive = recursive;
	SrmStatus result = request->execute();

	if (!(ignoreIfNotExists && result.code == SrmStatus::SRM_INVALID_PATH))
		CHECKSRMSTATUS(std::string("Invalid Result Code in SrmRmDir (") + surl + ")",
				       result.code, SrmStatus::SRM_SUCCESS);
}

const std::string& TestBase::testPath()
{
	static const std::string s(_testBasePath() + _testSpecificPath() + "/");
	return s;
}

const std::string& TestBase::_testBasePath()
{
	static std::string s(_endPointRoot() + "?SFN=/test/");
	return s;
}
