/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004.
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  GLite Data SRM Util CPP: SRM 2.2 Test Case
 *
 *  Authors: Zsolt Molnar <zsolt.molnar@cern.ch>
 *  Version info: $Id: TestExponentialBackoffLogic.cpp,v 1.2 2009/07/30 17:40:34 molnarzs Exp $
 *  Release: $Name:  $
 *
 */

#include "TestExponentialBackoffLogic.h"
#include "../src/BackoffLogicExecutor.h"
#include <iostream>
using namespace glite::data::srm::util;

void TestExponentialBackoffLogic::setUp()
{
	std::cout << std::endl;
	_logic.reset(new ExponentialBackoffLogic<AlgorithmConstants>());
	_c = AlgorithmConstants::c();
	_base = AlgorithmConstants::base();
}

void TestExponentialBackoffLogic::tearDown()
{
	// EMPTY
}

void TestExponentialBackoffLogic::testSimpleTimeoutNoRandomness()
{
	CPPUNIT_ASSERT(!_logic->timeout());

	for (int i = 1; i <= 3; ++i) {
		double expected = _c * (std::pow(_base, static_cast<double>(i)) - 1);
		std::cout << "\tWaiting for " << expected << " seconds..." << std::endl;
		boost::posix_time::ptime start(boost::posix_time::second_clock::universal_time());
		_logic->sleep();
		CPPUNIT_ASSERT_EQUAL(static_cast<long>(expected), _elapsed(start));
		CPPUNIT_ASSERT(!_logic->timeout());
	}
	// This should force timeout
	boost::posix_time::ptime start(boost::posix_time::second_clock::universal_time());
	_logic->sleep();
	CPPUNIT_ASSERT(_logic->timeout());
	// The sleep would anyway exceed the global timeout, so it must return immediately
	CPPUNIT_ASSERT_EQUAL((long)0, _elapsed(start));
}

void TestExponentialBackoffLogic::testEWTNoRandomness()
{
	CPPUNIT_ASSERT(!_logic->timeout());

	// Start with exponential timeout
	for (int i = 1; i <= 2; ++i) {
		double expected = _c * (std::pow(_base, static_cast<double>(i)) - 1);
		std::cout << "\tWaiting for " << expected << " seconds..." << std::endl;
		boost::posix_time::ptime start(boost::posix_time::second_clock::universal_time());
		_logic->sleep();
		CPPUNIT_ASSERT_EQUAL(static_cast<long>(expected), _elapsed(start));
		CPPUNIT_ASSERT(!_logic->timeout());
	}

	// Now, set eWT
	int eWT = 2;
	_logic->setSleepTime(eWT);
	boost::posix_time::ptime start_1(boost::posix_time::second_clock::universal_time());
	CPPUNIT_ASSERT(!_logic->timeout());
	std::cout << "\tWaiting for " << eWT << " seconds..." << std::endl;
	_logic->sleep();
	CPPUNIT_ASSERT_EQUAL(static_cast<long>(eWT), _elapsed(start_1));
	CPPUNIT_ASSERT(!_logic->timeout());

	// Check if the previous operation reset the exponential logic
	for (int i = 1; i <= 2; ++i) {
		double expected = _c * (std::pow(_base, static_cast<double>(i)) - 1);
		std::cout << "\tWaiting for " << expected << " seconds..." << std::endl;
		boost::posix_time::ptime start(boost::posix_time::second_clock::universal_time());
		_logic->sleep();
		CPPUNIT_ASSERT_EQUAL(static_cast<long>(expected), _elapsed(start));
		CPPUNIT_ASSERT(!_logic->timeout());
	}
}

void TestExponentialBackoffLogic::testReset()
{
	CPPUNIT_ASSERT(!_logic->timeout());

	for (int i = 1; i <= 2; ++i) {
		double expected = _c * (std::pow(_base, static_cast<double>(i)) - 1);
		std::cout << "\tWaiting for " << expected << " seconds..." << std::endl;
		boost::posix_time::ptime start(boost::posix_time::second_clock::universal_time());
		_logic->sleep();
		CPPUNIT_ASSERT_EQUAL(static_cast<long>(expected), _elapsed(start));
		CPPUNIT_ASSERT(!_logic->timeout());
	}

	_logic->reset();

	for (int i = 1; i <= 2; ++i) {
		double expected = _c * (std::pow(_base, static_cast<double>(i)) - 1);
		std::cout << "\tWaiting for " << expected << " seconds..." << std::endl;
		boost::posix_time::ptime start(boost::posix_time::second_clock::universal_time());
		_logic->sleep();
		CPPUNIT_ASSERT_EQUAL(static_cast<long>(expected), _elapsed(start));
		CPPUNIT_ASSERT(!_logic->timeout());
	}

	// Test reset after setting eWT
	_logic->setSleepTime(9);
	_logic->reset();

	for (int i = 1; i <= 2; ++i) {
		double expected = _c * (std::pow(_base, static_cast<double>(i)) - 1);
		std::cout << "\tWaiting for " << expected << " seconds..." << std::endl;
		boost::posix_time::ptime start(boost::posix_time::second_clock::universal_time());
		_logic->sleep();
		CPPUNIT_ASSERT_EQUAL(static_cast<long>(expected), _elapsed(start));
		CPPUNIT_ASSERT(!_logic->timeout());
	}
}

void TestExponentialBackoffLogic::testSyncExecutorNoTimeoutNoEWT()
{
	MockRequest request(3, 0);
	glite::data::srm::util::SrmStatus result;
	// The exp. wait times are: 1 + 3 + 7 = 11 sec altogether
	long expected = 11;
	std::cout << "\tWaiting for " << expected << " seconds..." << std::endl;
	boost::posix_time::ptime start(boost::posix_time::second_clock::universal_time());
	result = BackoffLogicExecutor::executeSync(request);
	CPPUNIT_ASSERT_EQUAL(expected, _elapsed(start));
	CPPUNIT_ASSERT_EQUAL(SrmStatus::SRM_SUCCESS, result.code);
}

void TestExponentialBackoffLogic::testAsyncExecutorNoTimeoutNoEWT()
{
	MockRequest request(3, 0);
	// The exp. wait times are: 1 + 3 + 7 = 11 sec altogether
	long expected = 11;
	std::cout << "\tWaiting for " << expected << " seconds..." << std::endl;
	boost::posix_time::ptime start(boost::posix_time::second_clock::universal_time());
	BackoffLogicExecutor::executeAsync(request);
	CPPUNIT_ASSERT_EQUAL(expected, _elapsed(start));
	CPPUNIT_ASSERT_EQUAL(SrmStatus::SRM_SUCCESS, request.status.code);
	CPPUNIT_ASSERT(!request.isDoneCalled);
}

void TestExponentialBackoffLogic::testSyncExecutorNoTimeoutWithEWT()
{
	MockRequestWithEWT request(3, 0);
	glite::data::srm::util::SrmStatus result;
	// The exp. wait times are: 2 + 2 + 2 = 6 sec altogether (2 sec is the EWT)
	long expected = 6;
	std::cout << "\tWaiting for " << expected << " seconds..." << std::endl;
	boost::posix_time::ptime start(boost::posix_time::second_clock::universal_time());
	result = BackoffLogicExecutor::executeSync(request);
	CPPUNIT_ASSERT_EQUAL(expected, _elapsed(start));
	CPPUNIT_ASSERT_EQUAL(SrmStatus::SRM_SUCCESS, result.code);
}

void TestExponentialBackoffLogic::testAsyncExecutorNoTimeoutWithEWT()
{
	MockRequestWithEWT request(3, 0);
	// The exp. wait times are: 2 + 2 + 2 = 6 sec altogether (2 sec is the EWT)
	long expected = 6;
	std::cout << "\tWaiting for " << expected << " seconds..." << std::endl;
	boost::posix_time::ptime start(boost::posix_time::second_clock::universal_time());
	BackoffLogicExecutor::executeAsync(request);
	CPPUNIT_ASSERT_EQUAL(expected, _elapsed(start));
	CPPUNIT_ASSERT_EQUAL(SrmStatus::SRM_SUCCESS, request.status.code);
	CPPUNIT_ASSERT(!request.isDoneCalled);
}

void TestExponentialBackoffLogic::testAsyncExecutorNoTimeoutNoEWTRequestQueued()
{
	MockRequest request(1, 2);
	// The exp. wait times are: 1 + 3 + 7 = 11 sec altogether
	long expected = 11;
	std::cout << "\tWaiting for " << expected << " seconds..." << std::endl;
	boost::posix_time::ptime start(boost::posix_time::second_clock::universal_time());
	BackoffLogicExecutor::executeAsync(request);
	BackoffLogicExecutor::updateAsyncUntilQueued(request);
	CPPUNIT_ASSERT_EQUAL(expected, _elapsed(start));
	CPPUNIT_ASSERT_EQUAL(SrmStatus::SRM_SUCCESS, request.status.code);
}

void TestExponentialBackoffLogic::testAsyncExecutorNoTimeoutWithEWTRequestQueued()
{
	MockRequestWithEWT request(1, 2);
	// The exp. wait times are: 2 + 2 + 2 = 6 sec altogether (2 sec is the EWT)
	long expected = 6;
	std::cout << "\tWaiting for " << expected << " seconds..." << std::endl;
	boost::posix_time::ptime start(boost::posix_time::second_clock::universal_time());
	BackoffLogicExecutor::executeAsync(request);
	BackoffLogicExecutor::updateAsyncUntilQueued(request);
	CPPUNIT_ASSERT_EQUAL(expected, _elapsed(start));
	CPPUNIT_ASSERT_EQUAL(SrmStatus::SRM_SUCCESS, request.status.code);
}

void TestExponentialBackoffLogic::testMaxSleepTime()
{
	MockRequestWithEWT request(0, 1, 20);
	long expected = 10;
	std::cout << "\tWaiting for " << expected << " seconds..." << std::endl;
	boost::posix_time::ptime start(boost::posix_time::second_clock::universal_time());
	BackoffLogicExecutor::executeAsync(request);
	BackoffLogicExecutor::updateAsyncUntilQueued(request);
	CPPUNIT_ASSERT_EQUAL(expected, _elapsed(start));
	CPPUNIT_ASSERT_EQUAL(SrmStatus::SRM_SUCCESS, request.status.code);
}

void TestExponentialBackoffLogic::testNoBackoffIfTimeoutWouldOccur_Sync()
{
	MockRequest request(4, 0);
	// The exp. wait times are: 1 + 3 + 7 = 11 sec altogether
	long expected = 11;
	std::cout << "\tWaiting for " << expected << " seconds..." << std::endl;
	boost::posix_time::ptime start(boost::posix_time::second_clock::universal_time());
	glite::data::srm::util::SrmStatus result =
		BackoffLogicExecutor::executeSync(request);
	CPPUNIT_ASSERT_EQUAL(expected, _elapsed(start));
	CPPUNIT_ASSERT_EQUAL(SrmStatus::SRM_REQUEST_TIMED_OUT, result.code);
}

void TestExponentialBackoffLogic::testNoBackoffIfTimeoutWouldOccur_Async()
{
	MockRequest request(2, 3);
	// The exp. wait times are: 1 + 3 + 7 = 11 sec altogether
	long expected = 11;
	std::cout << "\tWaiting for " << expected << " seconds..." << std::endl;
	boost::posix_time::ptime start(boost::posix_time::second_clock::universal_time());
	BackoffLogicExecutor::executeAsync(request);
	BackoffLogicExecutor::updateAsyncUntilQueued(request);
	CPPUNIT_ASSERT_EQUAL(expected, _elapsed(start));
	CPPUNIT_ASSERT_EQUAL(SrmStatus::SRM_REQUEST_TIMED_OUT, request.status.code);
	CPPUNIT_ASSERT(request.isAbortCalled);
}

void TestExponentialBackoffLogic::testNoBackoffIfTimeoutWouldOccur_eWT()
{
	MockRequestWithEWT request(1, 3, 5);
	// The exp. wait times are: 5 + 5 = 10 < 12 sec altogether
	long expected = 10;
	std::cout << "\tWaiting for " << expected << " seconds..." << std::endl;
	boost::posix_time::ptime start(boost::posix_time::second_clock::universal_time());
	BackoffLogicExecutor::executeAsync(request);
	BackoffLogicExecutor::updateAsyncUntilQueued(request);
	CPPUNIT_ASSERT_EQUAL(expected, _elapsed(start));
	CPPUNIT_ASSERT_EQUAL(SrmStatus::SRM_REQUEST_TIMED_OUT, request.status.code);
	CPPUNIT_ASSERT(request.isAbortCalled);
}

long TestExponentialBackoffLogic::_elapsed(const boost::posix_time::ptime& startTime) const
{
	boost::posix_time::ptime now(boost::posix_time::second_clock::universal_time());
	boost::posix_time::time_period elapsed(startTime, now);
	return elapsed.length().total_seconds();
}

TestExponentialBackoffLogic::MockRequest::MockRequest(const int intErr, const int queued)
	: MockRequest::_BASETYPE(*this, NULL),
	  isAbortCalled(false),
	  isDoneCalled(false),
	  _intErr(intErr),
	  _queued(queued),
	  _logic(new ExponentialBackoffLogic<AlgorithmConstants>)
{
	// EMPTY
}

glite::data::srm::util::SrmStatus TestExponentialBackoffLogic::MockRequest::executeOneCall()
{
	if (_intErr) {
		--_intErr;
		status.code = SrmStatus::SRM_INTERNAL_ERROR;
	} else if (_queued) {
		--_queued;
		status.code = (_queued / 2) ? SrmStatus::SRM_REQUEST_QUEUED : SrmStatus::SRM_REQUEST_INPROGRESS;
	} else {
		status.code = SrmStatus::SRM_SUCCESS;
	}

	return status;
}

void TestExponentialBackoffLogic::MockRequest::abort()
{
	isAbortCalled = true;
}

void TestExponentialBackoffLogic::MockRequest::done()
{
	isDoneCalled = true;
}

void TestExponentialBackoffLogic::MockRequest::update()
{
	if (_queued) {
		--_queued;
		status.code = (_queued / 2) ? SrmStatus::SRM_REQUEST_QUEUED : SrmStatus::SRM_REQUEST_INPROGRESS;
	} else {
		status.code = SrmStatus::SRM_SUCCESS;
	}
}

IBackoffLogic::Ptr TestExponentialBackoffLogic::MockRequest::getBackoffLogic()
{
	return _logic;
}

int TestExponentialBackoffLogic::MockRequest::estimatedWaitTime() const
{
	return -1;
}

TestExponentialBackoffLogic::MockRequestWithEWT::MockRequestWithEWT(
	const int intErr, const int queued, const unsigned int eWT
)
	: TestExponentialBackoffLogic::MockRequest(intErr, queued),
	  _eWT(eWT)
{
	// EMPTY
}

int TestExponentialBackoffLogic::MockRequestWithEWT::estimatedWaitTime() const
{
	return _eWT;
}

void TestExponentialBackoffLogic::testElapsedCanDetectLongTimeout_65664()
{
    CPPUNIT_ASSERT(_logic->_elapsed() < 60);
    std::cout << "Sleeping for 70 sec..." << std::endl;
    sleep(70);

    std::cout << "ELAPSED: " << _logic->_elapsed() << std::endl;
    CPPUNIT_ASSERT(_logic->_elapsed() >= 70);
}

void TestExponentialBackoffLogic::testRequestTimeoutInLogic_67669()
{
    CPPUNIT_ASSERT(!_logic->timeout());     
    _logic->forceTimeout();
    CPPUNIT_ASSERT(_logic->timeout());
}


void TestExponentialBackoffLogic::testRequestTimeoutAsync_67669()
{
    MockRequestWithEWT request(4, 1, 0);
    long expected = 0;
    boost::posix_time::ptime start(boost::posix_time::second_clock::universal_time());  
    glite::data::srm::util::SrmStatus result = BackoffLogicExecutor::executeAsync(request);
    CPPUNIT_ASSERT_EQUAL(expected, _elapsed(start));
    CPPUNIT_ASSERT_EQUAL(SrmStatus::SRM_REQUEST_TIMED_OUT, result.code);
}

void TestExponentialBackoffLogic::testRequestTimeoutSync_67669()
{
    MockRequestWithEWT request(4, 1, 0);
    long expected = 0;
    boost::posix_time::ptime start(boost::posix_time::second_clock::universal_time());  
	glite::data::srm::util::SrmStatus result = BackoffLogicExecutor::executeSync(request);
    CPPUNIT_ASSERT_EQUAL(expected, _elapsed(start));
    CPPUNIT_ASSERT_EQUAL(SrmStatus::SRM_REQUEST_TIMED_OUT, result.code);
}

void TestExponentialBackoffLogic::testRequestTimeoutBackoff_67669()
{
    MockRequestWithEWT request(4, 1, 0);
    long expected = 0;
    boost::posix_time::ptime start(boost::posix_time::second_clock::universal_time());  
    glite::data::srm::util::SrmStatus result = BackoffLogicExecutor::updateAsyncUntilQueued(request);
    CPPUNIT_ASSERT_EQUAL(expected, _elapsed(start));
    CPPUNIT_ASSERT_EQUAL(SrmStatus::SRM_REQUEST_TIMED_OUT, result.code);
}

