/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004.
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  GLite Data SRM Util CPP: SRM 2.2 Test Case
 *
 *  Authors: Zsolt Molnar <zsolt.molnar@cern.ch>
 *  Version info: $Id: TestExponentialBackoffLogic.h,v 1.2 2009/07/30 17:40:34 molnarzs Exp $
 *  Release: $Name:  $
 *
 */

#ifndef GLITE_DATA_SRM_UTIL_CPP_TEST_EXPONENTIAL_BACKOFF_LOGIC_H_
#define GLITE_DATA_SRM_UTIL_CPP_TEST_EXPONENTIAL_BACKOFF_LOGIC_H_

#include "../src/ExponentialBackoffLogic.h"
#include "../src/SrmAsyncOperation.h"
#include "cppunit_wrapper.h"
#include "glite/data/srm/util/types.h"

#include <boost/date_time/posix_time/posix_time.hpp>
#include <boost/scoped_ptr.hpp>

using namespace glite::data::srm::util;

/**
 * Test for Savannah bug #44018: FTA RFE: should not retry on overloaded SRM service
 *
 * Test the ExponentialBackoffLogic class. Th etest does not require web service,
 * so in case of executing only this test class, the test executable can be called
 * with the -n flag.
 */
class TestExponentialBackoffLogic : public CppUnit::TestCase
{
public:
    void setUp ();
    void tearDown ();

private:
    CPPUNIT_TEST_SUITE( TestExponentialBackoffLogic );
    CPPUNIT_TEST( testSimpleTimeoutNoRandomness );
    CPPUNIT_TEST( testEWTNoRandomness );
    CPPUNIT_TEST( testReset );
    CPPUNIT_TEST( testSyncExecutorNoTimeoutNoEWT );
    CPPUNIT_TEST( testSyncExecutorNoTimeoutWithEWT );
    CPPUNIT_TEST( testAsyncExecutorNoTimeoutNoEWT );
    CPPUNIT_TEST( testAsyncExecutorNoTimeoutWithEWT );
    CPPUNIT_TEST( testAsyncExecutorNoTimeoutNoEWTRequestQueued );
    CPPUNIT_TEST( testAsyncExecutorNoTimeoutWithEWTRequestQueued );
    CPPUNIT_TEST( testMaxSleepTime );
    CPPUNIT_TEST( testNoBackoffIfTimeoutWouldOccur_Sync );
    CPPUNIT_TEST( testNoBackoffIfTimeoutWouldOccur_Async );
    CPPUNIT_TEST( testNoBackoffIfTimeoutWouldOccur_eWT );
    CPPUNIT_TEST( testElapsedCanDetectLongTimeout_65664 );
    CPPUNIT_TEST(testRequestTimeoutInLogic_67669);
    CPPUNIT_TEST(testRequestTimeoutAsync_67669);
    CPPUNIT_TEST(testRequestTimeoutSync_67669);
    CPPUNIT_TEST(testRequestTimeoutBackoff_67669);
	CPPUNIT_TEST_SUITE_END();

	void testSimpleTimeoutNoRandomness();
	void testEWTNoRandomness();
	void testReset();
	void testSyncExecutorNoTimeoutNoEWT();
	void testSyncExecutorNoTimeoutWithEWT();
	void testAsyncExecutorNoTimeoutNoEWT();
	void testAsyncExecutorNoTimeoutWithEWT();
	void testAsyncExecutorNoTimeoutNoEWTRequestQueued();
	void testAsyncExecutorNoTimeoutWithEWTRequestQueued();
	void testMaxSleepTime();
	void testNoBackoffIfTimeoutWouldOccur_Sync();
	void testNoBackoffIfTimeoutWouldOccur_Async();
	void testNoBackoffIfTimeoutWouldOccur_eWT();
    void testRequestTimeoutInLogic_67669();
    void testRequestTimeoutAsync_67669();
    void testRequestTimeoutSync_67669();
    void testRequestTimeoutBackoff_67669();
    /**
     * See: https://savannah.cern.ch/bugs/index.php?65664
     * FTA was looping when the timeout was longer than 60 sec.
     */
    void testElapsedCanDetectLongTimeout_65664();

	long _elapsed(const boost::posix_time::ptime& startTime) const;

	struct AlgorithmConstants
	{
		static double c() { return 1;  /* sec */ }
		static double timeout() { return 12;  /* sec */}
		static double base() { return 2; }
		static double maxSleepTime() { return 10; }
	};

	struct Context
	{
		typedef void* Ptr;
	};

	struct MockRequestBase
	{
		MockRequestBase() {}
		MockRequestBase(Context::Ptr&) {}
		MockRequestBase(Context::Ptr, std::string) {}
	};

	class MockRequest : public SrmAsyncOperation<MockRequest, MockRequestBase, Context>
	{
		typedef SrmAsyncOperation<MockRequest, MockRequestBase, Context> _BASETYPE;
	public:
		MockRequest(const int intErr, const int queued);

		glite::data::srm::util::SrmStatus executeOneCall();
		void abort();
		void done();
		void update();
		IBackoffLogic::Ptr getBackoffLogic();
		int estimatedWaitTime() const;

		glite::data::srm::util::SrmStatus status;
		bool isAbortCalled;
		bool isDoneCalled;

	private:
		int _intErr;
		int _queued;
		IBackoffLogic::Ptr _logic;
	};

	class MockRequestWithEWT : public MockRequest
	{
	public:
		MockRequestWithEWT(const int intErr, const int queued, const unsigned int eWT = 2);
		int estimatedWaitTime() const;

	private:
		unsigned int _eWT;
	};

	double _c;
	double _base;
	boost::scoped_ptr<ExponentialBackoffLogic<AlgorithmConstants> > _logic;
};

#endif // #ifndef GLITE_DATA_SRM_UTIL_CPP_TEST_EXPONENTIAL_BACKOFF_LOGIC_H_

