/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004.
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  GLite Data SRM Util CPP: SRM 2.2 Test Case
 *
 *  Authors: Zsolt Molnar <zsolt.molnar@cern.ch>
 *  Version info: $Id: TestSrmBusy.h,v 1.2 2009/07/30 17:40:34 molnarzs Exp $
 *  Release: $Name:  $
 *
 */

#ifndef GLITE_DATA_SRM_UTIL_CPP_TESTSRMBUSY_H_
#define GLITE_DATA_SRM_UTIL_CPP_TESTSRMBUSY_H_

#include "TestBase_v2.h"
#include "glite/data/srm/util/types.h"

using namespace glite::data::srm::util;

/**
 * Test for Savannah bug #44018: FTA RFE: should not retry on overloaded SRM service
 *
 * General comments:
 * - The test is executed with SRM_SUCCESS return values only, because the tested logic
 *   does not depend on the final return status, it is passed as is.
 *
 * - File name rule:
 *
 *   srm_busy_SRM_INTERNAL_ERROR_<c>_<nIE>_SRM_REQUEST_QUEUED_<nRQ>_<eWT>.dat
 *
 *   - nIE: number of internal errors that should be returned
 *   - nRQ: number of SRM_REQUEST_QUEUED returned, after the nIE returned SRM_INTERNAL_ERROR
 *   - c: the wait times between polls are c * 2 ^ i - 1, where 0 <= i < nIE + nRQ
 *   - eWT: estimated wait time.
 *
 * - No real file operations are performed on the server side, to simplify the execution!
 */
class TestSrmBusy : public TestBase_v2
{
public:
    void setUp ();
    void tearDown ();

private:
    CPPUNIT_TEST_SUITE( TestSrmBusy );
    CPPUNIT_TEST( testSupported );
	CPPUNIT_TEST( testCallReturns_IntErr_AbortFiles );
	CPPUNIT_TEST( testCallReturns_IntErr_Rm );
	CPPUNIT_TEST( testCallReturns_IntErr_Ls );
	CPPUNIT_TEST( testCallReturns_IntErr_Stat );
	CPPUNIT_TEST( testCallReturns_IntErr_PrepareToGet );
	CPPUNIT_TEST( testCallReturns_IntErr_PrepareToPut );
	CPPUNIT_TEST( testCallReturns_IntErr_BringOnline );
	CPPUNIT_TEST( testCallReturns_IntErr_GetSpaceMetaData );
	CPPUNIT_TEST( testCallReturns_IntErr_GetSpaceTokens );
	CPPUNIT_TEST( testCallReturns_IntErr_Copy );
	CPPUNIT_TEST( testCallReturns_IntErr_MkDir );
	CPPUNIT_TEST( testCallReturns_IntErr_RmDir );
	CPPUNIT_TEST( testAsyncCallReturns_IntErr_ReqQueued_noEWT_Stat );
	CPPUNIT_TEST( testCallReturns_IntErr_GeneralTimeout_Rm );
	CPPUNIT_TEST( testCallReturns_IntErr_GeneralTimeout_PrepareToGet );
	CPPUNIT_TEST( testCallReturns_IntErr_GeneralTimeout_Stat );
	CPPUNIT_TEST( testCallReturns_IntErr_ReqQueued_noEWT_GeneralTimeout_Stat );
	CPPUNIT_TEST( testAsynchCallReturns_ReqQueued_withEWT_PrepareToGet );
	CPPUNIT_TEST( testAsynchCallReturns_ReqQueued_withEWT_GeneralTimeout_Copy );
	//CPPUNIT_TEST( testAsynchCallReturns_ReqQueued_PollReset );
    CPPUNIT_TEST_SUITE_END();

    /**
     * Synch/asynch call returns SRM_INTERNAL_ERROR
     *
     * The test verifies that the general polling mechanism works for each SRM
     * operations.
     *
     * The file name rule is:
     *
     * srm_busy:SRM_INTERNAL_ERROR:<c>:<nIE>:SRM_REQUEST_QUEUED:0:-1.dat
     *
     * Client side:
     *   - checks the call returns with SRM_SUCCESS. Content, phase, etc. of the returned
     *     object is irrelevant here.
     * Server side:
     *   - checks if nIE number of polls received
     *   - the wait times between the polls follow the formula above
     *   - nIE times return SRM_INTERNAL_ERROR
     *   - in case of failure in the above checks, the server returns SRM_CUSTOM_STATUS,
     *     with description indicating the failed test phase. It will trigger the client-side
     *     unit test failure.
     *   - Finally, returns with SRM_SUCCESS
     *
     * The test is executed with the following calls:
     *    SrmStat, AbortFiles, GetSpaceMetaData, GetSpaceTokens, SrmMkDir,
     *    SrmRm, SrmRmDir, PrepareToGet, PrepareToPut, BringOnline, SrmCopy, SrmLs
     * Not tested call is ping, because we cannot specify rules, as Ping has no parameters.
     */
    void testCallReturns_IntErr_AbortFiles();
    void testCallReturns_IntErr_Rm();
    void testCallReturns_IntErr_Ls();
    void testCallReturns_IntErr_Stat();
    void testCallReturns_IntErr_PrepareToGet();
    void testCallReturns_IntErr_PrepareToPut();
    void testCallReturns_IntErr_BringOnline();
    void testCallReturns_IntErr_GetSpaceMetaData();
    void testCallReturns_IntErr_GetSpaceTokens();
    void testCallReturns_IntErr_Copy();
    void testCallReturns_IntErr_MkDir();
    void testCallReturns_IntErr_RmDir();

    /**
     * Asynch call returns SRM_INTERNAL_ERROR, then SRM_REQUEST_QUEUED, no eWT.
     *
     * The test verifies that the general polling mechanism works for asynch,
     * queued calls as well, that are called synchronously. In that sense, the
     * two status values are equal. No estimatedWaitTime (eWT) returned.
     *
     * The following operations work in this way: SrmStat.
     *
     * Client side:
     *   - checks if the the ls call returns with SRM_SUCCESS. Content, phase, etc. of
     *     the returned object is irrelevant here.
     * Server side:
     *   - checks if nIE + nRQ number of polls received
     *   - the wait times between the polls follow the formula above
     *   - nIE times return SRM_INTERNAL_ERROR, then nRQ times SRM_REQUEST_QUEUED
     *   - in case of failure in the above checks, the server returns SRM_CUSTOM_STATUS,
     *     with description indicating the failed test phase. It will trigger the client-side
     *     unit test failure.
     *   - Finally, returns with SRM_SUCCESS
     */
    void testAsyncCallReturns_IntErr_ReqQueued_noEWT_Stat();

    /**
     * Call returns SRM_INTERNAL_ERROR, general timeout occurs
     *
	 * Same like testCallReturns_IntErr, but nIE is kept so high that
	 * client side timeout occurs. Tested only with one sync (srmRm) and one async
	 * (PrepareToPut) calls as well as with the "mixed" one (Stat).
	 *
	 * Client side must detect timeout.
     */
    void testCallReturns_IntErr_GeneralTimeout_Rm();
    void testCallReturns_IntErr_GeneralTimeout_PrepareToGet();
    void testCallReturns_IntErr_GeneralTimeout_Stat();

    /**
     * Call returns SRM_INTERNAL_ERROR, then SRM_REQUEST_QUEUED, no eWT returned,
     * general timeout occurs
     *
	 * Same like testAsyncCallReturns_IntErr_ReqQueued_noEWT, but
	 * nRQ is kept so high that client side timeout occurs (no timeout in the nIE part).
	 *
	 * Only SrmStat falls into this category (async request executed synchronously)
	 *
	 * Client side must detect timeout.
     */
    void testCallReturns_IntErr_ReqQueued_noEWT_GeneralTimeout_Stat();

    /**
     * Async call returns SRM_REQUEST_QUEUED with eWT, no client timeout
     *
     * The file name rule is:
     *
     * srm_busy:SRM:INTERNAL_ERROR:<c>:0:SRM_REQUEST_QUEUED:<nRQ>:<eWT>.dat
     *
	 * Client side: checks if the call returns with SRM_SUCCESS.
	 * Server side: behaves similarly like in the earlier cases, but checks if the
	 * client returns really after the eWT.
	 *
	 * The test is executed with stat only.
     */
    void testAsynchCallReturns_ReqQueued_withEWT_PrepareToGet();

    /**
     * Async call returns SRM_REQUEST_QUEUED with eWT, client timeout occurs
     *
     * Same like testAsynchCallReturns_ReqQueued_withEWT but nRQ is so high that
     * client timeout occurs.
     *
 	 * Client side must detect timeout.
     */
    void testAsynchCallReturns_ReqQueued_withEWT_GeneralTimeout_Copy();

    /**
     * TODO
     * Async call returns SRM_REQUEST_QUEUED with no eWT, then with eWT
     *
     * The test checks if the server returns eWT in an exponential polling case
     * (no eWT was returned so far), it resets the exponential polling, ant the client
     * returns after the specified eWT.
     *
     * The file name rule changes here:
     *
     * srm_busy:SRM_REQUEST_QUEUED:<c>:<nRQ1>:SRM_REQUEST_QUEUED:<nRQ2>:<eWT>.dat
     *
     * where
     *   - nRQ1: SRM_REQUEST_QUEUED with no eWT
     *   - nRQ2: SRM_REQUEST_QUEUED with eWT
     *
     * The rest is the same as in the general description.
     *
	 * Client side: checks if the call returns with SRM_SUCCESS.
	 * The server returns the times, as usual.
	 *
	 * The test is executed with stat only.
     */
    void testAsynchCallReturns_ReqQueued_PollReset();

    /**
     * TODO: check if int. err returns ewt for all the calls
     */


private:
	virtual void _check_supported(const std::string& version);
	void _resetTestInServer();

	template<class REQUEST>
	void _testCallReturns_IntErr_Base_Async();

	template<class REQUEST>
	void _testCallReturns_IntErr_Base_Sync();

	template <class REQUEST>
	void _setupRequest(boost::shared_ptr<REQUEST>& request, const std::string& fname);

	template<class REQUEST>
	void _executeSyncRequest(boost::shared_ptr<REQUEST>& request,
				             SrmStatus::Code code = SrmStatus::SRM_SUCCESS);

	template<class REQUEST>
	void _executeAsyncRequest(boost::shared_ptr<REQUEST>& request,
						      SrmStatus::Code code = SrmStatus::SRM_SUCCESS);
};

#endif // #ifndef GLITE_DATA_SRM_UTIL_CPP_TESTSRMBUSY_H_
