/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004.
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  GLite Data SRM Util CPP: SRM 2.2 Test Case
 *
 *  Authors: Paolo Badino <paolo.badino@cern.ch>,
 *           Paolo Tedesco <paolo.tedesco@cern.ch>
 *           Zsolt Molnar <zsolt.molnar@cern.ch>
 *  Version info: $Id: TestSrmStat.cpp,v 1.2 2009/06/21 17:09:31 molnarzs Exp $
 *  Release: $Name:  $
 *
 */
#include "glite/data/srm/util/SrmExceptions.h"
#include "glite/data/srm/util/SrmStat.h"
#include "glite/data/agents/UserProxyEnv.h"

#include <memory>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <boost/scoped_ptr.hpp>
#include <cppunit/TestCase.h>
#include <cppunit/extensions/HelperMacros.h>

#include "TestSrmStat.h"

using boost::scoped_ptr;

void TestSrmStat::setUp()
{
	_d.TestBase::setUp();
    _d.put_file(_testFilePath_synch(), true);
}

void TestSrmStat::tearDown()
{
	_d.remove_file(_testFilePath_synch());
	_d.TestBase::tearDown();
}

void TestSrmStat::testStat_synch_request()
{
    // Create Request
    boost::shared_ptr<SrmStat> request(SrmStat::create(_d.m_ctx));
    std::string surl = _testFilePath_synch();
    request->setBackoffLogic(_d._logic);
    // Set Request Parameters
    request->files.resize(1);
    request->files[0].surl = surl;

    // Send Request
    SrmStatus result = request->execute();

    // Check Return Value
    CHECKSRMSTATUS("Invalid Result Code",result.code,SrmStatus::SRM_SUCCESS);

    // Check file Properties
    CPPUNIT_ASSERT_MESSAGE("Invalid SURL",   request->files[0].surl == surl);
    CPPUNIT_ASSERT_MESSAGE("Invalid Status", (request->files[0].status.code == SrmStatus::SRM_SUCCESS) || (request->files[0].status.code == SrmStatus::SRM_UNDEF));
    CPPUNIT_ASSERT_MESSAGE("Null Size"   , request->files[0].size > 0);
    CPPUNIT_ASSERT_MESSAGE("Invalid Size", (size_t)request->files[0].size == strlen(SAMPLE_TEXT));
    CPPUNIT_ASSERT_MESSAGE("Invalid File Locality", request->files[0].fileLocality == FileLocality::ONLINE);
    CPPUNIT_ASSERT_MESSAGE("Invalid File Type", request->files[0].fileType == FileType::FILE);
    CPPUNIT_ASSERT_MESSAGE("Invalid Storage Type", request->files[0].fileStorageType == FileStorageType::VOLATILE);

}

void TestSrmStat::testStat_ErrorNoFile()
{
    // Create Request
    boost::shared_ptr<SrmStat> request(SrmStat::create(_d.m_ctx));

    // Set Request Parameters
    request->files.resize(1);
    request->files[0].surl = _d.testPath() + "this-file-does-not-exist.txt";

    // Send Request
    SrmStatus result = request->execute();

    // Check Return Value
    CHECKSRMSTATUS("Invalid Result Code",result.code,SrmStatus::SRM_FAILURE);

    // Check file Properties
    CHECKSRMSTATUS("Invalid Status",request->files[0].status.code,SrmStatus::SRM_INVALID_PATH);

}

void TestSrmStat::testStat_ErrorNoFile_2()
{
    // Create Request
    boost::shared_ptr<SrmStat> request(SrmStat::create(_d.m_ctx));

    // Set Request Parameters
    request->files.resize(2);
    request->files[0].surl = _testFilePath_synch();
    request->files[1].surl = _d.testPath() + "this-file-does-not-exist.txt";

    // Send Request
    SrmStatus result = request->execute();

    // Check Return Value
    CHECKSRMSTATUS("Invalid Result Code",result.code,SrmStatus::SRM_PARTIAL_SUCCESS);

    // Check file Properties
    CPPUNIT_ASSERT_MESSAGE("Invalid Status #0", (request->files[0].status.code == SrmStatus::SRM_SUCCESS) ||
    		                                    (request->files[0].status.code == SrmStatus::SRM_UNDEF));
    CHECKSRMSTATUS("Invalid Status #1",request->files[1].status.code,SrmStatus::SRM_INVALID_PATH);

}

const std::string& TestSrmStat::_testFilePath_synch()
{
	static const std::string s(_d.testPath() + "file001.txt");
	return s;
}
