/*
 *  Copyright (c) Members of the EGEE Collaboration. 2004.
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  GLite Data SRM Util CPP: main function
 *
 *  Authors: Paolo Badino <paolo.badino@cern.ch>
 *           Paolo Tedesco <paolo.tedesco@cern.ch>
 *  Version info: $Id: TestSrmUtil.cpp,v 1.8 2009/06/21 17:09:31 molnarzs Exp $
 *  Release: $Name:  $
 *
 */
#include "glite/data/config/service/glite-config.h"
#include "glite/data/config/service/glite-config-register.hxx"

#include <stdlib.h>
#include <string>
#include <iostream>
#include <unistd.h>
#include <stdarg.h>
#include <cppunit/extensions/TestFactoryRegistry.h>
#include <cppunit/ui/text/TestRunner.h>
#include <cppunit/CompilerOutputter.h>
#include "test_helpers.h"

using std::cout;
using std::cerr;
using std::cin;
using std::endl;

namespace {
const char * const CONFIG_NAME = "glite-srm-util-test";
const char * const PROGNAME    = "glite_srm_util_test";
const char * const PROGVERSION = "0.0";
const char * const PROGAUTHOR  = "(C) EGEE";
}

/*
 * The SRM endpoints
 */
std::string g_srm_11_endpoint;
std::string g_srm_22_endpoint;

// -----------------------------------------------------------------------------
// -----------------------------------------------------------------------------
// Configuration for the Unit Test
// -----------------------------------------------------------------------------
// -----------------------------------------------------------------------------

#include "glite/data/config/service/ComponentConfiguration.hxx"
#include "log4cpp/Category.hh"

using namespace glite::config;

/**
 * SrmUtilTestComponent
 */
class SrmUtilTestComponent: public ComponentConfiguration {
public:
    /**
     * Initialize the Component
     */
    virtual int init(const Params& params){return 0;}

    /**
     * Configure the Component
     */
    virtual int config(const Params& params){return 0;}

    /**
     * Start the Component
     */
    virtual int start(){
        m_logger.log(log4cpp::Priority::INFO,"Test Started");
        return 0;
    }

    /**
     * Stop the Component
     */
    virtual int stop(){
        m_logger.log(log4cpp::Priority::INFO,"Test Stopped");
        return 0;
    }

    /**
     * Finalize the Component
     */
    virtual int fini(){return 0;}

    /**
     * Return Singleton Instance
     */
    static SrmUtilTestComponent& instance() {
        static SrmUtilTestComponent s_instance;
        return s_instance;
    }

    /**
     * debug
     */
    void debug(const char *fmt, va_list ap){
        m_logger.logva(log4cpp::Priority::DEBUG,fmt,ap);
    }
    /**
     * info
     */
    void info(const char *fmt, va_list ap){
        m_logger.logva(log4cpp::Priority::INFO,fmt,ap);
    }
    /**
     * error
     */
    void error(const char *fmt, va_list ap){
        m_logger.logva(log4cpp::Priority::ERROR,fmt,ap);
    }

private:
    /**
     * Constructor
     */
    SrmUtilTestComponent():ComponentConfiguration(CONFIG_NAME),
        m_logger(log4cpp::Category::getInstance(CONFIG_NAME)){
    }
    /**
     * Destructor
     */
    virtual ~SrmUtilTestComponent(){}

private:

    /**
     * The Logger
     */
    mutable log4cpp::Category& m_logger;
};

/*
 * debug
 */
void test_debug(const char *fmt, ...){
    va_list ap;
	va_start(ap, fmt);
	SrmUtilTestComponent::instance().debug(fmt,ap);
    va_end(ap);
}
/*
 * info
 */
void test_info(const char *fmt, ...){
	va_list ap;
	va_start(ap, fmt);
	SrmUtilTestComponent::instance().info(fmt,ap);
	va_end(ap);
}
/*
 * error
 */
void test_error(const char *fmt, ...){
	va_list ap;
	va_start(ap, fmt);
	SrmUtilTestComponent::instance().error(fmt,ap);
	va_end(ap);
}

/**
 * before_call
 */
void test_helpers::TestContext::before_call(
    const char * method){
    test_debug("# * Invoking %s",method);
}

/**
 * on_success
 */
void test_helpers::TestContext::on_success(
    const char * method,
    const char * address,
    const u::CallInfo * info){
    test_debug("#   ---> %s success [%s]",method,address);
    if (info){
        if (!info->token().empty()){
            test_debug("#        request token = [%s]",info->token().c_str());
        }
        u::CallInfo::AdditionalInfos::const_iterator it;
        const u::CallInfo::AdditionalInfos& infos = info->additionalInfos();
        for (it = infos.begin();it != infos.end();++it){
            test_debug("#        %s : [%s]",it->key.c_str(),it->value.c_str());
        }
    }
}

/**
 * on_failure
 */
void test_helpers::TestContext::on_failure(
    const char * method,
    const char * address,
    const u::CallInfo * info){
    test_error("#   ---> %s failed [%s]",method,address);
    if (info){
        if (!info->token().empty()){
            test_error("#        request token = [%s]",info->token().c_str());
        }
        u::CallInfo::AdditionalInfos::const_iterator it;
        const u::CallInfo::AdditionalInfos& infos = info->additionalInfos();
        for (it = infos.begin();it != infos.end();++it){
            test_error("#        %s : [%s]",it->key.c_str(),it->value.c_str());
        }
    }
}

/**
 * on_warning
 */
void test_helpers::TestContext::on_warning(
    const char * scope,
    const char * description,
    const char * address){
    test_error("#   warning : %s - %s [%s]",scope,description,address);
}


// -----------------------------------------------------------------------------
// -----------------------------------------------------------------------------

/*
 * print_usage_and_die
 *
 * Print Program Usage and ther exit
 */
void print_usage_and_die(){
    cout << endl;
    cout << "<" << PROGNAME << "> Version " << PROGVERSION << " by " << PROGAUTHOR;
    cout << endl;
    cout << "usage: " << PROGNAME << "[-l <glite_location_user>] [-s] -1 <srm1.1 endpoint> -2 <srm2.2 endpoint>" << endl;
    cout << " Where:" << endl;
    cout << " -l : specify the glite location. Configuration files should be located under the etc folder on that path" << endl;
    cout << " -s : Enable the Compiler Outputter" << endl;
    cout << " -1 : SRM 1.1 enpoint" << endl;
    cout << " -2 : SRM 2.2 enpoint" << endl;
    cout << " -c : Configuration name" << endl;
    cout << " -n : Test does not require web service (partial test execution case)" << endl;
    cout << " Where:" << endl;
    cout << " Optional parameters:" << endl;
    cout << "  -h : print this screen" << endl;
    exit(-1);
}

/**
 * check_status
 */
void check_status(const char * msg,const SrmStatus::Code& value,const SrmStatus::Code& expected) {
    if (expected != value){
        std::stringstream message;
        message << msg << " - got : " <<
            SrmStatus::toString (value) <<
            " expected : " << SrmStatus::toString (expected);
        CPPUNIT_FAIL(message.str().c_str());
    }
}

long elapsedTime(const boost::posix_time::ptime& startTime)
{
	boost::posix_time::ptime now(boost::posix_time::second_clock::universal_time());
	boost::posix_time::time_period elapsed(startTime, now);
	return elapsed.length().seconds();
}

/*
 * main
 *
 * Transfer FTS Main function
 */
int main(int argc, char* argv[]) {

    cout << "GLite Data SRM util cpp test program." << endl;

    // Get Options
    const char * location = 0;
    const char * config   = CONFIG_NAME;
    bool selfTest = false;
    bool noWebService = false;

    int flag;
    while ((flag = getopt (argc, argv, "l:1:2:c:shn")) != -1) {
        switch (flag) {
            case 'l':
                location = optarg;
            break;
            case 'c':
                config = optarg;
            break;
            case 's':
                selfTest = true;
            break;
            case '1':
                if(0 != optarg){
                    g_srm_11_endpoint = optarg;
                }
            break;
            case '2':
                if(0 != optarg){
                    g_srm_22_endpoint = optarg;
                }
            break;
            // No web service needed
            case 'n':
            	noWebService = true;
            	break;
            case 'h':
            default:
                print_usage_and_die();
            break;
        } // End switch
    } // End while

    if (!noWebService) {
		if(g_srm_11_endpoint.empty()){
			print_usage_and_die();
		}
		if(g_srm_22_endpoint.empty()){
			print_usage_and_die();
		}

		if(0 != location){
			// Set "GLite Location User"
			setenv("GLITE_LOCATION",location,1);
		}
		// Register the Static Component
		glite_config_register(&(SrmUtilTestComponent::instance()));
		// Initialize the Application
		int initres = glite_config_initialize(config);
		if(initres < 0){
			cerr << "Initialization failed " << endl;
			return -1;
		}
    }

    CPPUNIT_NS :: TestResult testresult;
    CPPUNIT_NS :: TestResultCollector collectedresults;
    testresult.addListener (&collectedresults);
    CPPUNIT_NS :: BriefTestProgressListener progress;
    testresult.addListener (&progress);
    CPPUNIT_NS :: TestRunner testrunner;
    testrunner.addTest (CPPUNIT_NS :: TestFactoryRegistry :: getRegistry ().makeTest ());
    testrunner.run (testresult);
    CPPUNIT_NS :: CompilerOutputter compileroutputter (&collectedresults, std::cerr);
    compileroutputter.write ();

    // Run the test
    bool wasSucessful = collectedresults.wasSuccessful();

    if (!noWebService) {
    	// Finalize
    	glite_config_finalize();
    }

    return wasSucessful ? 0 : 1;
}
