/*
 *  Copyright (c) Members of the EGEE Collaboration. 2004.
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  GLite Data SRM Util CPP: Srm 1.1. Server
 *
 *  Authors: Paolo Badino <paolo.badino@cern.ch>, Paolo Tedesco <paolo.tedesco@cern.ch>
 *  Version info: $Id: server.cpp,v 1.16 2009/07/30 17:40:34 molnarzs Exp $
 *  Release: $Name:  $
 *
 */

#ifndef SOAP_FMAC3S
#define SOAP_FMAC3S
#endif
#ifndef SOAP_FMAC4S
#define SOAP_FMAC4S
#endif

#include "glite/data/srm/cpp/envH.h"
#include "glite/data/srm/cpp/srmH.h"
#include "glite/data/agents/url-utility.h"
#include "glite/data/agents/AgentExceptions.h"

#include <vector>
#include <map>
#include <sstream>
#include <boost/shared_ptr.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/regex.hpp>
#include <libgen.h>
#include <string.h>
#include <pthread.h>

extern int errno;

#include <log4cpp/Category.hh>
#define TRACE_LOG(a) log4cpp::Category::getInstance("srm-service-stub").debugStream() << a
#define TRACE_ERR(a) log4cpp::Category::getInstance("srm-service-stub").errorStream() << a

namespace {
const char * const STATUS_PENDING = "Pending";
const char * const STATUS_READY   = "Ready";
const char * const STATUS_ACTIVE  = "Active";
const char * const STATUS_RUNNING = "Running";
const char * const STATUS_DONE    = "Done";
const char * const STATUS_FAILED  = "Failed";

const char * const TYPE_PUT       = "put";
const char * const TYPE_GET       = "get";
const char * const TYPE_COPY      = "copy";

const char * const PROTOCOL_FILE  = "file";

const char * const SFN            = "SFN";
}

using namespace glite::data::agents;
extern std::string g_pool_path;

// the lock
extern pthread_mutex_t  g_lock;

namespace srm{

/**
 * Generic rule to return file status based on name
 */
class FileStatusRule{
public:
    /**
     * singleton instance
     */
    static FileStatusRule& instance(){
        static FileStatusRule m_instance;
        return m_instance;
    }
    /**
     * is_match
     * return true if file name matches the regular expression pattern and
     * (optionally) the operation name, e.g.
     * "this-one-returns-Ready-the file is ready-on-get","get" ->
	 * 		true,status = "Ready",message = "the file is ready"
	 * @param file_name [IN] the file name to match
	 * @param operation [IN] optional operation name that acts as a filter, use
	 * 		empty string to disable checking
	 * @param status [OUT] the file status found in the file name
	 * @param message [OUT] the file message found in the file name
	 * @return true if the file name matches the pattern
	 * 		this-one-returns-(status)[-(message)][-on-(operation)], false
	 * 		otherwise
     */
    bool is_match(const std::string& file_name,const std::string& operation,std::string& status,std::string& message) {
        boost::match_results<std::string::const_iterator> what;
        boost::match_flag_type flags = boost::match_default;
        bool is_match = boost::regex_search(file_name.begin(),file_name.end(),what,m_regex,flags);
        if (is_match){
            status = what[1];
            message = what[3];
            const std::string& op_name = what[5];
            if ((0 != operation.length()) && (0 != op_name.length())){
                is_match = (operation == op_name);
            }
        }
        return is_match;
    }

protected:
    FileStatusRule() : m_regex("this-one-returns-(Pending|Ready|Active|Running|Done|Failed)(-([\\w\\s]*))?(-on-(\\l+))?") {
    }
    boost::regex m_regex;
};

/**
 * File Request Object
 */
struct FileRequest{
public:
    typedef boost::shared_ptr<FileRequest> Ptr;

    int         id;
    std::string filename;
    std::string surl;
    std::string turl;
    std::string status;
    std::string message;
    std::string protocol;
    long long   size;
    int         permissions;

    FileRequest():id(0),size(0),permissions(0){}
};

/**
 * Request Object
 */
struct Request{
public:
    typedef boost::shared_ptr<Request> Ptr;

    int  id;
    std::string type;
    std::string status;
    std::vector<FileRequest::Ptr> fileRequests;
    time_t start_time;
    time_t finish_time;

    Request():id(0),start_time(-1),finish_time(-1){}
};

/**
 * Request Map
 */
typedef std::map<int,Request::Ptr> RequestMap;
RequestMap g_RequestMap;

/**
 * Store a Request
 */
int store_request(Request::Ptr request){
    static int s_id = 0;
    int stored = -1;
    // lock
    //pthread_mutex_lock(&g_lock);
    // Assign the id
    request->id = ++s_id;
    // Store the Request
    std::pair<RequestMap::iterator, bool> result =
        g_RequestMap.insert(RequestMap::value_type(request->id,request));
    if(!result.second){
	stored = -1;
    } else {
        // Set start time
        request->start_time = time(0);
        stored = 0;;
    }
    // unlock
    //pthread_mutex_unlock(&g_lock);
    return stored;
}

/**
 * Return a Request
 */
Request::Ptr get_request(int id){
    Request::Ptr ptr;
    // lock
    //pthread_mutex_lock(&g_lock);
    RequestMap::const_iterator it = g_RequestMap.find(id);
    if(it != g_RequestMap.end()){
        ptr = (*it).second;
    }
    // unlock
    //pthread_mutex_unlock(&g_lock);
    return ptr;
}

/**
 * Return a Request
 */
void purge_requests(){
    // lock
    //pthread_mutex_lock(&g_lock);

    time_t current;
    time(&current);

    std::vector<int> ids;
    Request::Ptr ptr;
    RequestMap::iterator it;
    for(it = g_RequestMap.begin(); it != g_RequestMap.end(); ++it){
        // Check the requests that has been completed more than 30 minutes ago
        ptr = (*it).second;
        if((ptr->status == STATUS_DONE) ||
           (ptr->status == STATUS_FAILED)){
            if((ptr->finish_time != (time_t)-1) &&
               (ptr->finish_time > (current + 1800))){
                // This request should be removed. Cannot use erase since it
                // looks like it return void instead of iterator (?)
                ids.push_back(ptr->id);
            }
        }
    }
    std::vector<int>::iterator iit;
    for(iit = ids.begin(); iit != ids.end(); ++iit){
        g_RequestMap.erase((*iit));
    }

    // unlock
    //pthread_mutex_unlock(&g_lock);
}

// -----------------------------------------------------------------------------
// Stub implementation: same file on the file system
// -----------------------------------------------------------------------------
/*
 * Extract the file path from a surl
 */
std::string path_from_surl(const std::string& surl){
    std::string file_path;
    try{
        std::string  hostname;
        unsigned int port = 0;
        std::string  protocol;
        std::string  path;
        std::string  query_str;

        // Parse surl
        parse_url(surl,hostname,port,protocol,path,query_str);

        // Check query string
        if(query_str.empty()){
            // SURL in compact form: take the path
            file_path = path;
        } else {
            // Parse Query string
            std::vector<std::pair<std::string,std::string> > values;
            parse_query_string(query_str,values);

            // Take the SFN
            std::vector<std::pair<std::string,std::string> >::iterator it;
            for(it = values.begin(); it != values.end(); ++it){
                if((*it).first == SFN){
                    file_path = (*it).second;
                    break;
                }
            }
        }
    }catch (const LogicError& exc){
        // Invalid SURL
        file_path = "";
    }

    // Append the file_path the the pool directory path
    if(false == file_path.empty()){
        if(*(g_pool_path.rbegin()) != '/'){
            file_path = g_pool_path + "/" + file_path;
        } else {
            file_path = g_pool_path + file_path;
        }
    }

    return file_path;
}

/*
 * Create a Directory
 */
int create_dir_path(const std::string& file_path){
    int result = -1;
    // Get the dirname for that path
    char * dir_name_str = strdup(file_path.c_str());
    dirname(dir_name_str);
    std::string dir_name = dir_name_str;
    free(dir_name_str);

    // Check if dirname exist
    int r = access(dir_name.c_str(), R_OK);
    if(0 == r){
        // Diractory exists, check if it's writable
        r = access(dir_name.c_str(), W_OK);
        if(0 == r){
            // Ok
            result = 0;
        } else {
            // Directory exists, but it's not writable
            result = -2;
        }
    } else {
        // Directory doesn't exist: try to create the parent dir
        r = create_dir_path(dir_name);
        if(0 == r){
            // Ok, parent exists or has been created: Create teh directory
            r = mkdir(dir_name.c_str(),0700);
            if(0 == r){
                // Ok
                result = 0;
            } else {
                // Can't create directory
                result = -1;
            }
        } else {
            // Can't create parent directory
            result = -1;
        }
    }
    return result;
}

/*
 * Delete a File or directory
 */
void process_delete_request(const std::string& surl, std::string& error){
    // Get file path
    std::string file_path = path_from_surl(surl);
    if(false == file_path.empty()){
        // Get the stat of the file
        struct stat64 stat_buf;
        int r = stat64(file_path.c_str(),&stat_buf);
        if(0 == r){
            if(S_ISDIR(stat_buf.st_mode)){
                // It's a directory: it should be empty!
                r = rmdir(file_path.c_str());
                if(r == -1){
                    error = strerror(errno);
                }
            } else {
                // It's a file: unlink it
                r = unlink(file_path.c_str());
                if(r == -1){
                    error = strerror(errno);
                }
            }
        } else {
            error = strerror(errno);
        }
    }
}

/*
 * update_request_status
 */
void update_request_status(Request::Ptr req){
    if((req->type == TYPE_GET) || (req->type == TYPE_PUT)){
        bool has_pending = false;
        bool has_ready   = false;
        bool has_running = false;
        bool has_done    = false;
        bool has_failed  = false;
        std::vector<FileRequest::Ptr>::iterator it;
        for(it = req->fileRequests.begin();it != req->fileRequests.end(); ++it){
            FileRequest::Ptr file = (*it);
            // Check File Status
            if(file->status == STATUS_PENDING){
                has_pending = true;
            } else if(file->status == STATUS_READY){
                has_ready = true;
            } else if(file->status == STATUS_RUNNING){
                has_running = true;
            } else if(file->status == STATUS_DONE){
                has_done = true;
            } else if(file->status == STATUS_FAILED){
                has_failed = true;
            }
        }
        if(has_pending){
            // There are still some pending
            req->status = STATUS_PENDING;
        } else if(has_ready){
            // There are still some ready
            req->status = STATUS_READY;
        } else if(has_running){
            // There are still some running
            req->status = STATUS_READY;
        } else if(has_failed){
            // There are some failures
            req->status = STATUS_FAILED;
            req->finish_time = time(0);
        } else {
            // Can be only done
            req->status = STATUS_DONE;
            req->finish_time = time(0);
        }
   } else if(req->type == TYPE_COPY){
        bool has_pending = false;
        bool has_active = false;
        bool has_done    = false;
        bool has_failed  = false;
        std::vector<FileRequest::Ptr>::iterator it;
        for(it = req->fileRequests.begin();it != req->fileRequests.end(); ++it){
            FileRequest::Ptr file = (*it);
            // Check File Status
            if(file->status == STATUS_PENDING){
                has_pending = true;
            } else if(file->status == STATUS_ACTIVE){
                has_active = true;
            } else if(file->status == STATUS_DONE){
                has_done = true;
            } else if(file->status == STATUS_FAILED){
                has_failed = true;
            }
        }
        if(has_active){
            // There are still some pending
            req->status = STATUS_ACTIVE;
        } else if(has_pending){
            // There are still some running
            req->status = STATUS_PENDING;
        } else if(has_failed){
            // There are some failures
            req->status = STATUS_FAILED;
            req->finish_time = time(0);
        } else {
            // Can be only done
            req->status = STATUS_DONE;
            req->finish_time = time(0);
        }
   }
}

/*
 * process_put_request
 */
void process_put_request(Request::Ptr req){

    // Check Request Status
    if((req->status == STATUS_DONE)   ||
       (req->status == STATUS_READY)  ||
       (req->status == STATUS_RUNNING)||
       (req->status == STATUS_FAILED)){
        // Nothing To Do
    } else if (req->status == STATUS_PENDING) {
        // Check the files
        std::vector<FileRequest::Ptr>::iterator it;
        for(it = req->fileRequests.begin();it != req->fileRequests.end(); ++it){
            FileRequest::Ptr file = (*it);
            // Check File Status
            if(file->status == STATUS_PENDING){
                // Check if the file exists
                std::string file_path = path_from_surl(file->surl);

                std::string status_from_name;
                std::string message_from_name;

                if(file_path.empty()){
                    file->status  = STATUS_FAILED;
                    file->message = "Invalid SURL passed";
                } else if(std::string::npos != file_path.find("this-one-should-stay-pending")){
                    // File should remain pending
                    continue;
                } else if(std::string::npos != file_path.find("this-one-is-busy")){
                    // File is busy
                    file->status  = STATUS_FAILED;
                    file->message = "file busy";
                } else if (FileStatusRule::instance().is_match(file_path,TYPE_PUT,status_from_name,message_from_name)) {
                    file->status  = status_from_name;
                    file->message = message_from_name;
                } else {
                    // Check if file exists
                    struct stat64 stat_buf;
                    int r = stat64(file_path.c_str(),&stat_buf);
                    if(0 != r){
                        // Get the directory name
                        char * base_name_str = strdup(file_path.c_str());
                        basename(base_name_str);
                        std::string base_name = base_name_str;
                        free(base_name_str);

                        // Create the directory structure
                        r = create_dir_path(file_path);
                        if(0 != r){
                            file->status  = STATUS_FAILED;
                            file->message = "Cannot create directory for file";
                        } else {
                            // Touch the file
                            int fd = creat(file_path.c_str(), file->permissions);
                            if(-1 != fd){
                                // Set the file properties
                                file->turl = (std::string)PROTOCOL_FILE + "://" + file_path;
                                file->permissions = stat_buf.st_mode;
                                file->status  = STATUS_READY;
                                close(fd);
                            } else {
                                file->status  = STATUS_FAILED;
                                file->message = "Cannot create file";
                            }
                        }
                    } else {
                        file->status  = STATUS_FAILED;
                        file->message = "Requested file already exists";
                    }
                }
            } else{
                // Nothing to Do
            }
        } // End for each file

        // Update request status
        update_request_status(req);
    }
}

/*
 * process_get_request
 */
void process_get_request(Request::Ptr req){

    // Check Request Status
    if((req->status == STATUS_DONE)   ||
       (req->status == STATUS_READY)  ||
       (req->status == STATUS_RUNNING)||
       (req->status == STATUS_FAILED)){
        // Nothing To Do
    } else if (req->status == STATUS_PENDING) {
        // Check the files
        std::vector<FileRequest::Ptr>::iterator it;
        for(it = req->fileRequests.begin();it != req->fileRequests.end(); ++it){
            FileRequest::Ptr file = (*it);
            // Check File Status
            if(file->status == STATUS_PENDING){
                // Check if the file exists
                std::string file_path = path_from_surl(file->surl);

                std::string status_from_name;
                std::string message_from_name;

                if(file_path.empty()){
                    file->status  = STATUS_FAILED;
                    file->message = "Invalid SURL passed";
                } else if(std::string::npos != file_path.find("this-one-should-stay-pending")){
                    // File should remain pending
                    continue;
                } else if (FileStatusRule::instance().is_match(file_path,TYPE_GET,status_from_name,message_from_name)) {
                    file->status  = status_from_name;
                    file->message = message_from_name;
                } else {
                    // Check if file exists
                    struct stat64 stat_buf;
                    int r = stat64(file_path.c_str(),&stat_buf);
                    if(0 == r){
                        if(S_ISDIR(stat_buf.st_mode)){
                            file->status  = STATUS_FAILED;
                            file->message = "Requested path is a directory";
                        } else {
                            // Set the file properties
                            file->turl = (std::string)PROTOCOL_FILE + "://" + file_path;
                            if (std::string::npos != file_path.find("this-one-returns-size-zero-on-get")){
                                file->size = 0;
                            } else if (std::string::npos != file_path.find("this-one-returns-wrong-size-on-get")){
                                file->size = (stat_buf.st_size >> 1);
                            } else {
                                file->size = stat_buf.st_size;
                            }
                            file->permissions = stat_buf.st_mode;
                            file->status  = STATUS_READY;
                        }
                    } else {
                        file->status  = STATUS_FAILED;
                        file->message = "No such file";
                    }
                }
            } else{
                // Nothing to Do
            }
        } // End for each file

        // Update request status
        update_request_status(req);
    }
}

/*
 * process_copy_request
 */
void process_copy_request(Request::Ptr req){

    // Check Request Status
    if((req->status == STATUS_DONE)   ||
       (req->status == STATUS_FAILED)){
        // Nothing To Do
    } else if((req->status == STATUS_PENDING) ||
              (req->status == STATUS_ACTIVE)){
        // Check the files
        std::vector<FileRequest::Ptr>::iterator it;
        for(it = req->fileRequests.begin();it != req->fileRequests.end(); ++it){
            FileRequest::Ptr file = (*it);
            // Check File Status
            if(file->status == STATUS_PENDING){
                // Check the source file
                std::string src_file_path = path_from_surl(file->surl);

                std::string status_from_name;
                std::string message_from_name;

                if(src_file_path.empty()){
                    file->status  = STATUS_FAILED;
                    file->message = "Invalid source SURL passed";
                } else if(std::string::npos != src_file_path.find("this-one-should-stay-pending")){
                    // File should remain pending
                    continue;
                } else if (FileStatusRule::instance().is_match(src_file_path,TYPE_COPY,status_from_name,message_from_name)) {
                        file->status  = status_from_name;
                        file->message = message_from_name;
                } else {
                    // Check if source file exists
                    struct stat64 stat_buf;
                    int r = stat64(src_file_path.c_str(),&stat_buf);
                    if(0 == r){
                        if(S_ISDIR(stat_buf.st_mode)){
                            file->status  = STATUS_FAILED;
                            file->message = "Requested path is a directory";
                        } else {
                            // Set the file properties
                            file->size = stat_buf.st_size;
                            file->permissions = stat_buf.st_mode;
                        }
                    } else {
                        file->status  = STATUS_FAILED;
                        file->message = "No such file";
                    }
                }
                if(file->status != STATUS_FAILED){
                    // Check the source file
                    std::string dst_file_path = path_from_surl(file->turl);

                    if(dst_file_path.empty()){
                        file->status  = STATUS_FAILED;
                        file->message = "Invalid destination SURL passed";
                    } else if(std::string::npos != dst_file_path.find("this-one-should-stay-pending")){
                        // File should remain pending
                        continue;
                    } else if (FileStatusRule::instance().is_match(dst_file_path,TYPE_GET,status_from_name,message_from_name)) {
                        file->status  = status_from_name;
                        file->message = message_from_name;
                    } else {
                        // Check if dest file exists
                        struct stat64 stat_buf;
                        int r = stat64(dst_file_path.c_str(),&stat_buf);
                        if(0 == r){
                            // Destination file already exist
                            file->status  = STATUS_FAILED;
                            file->message = "File already exists";
                        } else {
                            // Create the directory structure
                            r = create_dir_path(dst_file_path);
                            if(0 != r){
                                file->status  = STATUS_FAILED;
                                file->message = "Cannot create directory for destination file";
                            } else {
                                // execute a copy command
                                std::string cmd =
                                    (std::string)"cp " + src_file_path + " " + dst_file_path;
                                r = system(cmd.c_str());
                                file->status  = STATUS_ACTIVE;
                            }
                        }
                    }
                }
            } else if(file->status == STATUS_ACTIVE){
                // Copy the file from source to destination

                // Get the source name
                std::string src_file_path = path_from_surl(file->surl);
                if(src_file_path.empty()){
                    // This should never happen
                    file->status  = STATUS_FAILED;
                    file->message = "Invalid source SURL passed";
                }
                // Get the dest name
                std::string dst_file_path = path_from_surl(file->turl);
                if(dst_file_path.empty()){
                    // This should never happen
                    file->status  = STATUS_FAILED;
                    file->message = "Invalid destination SURL passed";
                }

                // Check that the destination has been copied correctly
                struct stat64 dst_stat_buf;
                int r = stat64(dst_file_path.c_str(),&dst_stat_buf);
                if(0 != r){
                    // Destination file already exist
                    file->status  = STATUS_FAILED;
                    file->message = "Failed to copy file";
                } else {
                    // Check that the size is the same
                    if(dst_stat_buf.st_size == file->size){
                        file->status  = STATUS_DONE;
                    } else {
                        file->status  = STATUS_FAILED;
                        file->message = "Failed to copy file, different sizes";
                    }
                }
            } else {
                // Nothing to Do
            }
        } // End for each file

        // Update request status
        update_request_status(req);
    }
}

/*
 * finish_file
 */
void finish_file(Request::Ptr req, FileRequest::Ptr file){
    if((req->type == TYPE_GET) ||
       (req->type == TYPE_PUT)){
        if((file->status == STATUS_READY)||(file->status == STATUS_RUNNING)){
            // Complete request
            file->status = STATUS_DONE;
        } else {
            // Nothing to do
        }
    } else {
        // Nothing to do
    }
}

/*
 * abort_file
 */
void abort_file(Request::Ptr req, FileRequest::Ptr file){
    if(req->type == TYPE_GET){
        if((file->status == STATUS_PENDING)||
           (file->status == STATUS_READY)||
           (file->status == STATUS_RUNNING)){
            // Complete request
            file->status  = STATUS_FAILED;
            file->message = "Aborted by the user";
        } else {
            // Nothing to do
        }
    } else if(req->type == TYPE_PUT){
        if(file->status == STATUS_PENDING){
            file->status  = STATUS_FAILED;
            file->message = "Aborted by the user";
        } else if(
           (file->status == STATUS_READY)||
           (file->status == STATUS_RUNNING)){
            // Unlink the destination file
            std::string file_path = path_from_surl(file->surl);
            if(!file_path.empty()){
                // Unlink file
                unlink(file_path.c_str());
            }
            file->status  = STATUS_FAILED;
            file->message = "Aborted by the user";
        } else {
            // Nothing to do
        }
    } else if(req->type == TYPE_COPY){
        if(file->status == STATUS_PENDING){
            file->status  = STATUS_FAILED;
            file->message = "Aborted by the user";
        } else if(
           (file->status == STATUS_ACTIVE)){
            // Unlink the destination file
            std::string file_path = path_from_surl(file->turl);
            if(!file_path.empty()){
                // Unlink file
                unlink(file_path.c_str());
            }
            file->status  = STATUS_FAILED;
            file->message = "Aborted by the user";
        } else {
            // Nothing to do
        }
    }
}

// -----------------------------------------------------------------------------
// SRM 1.1 translation routines
// -----------------------------------------------------------------------------

/*
 * File Status
 */
srm2__RequestFileStatus* request_file_status(struct soap * soap, FileRequest::Ptr request){
    srm2__RequestFileStatus * filestatus = soap_new_srm2__RequestFileStatus(soap,-1);

    // Id
    filestatus->fileId = request->id;
    // State
    filestatus->state  = soap_strdup(soap,request->status.c_str());
    // SURL
    if(false == request->surl.empty()){
        filestatus->SURL = soap_strdup(soap,request->surl.c_str());
    } else {
        filestatus->SURL = 0;
    }
    // TURL
    if(false == request->turl.empty()){
        filestatus->TURL = soap_strdup(soap,request->turl.c_str());
    } else {
        filestatus->TURL = 0;
    }
    // Est. Time To Start
    filestatus->estSecondsToStart  = 0;
    // Source Filename
    if(false == request->filename.empty()){
        filestatus->sourceFilename = soap_strdup(soap,request->filename.c_str());
    } else {
        filestatus->sourceFilename = 0;
    }
    // Dest Filename
    if(false == request->turl.empty()){
        filestatus->destFilename = soap_strdup(soap,request->turl.c_str());
    } else {
        filestatus->destFilename = 0;
    }
    // Size
    if(request->size > 0){
        filestatus->size = request->size;
    } else {
        filestatus->size = 0;
    }
    // Owner
    filestatus->owner = 0;
    // Group
    filestatus->group = 0;
    // Permission
    if(request->permissions > 0){
        filestatus->permMode = request->permissions;
    } else {
        filestatus->permMode = 0;
    }
    // Not Used
    filestatus->queueOrder    = 0;
    filestatus->checksumType  = 0;
    filestatus->checksumValue = 0;
    filestatus->isPinned      = false;
    filestatus->isPermanent   = false;
    filestatus->isCached      = true;

    return filestatus;
}

/*
 * get_file_stat
 */
srm2__FileMetaData* get_file_stat(struct soap * soap, const std::string& surl, std::string& message){

    srm2__FileMetaData * filestat = 0;
    // Get File Path
    std::string file_path = path_from_surl(surl);
    if(false == file_path.empty()){
        // Get the Stats of the file
        struct stat64 stat_buf;
        int r = stat64(file_path.c_str(),&stat_buf);
        if(0 == r){
            filestat              = soap_new_srm2__FileMetaData(soap,-1);
            filestat->SURL        = soap_strdup(soap,surl.c_str());
            filestat->size        = stat_buf.st_size;
            filestat->owner       = soap_strdup(soap,(boost::lexical_cast<std::string>(stat_buf.st_uid)).c_str());
            filestat->group       = soap_strdup(soap,(boost::lexical_cast<std::string>(stat_buf.st_gid)).c_str());
            filestat->permMode    = stat_buf.st_mode;
            filestat->isPinned    = false;
            filestat->isPermanent = false;
            filestat->isCached    = true;

            // Not Used
            filestat->checksumType  = 0;
            filestat->checksumValue = 0;
        } else {
            message = "Requested file doesn't exist";
        }
    } else {
        message = "Not a valid SURL";
    }
    return filestat;
}

static const int DEFAULT_RETRYDELTATIME = 1;

/*
 * Request Status
 */
srm2__RequestStatus * request_status(struct soap * soap, Request::Ptr req){

    TRACE_LOG(("Status of Request %d is %s",req->id,req->status.c_str()));

    srm2__RequestStatus * result = soap_new_srm2__RequestStatus(soap,-1);
    result->requestId      = req->id;
    result->type           = soap_strdup(soap,req->type.c_str());
    result->state          = soap_strdup(soap,req->status.c_str());
    result->submitTime     = &(req->start_time);
    result->startTime      = &(req->start_time);
    if(-1 != req->finish_time){
        result->retryDeltaTime = 0;
        result->finishTime     = &(req->finish_time);
    } else {
        result->retryDeltaTime = DEFAULT_RETRYDELTATIME;
        result->finishTime     = 0;
    }
    result->estTimeToStart = 0;
    result->errorMessage   = 0;
    result->fileStatuses   = soap_new_srm2ArrayOfRequestFileStatus(soap,-1);
    result->fileStatuses->__size = req->fileRequests.size();
    result->fileStatuses->__ptr = (srm2__RequestFileStatus**)soap_malloc(soap, req->fileRequests.size() * sizeof(srm2__RequestFileStatus*));
    std::vector<FileRequest::Ptr>::iterator it;
    int i;
    for(it = req->fileRequests.begin(), i = 0; it != req->fileRequests.end(); ++it, ++i){
        // Create File Status Object
        result->fileStatuses->__ptr[i] = request_file_status(soap,(*it));
        TRACE_LOG(("File %d status is %s",i,(*it)->status.c_str()));
    }
    // Check error messages
    std::stringstream error;
    for(it = req->fileRequests.begin(); it != req->fileRequests.end(); ++it){
        if(false == (*it)->message.empty()){
            error << "\nRequestFileStatus#"
                  << (*it)->id << " : " << (*it)->message;
        }
    }
    // Set Error Message
    std::string str = error.str();
    if(false == str.empty()){
        result->errorMessage = soap_strdup(soap,str.c_str());
    }
    return result;
}

/*
 * Request Status
 */
srm2__RequestStatus * request_status(struct soap * soap, Request::Ptr req, FileRequest::Ptr file){
    TRACE_LOG(("Status of Request %d is %s",req->id,req->status.c_str()));
    TRACE_LOG(("File %d status is %s",0,file->status.c_str()));

    srm2__RequestStatus * result = soap_new_srm2__RequestStatus(soap,-1);
    result->requestId      = req->id;
    result->type           = soap_strdup(soap,req->type.c_str());
    result->state          = soap_strdup(soap,req->status.c_str());
    result->submitTime     = &(req->start_time);
    result->startTime      = &(req->start_time);
    if(-1 != req->finish_time){
        result->retryDeltaTime = 0;
        result->finishTime     = &(req->finish_time);
    } else {
        result->retryDeltaTime = DEFAULT_RETRYDELTATIME;
        result->finishTime     = 0;
    }
    result->estTimeToStart = 0;
    result->errorMessage   = 0;
    result->retryDeltaTime = 10;
    result->fileStatuses   = soap_new_srm2ArrayOfRequestFileStatus(soap,-1);
    result->fileStatuses->__size = 1;
    result->fileStatuses->__ptr = (srm2__RequestFileStatus**)soap_malloc(soap, 1 * sizeof(srm2__RequestFileStatus*));
    // Create File Status Object
    result->fileStatuses->__ptr[0] = request_file_status(soap,file);
    if(file->message.empty()){
        result->errorMessage = 0;
    } else {
        // Set Error Message
        std::stringstream error;
        error << "\nRequestFileStatus#"
              << file->id << " : " << file->message;
        result->errorMessage = soap_strdup(soap,error.str().c_str());
    }
    return result;
}

// -----------------------------------------------------------------------------
//
// SRM 1.1 Stub Implementation
//
// -----------------------------------------------------------------------------

/*
 * srm__ping
 */
int srm__ping(struct soap * soap, bool &_Result) {
    TRACE_LOG(("srm__ping called"));

    _Result = true;

    TRACE_LOG(("Done: srm__ping"));
    return SOAP_OK;
}

/*
 * srm__getProtocols
 */
int srm__getProtocols(struct soap * soap, struct srm__getProtocolsResponse & out) {
    TRACE_LOG(("srm__getProtocols called"));

    out._Result = soap_new_srm4ArrayOfstring(soap,-1);
    out._Result->__size = 1;
    out._Result->__ptr = (char**)soap_malloc(soap, 1*sizeof(char*));
    out._Result->__ptr[0] = soap_strdup(soap,PROTOCOL_FILE);

    TRACE_LOG(("Done: srm__getProtocols"));
    return SOAP_OK;
}

/*
 * srm__put
 */
int srm__put(struct soap * soap, srm4ArrayOfstring *_arg0, srm4ArrayOfstring *_arg1, srm5ArrayOflong *_arg2, srm5ArrayOfboolean *_arg3, srm4ArrayOfstring *_arg4, struct srm__putResponse &out) {
    TRACE_LOG(("srm__put called"));

    int i = 0;
    int result  = SOAP_FAULT;

    TRACE_LOG(("Request to Put %d Files",_arg0->__size));
    for(i = 0; i < _arg0->__size; i++){
        TRACE_LOG(("File #%d: ",i));
        TRACE_LOG((" %s",_arg0->__ptr[i]));
        TRACE_LOG((" Src %s",_arg1->__ptr[i]));
        TRACE_LOG((" Size %lld",_arg2->__ptr[i]));
        TRACE_LOG((" Permanent %d",_arg3->__ptr[i]));
        TRACE_LOG((" Protocol  %s",_arg4->__ptr[i]));
    }

    // Create Request Object
    Request::Ptr req(new Request());
    req->type   = TYPE_PUT;
    req->status = STATUS_PENDING;

    // Get Protocol
    std::string protocol = PROTOCOL_FILE;
    if(_arg4->__size > 0){
        if(0 != _arg4->__ptr[0]){
            protocol = _arg4->__ptr[0];
        }
    }
    // For each file
    for(i = 0; i < _arg0->__size; i++){
        FileRequest::Ptr f(new FileRequest());
        f->id = i;
        if(0 != _arg0->__ptr[i]){
            f->filename = _arg0->__ptr[i];
        }
        if(0 != _arg1->__ptr[i]){
            f->surl = _arg1->__ptr[i];
        }
        if(_arg2->__ptr[i] > 0){
            f->size = _arg2->__ptr[i];
        }
        if(_arg3->__ptr[i] > 0){
            f->permissions = 0644;
        }
        f->status = STATUS_PENDING;
        f->protocol = protocol;
        // Store file
        req->fileRequests.push_back(f);
    }

    // Insert Request
    if(0 == store_request(req)){
        out._Result = request_status(soap,req);
        result = SOAP_OK;
    } else {
        // Set error Message
        result = soap_receiver_fault(soap,"Failed to store request id",0);
    }

    TRACE_LOG(("Done: srm__put"));
    return result;
}

/*
 * srm__get
 */
int srm__get(struct soap * soap, srm4ArrayOfstring *_arg0, srm4ArrayOfstring *_arg1, struct srm__getResponse &out) {
    TRACE_LOG(("srm__get called"));

    int i = 0;
    int result  = SOAP_FAULT;
    TRACE_LOG("Request to Get " << _arg0->__size << " files");
    for(i = 0; i < _arg0->__size; i++){
        TRACE_LOG("File #" << i << ": " << _arg0->__ptr[i] << " - Protocol " << _arg1->__ptr[i]);
    }

    // Create Request Object
    Request::Ptr req = Request::Ptr(new Request());
    req->type = TYPE_GET;
    req->status = STATUS_PENDING;

    // Get Protocol
    std::string protocol = PROTOCOL_FILE;
    if(_arg1->__size > 0){
        if(0 != _arg1->__ptr[0]){
            protocol = _arg1->__ptr[0];
        }
    }

    // For each file
    for(i = 0; i < _arg0->__size; i++){
        FileRequest::Ptr f(new FileRequest());
        f->id = i;
        if(0 != _arg0->__ptr[i]){
            f->filename = _arg0->__ptr[i];
            f->surl = _arg0->__ptr[i];
        }
        f->status = STATUS_PENDING;
        f->protocol = protocol;
        // Store file
        req->fileRequests.push_back(f);
    }

    // Insert Request
    if(0 == store_request(req)){
        out._Result = request_status(soap,req);
        result = SOAP_OK;
    } else {
        // Set error Message
        result = soap_receiver_fault(soap,"Failed to store request id",0);
    }

    TRACE_LOG(("Done: srm__get"));
    return result;
}

/*
 * srm__copy
 */
int srm__copy(struct soap * soap, srm4ArrayOfstring *_arg0, srm4ArrayOfstring *_arg1, srm5ArrayOfboolean *_arg2, struct srm__copyResponse &out) {
    TRACE_LOG(("srm__copy called"));

    int i = 0;
    int result  = SOAP_FAULT;

    TRACE_LOG(("Request to Copy %d Files",_arg0->__size));
    for(i = 0; i < _arg0->__size; i++){
        TRACE_LOG(("File #%d: ",i));
        TRACE_LOG((" Src %s",_arg0->__ptr[i]));
        TRACE_LOG((" Destination %s",_arg1->__ptr[i]));
    }

    // Create Request Object
    Request::Ptr req(new Request());
    req->type   = TYPE_COPY;
    req->status = STATUS_PENDING;

    // For each file
    for(i = 0; i < _arg0->__size; i++){
        FileRequest::Ptr f(new FileRequest());
        f->id = i;
        if(0 != _arg0->__ptr[i]){
            f->filename = _arg0->__ptr[i];
            f->surl = _arg0->__ptr[i];
        }
        if(0 != _arg1->__ptr[i]){
            // For copy, store the destination on the turl
            f->turl = _arg1->__ptr[i];
        }
        f->status = STATUS_PENDING;
        // Store file
        req->fileRequests.push_back(f);
    }

    // Insert Request
    if(0 == store_request(req)){
        out._Result = request_status(soap,req);
        result = SOAP_OK;
    } else {
        // Set error Message
        result = soap_receiver_fault(soap,"Failed to store request id",0);
    }

    TRACE_LOG(("Done: srm__copy"));
    return result;
}

/*
 * srm__getRequestStatus
 */
int srm__getRequestStatus(struct soap * soap, int _arg0, struct srm__getRequestStatusResponse &out) {
    TRACE_LOG(("srm__getRequestStatus called"));

    TRACE_LOG(("Checking status of request %d",_arg0));

    int result = SOAP_OK;
    std::string error;

    // Find Request
    Request::Ptr req = get_request(_arg0);
    if(req.get() != 0){
        // Check the Request
        if(req->type == TYPE_PUT){
            // process Put pequest
            process_put_request(req);
        } else if(req->type == TYPE_GET){
            // process Get pequest
            process_get_request(req);
        } else if(req->type == TYPE_COPY){
            // process Copy pequest
            process_copy_request(req);
        } else {
            // Invalid Request Type
            result = SOAP_FAULT;
            error = "Invalid Request Type";
        }
        if(result == SOAP_OK){
            // Get Return value
            out._Result = request_status(soap,req);

            // Purge Old requests
            purge_requests();
        }
    } else {
        // No Request Found
        result = SOAP_FAULT;
        error = "No Request Found";
    }

    if(result != SOAP_OK){
        // Set error Message
        result = soap_receiver_fault(soap,error.c_str(),0);
    }

    TRACE_LOG(("Done: srm__getRequestStatus"));
    return result;
}

/*
 * srm__advisoryDelete
 */
int srm__advisoryDelete(struct soap * soap, srm4ArrayOfstring *_arg0, struct srm__advisoryDeleteResponse &out) {
    TRACE_LOG(("srm__advisoryDelete called"));

    int result = SOAP_FAULT;
    std::string error;
    for(int i = 0; i < _arg0->__size; ++i){
        if(0 != _arg0->__ptr[i]){
            TRACE_LOG(("Deleting SURL #%d: <%s>",i,_arg0->__ptr[i]));
            // Process Delete Request
            std::string file_error;
            process_delete_request(_arg0->__ptr[i],file_error);
            if(!file_error.empty()){
                error += (std::string)"Cannot delete " + _arg0->__ptr[i] + ": " + file_error + "\n";
            }
        }
    }
    if(!error.empty()){
        // Set error Message
        result = soap_receiver_fault(soap,error.c_str(),0);
    } else {
        result = SOAP_OK;
    }

    TRACE_LOG(("Done: srm__advisoryDelete"));
    return result;
}

/*
 * srm__setFileStatus
 */
int srm__setFileStatus(struct soap * soap, int _arg0, int _arg1, char *_arg2, struct srm__setFileStatusResponse &out) {
    TRACE_LOG(("srm__setFileStatus called"));

    int result = SOAP_FAULT;
    std::string error;

    // The Status can be only Running or Done
    std::string new_state;
    if(0 != _arg2){
        new_state = _arg2;
        if((new_state == STATUS_RUNNING) ||
           (new_state == STATUS_DONE)){
            result = SOAP_OK;
        }
    }
    if(result != SOAP_OK){
        error = "Invalid value for status";
        // set error message
        return soap_sender_fault(soap,error.c_str(),0);
    }
    TRACE_LOG(("Requested status is %s",new_state.c_str()));

    // Find Request
    FileRequest::Ptr file;
    Request::Ptr req = get_request(_arg0);
    if(req.get() != 0){
        // Find the file
        std::vector<FileRequest::Ptr>::iterator it;
        for(it = req->fileRequests.begin(); it != req->fileRequests.end(); ++it){
            if((*it)->id == _arg1){
                file = (*it);
                break;
            }
        }
        if(0 != file.get()){
            // Set File Status
            if(new_state == STATUS_RUNNING){
                if(req->type == TYPE_COPY) {
                    // Error, Can't Modify teh status to running
                    error = "Can't set file to Running on srmCopy";
                } else {
                    if(file->status == STATUS_READY){
                        file->status = STATUS_RUNNING;
                    } else if(file->status == STATUS_ACTIVE){
                        // Ignore
                    } else {
                        // Error, File not Ready
                        error = "File not Ready";
                    }
                }
            } else if(new_state == STATUS_DONE){
                if(req->type == TYPE_COPY) {
                    // Cancel request
                    abort_file(req,file);
                } else {
                    if(file->status == STATUS_RUNNING){
                        // Complete request
                        finish_file(req,file);
                    } else if(file->status == STATUS_READY){
                        if(req->type == TYPE_GET) {
                            // Complete request
                            finish_file(req,file);
                        } else {
                            // Cancel request
                            abort_file(req,file);
                        }
                    } else {
                        // Cancel request
                        abort_file(req,file);
                    }
                }
            }
            // Update Request Status
            update_request_status(req);
        } else {
            // Invalid Request Type
            result = SOAP_FAULT;
            error = "No File ID found";
        }
    } else {
        // No Request Found
        result = SOAP_FAULT;
        error = "No Request ID found";
    }
    if(result == SOAP_OK){
        // Get Return value
        out._Result = request_status(soap,req,file);

        // Purge Old requests
        purge_requests();
    } else {
        // Set error Message
        result = soap_receiver_fault(soap,error.c_str(),0);
    }

    TRACE_LOG(("Done: srm__setFileStatus"));
    return result;
}

/*
 * srm__getFileMetaData
 */
int srm__getFileMetaData(struct soap * soap, srm4ArrayOfstring *_arg0, struct srm__getFileMetaDataResponse &out) {
    TRACE_LOG(("srm__getFileMetaData called"));

    std::string error;
    int count_error = 0;
    int result = SOAP_FAULT;
    if(_arg0->__size > 0){
        out._Result = soap_new_srm2ArrayOfFileMetaData(soap,-1);
        out._Result->__size = _arg0->__size;
        out._Result->__ptr = (srm2__FileMetaData**)soap_malloc(soap, out._Result->__size * sizeof(srm2__FileMetaData*));
        for(int i = 0; i < _arg0->__size; ++i){
            if(0 != _arg0->__ptr[i]){
                TRACE_LOG(("Checking Stats for SURL #%d: <%s>",i,_arg0->__ptr[i]));
                std::string message;
                srm2__FileMetaData* file_stat = get_file_stat(soap,_arg0->__ptr[i],message);
                if(!message.empty()){
                    error += (std::string)"Error for file " + _arg0->__ptr[i] + ":" + message;
                    ++count_error;
                }
                out._Result->__ptr[i] = file_stat;
            } else {
                out._Result->__ptr[i] = 0;
            }
        }
    }
    if(error.empty()){
        result = SOAP_OK;
    } else {
        if(count_error == _arg0->__size){
            // All file missing: Set error Message
            result = soap_receiver_fault(soap,error.c_str(),0);
        } else {
            // Just skip the missing files
            result = SOAP_OK;
        }
    }
    TRACE_LOG(("Done: srm__getFileMetaData"));
    return result;
}

// -----------------------------------------------------------------------------
// Not Implemented
// -----------------------------------------------------------------------------
int srm__pin(struct soap * soap, srm4ArrayOfstring *_arg0, struct srm__pinResponse &out) {
    return SOAP_NO_METHOD ;
}
int srm__unPin(struct soap * soap, srm4ArrayOfstring *_arg0, int _arg1, struct srm__unPinResponse &out) {
    return SOAP_NO_METHOD ;
}
int srm__mkPermanent(struct soap * soap, srm4ArrayOfstring *_arg0, struct srm__mkPermanentResponse &out) {
    return SOAP_NO_METHOD ;
}
int srm__getEstGetTime(struct soap * soap, srm4ArrayOfstring *_arg0, srm4ArrayOfstring *_arg1, struct srm__getEstGetTimeResponse &out) {
    return SOAP_NO_METHOD ;
}
int srm__getEstPutTime(struct soap * soap, srm4ArrayOfstring *_arg0, srm4ArrayOfstring *_arg1, srm5ArrayOflong *_arg2, srm5ArrayOfboolean *_arg3, srm4ArrayOfstring *_arg4, struct srm__getEstPutTimeResponse & out) {
    return SOAP_NO_METHOD ;
}

} // end namespace srm
