/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004.
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  GLite Data SRM Util CPP: Common base of the specific test handlers
 *
 *  Authors: Zsolt Molnar <zsolt.molnar@cern.ch - http://www.zsoltmolnar.hu>
 *  Version info: $Id: SrmBusyTestHandler.cpp,v 1.2 2009/07/30 17:40:34 molnarzs Exp $
 *  Release: $Name:  $
 *
 */
#include "SrmBusyTestHandler.h"
#include "rules.h"
#include "log.h"

#include <boost/tokenizer.hpp>
#include <boost/lexical_cast.hpp>

namespace srm2 {

bool SrmBusyTestHandlerBase::_ongoing = false;
bool SrmBusyTestHandlerBase::_firstVisit = true;
int SrmBusyTestHandlerBase::_firstCount = 0;
int SrmBusyTestHandlerBase::_secondCount = 0;
bool SrmBusyTestHandlerBase::_isFirstPart = true;
int SrmBusyTestHandlerBase::_n = 0;
boost::scoped_ptr<boost::posix_time::ptime> SrmBusyTestHandlerBase::_timer;
long SrmBusyTestHandlerBase::_expectedBackoffTime = 0;

TestData<SrmBusyTestTraits>* SrmBusyTestDataExtractor::extract(const std::string& tData)
{
	typedef boost::tokenizer< boost::char_separator<char> > Tokenizer;
	boost::char_separator<char> sep(":");
	Tokenizer tokenizer(tData, sep);
	std::vector<std::string> tokens;
	std::copy(tokenizer.begin(), tokenizer.end(), std::back_inserter(tokens));
	static const int NUMBER_OF_TOKENS_IN_SRM_BUSY_TEST_DATA = 7;
	static const std::string SRM_BUSY_TEST_LABEL = "srm_busy";
	static const std::string SRM_BUSY_TEST_RESET = "TEST_RESET";
	static const size_t FIRST_RETURN_CODE_PLACE = 1;
	static const int C_PLACE = 2;
	static const int NIE_PLACE = 3;
	static const size_t SECOND_RETURN_CODE_PLACE = 4;
	static const int NRQ_PLACE = 5;
	static const int EWT_PLACE = 6;

	// Check is TEST_RESET request arrives
	if (tokens.size() == 2 &&
	    tokens[0] == SRM_BUSY_TEST_LABEL &&
	    tokens[1] == SRM_BUSY_TEST_RESET)
	{
		return new TestData<SrmBusyTestTraits>();
	}

	if (tokens.size() != NUMBER_OF_TOKENS_IN_SRM_BUSY_TEST_DATA ||
		tokens[0] != SRM_BUSY_TEST_LABEL)
	{
		return NULL;
	}

	FileStatusRule& sr = FileStatusRule::instance();
	srm2__TStatusCode firstReturnCode, secondReturnCode;
	bool res = sr.get_result(tokens.at(FIRST_RETURN_CODE_PLACE), firstReturnCode);
	res &= sr.get_result(tokens.at(SECOND_RETURN_CODE_PLACE), secondReturnCode);

	if (!res) {
		TRACE_LOG("SrmBusyTestDataExtractor::extract: rule cannot be applied");
		return NULL;
	}

	int c = boost::lexical_cast<int>(tokens.at(C_PLACE));
	int nIE = boost::lexical_cast<int>(tokens.at(NIE_PLACE));
	int nRQ = boost::lexical_cast<int>(tokens.at(NRQ_PLACE));
	int eWT = boost::lexical_cast<int>(tokens.at(EWT_PLACE));

	return new TestData<SrmBusyTestTraits>(tData, firstReturnCode, c, nIE, secondReturnCode, nRQ, eWT);
}

// -----------------------------------------------------------------------------
// The specific test data extractors for the srm busy tests.
// -----------------------------------------------------------------------------
TestData<SrmBusyTestTraits>* TestDataExtractor<SrmAbortFilesTraits, SrmBusyTestTraits>::extract(const SrmAbortFilesTraits::RequestType& request)
{
	return SrmBusyTestDataExtractor::extract(request.arrayOfSURLs->urlArray.at(0));
}

TestData<SrmBusyTestTraits>* TestDataExtractor<SrmRmTraits, SrmBusyTestTraits>::extract(const SrmRmTraits::RequestType& request)
{
	return SrmBusyTestDataExtractor::extract(request.arrayOfSURLs->urlArray.at(0));
}

TestData<SrmBusyTestTraits>* TestDataExtractor<SrmLsTraits, SrmBusyTestTraits>::extract(const SrmLsTraits::RequestType& request)
{
	return SrmBusyTestDataExtractor::extract(request.arrayOfSURLs->urlArray.at(0));
}

TestData<SrmBusyTestTraits>* TestDataExtractor<SrmStatusOfLsTraits, SrmBusyTestTraits>::extract(const SrmStatusOfLsTraits::RequestType& request)
{
	return SrmBusyTestDataExtractor::extract(request.requestToken);
}

TestData<SrmBusyTestTraits>* TestDataExtractor<SrmPrepareToGetTraits, SrmBusyTestTraits>::extract(const SrmPrepareToGetTraits::RequestType& request)
{
	return SrmBusyTestDataExtractor::extract(request.arrayOfFileRequests->requestArray.at(0)->sourceSURL);
}

TestData<SrmBusyTestTraits>* TestDataExtractor<SrmStatusOfGetTraits, SrmBusyTestTraits>::extract(const SrmStatusOfGetTraits::RequestType& request)
{
	return SrmBusyTestDataExtractor::extract(request.requestToken);
}

TestData<SrmBusyTestTraits>* TestDataExtractor<SrmPrepareToPutTraits, SrmBusyTestTraits>::extract(const SrmPrepareToPutTraits::RequestType& request)
{
	assert(request.arrayOfFileRequests);
	assert(request.arrayOfFileRequests->requestArray.at(0)->targetSURL);
	return SrmBusyTestDataExtractor::extract(*(request.arrayOfFileRequests->requestArray.at(0)->targetSURL));
}

TestData<SrmBusyTestTraits>* TestDataExtractor<SrmBringOnlineTraits, SrmBusyTestTraits>::extract(const SrmBringOnlineTraits::RequestType& request)
{
	assert(request.arrayOfFileRequests);
	assert(request.arrayOfFileRequests->requestArray.at(0));
	return SrmBusyTestDataExtractor::extract(request.arrayOfFileRequests->requestArray.at(0)->sourceSURL);
}

TestData<SrmBusyTestTraits>* TestDataExtractor<SrmGetSpaceMetaDataTraits, SrmBusyTestTraits>::extract(const SrmGetSpaceMetaDataTraits::RequestType& request)
{
	assert(request.arrayOfSpaceTokens);
	return SrmBusyTestDataExtractor::extract(request.arrayOfSpaceTokens->stringArray.at(0));
}

TestData<SrmBusyTestTraits>* TestDataExtractor<SrmGetSpaceTokensTraits, SrmBusyTestTraits>::extract(const SrmGetSpaceTokensTraits::RequestType& request)
{
	assert(request.userSpaceTokenDescription);
	return SrmBusyTestDataExtractor::extract(*(request.userSpaceTokenDescription));
}

TestData<SrmBusyTestTraits>* TestDataExtractor<SrmCopyTraits, SrmBusyTestTraits>::extract(const SrmCopyTraits::RequestType& request)
{
	assert(request.arrayOfFileRequests);
	assert(request.arrayOfFileRequests->requestArray.at(0));
	return SrmBusyTestDataExtractor::extract(request.arrayOfFileRequests->requestArray.at(0)->sourceSURL);
}

TestData<SrmBusyTestTraits>* TestDataExtractor<SrmStatusOfCopyTraits, SrmBusyTestTraits>::extract(const SrmStatusOfCopyTraits::RequestType& request)
{
	return SrmBusyTestDataExtractor::extract(request.requestToken);
}

TestData<SrmBusyTestTraits>* TestDataExtractor<SrmMkDirTraits, SrmBusyTestTraits>::extract(const SrmMkDirTraits::RequestType& request)
{
	return SrmBusyTestDataExtractor::extract(request.SURL);
}

TestData<SrmBusyTestTraits>* TestDataExtractor<SrmRmDirTraits, SrmBusyTestTraits>::extract(const SrmRmDirTraits::RequestType& request)
{
	return SrmBusyTestDataExtractor::extract(request.SURL);
}

TestData<SrmBusyTestTraits>* TestDataExtractor<SrmAbortRequestTraits, SrmBusyTestTraits>::extract(const SrmAbortRequestTraits::RequestType& request)
{
	return SrmBusyTestDataExtractor::extract(request.requestToken);
}

TestData<SrmBusyTestTraits>* TestDataExtractor<SrmReleaseFilesTraits, SrmBusyTestTraits>::extract(const SrmReleaseFilesTraits::RequestType& request)
{
	return SrmBusyTestDataExtractor::extract(*(request.requestToken));
}

// -----------------------------------------------------------------------------
// The specific reply generators for the srm busy tests.
// -----------------------------------------------------------------------------
namespace {

template<class RESPONSE>
void _setEWT(soap* soap, RESPONSE* r, const unsigned int eWT)
{
	assert(r);

    if (eWT == -1) {
    	r->remainingTotalRequestTime = 0;
    } else {
    	r->remainingTotalRequestTime =
			static_cast<int*>(soap_malloc(soap, sizeof(int)));
		*(r->remainingTotalRequestTime) = eWT;
    }
}

} // namespace

void SrmBusyTestReplyGenerator<SrmAbortFilesTraits>::generateReply(SrmAbortFilesTraits& traits, srm2__TStatusCode code, const unsigned int, const std::string&)
{
	srm2__srmAbortFilesResponse_& out = traits.reply;
	struct soap* soap = traits.soap;
    out.srmAbortFilesResponse = soap_new_srm2__srmAbortFilesResponse(soap,-1);
    out.srmAbortFilesResponse->returnStatus = soap_new_srm2__TReturnStatus(soap,-1);
    out.srmAbortFilesResponse->returnStatus->statusCode = code;
    out.srmAbortFilesResponse->returnStatus->explanation = 0;
    out.srmAbortFilesResponse->arrayOfFileStatuses = soap_new_srm2__ArrayOfTSURLReturnStatus(soap,-1);
}

void SrmBusyTestReplyGenerator<SrmRmTraits>::generateReply(SrmRmTraits& traits, srm2__TStatusCode code, const unsigned int, const std::string&)
{
	srm2__srmRmResponse_& out = traits.reply;
	struct soap* soap = traits.soap;
    out.srmRmResponse = soap_new_srm2__srmRmResponse(soap,-1);
    out.srmRmResponse->returnStatus = soap_new_srm2__TReturnStatus(soap,-1);
    out.srmRmResponse->returnStatus->statusCode  = code;
    out.srmRmResponse->returnStatus->explanation = 0;
    out.srmRmResponse->arrayOfFileStatuses = soap_new_srm2__ArrayOfTSURLReturnStatus(soap,-1);
}

void SrmBusyTestReplyGenerator<SrmLsTraits>::generateReply(SrmLsTraits& traits, srm2__TStatusCode code, const unsigned int, const std::string& token)
{
	srm2__srmLsResponse_& out = traits.reply;
	struct soap* soap = traits.soap;
    out.srmLsResponse = soap_new_srm2__srmLsResponse(soap, -1);
    out.srmLsResponse->returnStatus = soap_new_srm2__TReturnStatus(soap,-1);
    out.srmLsResponse->returnStatus->statusCode = code;
    out.srmLsResponse->returnStatus->explanation = 0;
    out.srmLsResponse->details = 0;
	out.srmLsResponse->requestToken = soap_new_std__string(soap, -1);
	*(out.srmLsResponse->requestToken) = token;
}

void SrmBusyTestReplyGenerator<SrmStatusOfLsTraits>::generateReply(SrmStatusOfLsTraits& traits, srm2__TStatusCode code, const unsigned int, const std::string&)
{
	srm2__srmStatusOfLsRequestResponse_& out = traits.reply;
	struct soap* soap = traits.soap;
    out.srmStatusOfLsRequestResponse = soap_new_srm2__srmStatusOfLsRequestResponse(soap,-1);
    out.srmStatusOfLsRequestResponse->returnStatus = soap_new_srm2__TReturnStatus(soap,-1);
    out.srmStatusOfLsRequestResponse->returnStatus->explanation = 0;
    out.srmStatusOfLsRequestResponse->returnStatus->statusCode = code;
    out.srmStatusOfLsRequestResponse->details = 0;
}

void SrmBusyTestReplyGenerator<SrmPrepareToGetTraits>::generateReply(SrmPrepareToGetTraits& traits, srm2__TStatusCode code, const unsigned int eWT, const std::string& token)
{
	srm2__srmPrepareToGetResponse_& out = traits.reply;
	struct soap* soap = traits.soap;
	out.srmPrepareToGetResponse = soap_new_srm2__srmPrepareToGetResponse(soap,-1);
    out.srmPrepareToGetResponse->returnStatus = soap_new_srm2__TReturnStatus(soap,-1);
    out.srmPrepareToGetResponse->returnStatus->statusCode = code;
    out.srmPrepareToGetResponse->returnStatus->explanation = 0;
    _setEWT(soap, out.srmPrepareToGetResponse, eWT);
    out.srmPrepareToGetResponse->arrayOfFileStatuses = 0;
	out.srmPrepareToGetResponse->requestToken = soap_new_std__string(soap, -1);
	*(out.srmPrepareToGetResponse->requestToken) = token;
}

void SrmBusyTestReplyGenerator<SrmStatusOfGetTraits>::generateReply(SrmStatusOfGetTraits& traits, srm2__TStatusCode code, const unsigned int eWT, const std::string& token)
{
	srm2__srmStatusOfGetRequestResponse_& out = traits.reply;
	struct soap* soap = traits.soap;
	out.srmStatusOfGetRequestResponse = soap_new_srm2__srmStatusOfGetRequestResponse(soap,-1);
    out.srmStatusOfGetRequestResponse->returnStatus = soap_new_srm2__TReturnStatus(soap,-1);
    out.srmStatusOfGetRequestResponse->returnStatus->statusCode = code;
    out.srmStatusOfGetRequestResponse->returnStatus->explanation = 0;
    _setEWT(soap, out.srmStatusOfGetRequestResponse, eWT);
    out.srmStatusOfGetRequestResponse->arrayOfFileStatuses       = 0;
}

void SrmBusyTestReplyGenerator<SrmPrepareToPutTraits>::generateReply(SrmPrepareToPutTraits& traits, srm2__TStatusCode code, const unsigned int eWT, const std::string& token)
{
	srm2__srmPrepareToPutResponse_& out = traits.reply;
	struct soap* soap = traits.soap;

    out.srmPrepareToPutResponse = soap_new_srm2__srmPrepareToPutResponse(soap,-1);
    out.srmPrepareToPutResponse->returnStatus = soap_new_srm2__TReturnStatus(soap,-1);
    out.srmPrepareToPutResponse->returnStatus->statusCode = code;
    out.srmPrepareToPutResponse->returnStatus->explanation = 0;
    _setEWT(soap, out.srmPrepareToPutResponse, eWT);
    out.srmPrepareToPutResponse->arrayOfFileStatuses       = 0;
	out.srmPrepareToPutResponse->requestToken = soap_new_std__string(soap, -1);
	*(out.srmPrepareToPutResponse->requestToken) = token;
}

void SrmBusyTestReplyGenerator<SrmBringOnlineTraits>::generateReply(SrmBringOnlineTraits& traits, srm2__TStatusCode code, const unsigned int eWT, const std::string& token)
{
	srm2__srmBringOnlineResponse_& out = traits.reply;
	struct soap* soap = traits.soap;

    out.srmBringOnlineResponse = soap_new_srm2__srmBringOnlineResponse(soap,-1);
    out.srmBringOnlineResponse->returnStatus = soap_new_srm2__TReturnStatus(soap,-1);
    out.srmBringOnlineResponse->returnStatus->statusCode = code;
    out.srmBringOnlineResponse->returnStatus->explanation = 0;
    _setEWT(soap, out.srmBringOnlineResponse, eWT);
    out.srmBringOnlineResponse->remainingDeferredStartTime= 0;
    out.srmBringOnlineResponse->arrayOfFileStatuses       = 0;
	out.srmBringOnlineResponse->requestToken = soap_new_std__string(soap, -1);
	*(out.srmBringOnlineResponse->requestToken) = token;
}

void SrmBusyTestReplyGenerator<SrmGetSpaceMetaDataTraits>::generateReply(SrmGetSpaceMetaDataTraits& traits, srm2__TStatusCode code, const unsigned int, const std::string& token)
{
	srm2__srmGetSpaceMetaDataResponse_& out = traits.reply;
	struct soap* soap = traits.soap;

    out.srmGetSpaceMetaDataResponse = soap_new_srm2__srmGetSpaceMetaDataResponse(soap,-1);
    out.srmGetSpaceMetaDataResponse->returnStatus = soap_new_srm2__TReturnStatus(soap,-1);
    out.srmGetSpaceMetaDataResponse->returnStatus->statusCode  = code;
    out.srmGetSpaceMetaDataResponse->returnStatus->explanation = 0;
    out.srmGetSpaceMetaDataResponse->arrayOfSpaceDetails = 0;
}

void SrmBusyTestReplyGenerator<SrmGetSpaceTokensTraits>::generateReply(SrmGetSpaceTokensTraits& traits, srm2__TStatusCode code, const unsigned int, const std::string& token)
{
	srm2__srmGetSpaceTokensResponse_& out = traits.reply;
	struct soap* soap = traits.soap;

    out.srmGetSpaceTokensResponse = soap_new_srm2__srmGetSpaceTokensResponse(soap,-1);
    out.srmGetSpaceTokensResponse->returnStatus = soap_new_srm2__TReturnStatus(soap,-1);
    out.srmGetSpaceTokensResponse->returnStatus->statusCode  = code;
    out.srmGetSpaceTokensResponse->returnStatus->explanation = 0;
    out.srmGetSpaceTokensResponse->arrayOfSpaceTokens        = 0;
}

void SrmBusyTestReplyGenerator<SrmCopyTraits>::generateReply(SrmCopyTraits& traits, srm2__TStatusCode code, const unsigned int eWT, const std::string& token)
{
	srm2__srmCopyResponse_& out = traits.reply;
	struct soap* soap = traits.soap;

    out.srmCopyResponse = soap_new_srm2__srmCopyResponse(soap,-1);
    out.srmCopyResponse->returnStatus = soap_new_srm2__TReturnStatus(soap,-1);
    out.srmCopyResponse->returnStatus->statusCode = code;
    out.srmCopyResponse->returnStatus->explanation = 0;
    _setEWT(soap, out.srmCopyResponse, eWT);
    out.srmCopyResponse->arrayOfFileStatuses       = 0;
	out.srmCopyResponse->requestToken = soap_new_std__string(soap, -1);
	*(out.srmCopyResponse->requestToken) = token;
}

void SrmBusyTestReplyGenerator<SrmStatusOfCopyTraits>::generateReply(SrmStatusOfCopyTraits& traits, srm2__TStatusCode code, const unsigned int eWT, const std::string& token)
{
	srm2__srmStatusOfCopyRequestResponse_& out = traits.reply;
	struct soap* soap = traits.soap;

    out.srmStatusOfCopyRequestResponse = soap_new_srm2__srmStatusOfCopyRequestResponse(soap,-1);
    out.srmStatusOfCopyRequestResponse->returnStatus = soap_new_srm2__TReturnStatus(soap,-1);
    out.srmStatusOfCopyRequestResponse->returnStatus->statusCode = code;
    out.srmStatusOfCopyRequestResponse->returnStatus->explanation = 0;
    _setEWT(soap, out.srmStatusOfCopyRequestResponse, eWT);
    out.srmStatusOfCopyRequestResponse->arrayOfFileStatuses       = 0;
}

void SrmBusyTestReplyGenerator<SrmMkDirTraits>::generateReply(SrmMkDirTraits& traits, srm2__TStatusCode code, const unsigned int, const std::string& token)
{
	srm2__srmMkdirResponse_& out = traits.reply;
	struct soap* soap = traits.soap;

    out.srmMkdirResponse = soap_new_srm2__srmMkdirResponse(soap,-1);
    out.srmMkdirResponse->returnStatus = soap_new_srm2__TReturnStatus(soap,-1);
    out.srmMkdirResponse->returnStatus->statusCode = code;
    out.srmMkdirResponse->returnStatus->explanation = 0;
}

void SrmBusyTestReplyGenerator<SrmRmDirTraits>::generateReply(SrmRmDirTraits& traits, srm2__TStatusCode code, const unsigned int, const std::string& token)
{
	srm2__srmRmdirResponse_& out = traits.reply;
	struct soap* soap = traits.soap;

    out.srmRmdirResponse = soap_new_srm2__srmRmdirResponse(soap,-1);
    out.srmRmdirResponse->returnStatus = soap_new_srm2__TReturnStatus(soap,-1);
    out.srmRmdirResponse->returnStatus->statusCode = code;
    out.srmRmdirResponse->returnStatus->explanation = 0;
}

void SrmBusyTestReplyGenerator<SrmAbortRequestTraits>::generateReply(SrmAbortRequestTraits& traits, srm2__TStatusCode, const unsigned int, const std::string&)
{
	srm2__srmAbortRequestResponse_& out = traits.reply;
	struct soap* soap = traits.soap;

    out.srmAbortRequestResponse = soap_new_srm2__srmAbortRequestResponse(soap,-1);
    out.srmAbortRequestResponse->returnStatus = soap_new_srm2__TReturnStatus(soap,-1);
    out.srmAbortRequestResponse->returnStatus->statusCode = srm2__TStatusCode__SRM_USCORESUCCESS;
    out.srmAbortRequestResponse->returnStatus->explanation = 0;
}

void SrmBusyTestReplyGenerator<SrmReleaseFilesTraits>::generateReply(SrmReleaseFilesTraits& traits, srm2__TStatusCode, const unsigned int, const std::string&)
{
	srm2__srmReleaseFilesResponse_& out = traits.reply;
	struct soap* soap = traits.soap;

    out.srmReleaseFilesResponse = soap_new_srm2__srmReleaseFilesResponse(soap,-1);
    out.srmReleaseFilesResponse->returnStatus = soap_new_srm2__TReturnStatus(soap,-1);
    out.srmReleaseFilesResponse->returnStatus->statusCode = srm2__TStatusCode__SRM_USCORESUCCESS;
    out.srmReleaseFilesResponse->returnStatus->explanation = 0;
    out.srmReleaseFilesResponse->arrayOfFileStatuses = 0;
}

} // namespace srm2
