/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004.
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  GLite Data SRM Util CPP: Common base of the specific test handlers
 *
 *  Authors: Zsolt Molnar <zsolt.molnar@cern.ch - http://www.zsoltmolnar.hu>
 *  Version info: $Id: SrmBusyTestHandler.h,v 1.2 2009/07/30 17:40:34 molnarzs Exp $
 *  Release: $Name:  $
 *
 */
#ifndef GLITE_DATA_SRM_UTIL_CPP_TEST_SRMBUSYTESTHANDLER_H_
#define GLITE_DATA_SRM_UTIL_CPP_TEST_SRMBUSYTESTHANDLER_H_

#include "TestHandler.h"
#include "log.h"
#include "utils.h"

#include <cmath>
#include <boost/timer.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/scoped_ptr.hpp>
#include <boost/lexical_cast.hpp>

namespace srm2 {

struct SrmBusyTestTraits {};

template<>
class TestData<SrmBusyTestTraits>
{
public:
	TestData(const std::string& origStr, srm2__TStatusCode fst, int c, int nIE, srm2__TStatusCode sst, int nRQ, int eWT)
		: _firstStatusCode(fst),
		  _c(c),
		  _nIE(nIE),
		  _secondStatusCode(sst),
		  _nRQ(nRQ),
		  _eWT(eWT),
		  _origStr(origStr),
		  _isReset(false)
	{ /* EMPTY */ }

	// default constructed test data will instruct the test handler to reset the test
	TestData()
		: _isReset(true)
	{ /* EMPTY */ }

	srm2__TStatusCode firstStatusCode() const {return _firstStatusCode;}
	int c() const {return _c;}
	int nIE() const {return _nIE;}
	srm2__TStatusCode secondStatusCode() const {return _secondStatusCode;}
	int nRQ() const {return _nRQ;}
	int eWT() const {return _eWT;}
	const std::string& origStr() const {return _origStr;}
	const bool isReset() const {return _isReset;}

private:
	srm2__TStatusCode _firstStatusCode;
	int _c;
	int _nIE;
	srm2__TStatusCode _secondStatusCode;
	int _nRQ;
	int _eWT;
	std::string _origStr;
	bool _isReset;
};

// -----------------------------------------------------------------------------
// The specific Srm call traits
// -----------------------------------------------------------------------------
typedef SrmOpTraits<srm2__srmAbortFilesRequest, srm2__srmAbortFilesResponse_> SrmAbortFilesTraits;
typedef SrmOpTraits<srm2__srmRmRequest, srm2__srmRmResponse_> SrmRmTraits;
typedef SrmOpTraits<srm2__srmLsRequest, srm2__srmLsResponse_> SrmLsTraits;
typedef SrmOpTraits<srm2__srmStatusOfLsRequestRequest, srm2__srmStatusOfLsRequestResponse_> SrmStatusOfLsTraits;
typedef SrmOpTraits<srm2__srmPrepareToGetRequest, srm2__srmPrepareToGetResponse_> SrmPrepareToGetTraits;
typedef SrmOpTraits<srm2__srmStatusOfGetRequestRequest, srm2__srmStatusOfGetRequestResponse_> SrmStatusOfGetTraits;
typedef SrmOpTraits<srm2__srmPrepareToPutRequest, srm2__srmPrepareToPutResponse_> SrmPrepareToPutTraits;
typedef SrmOpTraits<srm2__srmBringOnlineRequest, srm2__srmBringOnlineResponse_> SrmBringOnlineTraits;
typedef SrmOpTraits<srm2__srmGetSpaceMetaDataRequest, srm2__srmGetSpaceMetaDataResponse_> SrmGetSpaceMetaDataTraits;
typedef SrmOpTraits<srm2__srmGetSpaceTokensRequest, srm2__srmGetSpaceTokensResponse_> SrmGetSpaceTokensTraits;
typedef SrmOpTraits<srm2__srmCopyRequest, srm2__srmCopyResponse_> SrmCopyTraits;
typedef SrmOpTraits<srm2__srmStatusOfCopyRequestRequest, srm2__srmStatusOfCopyRequestResponse_> SrmStatusOfCopyTraits;
typedef SrmOpTraits<srm2__srmMkdirRequest, srm2__srmMkdirResponse_> SrmMkDirTraits;
typedef SrmOpTraits<srm2__srmRmdirRequest, srm2__srmRmdirResponse_> SrmRmDirTraits;
typedef SrmOpTraits<srm2__srmAbortRequestRequest, srm2__srmAbortRequestResponse_> SrmAbortRequestTraits;
typedef SrmOpTraits<srm2__srmReleaseFilesRequest, srm2__srmReleaseFilesResponse_> SrmReleaseFilesTraits;


// -----------------------------------------------------------------------------
// The specific reply generators for the srm busy tests.
// -----------------------------------------------------------------------------
template <class SrmCallTraits> struct SrmBusyTestReplyGenerator {};

template<>
struct SrmBusyTestReplyGenerator<SrmAbortFilesTraits>
{
	static void generateReply(SrmAbortFilesTraits& traits, srm2__TStatusCode code, const unsigned int eWT, const std::string& token = "");
};

template<>
struct SrmBusyTestReplyGenerator<SrmRmTraits>
{
	static void generateReply(SrmRmTraits& traits, srm2__TStatusCode code, const unsigned int eWT, const std::string& token = "");
};

template<>
struct SrmBusyTestReplyGenerator<SrmLsTraits>
{
	static void generateReply(SrmLsTraits& traits, srm2__TStatusCode code, const unsigned int eWT, const std::string& token);
};

template<>
struct SrmBusyTestReplyGenerator<SrmStatusOfLsTraits>
{
	static void generateReply(SrmStatusOfLsTraits& traits, srm2__TStatusCode code, const unsigned int eWT, const std::string& token = "");
};

template<>
struct SrmBusyTestReplyGenerator<SrmPrepareToGetTraits>
{
	static void generateReply(SrmPrepareToGetTraits& traits, srm2__TStatusCode code, const unsigned int eWT, const std::string& token = "");
};

template<>
struct SrmBusyTestReplyGenerator<SrmStatusOfGetTraits>
{
	static void generateReply(SrmStatusOfGetTraits& traits, srm2__TStatusCode code, const unsigned int eWT, const std::string& token = "");
};

template<>
struct SrmBusyTestReplyGenerator<SrmPrepareToPutTraits>
{
	static void generateReply(SrmPrepareToPutTraits& traits, srm2__TStatusCode code, const unsigned int eWT, const std::string& token = "");
};

template<>
struct SrmBusyTestReplyGenerator<SrmBringOnlineTraits>
{
	static void generateReply(SrmBringOnlineTraits& traits, srm2__TStatusCode code, const unsigned int eWT, const std::string& token = "");
};

template<>
struct SrmBusyTestReplyGenerator<SrmGetSpaceMetaDataTraits>
{
	static void generateReply(SrmGetSpaceMetaDataTraits& traits, srm2__TStatusCode code, const unsigned int eWT, const std::string& token = "");
};

template<>
struct SrmBusyTestReplyGenerator<SrmGetSpaceTokensTraits>
{
	static void generateReply(SrmGetSpaceTokensTraits& traits, srm2__TStatusCode code, const unsigned int eWT, const std::string& token = "");
};

template<>
struct SrmBusyTestReplyGenerator<SrmCopyTraits>
{
	static void generateReply(SrmCopyTraits& traits, srm2__TStatusCode code, const unsigned int eWT, const std::string& token = "");
};

template<>
struct SrmBusyTestReplyGenerator<SrmStatusOfCopyTraits>
{
	static void generateReply(SrmStatusOfCopyTraits& traits, srm2__TStatusCode code, const unsigned int eWT, const std::string& token = "");
};

template<>
struct SrmBusyTestReplyGenerator<SrmMkDirTraits>
{
	static void generateReply(SrmMkDirTraits& traits, srm2__TStatusCode code, const unsigned int eWT, const std::string& token = "");
};

template<>
struct SrmBusyTestReplyGenerator<SrmRmDirTraits>
{
	static void generateReply(SrmRmDirTraits& traits, srm2__TStatusCode code, const unsigned int eWT, const std::string& token = "");
};

template<>
struct SrmBusyTestReplyGenerator<SrmAbortRequestTraits>
{
	static void generateReply(SrmAbortRequestTraits& traits, srm2__TStatusCode code, const unsigned int eWT, const std::string& token = "");
};

template<>
struct SrmBusyTestReplyGenerator<SrmReleaseFilesTraits>
{
	static void generateReply(SrmReleaseFilesTraits& traits, srm2__TStatusCode code, const unsigned int eWT, const std::string& token = "");
};


/**
 * Handle the SrmBusy test case, for all call cases.
 */
class SrmBusyTestHandlerBase : public ITestHandler
{
protected:
	virtual ~SrmBusyTestHandlerBase() { /* EMPTY */ }

	static long _Delta() { return 2; }

	static bool _ongoing;
	static bool _firstVisit;
	static int _firstCount;
	static int _secondCount;
	static bool _isFirstPart;
	static int _n;
	static boost::scoped_ptr<boost::posix_time::ptime> _timer;
	static long _expectedBackoffTime;
};


template <class SrmCallTraits>
class SrmBusyTestHandler : public SrmBusyTestHandlerBase
{
public:
	// Create a test handler instance. Re-create it if the test has been finished.
	// There may run only one SrmBusy unit test instance, no parallel testing of
	// the SRM calls allowed. The test handler instances share the same state machine.
	// The state machine is not aware if the test calls change in between, so
	// for example, an SrmLsUpdateStatus call will continue with the same timers,
	// numbers, etc. that a previous SrmLs call used. It is normal behaviour,
	// as the call pair is considered as two stages of the same Ls operation.
	// The test, however, will not check if an SrmLs is followed by (for example)
	// a PrepareToGetUpdateStatius. The client side logic must be error-free
	// from this respect!
	static SrmBusyTestHandler* instance(TestData<SrmBusyTestTraits>* sbtt,
										SrmCallTraits* callTraits)
	{
		assert(sbtt);
		assert(callTraits);
		static typename boost::shared_ptr<SrmBusyTestHandler> self;

		if (!self.get()) {
			self.reset(new SrmBusyTestHandler(sbtt, callTraits));
		} else {
			self->updateTraits(callTraits, sbtt);
		}

		// Restart the test
		if (!self->_ongoing) {
			_firstCount = sbtt->nIE();
			_secondCount = sbtt->nRQ();
			_firstVisit = true;
		}

		return self.get();
	}

	virtual void handle()
	{
		if (_sbtt->isReset()) {
			TRACE_LOG("Test reset request arrived");
			SrmBusyTestReplyGenerator<SrmCallTraits>::generateReply(*_callTraits, srm2__TStatusCode__SRM_USCORESUCCESS, -1, _sbtt->origStr());
			_ongoing = false;
			return;
		}
		// Check if it is the very first event handling. In this case,
		// do not check the backoff time. Set the ongoing status to true.
		if (_firstVisit) {
			TRACE_LOG("SrmBusyTestHandler::handle: test restarted");
			_ongoing = true;
			_firstVisit = false;
			_isFirstPart = true;
			_n = 1;
			_timer.reset(new boost::posix_time::ptime(boost::posix_time::second_clock::universal_time()));
		} else {
			// Get the actual time if it is not the first event handling
			// Check if it corresponds the expected time
			if (std::abs(Utils::elapsedTime(*_timer) - _expectedBackoffTime) > _Delta()) {
				// If not, return SRM_CUSTOM_STATUS, and test failed (reset)
				TRACE_LOG("SrmBusyTestHandler::handle detected unexpected backoff time");
				TRACE_LOG("Expected: " + boost::lexical_cast<std::string>(_expectedBackoffTime));
				TRACE_LOG("Real client backoff: " + boost::lexical_cast<std::string>(Utils::elapsedTime(*_timer)));
				SrmBusyTestReplyGenerator<SrmCallTraits>::generateReply(*_callTraits, srm2__TStatusCode__SRM_USCORECUSTOM_USCORESTATUS, -1, _sbtt->origStr());
				_ongoing = false;
				return;
			}
		}
		// if the test is ongoing:
		// Update counters. Switch to the second one if it reached zero.
		if (_isFirstPart && _firstCount == 0)
			_isFirstPart = false;

		--(_isFirstPart ? _firstCount : _secondCount);

		// Check if the test is over (all counters dropped to zero)
		if (_firstCount + _secondCount == -1) {
			_ongoing = false;
			TRACE_LOG("SrmBusyTestHandler::handle test is over successfully");
			SrmBusyTestReplyGenerator<SrmCallTraits>::generateReply(*_callTraits, srm2__TStatusCode__SRM_USCORESUCCESS, -1, _sbtt->origStr());
			return;
		}
		// Update the expected backoff time. If eWT is set, and we are in the second phase,
		// use that, else calculate, according to the actual return count / constant.
		_timer.reset(new boost::posix_time::ptime(boost::posix_time::second_clock::universal_time()));
		static const int NO_EWT_GIVEN = -1;

		_expectedBackoffTime =
			(!_isFirstPart && _sbtt->eWT() != NO_EWT_GIVEN) ?
				_sbtt->eWT() :
				_sbtt->c() * (static_cast<long>(std::pow(2.0, _n)) - 1);

		++_n;
		srm2__TStatusCode code = _isFirstPart ? _sbtt->firstStatusCode() : _sbtt->secondStatusCode();
		TRACE_LOG("SrmBusyTestHandler::handle returns code " + boost::lexical_cast<std::string>(code));
		SrmBusyTestReplyGenerator<SrmCallTraits>::generateReply(*_callTraits, code, _sbtt->eWT(), _sbtt->origStr());
	}

	virtual ~SrmBusyTestHandler() { /* EMPTY */ }

	void updateTraits(SrmCallTraits* callTraits, TestData<SrmBusyTestTraits>* sbtt)
	{
		assert(callTraits);
		_callTraits = callTraits;
		_sbtt.reset(sbtt);
	}

private:
	SrmBusyTestHandler(TestData<SrmBusyTestTraits>* sbtt,
			           SrmCallTraits* callTraits)
		: _sbtt(sbtt),
		  _callTraits(callTraits)
    {
		assert(sbtt);
		assert(callTraits);
	}

	boost::scoped_ptr<TestData<SrmBusyTestTraits> > _sbtt;
	SrmCallTraits* _callTraits;
};

template<class SrmCallTraits>
struct TestHandlerFactory<SrmCallTraits, SrmBusyTestTraits>
{
	static ITestHandler* getTestHandler(SrmCallTraits* traits)
	{
		assert(traits);
		TestData<SrmBusyTestTraits>* sbtt = TestDataExtractor<SrmCallTraits, SrmBusyTestTraits>::extract(traits->request);

		if (sbtt) {
			TRACE_LOG("TestHandlerFactory::createTestHandler extracted test data successfully.");
			return SrmBusyTestHandler<SrmCallTraits>::instance(sbtt, traits);
		}

		return NULL;
	}
};

/**
 * Retrieve the test data corresponding to a file name string, according to the
 * rules of the SrmBusy test.
 */
struct SrmBusyTestDataExtractor
{
	static TestData<SrmBusyTestTraits>* extract(const std::string& tData);
};

// -----------------------------------------------------------------------------
// The specific test data extractors: get the data of a specific test from a
// specific request type
// They are all template specializations.
// -----------------------------------------------------------------------------
template<>
struct TestDataExtractor<SrmAbortFilesTraits, SrmBusyTestTraits>
{
	static TestData<SrmBusyTestTraits>* extract(const SrmAbortFilesTraits::RequestType& request);
};

template<>
struct TestDataExtractor<SrmRmTraits, SrmBusyTestTraits>
{
	static TestData<SrmBusyTestTraits>* extract(const SrmRmTraits::RequestType& request);
};

template<>
struct TestDataExtractor<SrmLsTraits, SrmBusyTestTraits>
{
	static TestData<SrmBusyTestTraits>* extract(const SrmLsTraits::RequestType& request);
};

template<>
struct TestDataExtractor<SrmStatusOfLsTraits, SrmBusyTestTraits>
{
	static TestData<SrmBusyTestTraits>* extract(const SrmStatusOfLsTraits::RequestType& request);
};

template<>
struct TestDataExtractor<SrmPrepareToGetTraits, SrmBusyTestTraits>
{
	static TestData<SrmBusyTestTraits>* extract(const SrmPrepareToGetTraits::RequestType& request);
};

template<>
struct TestDataExtractor<SrmStatusOfGetTraits, SrmBusyTestTraits>
{
	static TestData<SrmBusyTestTraits>* extract(const SrmStatusOfGetTraits::RequestType& request);
};

template<>
struct TestDataExtractor<SrmPrepareToPutTraits, SrmBusyTestTraits>
{
	static TestData<SrmBusyTestTraits>* extract(const SrmPrepareToPutTraits::RequestType& request);
};

template<>
struct TestDataExtractor<SrmBringOnlineTraits, SrmBusyTestTraits>
{
	static TestData<SrmBusyTestTraits>* extract(const SrmBringOnlineTraits::RequestType& request);
};

template<>
struct TestDataExtractor<SrmGetSpaceMetaDataTraits, SrmBusyTestTraits>
{
	static TestData<SrmBusyTestTraits>* extract(const SrmGetSpaceMetaDataTraits::RequestType& request);
};

template<>
struct TestDataExtractor<SrmGetSpaceTokensTraits, SrmBusyTestTraits>
{
	static TestData<SrmBusyTestTraits>* extract(const SrmGetSpaceTokensTraits::RequestType& request);
};

template<>
struct TestDataExtractor<SrmCopyTraits, SrmBusyTestTraits>
{
	static TestData<SrmBusyTestTraits>* extract(const SrmCopyTraits::RequestType& request);
};

template<>
struct TestDataExtractor<SrmStatusOfCopyTraits, SrmBusyTestTraits>
{
	static TestData<SrmBusyTestTraits>* extract(const SrmStatusOfCopyTraits::RequestType& request);
};

template<>
struct TestDataExtractor<SrmMkDirTraits, SrmBusyTestTraits>
{
	static TestData<SrmBusyTestTraits>* extract(const SrmMkDirTraits::RequestType& request);
};

template<>
struct TestDataExtractor<SrmRmDirTraits, SrmBusyTestTraits>
{
	static TestData<SrmBusyTestTraits>* extract(const SrmRmDirTraits::RequestType& request);
};

template<>
struct TestDataExtractor<SrmAbortRequestTraits, SrmBusyTestTraits>
{
	static TestData<SrmBusyTestTraits>* extract(const SrmAbortRequestTraits::RequestType& request);
};

template<>
struct TestDataExtractor<SrmReleaseFilesTraits, SrmBusyTestTraits>
{
	static TestData<SrmBusyTestTraits>* extract(const SrmReleaseFilesTraits::RequestType& request);
};

} // namespace srm2

#endif // #ifndef GLITE_DATA_SRM_UTIL_CPP_TEST_SRMBUSYTESTHANDLER_H_
