/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004.
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  GLite Data SRM Util CPP: Srm 2.2 Server pattern based rules
 *
 *  Authors: Paolo Tedesco <paolo.tedesco@cern.ch>
 *  Version info: $Id: rules.cpp,v 1.4 2009/07/30 17:40:34 molnarzs Exp $
 *  Release: $Name:  $
 *
 */

#include "rules.h"
#include "log.h"

using namespace srm2;

/**
 * singleton instance
 */
ChecksumRule& ChecksumRule::instance(){
    static ChecksumRule s_instance;
    return s_instance;
}

/**
 * get checksum information from surl
 */
ChecksumRule::ChecksumInfo ChecksumRule::get(const std::string& surl)
{
	TRACE_LOG("ChecksumRule::get - parameter: " << surl);
    ChecksumInfo ret;
    boost::match_results<std::string::const_iterator> what;
    boost::match_flag_type flags = boost::match_default;
    if (boost::regex_search(surl.begin(),surl.end(),what,m_regex,flags)) {
    	TRACE_LOG("ChecksumRule::get - rule match");
        ret.type    = what[1];
        ret.value   = what[2];
    } else {
    	TRACE_LOG("ChecksumRule::get - rule NOT match");
    }

    return ret;
}

/**
 * constructor
 */
ChecksumRule::ChecksumRule() : m_regex("cksm-(\\w+)-([0-9a-fA-F]+)"){
}

//------------------------------------------------------------------------------
//                              FileLocalityRule
//------------------------------------------------------------------------------

/**
 * singleton instance
 */
FileLocalityRule& FileLocalityRule::instance(){
    static FileLocalityRule s_instance;
    return s_instance;
}

/**
 * get file locality from surl
 */
srm2::srm2__TFileLocality FileLocalityRule::get(const std::string& surl){
    srm2::srm2__TFileLocality ret = srm2__TFileLocality__ONLINE;
    boost::match_results<std::string::const_iterator> what;
    boost::match_flag_type flags = boost::match_default;
    bool is_match = boost::regex_search(surl.begin(),surl.end(),what,m_regex,flags);
    if (is_match){
        const std::string& locality = what[1];
        StringToLocalityMap::const_iterator it = m_map.find(locality);
        if (m_map.end() != it){
            ret = it->second;
        }
    }
    return ret;
}

/**
 * constructor
 */
FileLocalityRule::FileLocalityRule() : m_regex("locality-([\\u_]+)"){
    m_map["ONLINE"]                 = srm2__TFileLocality__ONLINE;
    m_map["NEARLINE"]               = srm2__TFileLocality__NEARLINE;
    m_map["ONLINE_AND_NEARLINE"]    = srm2__TFileLocality__ONLINE_USCOREAND_USCORENEARLINE;
    m_map["LOST"]                   = srm2__TFileLocality__LOST;
    m_map["NONE"]                   = srm2__TFileLocality__NONE;
    m_map["UNAVAILABLE"]            = srm2__TFileLocality__UNAVAILABLE;
}

//------------------------------------------------------------------------------
//                              StatusRule
//------------------------------------------------------------------------------

/**
 * constructor
 */
StatusRule::StatusRule(){
    m_map["SRM_SUCCESS"]                = srm2__TStatusCode__SRM_USCORESUCCESS;
    m_map["SRM_FAILURE"]                = srm2__TStatusCode__SRM_USCOREFAILURE;
    m_map["SRM_AUTHENTICATION_FAILURE"] = srm2__TStatusCode__SRM_USCOREAUTHENTICATION_USCOREFAILURE;
    m_map["SRM_AUTHORIZATION_FAILURE"]  = srm2__TStatusCode__SRM_USCOREAUTHORIZATION_USCOREFAILURE;
    m_map["SRM_INVALID_REQUEST"]        = srm2__TStatusCode__SRM_USCOREINVALID_USCOREREQUEST;
    m_map["SRM_INVALID_PATH"]           = srm2__TStatusCode__SRM_USCOREINVALID_USCOREPATH;
    m_map["SRM_FILE_LIFETIME_EXPIRED"]  = srm2__TStatusCode__SRM_USCOREFILE_USCORELIFETIME_USCOREEXPIRED;
    m_map["SRM_SPACE_LIFETIME_EXPIRED"] = srm2__TStatusCode__SRM_USCORESPACE_USCORELIFETIME_USCOREEXPIRED;
    m_map["SRM_EXCEED_ALLOCATION"]      = srm2__TStatusCode__SRM_USCOREEXCEED_USCOREALLOCATION;
    m_map["SRM_NO_USER_SPACE"]          = srm2__TStatusCode__SRM_USCORENO_USCOREUSER_USCORESPACE;
    m_map["SRM_NO_FREE_SPACE"]          = srm2__TStatusCode__SRM_USCORENO_USCOREFREE_USCORESPACE;
    m_map["SRM_DUPLICATION_ERROR"]      = srm2__TStatusCode__SRM_USCOREDUPLICATION_USCOREERROR;
    m_map["SRM_NON_EMPTY_DIRECTORY"]    = srm2__TStatusCode__SRM_USCORENON_USCOREEMPTY_USCOREDIRECTORY;
    m_map["SRM_TOO_MANY_RESULTS"]       = srm2__TStatusCode__SRM_USCORETOO_USCOREMANY_USCORERESULTS;
    m_map["SRM_INTERNAL_ERROR"]         = srm2__TStatusCode__SRM_USCOREINTERNAL_USCOREERROR;
    m_map["SRM_FATAL_INTERNAL_ERROR"]   = srm2__TStatusCode__SRM_USCOREFATAL_USCOREINTERNAL_USCOREERROR;
    m_map["SRM_NOT_SUPPORTED"]          = srm2__TStatusCode__SRM_USCORENOT_USCORESUPPORTED;
    m_map["SRM_REQUEST_QUEUED"]         = srm2__TStatusCode__SRM_USCOREREQUEST_USCOREQUEUED;
    m_map["SRM_REQUEST_INPROGRESS"]     = srm2__TStatusCode__SRM_USCOREREQUEST_USCOREINPROGRESS;
    m_map["SRM_REQUEST_SUSPENDED"]      = srm2__TStatusCode__SRM_USCOREREQUEST_USCORESUSPENDED;
    m_map["SRM_ABORTED"]                = srm2__TStatusCode__SRM_USCOREABORTED;
    m_map["SRM_RELEASED"]               = srm2__TStatusCode__SRM_USCORERELEASED;
    m_map["SRM_FILE_PINNED"]            = srm2__TStatusCode__SRM_USCOREFILE_USCOREPINNED;
    m_map["SRM_FILE_IN_CACHE"]          = srm2__TStatusCode__SRM_USCOREFILE_USCOREIN_USCORECACHE;
    m_map["SRM_SPACE_AVAILABLE"]        = srm2__TStatusCode__SRM_USCORESPACE_USCOREAVAILABLE;
    m_map["SRM_LOWER_SPACE_GRANTED"]    = srm2__TStatusCode__SRM_USCORELOWER_USCORESPACE_USCOREGRANTED;
    m_map["SRM_DONE"]                   = srm2__TStatusCode__SRM_USCOREDONE;
    m_map["SRM_PARTIAL_SUCCESS"]        = srm2__TStatusCode__SRM_USCOREPARTIAL_USCORESUCCESS;
    m_map["SRM_REQUEST_TIMED_OUT"]      = srm2__TStatusCode__SRM_USCOREREQUEST_USCORETIMED_USCOREOUT;
    m_map["SRM_LAST_COPY"]              = srm2__TStatusCode__SRM_USCORELAST_USCORECOPY;
    m_map["SRM_FILE_BUSY"]              = srm2__TStatusCode__SRM_USCOREFILE_USCOREBUSY;
    m_map["SRM_FILE_LOST"]              = srm2__TStatusCode__SRM_USCOREFILE_USCORELOST;
    m_map["SRM_FILE_UNAVAILABLE"]       = srm2__TStatusCode__SRM_USCOREFILE_USCOREUNAVAILABLE;
    m_map["SRM_CUSTOM_STATUS"]          = srm2__TStatusCode__SRM_USCORECUSTOM_USCORESTATUS;
}

/**
 * destructor
 */
StatusRule::~StatusRule(){
}

/**
 * translate a string to an srm status code
 * @param srm_result [IN] the string to translate
 * @param code [OUT] the srm status code
 * @return true if found, false otherwise
 */
bool StatusRule::get_result(const std::string& srm_result,srm2__TStatusCode& code){
    return translate_string(srm_result,m_map,code);
}

//------------------------------------------------------------------------------
//                              FileStatusRule
//------------------------------------------------------------------------------

/**
 * singleton instance
 */
FileStatusRule& FileStatusRule::instance(){
    static FileStatusRule s_instance;
    return s_instance;
}

/**
 * is_match
 * return true if file name matches the regular expression pattern and
 * (optionally) the operation name, e.g.
 * "this-one-returns-SRM_NO_FREE_SPACE-on-get", "get" ->
 *      true, code = srm2__TStatusCode__SRM_USCORENO_USCOREFREE_USCORESPACE
 * @param file_name [IN] the file name to match
 * @param operation [IN] optional operation name that acts as a filter, use
 *      empty string to disable checking
 * @param code [OUT] the srm2__TStatusCode found in the file name
 * @return true if the file name matches the pattern
 *      this-one-returns-(code)[-on-(operation)], false otherwise
 */
bool FileStatusRule::is_match(
    const std::string& file_name,
    const std::string& operation,
    srm2__TStatusCode& code) {
    boost::match_results<std::string::const_iterator> what;
    boost::match_flag_type flags = boost::match_default;
    bool is_match = boost::regex_search(file_name.begin(),file_name.end(),what,m_regex,flags);
    if (is_match){
        const std::string& srm_result = what[1];
        const std::string& op_name = what[3];
        is_match = get_result(srm_result,code);
        if (is_match && (0 != operation.length()) && (0 != op_name.length())){
            is_match = (operation == op_name);
        }
    }
    TRACE_LOG("Checking matches for file '" << file_name << "' on operation '" << operation << "' : is_match = " << is_match);
    return is_match;
}

/**
 * constructor
 */
FileStatusRule::FileStatusRule() : m_regex("this-one-returns-([\\u_]+)(-on-(\\l+))?") {
}

/**
 * destructor
 */
FileStatusRule::~FileStatusRule(){
}

//------------------------------------------------------------------------------
//                              SpaceMetaDataRule
//------------------------------------------------------------------------------

/**
 * singleton instance
 */
SpaceMetaDataRule& SpaceMetaDataRule::instance(){
    static SpaceMetaDataRule s_instance;
    return s_instance;
}

/**
 * is_token_error
 * @param token [IN] token
 * @param code [OUT] error code
 * @return true if token matches token error expression, false otherwise
 */
bool SpaceMetaDataRule::is_token_error(
    const std::string& token,
    srm2__TStatusCode& code){
    return is_error(token,"token",code);
}

/**
 * is_request_error
 * @param token [IN] token
 * @param code [OUT] error code
 * @return true if token matches request error expression, false otherwise
 */
bool SpaceMetaDataRule::is_request_error(
    const std::string& token,
    srm2__TStatusCode& code){
    return is_error(token,"request",code);
}


/**
 * is_token_rule
 * @param token [IN] token
 * @param code [OUT] error code
 * @return true if token matches expression, false otherwise
 */
bool SpaceMetaDataRule::is_token_rule(
    const std::string&      token,
    struct soap*            soap,
    srm2__TMetaDataSpace&   space){

    boost::match_results<std::string::const_iterator> what;
    boost::match_flag_type flags = boost::match_default;
    bool is_rule = boost::regex_search(token.begin(),token.end(),what,m_spaceRegex,flags);
    if (is_rule){
        const std::string& retention_policy     = what[1];
        const std::string& access_latency       = what[2];
        const std::string& owner                = what[3];
        const std::string& total_size           = what[4];
        const std::string& guaranteed_size      = what[5];
        const std::string& unused_size          = what[6];
        const std::string& lifetime_assigned    = what[7];
        const std::string& lifetime_left        = what[8];
        if (!retention_policy.empty() || !access_latency.empty()){
            space.retentionPolicyInfo = soap_new_srm2__TRetentionPolicyInfo(soap,-1);
            translate_retention_policy(retention_policy,space.retentionPolicyInfo->retentionPolicy);
            space.retentionPolicyInfo->accessLatency = 0;
            if (!access_latency.empty()){
                space.retentionPolicyInfo->accessLatency = (srm2__TAccessLatency*)soap_malloc(soap,sizeof(srm2__TAccessLatency));
                translate_access_latency(access_latency,*space.retentionPolicyInfo->accessLatency);
            }
        }
        if (!owner.empty()){
            space.owner = soap_new_std__string(soap,-1);
            *space.owner = owner;
        }
        convert_optional_numeric_value(soap,total_size,         space.totalSize);
        convert_optional_numeric_value(soap,guaranteed_size,    space.guaranteedSize);
        convert_optional_numeric_value(soap,unused_size,        space.unusedSize);
        convert_optional_numeric_value(soap,lifetime_assigned,  space.lifetimeAssigned);
        convert_optional_numeric_value(soap,lifetime_left,      space.lifetimeLeft);
    }
    TRACE_LOG("checking rule for token '" << token << "' : is_rule = " << is_rule);
    return is_rule;
}

/**
 * constructor
 */
SpaceMetaDataRule::SpaceMetaDataRule() :
    m_errorRegex("this-([\\w]+)-returns-([\\u_]+)"),
    m_spaceRegex("rp=([\\u_]*):al=([\\u_]*):o=([\\w_]*):ts=([-\\d]*):gs=([-\\d]*):us=([-\\d]*):la=([-\\d]*):ll=([-\\d]*)"){

    m_retentionPolicyMap["REPLICA"] = srm2__TRetentionPolicy__REPLICA;
    m_retentionPolicyMap["OUTPUT"] = srm2__TRetentionPolicy__OUTPUT;
    m_retentionPolicyMap["CUSTODIAL"] = srm2__TRetentionPolicy__CUSTODIAL;

    m_accessLatencyMap["ONLINE"] = srm2__TAccessLatency__ONLINE;
    m_accessLatencyMap["NEARLINE"] = srm2__TAccessLatency__NEARLINE;
}

bool SpaceMetaDataRule::is_error(
    const std::string& token,
    const std::string& type,
    srm2__TStatusCode& code){

    boost::match_results<std::string::const_iterator> what;
    boost::match_flag_type flags = boost::match_default;
    bool is_error = boost::regex_search(token.begin(),token.end(),what,m_errorRegex,flags);
    if (is_error){
        const std::string& error_type = what[1];
        const std::string& srm_result = what[2];
        if (error_type != type){
            is_error = false;
        } else {
            is_error = get_result(srm_result,code);
        }

    }
    TRACE_LOG("Checking error for token '" << token << "' type = '" << type << "' : is_error = " << is_error);
    return is_error;
}

bool SpaceMetaDataRule::translate_retention_policy(const std::string& in,srm2__TRetentionPolicy& out){
    return translate_string(in,m_retentionPolicyMap,out);
}
bool SpaceMetaDataRule::translate_access_latency(const std::string& in,srm2__TAccessLatency& out){
    return translate_string(in,m_accessLatencyMap,out);
}

/**
 * destructor
 */
SpaceMetaDataRule::~SpaceMetaDataRule(){
}


