/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004.
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  GLite Data SRM Util CPP: Srm 2.2 Server pattern based rules
 *
 *  Authors: Paolo Tedesco <paolo.tedesco@cern.ch>
 *  Version info: $Id: rules.h,v 1.4 2009/06/21 17:09:31 molnarzs Exp $
 *  Release: $Name:  $
 *
 */

#ifndef GLITE_DATA_SRM_UTIL_CPP_TEST_2_2_STUB_RULES_H_
#define GLITE_DATA_SRM_UTIL_CPP_TEST_2_2_STUB_RULES_H_

#include <glite/data/srm2/cpp/envH.h>
#include <glite/data/srm2/cpp/srm2H.h>

#include <map>
#include <string>
#include <boost/regex.hpp>
#include <boost/lexical_cast.hpp>

namespace srm2{

/**
 * Checksum rule
 */
class ChecksumRule {
public:
    struct ChecksumInfo{
        // the checksum type
        std::string type;
        // the checksum value
        std::string value;
    };
    /**
     * singleton instance
     */
    static ChecksumRule& instance();
    /**
     * get checksum information from surl
     */
    ChecksumInfo get(const std::string& surl);
private:
    /**
     * constructor
     */
    ChecksumRule();
    // regular expression
    boost::regex m_regex;
};

/**
 * FileLocality rule
 */
class FileLocalityRule{
public:
    /**
     * singleton instance
     */
    static FileLocalityRule& instance();
    /**
     * get file locality from surl
     */
    srm2::srm2__TFileLocality get(const std::string& surl);
private:
    /**
     * constructor
     */
    FileLocalityRule();
    // regular expression
    boost::regex m_regex;
    // map string -> srm2__TFileLocality
    typedef std::map<std::string,srm2__TFileLocality> StringToLocalityMap;
    StringToLocalityMap m_map;
};

/**
 * Base class for status rules
 */
class StatusRule{
public:
    /**
     * translate a string to an srm status code
     * @param srm_result [IN] the string to translate
     * @param code [OUT] the srm status code
     * @return true if found, false otherwise
     */
    bool get_result(const std::string& srm_result,srm2__TStatusCode& code);

protected:

    /**
     * constructor
     */
    StatusRule();

    /**
     * destructor
     */
    virtual ~StatusRule();

    /**
     * translate a string to an output type looking into a map
     */
    template<typename OutType>
    bool translate_string(const std::string& in,const std::map<std::string,OutType>& valuesmap,OutType& out){
        typename std::map<std::string,OutType>::const_iterator it = valuesmap.find(in);
        if (valuesmap.end() == it){
            return false; // no translation found
        } else {
            out = it->second;
            return true;
        }
    }

    // map string -> srm2__TStatusCode
    typedef std::map<std::string,srm2__TStatusCode> StringToStatusMap;
    StringToStatusMap m_map;
};

/**
 * Generic rule to return file status based on name
 */
class FileStatusRule : public StatusRule{

public:

    /**
     * singleton instance
     */
    static FileStatusRule& instance();

    /**
     * is_match
     * return true if file name matches the regular expression pattern and
     * (optionally) the operation name, e.g.
     * "this-one-returns-SRM_NO_FREE_SPACE-on-get", "get" ->
     *      true, code = srm2__TStatusCode__SRM_USCORENO_USCOREFREE_USCORESPACE
     * @param file_name [IN] the file name to match
     * @param operation [IN] optional operation name that acts as a filter, use
     *      empty string to disable checking
     * @param code [OUT] the srm2__TStatusCode found in the file name
     * @return true if the file name matches the pattern
     *      this-one-returns-(code)[-on-(operation)], false otherwise
     */
    bool is_match(
        const std::string& file_name,
        const std::string& operation,
        srm2__TStatusCode& code);

protected:

    /**
     * constructor
     */
    FileStatusRule();

    /**
     * destructor
     */
    virtual ~FileStatusRule();

    // regular expression
    boost::regex m_regex;
};

/**
 * SpaceMetaDataRule
 * applies patterns for GetSpaceMetaData operation
 * - request level failure: one of the tokens must be named "this-request-returns-(code)"
 * - token level failure: the token must be named "this-token-returns-(code)"
 * - token data: the token must be named
 * "rp=(rp):al=(al):o=(o):ts=(ts):gs=(gs):us=(us):la=(la):ll=(ll)", where:
 * rp = retention policy
 * al = access latency
 * o = owner
 * ts/gs/us = total/guaranteed/unused space
 * la/ll = lifetime assigned/left
 */
class SpaceMetaDataRule : public StatusRule{
public:

    /**
     * singleton instance
     */
    static SpaceMetaDataRule& instance();

    /**
     * is_token_error
     * @param token [IN] token
     * @param code [OUT] error code
     * @return true if token matches token error expression, false otherwise
     */
    bool is_token_error(
        const std::string& token,
        srm2__TStatusCode& code);

    /**
     * is_request_error
     * @param token [IN] token
     * @param code [OUT] error code
     * @return true if token matches request error expression, false otherwise
     */
    bool is_request_error(
        const std::string& token,
        srm2__TStatusCode& code);

    /**
     * is_token_rule
     * @param token [IN] token
     * @param code [OUT] error code
     * @return true if token matches expression, false otherwise
     */
    bool is_token_rule(
        const std::string&      token,
        struct soap*            soap,
        srm2__TMetaDataSpace&   space);

protected:

    template<typename OutType>
    void convert_optional_numeric_value (
        struct soap*        soap,
        const std::string&  in,
        OutType*&           out){
        if (!in.empty()){
            out = (OutType*)soap_malloc(soap,sizeof(OutType));
            try{
                *out = boost::lexical_cast<ULONG64>(in);
            } catch (const boost::bad_lexical_cast&){
                *out = 0;
            }
        }
    }

    /**
     * constructor
     */
    SpaceMetaDataRule();

    bool is_error(
        const std::string& token,
        const std::string& type,
        srm2__TStatusCode& code);

    bool translate_retention_policy(const std::string& in,srm2__TRetentionPolicy& out);

    bool translate_access_latency(const std::string& in,srm2__TAccessLatency& out);

    /**
     * destructor
     */
    virtual ~SpaceMetaDataRule();

    // regular expressions
    boost::regex m_errorRegex;
    boost::regex m_spaceRegex;

    typedef std::map<std::string,srm2__TRetentionPolicy> RetentionPolicyMap;
    RetentionPolicyMap m_retentionPolicyMap;

    typedef std::map<std::string,srm2__TAccessLatency> AccessLatencyMap;
    AccessLatencyMap m_accessLatencyMap;
};

}

#endif // GLITE_DATA_SRM_UTIL_CPP_TEST_2_2_STUB_RULES_H_


