/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004.
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  GLite Data SRM Util CPP: Srm 2.2 Server
 *
 *  Authors: Paolo Badino <paolo.badino@cern.ch>, Paolo Tedesco <paolo.tedesco@cern.ch>
 *  Version info: $Id: server.cpp,v 1.22 2009/07/30 17:40:34 molnarzs Exp $
 *  Release: $Name:  $
 *
 */

#ifndef SOAP_FMAC3S
#define SOAP_FMAC3S
#endif
#ifndef SOAP_FMAC4S
#define SOAP_FMAC4S
#endif

#include <glite/data/srm2/cpp/envH.h>
#include <glite/data/srm2/cpp/srm2H.h>
#include <glite/data/agents/url-utility.h>
#include <glite/data/agents/AgentExceptions.h>

#include <vector>
#include <map>
#include <sstream>
#include <boost/shared_ptr.hpp>
#include <boost/lexical_cast.hpp>
#include <libgen.h>
#include <string.h>
#include <dirent.h>
#include <pthread.h>
#include <boost/regex.hpp>

#include "rules.h"
#include "log.h"
#include "SrmBusyTestHandler.h"

extern int errno;


namespace {
const char * const PROTOCOL_FILE  = "file";
const char * const SFN            = "SFN";

}

using namespace glite::data::agents;
extern std::string g_pool_path;

// the lock
extern pthread_mutex_t  g_lock;

namespace srm2{


// -----------------------------------------------------------------------------
// Stub implementation: Request Handling
// -----------------------------------------------------------------------------

/**
 * File Request Object
 */
struct FileRequest{
public:
    typedef boost::shared_ptr<FileRequest> Ptr;

    srm2__TStatusCode status;
    std::string       explanation;
    long long         size;
    int               permissions;

    FileRequest():status(srm2__TStatusCode__SRM_USCORECUSTOM_USCORESTATUS),
        size(0),permissions(0){}
};

/**
 * Put File Request Object
 */
struct PutFileRequest : public FileRequest{
public:
    typedef boost::shared_ptr<PutFileRequest> Ptr;

    std::string surl;
    std::string turl;

    PutFileRequest():FileRequest(){}
};

/**
 * Get File Request Object
 */
struct GetFileRequest : public FileRequest{
public:
    typedef boost::shared_ptr<GetFileRequest> Ptr;

    std::string surl;
    std::string turl;

    GetFileRequest():FileRequest(){}
};

/**
 * BringOnline File Request Object
 */
struct BringOnlineFileRequest : public FileRequest{
public:
    typedef boost::shared_ptr<BringOnlineFileRequest> Ptr;

    std::string surl;

    BringOnlineFileRequest():FileRequest(){}
};

/**
 * Copy File Request Object
 */
struct CopyFileRequest : public FileRequest{
public:
    typedef boost::shared_ptr<CopyFileRequest> Ptr;

    std::string sourceSurl;
    std::string targetSurl;

    CopyFileRequest():FileRequest(){}
};

/**
 * Request Object
 */
struct Request{
public:
    typedef boost::shared_ptr<Request> Ptr;

    typedef enum{
        T_NONE = -1,
        T_PUT,
        T_GET,
        T_COPY,
        T_BRING_ONLINE
    } Type;

    Type              type;
    std::string       id;
    srm2__TStatusCode status;
    std::string       explanation;
    std::vector<FileRequest::Ptr> fileRequests;
    time_t            start_time;
    time_t            finish_time;

    Request():type(T_NONE),status(srm2__TStatusCode__SRM_USCORECUSTOM_USCORESTATUS),
        start_time(-1),finish_time(-1){}
};

/**
 * PutRequest Object
 */
struct PutRequest : public Request{
public:
    typedef boost::shared_ptr<PutRequest> Ptr;

    std::string protocol;
    bool overwrite;
    PutRequest():Request(),overwrite(false){
        type = Request::T_PUT;
    }
};

/**
 * GetRequest Object
 */
struct GetRequest : public Request{
public:
    typedef boost::shared_ptr<GetRequest> Ptr;

    std::string protocol;
    GetRequest():Request(){
        type = Request::T_GET;
    }
};

/**
 * BringOnlineRequest Object
 */
struct BringOnlineRequest : public Request{
public:
    typedef boost::shared_ptr<BringOnlineRequest> Ptr;

    std::string protocol;
    BringOnlineRequest():Request(){
        type = Request::T_BRING_ONLINE;
    }
};

/**
 * CopyRequest Object
 */
struct CopyRequest : public Request{
public:
    typedef boost::shared_ptr<CopyRequest> Ptr;

    bool overwrite;
    CopyRequest():Request(),overwrite(false){
        type = Request::T_COPY;
    }
};

/**
 * Request Map
 */
typedef std::map<std::string,Request::Ptr> RequestMap;
RequestMap g_RequestMap;

/**
 * Store a Request
 */
int store_request(Request::Ptr request){
    static int s_id = 0;
    int stored = -1;
    // lock
    //pthread_mutex_lock(&g_lock);
    // Assign the id
    request->id = boost::lexical_cast<std::string>(++s_id);
    // Store the Request
    std::pair<RequestMap::iterator, bool> result =
        g_RequestMap.insert(RequestMap::value_type(request->id,request));
    if(!result.second){
        stored = -1;
    } else {
        // Set start time
        request->start_time = time(0);
    stored = 0;
    }
    // unlock
    //pthread_mutex_unlock(&g_lock);
    return stored;
}

/**
 * Return a Request
 */
Request::Ptr get_request(std::string id){
    Request::Ptr ptr;
    // lock
    //pthread_mutex_lock(&g_lock);
    RequestMap::const_iterator it = g_RequestMap.find(id);
    if(it != g_RequestMap.end()){
        ptr = (*it).second;
    }
    // unlock
    //pthread_mutex_unlock(&g_lock);
    return ptr;
}

/**
 * Return a Request
 */
void purge_requests(){
    // lock
    //pthread_mutex_lock(&g_lock);

    time_t current;
    time(&current);

    std::vector<std::string> ids;
    Request::Ptr ptr;
    RequestMap::iterator it;
    for(it = g_RequestMap.begin(); it != g_RequestMap.end(); ++it){
        // Check the requests that has been completed more than 30 minutes ago
        ptr = (*it).second;
        if((ptr->finish_time != (time_t)-1) &&
           (ptr->finish_time > (current + 1800))){
            // This request should be removed. Cannot use erase since it
            // looks like it return void instead of iterator (?)
            ids.push_back(ptr->id);
        }
    }
    std::vector<std::string>::iterator iit;
    for(iit = ids.begin(); iit != ids.end(); ++iit){
        g_RequestMap.erase((*iit));
    }

    // unlock
    //pthread_mutex_unlock(&g_lock);
}

// -----------------------------------------------------------------------------
// Stub implementation: same file on the file system
// -----------------------------------------------------------------------------
/*
 * Extract the file path from a surl
 */
std::string path_from_surl(const std::string& surl){
    std::string file_path;
    try{
        std::string  hostname;
        unsigned int port = 0;
        std::string  protocol;
        std::string  path;
        std::string  query_str;

        // Parse surl
        parse_url(surl,hostname,port,protocol,path,query_str);

        // Check query string
        if(query_str.empty()){
            // SURL in compact form: take the path
            file_path = path;
        } else {
            // Parse Query string
            std::vector<std::pair<std::string,std::string> > values;
            parse_query_string(query_str,values);

            // Take the SFN
            std::vector<std::pair<std::string,std::string> >::iterator it;
            for(it = values.begin(); it != values.end(); ++it){
                if((*it).first == SFN){
                    file_path = (*it).second;
                    break;
                }
            }
        }
    }catch (const LogicError& exc){
        // Invalid SURL
        file_path = "";
    }

    // Append the file_path the the pool directory path
    if(false == file_path.empty()){
        if(*(g_pool_path.rbegin()) != '/'){
            file_path = g_pool_path + "/" + file_path;
        } else {
            file_path = g_pool_path + file_path;
        }
    }
    return file_path;
}

/*
 * Get the parent Directory name
 */
std::string get_dirname(const std::string& dir_path){
   char * parent_name_str = strdup(dir_path.c_str());
    dirname(parent_name_str);
    std::string parent_name = parent_name_str;
    free(parent_name_str);
    return parent_name;
}

/*
 * Create a Directory
 */
int create_dir_path(const std::string& dir_name, bool recursive){
    int r = -1;
    if(false == recursive){
        // Ok, parent exists or has been created: Create teh directory
        r = mkdir(dir_name.c_str(),0700);
    } else {
        // Get the parent for that dir
        std::string parent_name = get_dirname(dir_name);
        // Check if parent_name_str exist
        r = access(parent_name.c_str(), R_OK);
        if(0 == r){
            // Ok, parent exists: Create the directory
            r = mkdir(dir_name.c_str(),0700);
        } else {
            // Directory doesn't exist: try to create the parent dir
            r = create_dir_path(parent_name,true);
            if(0 == r){
                // Ok, parent exists or has been created: Create the directory
                r= mkdir(dir_name.c_str(),0700);
            }
        }
    }
    if(r != 0){
        return errno;
    }
    return 0;
}

/*
 * Remove a Directory
 */
int rm_dir_path(const std::string& dir_name, bool recursive){
    int r = -1;
    if(false == recursive){
        // Remove the directory
        r = rmdir(dir_name.c_str());
    } else {
        // Open the directory
        DIR * dir = opendir(dir_name.c_str());
        if(0 == dir){
            r = errno;
        } else {
            // For each entry
            struct dirent * dir_ent = 0;
            do {
                // Read dir
                dir_ent = readdir(dir);
                if(0 != dir_ent){
                    std::string child_name = dir_ent->d_name;
                    if((child_name == ".") ||
                       (child_name == "..")){
                        continue;
                    }
                    if(S_ISDIR(dir_ent->d_type)){
                        // directory: call rm_dir_path recursively
                        r = rm_dir_path(child_name,true);
                        if(0 != r){
                            break;
                        }
                    } else {
                        // file or symlink: unlink it
                        unlink(child_name.c_str());
                        if(0 != r){
                            break;
                        }
                    }
                }
            } while(dir_ent != 0);
            // close dir
            closedir(dir);

            // Remove the directory (now it should be empty)
            r = rmdir(dir_name.c_str());
        }
    }
    if(r != 0){
        return errno;
    }
    return 0;
}

/*
 * get_file_stat
 */
srm2__TMetaDataPathDetail * get_file_stat(struct soap * soap,const std::string& surl){

    srm2__TMetaDataPathDetail * detail = soap_new_srm2__TMetaDataPathDetail(soap,-1);
    detail->path                 = surl;
    detail->status               = soap_new_srm2__TReturnStatus(soap,-1);
    detail->status->statusCode   = srm2__TStatusCode__SRM_USCOREFAILURE;
    detail->status->explanation  = 0;
    detail->size                 = 0;
    detail->createdAtTime        = 0;
    detail->lastModificationTime = 0;
    detail->fileStorageType      = 0;
    detail->retentionPolicyInfo  = 0;
    detail->fileLocality         = 0;
    detail->arrayOfSpaceTokens   = 0;
    detail->type                 = 0;
    detail->lifetimeAssigned     = 0;
    detail->lifetimeLeft         = 0;
    detail->ownerPermission      = 0;
    detail->groupPermission      = 0;
    detail->otherPermission      = 0;
    detail->checkSumType         = 0;
    detail->checkSumValue        = 0;
    detail->arrayOfSubPaths      = 0;

    // Get File Path
    std::string error;
    std::string file_path = path_from_surl(surl);
    if(file_path.empty()){
        error = "Invalid SURL passed";
    } else {
        // Get file stat
        struct stat64 stat_buf;
        int r = stat64(file_path.c_str(),&stat_buf);
        if(0 == r){
            TRACE_LOG("Stat For file " << file_path);
            detail->status->statusCode = srm2__TStatusCode__SRM_USCORESUCCESS;
            // Set File properties
            detail->size                 = (ULONG64*)soap_malloc(soap,sizeof(ULONG64));
            *(detail->size)              = stat_buf.st_size;
            detail->createdAtTime        = (time_t*)soap_malloc(soap,sizeof(time_t));
            *(detail->createdAtTime)     = stat_buf.st_ctime;
            detail->lastModificationTime = (time_t*)soap_malloc(soap,sizeof(time_t));
            *(detail->lastModificationTime) = stat_buf.st_mtime;
            detail->fileStorageType      = (srm2__TFileStorageType*)soap_malloc(soap,sizeof(srm2__TFileStorageType));
            *(detail->fileStorageType)   = srm2__TFileStorageType__VOLATILE;
            detail->retentionPolicyInfo  = soap_new_srm2__TRetentionPolicyInfo(soap,-1);
            detail->retentionPolicyInfo->retentionPolicy = srm2__TRetentionPolicy__REPLICA;
            detail->retentionPolicyInfo->accessLatency = (srm2__TAccessLatency*)soap_malloc(soap,sizeof(srm2__TAccessLatency));
            *(detail->retentionPolicyInfo->accessLatency) = srm2__TAccessLatency__ONLINE;
            detail->fileLocality         = (srm2__TFileLocality*)soap_malloc(soap,sizeof(srm2__TFileLocality));
            *(detail->fileLocality)      = FileLocalityRule::instance().get(surl);
            detail->type                 = (srm2__TFileType*)soap_malloc(soap,sizeof(srm2__TFileType));

            if(S_ISDIR(stat_buf.st_mode)){
                *(detail->type) = srm2__TFileType__DIRECTORY;
            } else if(S_ISLNK(stat_buf.st_mode)){
                *(detail->type) = srm2__TFileType__LINK;
            } else {
                *(detail->type) = srm2__TFileType__FILE_;

                ChecksumRule::ChecksumInfo info = ChecksumRule::instance().get(surl);
                if (info.type.length()){
                    detail->checkSumType = soap_new_std__string(soap,-1);
                    *(detail->checkSumType) = info.type;
                }
                if (info.value.length()){
                    detail->checkSumValue = soap_new_std__string(soap,-1);
                    *(detail->checkSumValue) = info.value;
                }

                TRACE_LOG("Checksum: returned checksum for " << surl << ": "
                	<< (detail->checkSumType ? *(detail->checkSumType) : "(NULL)")
                	<< ":" << (detail->checkSumValue ? *(detail->checkSumValue) : "(NULL)"));
            }
            detail->ownerPermission          = soap_new_srm2__TUserPermission(soap,-1);
            detail->ownerPermission->userID  = boost::lexical_cast<std::string>(stat_buf.st_uid);
            detail->ownerPermission->mode    = (srm2__TPermissionMode)((stat_buf.st_mode & 0700) >> 6);
            detail->groupPermission          = soap_new_srm2__TGroupPermission(soap,-1);
            detail->groupPermission->groupID = boost::lexical_cast<std::string>(stat_buf.st_gid);
            detail->groupPermission->mode    = (srm2__TPermissionMode)((stat_buf.st_mode & 0070) >> 3);
            detail->otherPermission          = (srm2__TPermissionMode*)soap_malloc(soap,sizeof(srm2__TPermissionMode));
            *(detail->otherPermission)       = (srm2__TPermissionMode)(stat_buf.st_mode & 0007);
        } else {
            TRACE_LOG("File " << file_path << " doesn't exist");
            detail->status->statusCode = srm2__TStatusCode__SRM_USCOREINVALID_USCOREPATH;
            error = strerror(errno);
        }
    }
    if(false == error.empty()){
        detail->status->explanation = soap_new_std__string(soap,-1);
        *(detail->status->explanation) = error;
    }
    return  detail;
}

/*
 * update_request_status
 */
void update_request_status(PutRequest& req){

    if((req.status == srm2__TStatusCode__SRM_USCOREREQUEST_USCOREQUEUED) ||
       (req.status == srm2__TStatusCode__SRM_USCOREREQUEST_USCOREINPROGRESS) ||
       (req.status == srm2__TStatusCode__SRM_USCORESUCCESS) ||
       (req.status == srm2__TStatusCode__SRM_USCOREPARTIAL_USCORESUCCESS)){
        // Evaluate the status of the files
        bool has_queued     = false;
        bool has_inprogress = false;
        bool has_available  = false;
        bool has_success    = false;
        bool has_aborted    = false;
        bool has_failed     = false;
        // For each file
        std::vector<FileRequest::Ptr>::iterator it;
        for(it = req.fileRequests.begin();it != req.fileRequests.end(); ++it){
            if((*it)->status == srm2__TStatusCode__SRM_USCOREREQUEST_USCOREQUEUED){
                has_queued = true;
            } else if((*it)->status == srm2__TStatusCode__SRM_USCOREREQUEST_USCOREINPROGRESS){
                has_inprogress = true;
            } else if((*it)->status == srm2__TStatusCode__SRM_USCORESPACE_USCOREAVAILABLE){
                has_available = true;
            } else if((*it)->status == srm2__TStatusCode__SRM_USCORESUCCESS){
                has_success = true;
            } else if((*it)->status == srm2__TStatusCode__SRM_USCOREABORTED){
                has_aborted = true;
            } else {
                has_failed = true;
            }
        }
        // Set request status accordingly
        if(true == has_queued){
            req.status = srm2__TStatusCode__SRM_USCOREREQUEST_USCOREQUEUED;
        } else if(true == has_inprogress){
            req.status = srm2__TStatusCode__SRM_USCOREREQUEST_USCOREINPROGRESS;
        } else if(true == has_available){
            if(true == has_failed){
                req.status = srm2__TStatusCode__SRM_USCOREPARTIAL_USCORESUCCESS;
            } else {
                req.status = srm2__TStatusCode__SRM_USCORESUCCESS;
            }
        } else if(true == has_success){
            req.status = srm2__TStatusCode__SRM_USCORESUCCESS;
        } else if(true == has_aborted){
            req.status = srm2__TStatusCode__SRM_USCOREABORTED;
        } else {
            req.status = srm2__TStatusCode__SRM_USCOREFAILURE;
        }
    } else {
        // Nothing To Do
    }
}

/*
 * update_request_status
 */
void update_request_status(GetRequest& req){

    if((req.status == srm2__TStatusCode__SRM_USCOREREQUEST_USCOREQUEUED) ||
       (req.status == srm2__TStatusCode__SRM_USCOREREQUEST_USCOREINPROGRESS) ||
       (req.status == srm2__TStatusCode__SRM_USCORESUCCESS) ||
       (req.status == srm2__TStatusCode__SRM_USCOREPARTIAL_USCORESUCCESS)){
        // Evaluate the status of the files
        bool has_queued     = false;
        bool has_inprogress = false;
        bool has_pinned     = false;
        bool has_released   = false;
        bool has_aborted    = false;
        bool has_failed     = false;
        // For each file
        std::vector<FileRequest::Ptr>::iterator it;
        for(it = req.fileRequests.begin();it != req.fileRequests.end(); ++it){
            if((*it)->status == srm2__TStatusCode__SRM_USCOREREQUEST_USCOREQUEUED){
                has_queued = true;
            } else if((*it)->status == srm2__TStatusCode__SRM_USCOREREQUEST_USCOREINPROGRESS){
                has_inprogress = true;
            } else if((*it)->status == srm2__TStatusCode__SRM_USCOREFILE_USCOREPINNED){
                has_pinned = true;
            } else if((*it)->status == srm2__TStatusCode__SRM_USCORERELEASED){
                has_released = true;
            } else if((*it)->status == srm2__TStatusCode__SRM_USCOREABORTED){
                has_aborted = true;
            } else {
                has_failed = true;
            }
        }
        // Set request status accordingly
        if(true == has_queued){
            req.status = srm2__TStatusCode__SRM_USCOREREQUEST_USCOREQUEUED;
        } else if(true == has_inprogress){
            req.status = srm2__TStatusCode__SRM_USCOREREQUEST_USCOREINPROGRESS;
        } else if(true == has_pinned){
            if(true == has_failed){
                req.status = srm2__TStatusCode__SRM_USCOREPARTIAL_USCORESUCCESS;
            } else {
                req.status = srm2__TStatusCode__SRM_USCORESUCCESS;
            }
        } else if(true == has_released){
            req.status = srm2__TStatusCode__SRM_USCORESUCCESS;
        } else if(true == has_aborted){
            req.status = srm2__TStatusCode__SRM_USCOREABORTED;
        } else {
            req.status = srm2__TStatusCode__SRM_USCOREFAILURE;
        }
    } else {
        // Nothing To Do
    }
}

/*
 * update_request_status
 */
void update_request_status(BringOnlineRequest& req){

    if((req.status == srm2__TStatusCode__SRM_USCOREREQUEST_USCOREQUEUED) ||
       (req.status == srm2__TStatusCode__SRM_USCOREREQUEST_USCOREINPROGRESS) ||
       (req.status == srm2__TStatusCode__SRM_USCORESUCCESS) ||
       (req.status == srm2__TStatusCode__SRM_USCOREPARTIAL_USCORESUCCESS)){
        // Evaluate the status of the files
        bool has_queued     = false;
        bool has_inprogress = false;
        bool has_success    = false;
        bool has_released   = false;
        bool has_aborted    = false;
        bool has_failed     = false;
        // For each file
        std::vector<FileRequest::Ptr>::iterator it;
        for(it = req.fileRequests.begin();it != req.fileRequests.end(); ++it){
            if((*it)->status == srm2__TStatusCode__SRM_USCOREREQUEST_USCOREQUEUED){
                has_queued = true;
            } else if((*it)->status == srm2__TStatusCode__SRM_USCOREREQUEST_USCOREINPROGRESS){
                has_inprogress = true;
            } else if((*it)->status == srm2__TStatusCode__SRM_USCORESUCCESS){
                has_success = true;
            } else if((*it)->status == srm2__TStatusCode__SRM_USCORERELEASED){
                has_released = true;
            } else if((*it)->status == srm2__TStatusCode__SRM_USCOREABORTED){
                has_aborted = true;
            } else {
                has_failed = true;
            }
        }
        // Set request status accordingly
        if(true == has_queued){
            req.status = srm2__TStatusCode__SRM_USCOREREQUEST_USCOREQUEUED;
        } else if(true == has_inprogress){
            req.status = srm2__TStatusCode__SRM_USCOREREQUEST_USCOREINPROGRESS;
        } else if(true == has_success){
            if(true == has_failed){
                req.status = srm2__TStatusCode__SRM_USCOREPARTIAL_USCORESUCCESS;
            } else {
                req.status = srm2__TStatusCode__SRM_USCORESUCCESS;
            }
        } else if(true == has_released){
            req.status = srm2__TStatusCode__SRM_USCORESUCCESS;
        } else if(true == has_aborted){
            req.status = srm2__TStatusCode__SRM_USCOREABORTED;
        } else {
            req.status = srm2__TStatusCode__SRM_USCOREFAILURE;
        }
    } else {
        // Nothing To Do
    }
}

/*
 * update_request_status
 */
void update_request_status(CopyRequest& req){

    if((req.status == srm2__TStatusCode__SRM_USCOREREQUEST_USCOREQUEUED) ||
       (req.status == srm2__TStatusCode__SRM_USCOREREQUEST_USCOREINPROGRESS) ||
       (req.status == srm2__TStatusCode__SRM_USCORESUCCESS) ||
       (req.status == srm2__TStatusCode__SRM_USCOREPARTIAL_USCORESUCCESS)){
        // Evaluate the status of the files
        bool has_queued     = false;
        bool has_inprogress = false;
        bool has_success    = false;
        bool has_aborted    = false;
        bool has_failed     = false;
        // For each file
        std::vector<FileRequest::Ptr>::iterator it;
        for(it = req.fileRequests.begin();it != req.fileRequests.end(); ++it){
            if((*it)->status == srm2__TStatusCode__SRM_USCOREREQUEST_USCOREQUEUED){
                has_queued = true;
            } else if((*it)->status == srm2__TStatusCode__SRM_USCOREREQUEST_USCOREINPROGRESS){
                has_inprogress = true;
            } else if((*it)->status == srm2__TStatusCode__SRM_USCORESUCCESS){
                has_success = true;
            } else if((*it)->status == srm2__TStatusCode__SRM_USCOREABORTED){
                has_aborted = true;
            } else {
                has_failed = true;
            }
        }
        // Set request status accordingly
        if(true == has_queued){
            req.status = srm2__TStatusCode__SRM_USCOREREQUEST_USCOREQUEUED;
        } else if(true == has_inprogress){
            req.status = srm2__TStatusCode__SRM_USCOREREQUEST_USCOREINPROGRESS;
        } else if(true == has_success){
            if(true == has_failed){
                req.status = srm2__TStatusCode__SRM_USCOREPARTIAL_USCORESUCCESS;
            } else {
                req.status = srm2__TStatusCode__SRM_USCORESUCCESS;
            }
        } else if(true == has_aborted){
            req.status = srm2__TStatusCode__SRM_USCOREABORTED;
        } else {
            req.status = srm2__TStatusCode__SRM_USCOREFAILURE;
        }
    } else {
        // Nothing To Do
    }
}

/*
 * process_request
 */
void process_request(PutRequest& req){
    // Check Request Status
    if(req.status == srm2__TStatusCode__SRM_USCOREREQUEST_USCOREQUEUED){
        // Request is Queued

        // For each file
        std::vector<FileRequest::Ptr>::iterator it;
        for(it = req.fileRequests.begin();it != req.fileRequests.end(); ++it){
            if((*it)->status == srm2__TStatusCode__SRM_USCOREREQUEST_USCOREQUEUED){
                // set file status to in progress
                (*it)->status = srm2__TStatusCode__SRM_USCOREREQUEST_USCOREINPROGRESS;
            }
        }
    } else if (req.status == srm2__TStatusCode__SRM_USCOREREQUEST_USCOREINPROGRESS){
        // Request is in progress

        // For Each File
        std::vector<FileRequest::Ptr>::iterator it;
        for(it = req.fileRequests.begin();it != req.fileRequests.end(); ++it){
            PutFileRequest& file = static_cast<PutFileRequest&>(*((*it).get()));
            if(file.status == srm2__TStatusCode__SRM_USCOREREQUEST_USCOREINPROGRESS){
                // Reserve space and set file to SpaceAvailable

                // Get File Path
                std::string file_path = path_from_surl(file.surl);
                if(file_path.empty()){
                    file.status  = srm2__TStatusCode__SRM_USCOREFAILURE;
                    file.explanation = "Invalid SURL passed";
                    continue;
                }
                if(std::string::npos != file_path.find("this-one-should-stay-pending")){
                    // File should remain pending
                    continue;
                } else if(std::string::npos != file_path.find("this-one-is-busy")){
                    // File is busy
                    file.status  = srm2__TStatusCode__SRM_USCOREFILE_USCOREBUSY;
                    file.explanation = "file or resource busy";
                    continue;
                } else if (std::string::npos != file_path.find("this-one-returns-no-free-space-on-put")) {
                    file.status = srm2__TStatusCode__SRM_USCORENO_USCOREFREE_USCORESPACE;
                    file.explanation = "no free space";
                    continue;
                }
                srm2__TStatusCode code = srm2__TStatusCode__SRM_USCOREFAILURE;
                if ( FileStatusRule::instance().is_match(file_path,"put",code) ){
                    file.status = code;
                    file.explanation = "code extracted from file path";
                    continue;
                }

                /* TBA : no user space */
                // Check if the path exists
                std::string dir_name = get_dirname(file_path);
                if(0 != access(dir_name.c_str(), W_OK)){
                    file.status = srm2__TStatusCode__SRM_USCOREINVALID_USCOREPATH;
                    file.explanation = strerror(errno);
                    continue;
                }
                // Check if file already exists
                struct stat64 stat_buf;
                int r = stat64(file_path.c_str(),&stat_buf);
                if(0 != r){
                    // Touch the file
                    int fd = creat(file_path.c_str(), file.permissions);
                    if(-1 == fd){
                        file.status  = srm2__TStatusCode__SRM_USCOREFAILURE;
                        file.explanation = strerror(errno);
                        continue;
                    }
                    // Set the file properties
                    file.turl = (std::string)PROTOCOL_FILE + "://" + file_path;
                    file.permissions = stat_buf.st_mode;
                    file.status  = srm2__TStatusCode__SRM_USCORESPACE_USCOREAVAILABLE;
                    close(fd);
                    TRACE_LOG("File " << file_path << " is now available via TURL " << file.turl);
                } else {
                    // File already exist: Check owervite flag on the request
                    if(false == req.overwrite){
                        file.status  = srm2__TStatusCode__SRM_USCOREDUPLICATION_USCOREERROR;
                        file.explanation = "File already exists";
                        continue;
                    }
                    int fd = open(file_path.c_str(), file.permissions, O_WRONLY | O_CREAT | O_TRUNC);
                    if(-1 == fd){
                        file.status  = srm2__TStatusCode__SRM_USCOREFAILURE;
                        file.explanation = strerror(errno);
                        continue;
                    }
                    // Set the file properties
                    file.turl = (std::string)PROTOCOL_FILE + "://" + file_path;
                    file.permissions = stat_buf.st_mode;
                    file.status  = srm2__TStatusCode__SRM_USCORESPACE_USCOREAVAILABLE;
                    close(fd);
                }
            }
        }
    } else {
        // Nothing To do
    }
    // Update request status
    update_request_status(req);
}

/*
 * process Get request
 */
void process_request(GetRequest& req){
    // Check Request Status
    if(req.status == srm2__TStatusCode__SRM_USCOREREQUEST_USCOREQUEUED){
        // Request is Queued

        // For each file
        std::vector<FileRequest::Ptr>::iterator it;
        for(it = req.fileRequests.begin();it != req.fileRequests.end(); ++it){
            if((*it)->status == srm2__TStatusCode__SRM_USCOREREQUEST_USCOREQUEUED){
                // set file status to in progress
                (*it)->status = srm2__TStatusCode__SRM_USCOREREQUEST_USCOREINPROGRESS;
            }
        }
    } else if (req.status == srm2__TStatusCode__SRM_USCOREREQUEST_USCOREINPROGRESS){
        // Request is in progress

        // For Each File
        std::vector<FileRequest::Ptr>::iterator it;
        for(it = req.fileRequests.begin();it != req.fileRequests.end(); ++it){
            GetFileRequest& file = static_cast<GetFileRequest&>(*((*it).get()));
            if(file.status == srm2__TStatusCode__SRM_USCOREREQUEST_USCOREINPROGRESS){
                // Reserve space and set file to SpaceAvailable

                // Get File Path
                std::string file_path = path_from_surl(file.surl);
                if(file_path.empty()){
                    file.status  = srm2__TStatusCode__SRM_USCOREFAILURE;
                    file.explanation = "Invalid SURL passed";
                    continue;
                }
                if(std::string::npos != file_path.find("this-one-should-stay-pending")){
                    // File should remain pending
                    continue;
                }

                srm2__TStatusCode code = srm2__TStatusCode__SRM_USCOREFAILURE;
                if ( FileStatusRule::instance().is_match(file_path,"get",code) ){
                    file.status = code;
                    file.explanation = "code extracted from file path";
                    continue;
                }

                // Check if file exists
                if(0 != access(file_path.c_str(),R_OK)){
                    // File does not exist or not readable
                    file.status  = srm2__TStatusCode__SRM_USCOREINVALID_USCOREPATH;
                    file.explanation = strerror(errno);
                    continue;
                }
                struct stat64 stat_buf;
                int r = stat64(file_path.c_str(),&stat_buf);
                if(0 != r){
                    file.status  = srm2__TStatusCode__SRM_USCOREINVALID_USCOREPATH;
                    file.explanation = strerror(errno);
                    continue;
                }
                // Set the file properties
                file.turl = (std::string)PROTOCOL_FILE + "://" + file_path;
                file.permissions = stat_buf.st_mode;

                if (std::string::npos != file_path.find("this-one-returns-size-zero-on-get")){
                    file.size    = 0;
                } else if (std::string::npos != file_path.find("this-one-returns-wrong-size-on-get")){
                    file.size   = (stat_buf.st_size >> 1);
                } else {
                    file.size    = stat_buf.st_size;
                }

                file.status  = srm2__TStatusCode__SRM_USCOREFILE_USCOREPINNED;
                TRACE_LOG("File " << file_path << " is now available via TURL " << file.turl);
            }
        }
    } else {
        // Nothing To do
    }
    // Update request status
    update_request_status(req);
}

/*
 * process BringOnline request
 */
void process_request(BringOnlineRequest& req){
    // Check Request Status
    if(req.status == srm2__TStatusCode__SRM_USCOREREQUEST_USCOREQUEUED){
        // Request is Queued

        // For each file
        std::vector<FileRequest::Ptr>::iterator it;
        for(it = req.fileRequests.begin();it != req.fileRequests.end(); ++it){
            if((*it)->status == srm2__TStatusCode__SRM_USCOREREQUEST_USCOREQUEUED){
                // set file status to in progress
                (*it)->status = srm2__TStatusCode__SRM_USCOREREQUEST_USCOREINPROGRESS;
            }
        }
    } else if (req.status == srm2__TStatusCode__SRM_USCOREREQUEST_USCOREINPROGRESS){
        // Request is in progress

        // For Each File
        std::vector<FileRequest::Ptr>::iterator it;
        for(it = req.fileRequests.begin();it != req.fileRequests.end(); ++it){
            BringOnlineFileRequest& file = static_cast<BringOnlineFileRequest&>(*((*it).get()));
            if(file.status == srm2__TStatusCode__SRM_USCOREREQUEST_USCOREINPROGRESS){
                // Get File Path
                std::string file_path = path_from_surl(file.surl);
                if(file_path.empty()){
                    file.status  = srm2__TStatusCode__SRM_USCOREFAILURE;
                    file.explanation = "Invalid SURL passed";
                    continue;
                }
                if(std::string::npos != file_path.find("this-one-should-stay-pending")){
                    // File should remain pending
                    continue;
                }

                srm2__TStatusCode code = srm2__TStatusCode__SRM_USCOREFAILURE;
                if ( FileStatusRule::instance().is_match(file_path,"bringonline",code) ){
                    file.status = code;
                    file.explanation = "code extracted from file path";
                    continue;
                }

                // Check if file exists
                if(0 != access(file_path.c_str(),R_OK)){
                    // File does not exist or not readable
                    file.status  = srm2__TStatusCode__SRM_USCOREINVALID_USCOREPATH;
                    file.explanation = strerror(errno);
                    continue;
                }
                struct stat64 stat_buf;
                int r = stat64(file_path.c_str(),&stat_buf);
                if(0 != r){
                    file.status  = srm2__TStatusCode__SRM_USCOREINVALID_USCOREPATH;
                    file.explanation = strerror(errno);
                    continue;
                }
                // Set the file properties
                file.permissions = stat_buf.st_mode;
                file.size    = stat_buf.st_size;
                file.status  = srm2__TStatusCode__SRM_USCORESUCCESS;
                TRACE_LOG("File " << file_path << " is already online");
            }
        }
    } else {
        // Nothing To do
    }
    // Update request status
    update_request_status(req);
}

/*
 * process Copy request
 */
void process_request(CopyRequest& req){
    // Check Request Status
    if(req.status == srm2__TStatusCode__SRM_USCOREREQUEST_USCOREQUEUED){
        // Request is Queued

        // For each file
        std::vector<FileRequest::Ptr>::iterator it;
        for(it = req.fileRequests.begin();it != req.fileRequests.end(); ++it){
            if((*it)->status == srm2__TStatusCode__SRM_USCOREREQUEST_USCOREQUEUED){
                // set file status to in progress
                (*it)->status = srm2__TStatusCode__SRM_USCOREREQUEST_USCOREINPROGRESS;
            }
        }
    } else if (req.status == srm2__TStatusCode__SRM_USCOREREQUEST_USCOREINPROGRESS){
        // Request is in progress

        // For Each File
        std::vector<FileRequest::Ptr>::iterator it;
        for(it = req.fileRequests.begin();it != req.fileRequests.end(); ++it){
            CopyFileRequest& file = static_cast<CopyFileRequest&>(*((*it).get()));
            if(file.status == srm2__TStatusCode__SRM_USCOREREQUEST_USCOREINPROGRESS){
                // Get File Paths
                std::string src_file_path = path_from_surl(file.sourceSurl);
                if(src_file_path.empty()){
                    file.status  = srm2__TStatusCode__SRM_USCOREFAILURE;
                    file.explanation = "Invalid Source SURL passed";
                    continue;
                }
                if(std::string::npos != src_file_path.find("this-one-should-stay-pending")){
                    // File should remain pending
                    continue;
                }

                srm2__TStatusCode code = srm2__TStatusCode__SRM_USCOREFAILURE;
                if ( FileStatusRule::instance().is_match(src_file_path,"copy",code) ){
                    file.status = code;
                    file.explanation = "code extracted from source file path";
                    continue;
                }

                std::string dst_file_path = path_from_surl(file.targetSurl);
                if(dst_file_path.empty()){
                    file.status  = srm2__TStatusCode__SRM_USCOREFAILURE;
                    file.explanation = "Invalid Target SURL passed";
                    continue;
                }
                if(std::string::npos != dst_file_path.find("this-one-should-stay-pending")){
                    // File should remain pending
                    continue;
                }
                if ( FileStatusRule::instance().is_match(dst_file_path,"copy",code) ){
                    file.status = code;
                    file.explanation = "code extracted from destination file path";
                    continue;
                }

                TRACE_LOG("Copy file " << dst_file_path << " from " << src_file_path);
                // Check if Source exists
                struct stat64 src_stat_buf;
                int r = stat64(src_file_path.c_str(),&src_stat_buf);
                if(0 != r){
                    // Source does not exist
                    TRACE_LOG("source file " << src_file_path << " does not exist");
                    file.status = srm2__TStatusCode__SRM_USCOREINVALID_USCOREPATH;
                    file.explanation = strerror(errno);
                    continue;
                }
                // Check if dest dir exists
                std::string dst_dir_name = get_dirname(dst_file_path);
                if(0 != access(dst_dir_name.c_str(), W_OK)){
                    r = create_dir_path(dst_dir_name,true);
                    if(0 != r){
                        TRACE_LOG("Cannot create dest path " << dst_dir_name);
                        file.status = srm2__TStatusCode__SRM_USCOREINVALID_USCOREPATH;
                        file.explanation = strerror(r);
                        continue;
                    }
                }
                // Check if destination file already exists
                struct stat64 dst_stat_buf;
                r = stat64(dst_file_path.c_str(),&dst_stat_buf);
                if(0 == r){
                    // File already exist: Check owervite flag on the request
                    if(false == req.overwrite){
                        // Fail since file already exists
                        TRACE_LOG("File " << dst_file_path << " already exists");
                        file.status = srm2__TStatusCode__SRM_USCOREDUPLICATION_USCOREERROR;
                        file.explanation = "File already exists";
                        continue;
                    }
                    // Remove the file
                    r = unlink(dst_file_path.c_str());
                    if(r == -1){
                        TRACE_LOG("Cannot delete existing file " << dst_file_path);
                        file.status = srm2__TStatusCode__SRM_USCOREDUPLICATION_USCOREERROR;
                        file.explanation = "Cannot delete existing file even if overwrite flag is set";
                        continue;
                    }
                }
                // execute a copy command
                std::string cmd =
                    (std::string)"cp " + src_file_path + " " + dst_file_path;
                r = system(cmd.c_str());
                if(0 != r){
                    TRACE_LOG("Copy failed");
                    file.status = srm2__TStatusCode__SRM_USCOREFAILURE;
                    file.explanation = "Failed to Copy file";
                    continue;
                }
                // Check destination file
                r = stat64(dst_file_path.c_str(),&dst_stat_buf);
                if(0 != r){
                    // Invalid destination
                    TRACE_LOG("Invalid destination: " << strerror(errno));
                    file.status = srm2__TStatusCode__SRM_USCOREFAILURE;
                    file.explanation = "Failed to Copy file. Invalid destination";
                    continue;
                }
                file.status  = srm2__TStatusCode__SRM_USCORESUCCESS;
                file.size    = dst_stat_buf.st_size;
                file.permissions = dst_stat_buf.st_mode;
                TRACE_LOG("File " << dst_file_path << " copied from " << src_file_path);
            }
        }
    } else {
        // Nothing To do
    }
    // Update request status
    update_request_status(req);
}

/*
 * complete_request
 */
void complete_request(struct soap * soap,PutRequest& req, srm2__srmPutDoneResponse& resp){

    // Create File Return Statuses
    resp.arrayOfFileStatuses = soap_new_srm2__ArrayOfTSURLReturnStatus(soap,-1);

    bool is_completed = true;
    // For Each File
    std::vector<FileRequest::Ptr>::iterator it;
    for(it = req.fileRequests.begin();it != req.fileRequests.end(); ++it){
        PutFileRequest& f= static_cast<PutFileRequest&>(*((*it).get()));
        // Create File Status
        srm2__TSURLReturnStatus * file = soap_new_srm2__TSURLReturnStatus(soap,-1);
        // Set SURL
        file->surl = f.surl;
        // Set Status
        std::string file_error;
        file->status = soap_new_srm2__TReturnStatus(soap,-1);
        if(f.status == srm2__TStatusCode__SRM_USCORESPACE_USCOREAVAILABLE){
            // TODO Check if file exists
            // Complete File
            f.explanation.clear();
            f.status = srm2__TStatusCode__SRM_USCORESUCCESS;
            file->status->statusCode = srm2__TStatusCode__SRM_USCORESUCCESS;
            TRACE_LOG("File " << f.turl << " completed");
            // Reset TURL
            f.turl.clear();
        } else if((f.status == srm2__TStatusCode__SRM_USCOREREQUEST_USCOREQUEUED) ||
                  (f.status == srm2__TStatusCode__SRM_USCOREREQUEST_USCOREINPROGRESS)){
            // Can't complete file
            file_error = "Cannot complete a file that i in Queued or InProgress";
            file->status->statusCode = srm2__TStatusCode__SRM_USCOREFAILURE;
            is_completed = false;
        } else if(f.status == srm2__TStatusCode__SRM_USCORERELEASED){
            // In theory, this should never happen
            f.status = srm2__TStatusCode__SRM_USCORERELEASED;
            file->status->statusCode = srm2__TStatusCode__SRM_USCORESUCCESS;
        } else if(f.status == srm2__TStatusCode__SRM_USCORESUCCESS){
            // File already completed
            file->status->statusCode = srm2__TStatusCode__SRM_USCORESUCCESS;
        } else {
            // Nothing To Do: return the same status
            file->status->statusCode = f.status;
            file_error = f.explanation;
        }
        if(false == file_error.empty()){
            file->status->explanation = soap_new_std__string(soap,-1);
            *(file->status->explanation) = file_error;
        } else {
            file->status->explanation = 0;
        }
        // Add File
        resp.arrayOfFileStatuses->statusArray.push_back(file);
    }

    // Update Request Status
    if(is_completed){
        req.status = srm2__TStatusCode__SRM_USCORESUCCESS;
        req.explanation.clear();
        TRACE_LOG("Request " << req.id << " completed");
    }

    // Set Request Status
    resp.returnStatus->statusCode = req.status;
    if(false == req.explanation.empty()){
        resp.returnStatus->explanation = soap_new_std__string(soap,-1);
        *(resp.returnStatus->explanation) = req.explanation;
    }
}

/*
 * complete Get request
 */
void complete_request(struct soap * soap,GetRequest& req, srm2__srmReleaseFilesResponse& resp){

    // Create File Return Statuses
    resp.arrayOfFileStatuses = soap_new_srm2__ArrayOfTSURLReturnStatus(soap,-1);

    // For Each File
    std::vector<FileRequest::Ptr>::iterator it;
    for(it = req.fileRequests.begin();it != req.fileRequests.end(); ++it){
        GetFileRequest& f= static_cast<GetFileRequest&>(*((*it).get()));
        // Create File Status
        srm2__TSURLReturnStatus * file = soap_new_srm2__TSURLReturnStatus(soap,-1);
        // Set SURL
        file->surl = f.surl;
        // Set Status
        file->status = soap_new_srm2__TReturnStatus(soap,-1);
        file->status->statusCode = srm2__TStatusCode__SRM_USCORESUCCESS;
        file->status->explanation = 0;
        // Release File
        f.explanation.clear();
        f.status = srm2__TStatusCode__SRM_USCORERELEASED;
        TRACE_LOG("File " << f.turl << " released");
        // Reset TURL
        f.turl.clear();
        // Add File
        resp.arrayOfFileStatuses->statusArray.push_back(file);
    }

    // Update Request Status
    req.status = srm2__TStatusCode__SRM_USCORESUCCESS;
    req.explanation.clear();
    TRACE_LOG("Request " << req.id << " release");

    // Set Request Status
    resp.returnStatus->statusCode = req.status;
    resp.returnStatus->explanation = 0;
}

/*
 * complete BringOnline request
 */
void complete_request(struct soap * soap,BringOnlineRequest& req, srm2__srmReleaseFilesResponse& resp){

    // Create File Return Statuses
    resp.arrayOfFileStatuses = soap_new_srm2__ArrayOfTSURLReturnStatus(soap,-1);

    // For Each File
    std::vector<FileRequest::Ptr>::iterator it;
    for(it = req.fileRequests.begin();it != req.fileRequests.end(); ++it){
        BringOnlineFileRequest& f= static_cast<BringOnlineFileRequest&>(*((*it).get()));
        // Create File Status
        srm2__TSURLReturnStatus * file = soap_new_srm2__TSURLReturnStatus(soap,-1);
        // Set SURL
        file->surl = f.surl;
        // Set Status
        file->status = soap_new_srm2__TReturnStatus(soap,-1);
        file->status->statusCode = srm2__TStatusCode__SRM_USCORESUCCESS;
        file->status->explanation = 0;
        // Release File
        f.explanation.clear();
        f.status = srm2__TStatusCode__SRM_USCORERELEASED;
        TRACE_LOG("File " << f.surl << " BringOnline released");
        // Add File
        resp.arrayOfFileStatuses->statusArray.push_back(file);
    }

    // Update Request Status
    req.status = srm2__TStatusCode__SRM_USCORESUCCESS;
    req.explanation.clear();
    TRACE_LOG("Request " << req.id << " release");

    // Set Request Status
    resp.returnStatus->statusCode = req.status;
    resp.returnStatus->explanation = 0;
}

/*
 * abort_request
 */
void abort_request(struct soap * soap,PutRequest& req, srm2__srmAbortRequestResponse& resp){

    // For Each File
    std::vector<FileRequest::Ptr>::iterator it;
    for(it = req.fileRequests.begin();it != req.fileRequests.end(); ++it){
        PutFileRequest& f = static_cast<PutFileRequest&>(*((*it).get()));
        if(f.status == srm2__TStatusCode__SRM_USCORESPACE_USCOREAVAILABLE){
            // Get File Path
            std::string file_path = path_from_surl(f.surl);
            if(false == file_path.empty()){
                // Delete File
                int r = unlink(file_path.c_str());
                if(0 != r){
                    TRACE_ERR("Failed To delete File " << file_path << ": reason is " << strerror(errno));
                }
            }
        }
        // Set file Status to Aborted
        f.status = srm2__TStatusCode__SRM_USCOREABORTED;
        f.explanation.clear();
    }
    // Set Request Status
    req.status = srm2__TStatusCode__SRM_USCOREABORTED;
    req.explanation.clear();

    // Set Request Status
    resp.returnStatus->statusCode = srm2__TStatusCode__SRM_USCORESUCCESS;
}

/*
 * abort_request
 */
void abort_request(struct soap * soap,GetRequest& req, srm2__srmAbortRequestResponse& resp){

    // For Each File
    std::vector<FileRequest::Ptr>::iterator it;
    for(it = req.fileRequests.begin();it != req.fileRequests.end(); ++it){
        GetFileRequest& f = static_cast<GetFileRequest&>(*((*it).get()));
        // Set file Status to Aborted
        f.status = srm2__TStatusCode__SRM_USCOREABORTED;
        f.explanation.clear();
    }
    // Set Request Status
    req.status = srm2__TStatusCode__SRM_USCOREABORTED;
    req.explanation.clear();

    // Set Request Status
    resp.returnStatus->statusCode = srm2__TStatusCode__SRM_USCORESUCCESS;
}

/*
 * abort_request
 */
void abort_request(struct soap * soap,BringOnlineRequest& req, srm2__srmAbortRequestResponse& resp){

    // For Each File
    std::vector<FileRequest::Ptr>::iterator it;
    for(it = req.fileRequests.begin();it != req.fileRequests.end(); ++it){
        BringOnlineFileRequest& f = static_cast<BringOnlineFileRequest&>(*((*it).get()));
        // Set file Status to Aborted
        f.status = srm2__TStatusCode__SRM_USCOREABORTED;
        f.explanation.clear();
    }
    // Set Request Status
    req.status = srm2__TStatusCode__SRM_USCOREABORTED;
    req.explanation.clear();

    // Set Request Status
    resp.returnStatus->statusCode = srm2__TStatusCode__SRM_USCORESUCCESS;
}


/*
 * abort_request
 */
void abort_request(struct soap * soap,CopyRequest& req, srm2__srmAbortRequestResponse& resp){

    // For Each File
    std::vector<FileRequest::Ptr>::iterator it;
    for(it = req.fileRequests.begin();it != req.fileRequests.end(); ++it){
        CopyFileRequest& f = static_cast<CopyFileRequest&>(*((*it).get()));
        if(f.status == srm2__TStatusCode__SRM_USCORESUCCESS){
            // Get File Path
            std::string file_path = path_from_surl(f.targetSurl);
            if(false == file_path.empty()){
                // Delete File
                int r = unlink(file_path.c_str());
                if(0 != r){
                    TRACE_ERR("Failed To delete File " << file_path << ": reason is " << strerror(errno));
                }
            }
        }
        // Set file Status to Aborted
        f.status = srm2__TStatusCode__SRM_USCOREABORTED;
        f.explanation.clear();
    }
    // Set Request Status
    req.status = srm2__TStatusCode__SRM_USCOREABORTED;
    req.explanation.clear();

    // Set Request Status
    resp.returnStatus->statusCode = srm2__TStatusCode__SRM_USCORESUCCESS;
}

/*
 * abort_files
 */
void abort_files(struct soap * soap,const std::vector<std::string>& urlArray, PutRequest& req, srm2__srmAbortFilesResponse& resp){

    bool has_success = false;
    bool has_failed  = false;
    // For Each File
    std::vector<std::string>::const_iterator sit;
    std::vector<FileRequest::Ptr>::iterator it;
    unsigned int count = 0;
    for(sit = urlArray.begin(); sit != urlArray.end(); ++sit, ++count){
        // Find File request Object
        for(it = req.fileRequests.begin();it != req.fileRequests.end(); ++it){
            PutFileRequest& f = static_cast<PutFileRequest&>(*((*it).get()));
            if(f.surl != (*sit)){
                continue;
            }
            if(f.status == srm2__TStatusCode__SRM_USCORESPACE_USCOREAVAILABLE){
                // Get File Path
                std::string file_path = path_from_surl(f.surl);
                if(false == file_path.empty()){
                    // Delete File
                    int r = unlink(file_path.c_str());
                    if(0 != r){
                        TRACE_ERR("Failed To delete File " << file_path << ". Reason is " << strerror(errno));
                    }
                }
            }
            // Set file Status to Aborted
            f.status = srm2__TStatusCode__SRM_USCOREABORTED;
            f.explanation.clear();
            // Set result
            resp.arrayOfFileStatuses->statusArray.at(count)->status->statusCode = srm2__TStatusCode__SRM_USCORESUCCESS;
            resp.arrayOfFileStatuses->statusArray.at(count)->status->explanation = 0;
            has_success = true;
            break;
        } // end for each file

        if(it == req.fileRequests.end()){
            // Invalid Path
            resp.arrayOfFileStatuses->statusArray.at(count)->status->statusCode = srm2__TStatusCode__SRM_USCOREINVALID_USCOREPATH;
            resp.arrayOfFileStatuses->statusArray.at(count)->status->explanation = soap_new_std__string(soap,-1);
            *(resp.arrayOfFileStatuses->statusArray.at(count)->status->explanation) = "No such file in the request";
            has_failed = true;
        }
    } // end for each surl

    // Set Request Status
    update_request_status(req);

    // Set Request Status
    if(true == has_success){
        if(true == has_failed){
            resp.returnStatus->statusCode = srm2__TStatusCode__SRM_USCOREPARTIAL_USCORESUCCESS;
        } else {
            resp.returnStatus->statusCode = srm2__TStatusCode__SRM_USCORESUCCESS;
        }
    } else {
        resp.returnStatus->statusCode = srm2__TStatusCode__SRM_USCOREFAILURE;
    }
}

/*
 * abort_files
 */
void abort_files(struct soap * soap,const std::vector<std::string>& urlArray, GetRequest& req, srm2__srmAbortFilesResponse& resp){

    bool has_success = false;
    bool has_failed  = false;
    // For Each File
    std::vector<std::string>::const_iterator sit;
    std::vector<FileRequest::Ptr>::iterator it;
    unsigned int count = 0;
    for(sit = urlArray.begin(); sit != urlArray.end(); ++sit, ++count){
        // Find File request Object
        for(it = req.fileRequests.begin();it != req.fileRequests.end(); ++it){
            GetFileRequest& f = static_cast<GetFileRequest&>(*((*it).get()));
            if(f.surl != (*sit)){
                continue;
            }
            // Set file Status to Aborted
            f.status = srm2__TStatusCode__SRM_USCOREABORTED;
            f.explanation.clear();
            // Set result
            resp.arrayOfFileStatuses->statusArray.at(count)->status->statusCode = srm2__TStatusCode__SRM_USCORESUCCESS;
            resp.arrayOfFileStatuses->statusArray.at(count)->status->explanation = 0;
            has_success = true;
            break;
        } // end for each file

        if(it == req.fileRequests.end()){
            // Invalid Path
            resp.arrayOfFileStatuses->statusArray.at(count)->status->statusCode = srm2__TStatusCode__SRM_USCOREINVALID_USCOREPATH;
            resp.arrayOfFileStatuses->statusArray.at(count)->status->explanation = soap_new_std__string(soap,-1);
            *(resp.arrayOfFileStatuses->statusArray.at(count)->status->explanation) = "No such file in the request";
            has_failed = true;
        }
    } // end for each surl

    // Set Request Status
    update_request_status(req);

    // Set Request Status
    if(true == has_success){
        if(true == has_failed){
            resp.returnStatus->statusCode = srm2__TStatusCode__SRM_USCOREPARTIAL_USCORESUCCESS;
        } else {
            resp.returnStatus->statusCode = srm2__TStatusCode__SRM_USCORESUCCESS;
        }
    } else {
        resp.returnStatus->statusCode = srm2__TStatusCode__SRM_USCOREFAILURE;
    }
}

/*
 * abort_files
 */
void abort_files(struct soap * soap,const std::vector<std::string>& urlArray, CopyRequest& req, srm2__srmAbortFilesResponse& resp){

    bool has_success = false;
    bool has_failed  = false;
    // For Each File
    std::vector<std::string>::const_iterator sit;
    std::vector<FileRequest::Ptr>::iterator it;
    unsigned int count = 0;
    for(sit = urlArray.begin(); sit != urlArray.end(); ++sit, ++count){
        // Find File request Object
        for(it = req.fileRequests.begin();it != req.fileRequests.end(); ++it){
            CopyFileRequest& f = static_cast<CopyFileRequest&>(*((*it).get()));
            if((f.targetSurl != (*sit)) && (f.sourceSurl != (*sit))){
                continue;
            }
            if(f.status == srm2__TStatusCode__SRM_USCORESUCCESS){
                // Get File Path
                std::string file_path = path_from_surl(f.targetSurl);
                if(false == file_path.empty()){
                    // Delete File
                    int r = unlink(file_path.c_str());
                    if(0 != r){
                        TRACE_ERR("Failed To delete File " << file_path << ". Reason is " << strerror(errno));
                    }
                }
            }
            // Set file Status to Aborted
            f.status = srm2__TStatusCode__SRM_USCOREABORTED;
            f.explanation.clear();
            // Set result
            resp.arrayOfFileStatuses->statusArray.at(count)->status->statusCode = srm2__TStatusCode__SRM_USCORESUCCESS;
            resp.arrayOfFileStatuses->statusArray.at(count)->status->explanation = 0;
            has_success = true;
            break;
        } // end for each file

        if(it == req.fileRequests.end()){
            // Invalid Path
            resp.arrayOfFileStatuses->statusArray.at(count)->status->statusCode = srm2__TStatusCode__SRM_USCOREINVALID_USCOREPATH;
            resp.arrayOfFileStatuses->statusArray.at(count)->status->explanation = soap_new_std__string(soap,-1);
            *(resp.arrayOfFileStatuses->statusArray.at(count)->status->explanation) = "No such file in the request";
            has_failed = true;
        }
    } // end for each surl

    // Set Request Status
    update_request_status(req);

    // Set Request Status
    if(true == has_success){
        if(true == has_failed){
            resp.returnStatus->statusCode = srm2__TStatusCode__SRM_USCOREPARTIAL_USCORESUCCESS;
        } else {
            resp.returnStatus->statusCode = srm2__TStatusCode__SRM_USCORESUCCESS;
        }
    } else {
        resp.returnStatus->statusCode = srm2__TStatusCode__SRM_USCOREFAILURE;
    }
}

/*
 * abort_files
 */
void abort_files(struct soap * soap,const std::vector<std::string>& urlArray, BringOnlineRequest& req, srm2__srmAbortFilesResponse& resp){

    bool has_success = false;
    bool has_failed  = false;
    // For Each File
    std::vector<std::string>::const_iterator sit;
    std::vector<FileRequest::Ptr>::iterator it;
    unsigned int count = 0;
    for(sit = urlArray.begin(); sit != urlArray.end(); ++sit, ++count){
        // Find File request Object
        for(it = req.fileRequests.begin();it != req.fileRequests.end(); ++it){
            BringOnlineFileRequest& f = static_cast<BringOnlineFileRequest&>(*((*it).get()));
            if(f.surl != (*sit)){
                continue;
            }
            // Set file Status to Aborted
            f.status = srm2__TStatusCode__SRM_USCOREABORTED;
            f.explanation.clear();
            // Set result
            resp.arrayOfFileStatuses->statusArray.at(count)->status->statusCode = srm2__TStatusCode__SRM_USCORESUCCESS;
            resp.arrayOfFileStatuses->statusArray.at(count)->status->explanation = 0;
            has_success = true;
            break;
        } // end for each file

        if(it == req.fileRequests.end()){
            // Invalid Path
            resp.arrayOfFileStatuses->statusArray.at(count)->status->statusCode = srm2__TStatusCode__SRM_USCOREINVALID_USCOREPATH;
            resp.arrayOfFileStatuses->statusArray.at(count)->status->explanation = soap_new_std__string(soap,-1);
            *(resp.arrayOfFileStatuses->statusArray.at(count)->status->explanation) = "No such file in the request";
            has_failed = true;
        }
    } // end for each surl

    // Set Request Status
    update_request_status(req);

    // Set Request Status
    if(true == has_success){
        if(true == has_failed){
            resp.returnStatus->statusCode = srm2__TStatusCode__SRM_USCOREPARTIAL_USCORESUCCESS;
        } else {
            resp.returnStatus->statusCode = srm2__TStatusCode__SRM_USCORESUCCESS;
        }
    } else {
        resp.returnStatus->statusCode = srm2__TStatusCode__SRM_USCOREFAILURE;
    }
}

// -----------------------------------------------------------------------------
//
// Translation Routines
//
// -----------------------------------------------------------------------------

/*
 * Get the Status of Put File Request
 */
srm2__ArrayOfTPutRequestFileStatus * status_of_put_files(struct soap * soap, PutRequest& req){
    srm2__ArrayOfTPutRequestFileStatus * array = soap_new_srm2__ArrayOfTPutRequestFileStatus(soap,-1);

    // For Each File
    std::vector<FileRequest::Ptr>::const_iterator it;
    for(it = req.fileRequests.begin(); it != req.fileRequests.end(); ++it){
        PutFileRequest& f = static_cast<PutFileRequest&>(*((*it).get()));
        srm2__TPutRequestFileStatus * file = soap_new_srm2__TPutRequestFileStatus(soap,-1);
        // Set SURL
        file->SURL = f.surl;
        // Set Status
        file->status = soap_new_srm2__TReturnStatus(soap,-1);
        file->status->statusCode = f.status;
        if(false == f.explanation.empty()){
            file->status->explanation = soap_new_std__string(soap,-1);
            *(file->status->explanation) = f.explanation;
        } else {
            file->status->explanation = 0;
        }
        // Set File size
        if(f.size > 0){
            file->fileSize = (ULONG64*)soap_malloc(soap,sizeof(ULONG64));
            *(file->fileSize) = f.size;
        } else {
            file->fileSize = 0;
        }
        // Set Turl
        if(false == f.turl.empty()){
            file->transferURL = soap_new_std__string(soap,-1);
            *(file->transferURL) = f.turl;
        } else {
            file->transferURL =  0;
        }
        // estimatedWaitTime
        file->estimatedWaitTime = (int*)soap_malloc(soap,sizeof(int));
        *(file->estimatedWaitTime) = 1;
        // Not Used
        file->remainingPinLifetime  = 0;
        file->remainingFileLifetime = 0;
        file->transferProtocolInfo  = 0;
        // Add File
        array->statusArray.push_back(file);
    }
    return array;
}

/*
 * Get the Status of Get File Request
 */
srm2__ArrayOfTGetRequestFileStatus * status_of_get_files(struct soap * soap, GetRequest& req){
    srm2__ArrayOfTGetRequestFileStatus * array = soap_new_srm2__ArrayOfTGetRequestFileStatus(soap,-1);

    // For Each File
    std::vector<FileRequest::Ptr>::const_iterator it;
    for(it = req.fileRequests.begin(); it != req.fileRequests.end(); ++it){
        GetFileRequest& f = static_cast<GetFileRequest&>(*((*it).get()));
        srm2__TGetRequestFileStatus * file = soap_new_srm2__TGetRequestFileStatus(soap,-1);
        // Set SURL
        file->sourceSURL = f.surl;
        // Set Status
        file->status = soap_new_srm2__TReturnStatus(soap,-1);
        file->status->statusCode = f.status;
        if(false == f.explanation.empty()){
            file->status->explanation = soap_new_std__string(soap,-1);
            *(file->status->explanation) = f.explanation;
        } else {
            file->status->explanation = 0;
        }
        // Set File size
        if(f.size > 0){
            file->fileSize = (ULONG64*)soap_malloc(soap,sizeof(ULONG64));
            *(file->fileSize) = f.size;
        } else {
            file->fileSize = 0;
        }
        // Set Turl
        if(false == f.turl.empty()){
            file->transferURL = soap_new_std__string(soap,-1);
            *(file->transferURL) = f.turl;
        } else {
            file->transferURL =  0;
        }
        // Estimated wait time
        file->estimatedWaitTime = (int*)soap_malloc(soap,sizeof(int));
        *(file->estimatedWaitTime) = 1;
        // Not Used
        file->remainingPinTime      = 0;
        file->transferProtocolInfo  = 0;
        // Add File
        array->statusArray.push_back(file);
    }
    return array;
}

/*
 * Get the Status of BringOnline File Request
 */
srm2__ArrayOfTBringOnlineRequestFileStatus * status_of_bringonline_files(struct soap * soap, BringOnlineRequest& req){
    srm2__ArrayOfTBringOnlineRequestFileStatus * array = soap_new_srm2__ArrayOfTBringOnlineRequestFileStatus(soap,-1);

    // For Each File
    std::vector<FileRequest::Ptr>::const_iterator it;
    for(it = req.fileRequests.begin(); it != req.fileRequests.end(); ++it){
        BringOnlineFileRequest& f = static_cast<BringOnlineFileRequest&>(*((*it).get()));
        srm2__TBringOnlineRequestFileStatus * file = soap_new_srm2__TBringOnlineRequestFileStatus(soap,-1);
        // Set SURL
        file->sourceSURL = f.surl;
        // Set Status
        file->status = soap_new_srm2__TReturnStatus(soap,-1);
        file->status->statusCode = f.status;
        if(false == f.explanation.empty()){
            file->status->explanation = soap_new_std__string(soap,-1);
            *(file->status->explanation) = f.explanation;
        } else {
            file->status->explanation = 0;
        }
        // Set File size
        if(f.size > 0){
            file->fileSize = (ULONG64*)soap_malloc(soap,sizeof(ULONG64));
            *(file->fileSize) = f.size;
        } else {
            file->fileSize = 0;
        }
        // estimatedWaitTime
        file->estimatedWaitTime = (int*)soap_malloc(soap,sizeof(int));
        *(file->estimatedWaitTime) = 1;
        // Not Used
        file->remainingPinTime  = 0;
        // Add File
        array->statusArray.push_back(file);
    }
    return array;
}

/*
 * Get the Status of Copy File Request
 */
srm2__ArrayOfTCopyRequestFileStatus * status_of_copy_files(struct soap * soap, CopyRequest& req){
    srm2__ArrayOfTCopyRequestFileStatus * array = soap_new_srm2__ArrayOfTCopyRequestFileStatus(soap,-1);

    // For Each File
    std::vector<FileRequest::Ptr>::const_iterator it;
    for(it = req.fileRequests.begin(); it != req.fileRequests.end(); ++it){
        CopyFileRequest& f = static_cast<CopyFileRequest&>(*((*it).get()));
        srm2__TCopyRequestFileStatus * file = soap_new_srm2__TCopyRequestFileStatus(soap,-1);
        // Set SURLs
        file->sourceSURL = f.sourceSurl;
        file->targetSURL = f.targetSurl;
        // Set Status
        file->status = soap_new_srm2__TReturnStatus(soap,-1);
        file->status->statusCode = f.status;
        if(false == f.explanation.empty()){
            file->status->explanation = soap_new_std__string(soap,-1);
            *(file->status->explanation) = f.explanation;
        } else {
            file->status->explanation = 0;
        }
        // Set File size
        if(f.size > 0){
            file->fileSize = (ULONG64*)soap_malloc(soap,sizeof(ULONG64));
            *(file->fileSize) = f.size;
        } else {
            file->fileSize = 0;
        }
        // estimatedWaitTime
        file->estimatedWaitTime = (int*)soap_malloc(soap,sizeof(int));
        *(file->estimatedWaitTime) = 1;
        // Not Used
        file->remainingFileLifetime = 0;
        // Add File
        array->statusArray.push_back(file);
    }
    return array;
}

// -----------------------------------------------------------------------------
//
// SRM 2.2 Stub Implementation
//
// -----------------------------------------------------------------------------

template <typename RequestTraits>
bool srm2_handle(typename RequestTraits::RequestType* request, typename RequestTraits::ResponseType& out, struct soap* soap)
{
	assert(request);
	assert(soap);
	RequestTraits traits(*request, out, soap);
	ITestHandler* handler = TestHandlerFactory<RequestTraits, SrmBusyTestTraits>::getTestHandler(&traits);

	if (handler) {
		handler->handle();
		return true;
	}

	return false;
}

/*
 * srm2__srmPing
 */
int srm2__srmPing(struct soap* soap, srm2__srmPingRequest *srmPingRequest, struct srm2__srmPingResponse_ &out){
    TRACE_LOG("srm2__srmPing called");

    out.srmPingResponse = soap_new_srm2__srmPingResponse(soap,-1);
    out.srmPingResponse->versionInfo = "v2.2";
    out.srmPingResponse->otherInfo = 0;

    TRACE_LOG("Done: srm2__srmPing");
    return SOAP_OK;
}

//..............................................................................
// Put Methods
//..............................................................................

/*
 * srm2__srmPrepareToPut
 */
int srm2__srmPrepareToPut(struct soap* soap, srm2__srmPrepareToPutRequest *srmPrepareToPutRequest, struct srm2__srmPrepareToPutResponse_ &out){
    TRACE_LOG("srm2__srmPrepareToPut called");
    int result = SOAP_OK;

    // Precondition Checks
    if((0 == srmPrepareToPutRequest) ||
       (0 == srmPrepareToPutRequest->arrayOfFileRequests) ||
       (true == srmPrepareToPutRequest->arrayOfFileRequests->requestArray.empty())){
        // Error
        return soap_sender_fault(soap,"Null request sent",0);
    }

    if (srm2_handle<SrmPrepareToPutTraits>(srmPrepareToPutRequest, out, soap))
    	return result;

    const std::vector<srm2__TPutFileRequest * > & requestArray = srmPrepareToPutRequest->arrayOfFileRequests->requestArray;
    TRACE_LOG("Request to Put " << (int)requestArray.size() << " Files");
    // Overwrite Mode
    if(srmPrepareToPutRequest->overwriteOption){
        switch(*(srmPrepareToPutRequest->overwriteOption)){
            case srm2__TOverwriteMode__NEVER :
                TRACE_LOG("\tOverwrite: Never");
            break;
            case srm2__TOverwriteMode__ALWAYS :
                TRACE_LOG("\tOverwrite: Always");
            break;
            case srm2__TOverwriteMode__WHEN_USCOREFILES_USCOREARE_USCOREDIFFERENT :
                TRACE_LOG("\tOverwrite: When Different");
            break;
        }
    }
    // Protocol
    if((0 != srmPrepareToPutRequest->transferParameters) &&
       (0 != srmPrepareToPutRequest->transferParameters->arrayOfTransferProtocols) &&
       (false == srmPrepareToPutRequest->transferParameters->arrayOfTransferProtocols->stringArray.empty())){
        TRACE_LOG("\tProtocol: " << srmPrepareToPutRequest->transferParameters->arrayOfTransferProtocols->stringArray.at(0));
    }
    // Token
    if(0 != srmPrepareToPutRequest->targetSpaceToken){
        TRACE_LOG("\tToken: " << srmPrepareToPutRequest->targetSpaceToken);
    }
    // For Each File
    std::vector<srm2__TPutFileRequest * >::const_iterator it;
    for(it = requestArray.begin(); it != requestArray.end(); ++it){
        const srm2__TPutFileRequest *  file = (*it);
        if(0 != file){
            TRACE_LOG("\tFile  " << (file->targetSURL ? file->targetSURL->c_str() : "(null)") );
            TRACE_LOG("\tSize " << (file->expectedFileSize ? *file->expectedFileSize : 0) << "ld");
        } else {
            TRACE_LOG("\tNull file");
        }
    }

    // check if some parameter was passed with the extra info array
    bool add_extra_info_details = false;
    std::stringstream extra_info_details;
    if (srmPrepareToPutRequest->storageSystemInfo){
        // check source space token
        TRACE_LOG("Reading storageSystemInfo extra info array");
        const std::vector<srm2__TExtraInfo * >& info = srmPrepareToPutRequest->storageSystemInfo->extraInfoArray;
        std::vector<srm2__TExtraInfo * >::const_iterator it;
        for (it = info.begin();it != info.end();++it){
            srm2__TExtraInfo * extra_info = (*it);
            const std::string& key = extra_info->key;
            if (key == "CopyPinLifetime"){
                TRACE_LOG("Reading parameter " << key);
                add_extra_info_details = true;
                std::string value = extra_info->value ? *(extra_info->value) : "<NULL>";
                extra_info_details << " [" << key << "=" << value << "]";
            } else {
                TRACE_LOG("Ignoring parameter " << key);
            }
        }
    }

    // Create Response
    out.srmPrepareToPutResponse = soap_new_srm2__srmPrepareToPutResponse(soap,-1);
    out.srmPrepareToPutResponse->returnStatus = soap_new_srm2__TReturnStatus(soap,-1);
    out.srmPrepareToPutResponse->returnStatus->statusCode = srm2__TStatusCode__SRM_USCOREFAILURE;
    out.srmPrepareToPutResponse->returnStatus->explanation = 0;
    out.srmPrepareToPutResponse->remainingTotalRequestTime = 0;
    out.srmPrepareToPutResponse->arrayOfFileStatuses       = 0;

    // Create Request Object
    PutRequest::Ptr req(new PutRequest());
    req->status = srm2__TStatusCode__SRM_USCOREREQUEST_USCOREQUEUED;
    // Overwrite Mode
    if(srmPrepareToPutRequest->overwriteOption){
        switch(*(srmPrepareToPutRequest->overwriteOption)){
            case srm2__TOverwriteMode__NEVER :
                req->overwrite = false;
            break;
            case srm2__TOverwriteMode__ALWAYS :
            case srm2__TOverwriteMode__WHEN_USCOREFILES_USCOREARE_USCOREDIFFERENT :
                req->overwrite = true;
            break;
        }
    }
    // Get Protocol
    std::string protocol = PROTOCOL_FILE;
    if((0 != srmPrepareToPutRequest->transferParameters) &&
       (0 != srmPrepareToPutRequest->transferParameters->arrayOfTransferProtocols) &&
       (false == srmPrepareToPutRequest->transferParameters->arrayOfTransferProtocols->stringArray.empty())){
        protocol = srmPrepareToPutRequest->transferParameters->arrayOfTransferProtocols->stringArray.at(0);
    }
    req->protocol = protocol;

    // For each file
    for(it = requestArray.begin(); it != requestArray.end(); ++it){
        const srm2__TPutFileRequest *  file = (*it);
        PutFileRequest::Ptr f(new PutFileRequest());
        if(0 != file){
            if(file->targetSURL) {
                f->surl = *(file->targetSURL);
                if(file->expectedFileSize) {
                    f->size = *(file->expectedFileSize);
                }
                f->permissions = 0644;
                f->status = srm2__TStatusCode__SRM_USCOREREQUEST_USCOREQUEUED;
            } else {
                f->status = srm2__TStatusCode__SRM_USCOREFAILURE;
                f->explanation = "Null SURL not supported";
            }
        } else {
            f->status = srm2__TStatusCode__SRM_USCOREFAILURE;
            f->explanation = "Null file";
        }
        // Store file
        req->fileRequests.push_back(f);
    }

    // Insert Request
    if(0 == store_request(req)){
        // Set Request Token
        out.srmPrepareToPutResponse->requestToken = soap_new_std__string(soap,-1);
        *(out.srmPrepareToPutResponse->requestToken) = req->id;
        // Set Request Status
        out.srmPrepareToPutResponse->returnStatus->statusCode = req->status;
        if(false == req->explanation.empty()){
            out.srmPrepareToPutResponse->returnStatus->explanation = soap_new_std__string(soap,-1);
            *(out.srmPrepareToPutResponse->returnStatus->explanation) = req->explanation;
        }
        // Set File Status Array
        out.srmPrepareToPutResponse->arrayOfFileStatuses = status_of_put_files(soap,*req);
    } else {
        // Set error Message
        result = soap_receiver_fault(soap,"Failed to store request",0);
    }

    if (add_extra_info_details){
        if (0 == out.srmPrepareToPutResponse->returnStatus->explanation){
            out.srmPrepareToPutResponse->returnStatus->explanation = soap_new_std__string(soap,-1);
        }
        std::stringstream tmp;
        tmp << *(out.srmPrepareToPutResponse->returnStatus->explanation);
        tmp << extra_info_details.str();
        *(out.srmPrepareToPutResponse->returnStatus->explanation) = tmp.str();
    }

    TRACE_LOG("Done: srm2__srmPrepareToPut");
    return result;
}

/*
 * srm2__srmStatusOfPutRequest
 */
int srm2__srmStatusOfPutRequest(struct soap* soap, srm2__srmStatusOfPutRequestRequest *srmStatusOfPutRequestRequest, struct srm2__srmStatusOfPutRequestResponse_ &out){
    TRACE_LOG("srm2__srmStatusOfPutRequest called");
    int result = SOAP_OK;

    // Precondition Checks
    if(0 == srmStatusOfPutRequestRequest){
        // Error
        return soap_sender_fault(soap,"Null request sent",0);
    }
    if(true == srmStatusOfPutRequestRequest->requestToken.empty()){
        // Error
        return soap_sender_fault(soap,"Empty token sent",0);
    }
    TRACE_LOG("Get Status of Put Request " << srmStatusOfPutRequestRequest->requestToken);

    // Create Response
    out.srmStatusOfPutRequestResponse = soap_new_srm2__srmStatusOfPutRequestResponse(soap,-1);
    out.srmStatusOfPutRequestResponse->returnStatus = soap_new_srm2__TReturnStatus(soap,-1);
    out.srmStatusOfPutRequestResponse->returnStatus->statusCode = srm2__TStatusCode__SRM_USCOREFAILURE;
    out.srmStatusOfPutRequestResponse->returnStatus->explanation = 0;
    out.srmStatusOfPutRequestResponse->remainingTotalRequestTime = 0;
    out.srmStatusOfPutRequestResponse->arrayOfFileStatuses       = 0;
    // Find Request
    Request::Ptr req = get_request(srmStatusOfPutRequestRequest->requestToken);
    if(0 == req.get()){
        // No such Request
        out.srmStatusOfPutRequestResponse->returnStatus->statusCode = srm2__TStatusCode__SRM_USCOREINVALID_USCOREREQUEST;
        out.srmStatusOfPutRequestResponse->returnStatus->explanation = soap_new_std__string(soap,-1);
        *(out.srmStatusOfPutRequestResponse->returnStatus->explanation) = "Invalid request token";
    } else if(req->type != Request::T_PUT) {
        // Invalid request Type
        out.srmStatusOfPutRequestResponse->returnStatus->statusCode = srm2__TStatusCode__SRM_USCOREINVALID_USCOREREQUEST;
        out.srmStatusOfPutRequestResponse->returnStatus->explanation = soap_new_std__string(soap,-1);
        *(out.srmStatusOfPutRequestResponse->returnStatus->explanation) = "Not a Put Request";
    } else {
        // Check if the request is the correct type
        PutRequest& put_req = static_cast<PutRequest&>(*(req.get()));
        // Process Put Request
        process_request(put_req);
        // Set Request Status
        out.srmStatusOfPutRequestResponse->returnStatus->statusCode = put_req.status;
        if(false == put_req.explanation.empty()){
            out.srmStatusOfPutRequestResponse->returnStatus->explanation = soap_new_std__string(soap,-1);
            *(out.srmStatusOfPutRequestResponse->returnStatus->explanation) = put_req.explanation;
        }
        // Set File Status Array
        out.srmStatusOfPutRequestResponse->arrayOfFileStatuses = status_of_put_files(soap,put_req);
    }

    TRACE_LOG("Done: srm2__srmStatusOfPutRequest");
    return result;
}

//..............................................................................
// Get Methods
//..............................................................................

/*
 * srm2__srmPrepareToGet
 */
int srm2__srmPrepareToGet(struct soap* soap, srm2__srmPrepareToGetRequest *srmPrepareToGetRequest, struct srm2__srmPrepareToGetResponse_ &out){
    TRACE_LOG("srm2__srmPrepareToGet called");
    int result = SOAP_OK;

    // Precondition Checks
    if((0 == srmPrepareToGetRequest) ||
       (0 == srmPrepareToGetRequest->arrayOfFileRequests) ||
       (true == srmPrepareToGetRequest->arrayOfFileRequests->requestArray.empty())){
        // Error
        return soap_sender_fault(soap,"Null request sent",0);
    }

    if (srm2_handle<SrmPrepareToGetTraits>(srmPrepareToGetRequest, out, soap))
     	return result;

    const std::vector<srm2__TGetFileRequest * > & requestArray = srmPrepareToGetRequest->arrayOfFileRequests->requestArray;
    TRACE_LOG("Request to Get " << (int)requestArray.size() << " Files");
    // Protocol
    if((0 != srmPrepareToGetRequest->transferParameters) &&
       (0 != srmPrepareToGetRequest->transferParameters->arrayOfTransferProtocols) &&
       (false == srmPrepareToGetRequest->transferParameters->arrayOfTransferProtocols->stringArray.empty())){
        TRACE_LOG("\tProtocol: " << srmPrepareToGetRequest->transferParameters->arrayOfTransferProtocols->stringArray.at(0));
    }
    // For Each File
    std::vector<srm2__TGetFileRequest * >::const_iterator it;
    for(it = requestArray.begin(); it != requestArray.end(); ++it){
        const srm2__TGetFileRequest *  file = (*it);
        if(0 != file){
            TRACE_LOG("\tFile  " << file->sourceSURL);
        } else {
            TRACE_LOG("\tNull file");
        }
    }

    // Create Response
    out.srmPrepareToGetResponse = soap_new_srm2__srmPrepareToGetResponse(soap,-1);
    out.srmPrepareToGetResponse->returnStatus = soap_new_srm2__TReturnStatus(soap,-1);
    out.srmPrepareToGetResponse->returnStatus->statusCode = srm2__TStatusCode__SRM_USCOREFAILURE;
    out.srmPrepareToGetResponse->returnStatus->explanation = 0;
    out.srmPrepareToGetResponse->remainingTotalRequestTime = 0;
    out.srmPrepareToGetResponse->arrayOfFileStatuses       = 0;

    // Create Request Object
    GetRequest::Ptr req(new GetRequest());
    req->status = srm2__TStatusCode__SRM_USCOREREQUEST_USCOREQUEUED;
    // Get Protocol
    std::string protocol = PROTOCOL_FILE;
    if((0 != srmPrepareToGetRequest->transferParameters) &&
       (0 != srmPrepareToGetRequest->transferParameters->arrayOfTransferProtocols) &&
       (false == srmPrepareToGetRequest->transferParameters->arrayOfTransferProtocols->stringArray.empty())){
        protocol = srmPrepareToGetRequest->transferParameters->arrayOfTransferProtocols->stringArray.at(0);
    }
    req->protocol = protocol;

    // For each file
    for(it = requestArray.begin(); it != requestArray.end(); ++it){
        const srm2__TGetFileRequest *  file = (*it);
        GetFileRequest::Ptr f(new GetFileRequest());
        if(0 != file){
            if(false == file->sourceSURL.empty()) {
                f->surl = file->sourceSURL;
                f->status = srm2__TStatusCode__SRM_USCOREREQUEST_USCOREQUEUED;
            } else {
                f->status = srm2__TStatusCode__SRM_USCOREFAILURE;
                f->explanation = "Empty Surl";
            }
        } else {
            f->status = srm2__TStatusCode__SRM_USCOREFAILURE;
            f->explanation = "Null file";
        }
        // Store file
        req->fileRequests.push_back(f);
    }

    // Insert Request
    if(0 == store_request(req)){
        // Set Request Token
        out.srmPrepareToGetResponse->requestToken = soap_new_std__string(soap,-1);
        *(out.srmPrepareToGetResponse->requestToken) = req->id;
        // Set Request Status
        out.srmPrepareToGetResponse->returnStatus->statusCode = req->status;
        if(false == req->explanation.empty()){
            out.srmPrepareToGetResponse->returnStatus->explanation = soap_new_std__string(soap,-1);
            *(out.srmPrepareToGetResponse->returnStatus->explanation) = req->explanation;
        }
        // Set File Status Array
        out.srmPrepareToGetResponse->arrayOfFileStatuses = status_of_get_files(soap,*req);
    } else {
        // Set error Message
        result = soap_receiver_fault(soap,"Failed to store request",0);
    }
    TRACE_LOG("Done: srm2__srmPrepareToGet");
    return result;
}
/*
 * srm2__srmStatusOfGetRequest
 */
int srm2__srmStatusOfGetRequest(struct soap* soap, srm2__srmStatusOfGetRequestRequest *srmStatusOfGetRequestRequest, struct srm2__srmStatusOfGetRequestResponse_ &out){
    TRACE_LOG("srm2__srmStatusOfGetRequest called");
    int result = SOAP_OK;

    // Precondition Checks
    if(0 == srmStatusOfGetRequestRequest){
        // Error
        return soap_sender_fault(soap,"Null request sent",0);
    }
    if(true == srmStatusOfGetRequestRequest->requestToken.empty()){
        // Error
        return soap_sender_fault(soap,"Empty token sent",0);
    }
    TRACE_LOG("Get Status of Get Request " << srmStatusOfGetRequestRequest->requestToken);

    if (srm2_handle<SrmStatusOfGetTraits>(srmStatusOfGetRequestRequest, out, soap))
     	return result;

    // Create Response
    out.srmStatusOfGetRequestResponse = soap_new_srm2__srmStatusOfGetRequestResponse(soap,-1);
    out.srmStatusOfGetRequestResponse->returnStatus = soap_new_srm2__TReturnStatus(soap,-1);
    out.srmStatusOfGetRequestResponse->returnStatus->statusCode = srm2__TStatusCode__SRM_USCOREFAILURE;
    out.srmStatusOfGetRequestResponse->returnStatus->explanation = 0;
    out.srmStatusOfGetRequestResponse->remainingTotalRequestTime = 0;
    out.srmStatusOfGetRequestResponse->arrayOfFileStatuses       = 0;
    // Find Request
    Request::Ptr req = get_request(srmStatusOfGetRequestRequest->requestToken);
    if(0 == req.get()){
        // No such Request
        out.srmStatusOfGetRequestResponse->returnStatus->statusCode = srm2__TStatusCode__SRM_USCOREINVALID_USCOREREQUEST;
        out.srmStatusOfGetRequestResponse->returnStatus->explanation = soap_new_std__string(soap,-1);
        *(out.srmStatusOfGetRequestResponse->returnStatus->explanation) = "Invalid request token";
    } else if(req->type != Request::T_GET) {
        // Invalid request Type
        out.srmStatusOfGetRequestResponse->returnStatus->statusCode = srm2__TStatusCode__SRM_USCOREINVALID_USCOREREQUEST;
        out.srmStatusOfGetRequestResponse->returnStatus->explanation = soap_new_std__string(soap,-1);
        *(out.srmStatusOfGetRequestResponse->returnStatus->explanation) = "Not a Get Request";
    } else {
        GetRequest& get_req = static_cast<GetRequest&>(*(req.get()));
        // Process Get Request
        process_request(get_req);
        // Set Request Status
        out.srmStatusOfGetRequestResponse->returnStatus->statusCode = get_req.status;
        if(false == get_req.explanation.empty()){
            out.srmStatusOfGetRequestResponse->returnStatus->explanation = soap_new_std__string(soap,-1);
            *(out.srmStatusOfGetRequestResponse->returnStatus->explanation) = get_req.explanation;
        }
        // Set File Status Array
        out.srmStatusOfGetRequestResponse->arrayOfFileStatuses = status_of_get_files(soap,get_req);
    }

    TRACE_LOG("Done: srm2__srmStatusOfGetRequest");
    return result;
}

//..............................................................................
// BringOnline Methods
//..............................................................................

/*
 * srm2__srmBringOnline
 */
int srm2__srmBringOnline(struct soap* soap, srm2__srmBringOnlineRequest *srmBringOnlineRequest, struct srm2__srmBringOnlineResponse_ &out){
    TRACE_LOG("srm2__srmBringOnline called");
    int result = SOAP_OK;

    // Precondition Checks
    if((0 == srmBringOnlineRequest) ||
       (0 == srmBringOnlineRequest->arrayOfFileRequests) ||
       (true == srmBringOnlineRequest->arrayOfFileRequests->requestArray.empty())){
        // Error
        return soap_sender_fault(soap,"Null request sent",0);
    }

    if (srm2_handle<SrmBringOnlineTraits>(srmBringOnlineRequest, out, soap))
      	return result;

    const std::vector<srm2__TGetFileRequest * > & requestArray = srmBringOnlineRequest->arrayOfFileRequests->requestArray;
    TRACE_LOG("Request to BringOnline " << (int)requestArray.size() << " Files");
    // Protocol
    if((0 != srmBringOnlineRequest->transferParameters) &&
       (0 != srmBringOnlineRequest->transferParameters->arrayOfTransferProtocols) &&
       (false == srmBringOnlineRequest->transferParameters->arrayOfTransferProtocols->stringArray.empty())){
        TRACE_LOG("\tProtocol: " << srmBringOnlineRequest->transferParameters->arrayOfTransferProtocols->stringArray.at(0));
    }
    // For Each File
    std::vector<srm2__TGetFileRequest * >::const_iterator it;
    for(it = requestArray.begin(); it != requestArray.end(); ++it){
        const srm2__TGetFileRequest *  file = (*it);
        if(0 != file){
            TRACE_LOG("\tFile  " << file->sourceSURL);
        } else {
            TRACE_LOG("\tNull file");
        }
    }

    // Create Response
    out.srmBringOnlineResponse = soap_new_srm2__srmBringOnlineResponse(soap,-1);
    out.srmBringOnlineResponse->returnStatus = soap_new_srm2__TReturnStatus(soap,-1);
    out.srmBringOnlineResponse->returnStatus->statusCode = srm2__TStatusCode__SRM_USCOREFAILURE;
    out.srmBringOnlineResponse->returnStatus->explanation = 0;
    out.srmBringOnlineResponse->remainingTotalRequestTime = 0;
    out.srmBringOnlineResponse->remainingDeferredStartTime= 0;
    out.srmBringOnlineResponse->arrayOfFileStatuses       = 0;

    // Create Request Object
    BringOnlineRequest::Ptr req(new BringOnlineRequest());
    req->status = srm2__TStatusCode__SRM_USCOREREQUEST_USCOREQUEUED;
    // Get Protocol
    std::string protocol = PROTOCOL_FILE;
    if((0 != srmBringOnlineRequest->transferParameters) &&
       (0 != srmBringOnlineRequest->transferParameters->arrayOfTransferProtocols) &&
       (false == srmBringOnlineRequest->transferParameters->arrayOfTransferProtocols->stringArray.empty())){
        protocol = srmBringOnlineRequest->transferParameters->arrayOfTransferProtocols->stringArray.at(0);
    }
    req->protocol = protocol;

    // For each file
    for(it = requestArray.begin(); it != requestArray.end(); ++it){
        const srm2__TGetFileRequest *  file = (*it);
        BringOnlineFileRequest::Ptr f(new BringOnlineFileRequest());
        if(0 != file){
            if(false == file->sourceSURL.empty()) {
                f->surl = file->sourceSURL;
                f->status = srm2__TStatusCode__SRM_USCOREREQUEST_USCOREQUEUED;
            } else {
                f->status = srm2__TStatusCode__SRM_USCOREFAILURE;
                f->explanation = "Empty Surl";
            }
        } else {
            f->status = srm2__TStatusCode__SRM_USCOREFAILURE;
            f->explanation = "Null file";
        }
        // Store file
        req->fileRequests.push_back(f);
    }

    // Insert Request
    if(0 == store_request(req)){
        // Set Request Token
        out.srmBringOnlineResponse->requestToken = soap_new_std__string(soap,-1);
        *(out.srmBringOnlineResponse->requestToken) = req->id;
        // Set Request Status
        out.srmBringOnlineResponse->returnStatus->statusCode = req->status;
        if(false == req->explanation.empty()){
            out.srmBringOnlineResponse->returnStatus->explanation = soap_new_std__string(soap,-1);
            *(out.srmBringOnlineResponse->returnStatus->explanation) = req->explanation;
        }
        // Set File Status Array
        out.srmBringOnlineResponse->arrayOfFileStatuses = status_of_bringonline_files(soap,*req);
    } else {
        // Set error Message
        result = soap_receiver_fault(soap,"Failed to store request",0);
    }
    TRACE_LOG("Done: srm2__srmBringOnline");
    return result;
}
/*
 * srm2__srmStatusOfBringOnlineRequest
 */
int srm2__srmStatusOfBringOnlineRequest(struct soap* soap, srm2__srmStatusOfBringOnlineRequestRequest *srmStatusOfBringOnlineRequestRequest, struct srm2__srmStatusOfBringOnlineRequestResponse_ &out){
    TRACE_LOG("srm2__srmStatusOfBringOnlineRequest called");
    int result = SOAP_OK;

    // Precondition Checks
    if(0 == srmStatusOfBringOnlineRequestRequest){
        // Error
        return soap_sender_fault(soap,"Null request sent",0);
    }
    if(true == srmStatusOfBringOnlineRequestRequest->requestToken.empty()){
        // Error
        return soap_sender_fault(soap,"Empty token sent",0);
    }
    TRACE_LOG("Get Status of BringOnline Request " << srmStatusOfBringOnlineRequestRequest->requestToken);

    // Create Response
    out.srmStatusOfBringOnlineRequestResponse = soap_new_srm2__srmStatusOfBringOnlineRequestResponse(soap,-1);
    out.srmStatusOfBringOnlineRequestResponse->returnStatus = soap_new_srm2__TReturnStatus(soap,-1);
    out.srmStatusOfBringOnlineRequestResponse->returnStatus->statusCode = srm2__TStatusCode__SRM_USCOREFAILURE;
    out.srmStatusOfBringOnlineRequestResponse->returnStatus->explanation = 0;
    out.srmStatusOfBringOnlineRequestResponse->remainingTotalRequestTime = 0;
    out.srmStatusOfBringOnlineRequestResponse->remainingDeferredStartTime= 0;
    out.srmStatusOfBringOnlineRequestResponse->arrayOfFileStatuses       = 0;
    // Find Request
    Request::Ptr req = get_request(srmStatusOfBringOnlineRequestRequest->requestToken);
    if(0 == req.get()){
        // No such Request
        out.srmStatusOfBringOnlineRequestResponse->returnStatus->statusCode = srm2__TStatusCode__SRM_USCOREINVALID_USCOREREQUEST;
        out.srmStatusOfBringOnlineRequestResponse->returnStatus->explanation = soap_new_std__string(soap,-1);
        *(out.srmStatusOfBringOnlineRequestResponse->returnStatus->explanation) = "Invalid request token";
    } else if(req->type != Request::T_BRING_ONLINE) {
        // Invalid request Type
        out.srmStatusOfBringOnlineRequestResponse->returnStatus->statusCode = srm2__TStatusCode__SRM_USCOREINVALID_USCOREREQUEST;
        out.srmStatusOfBringOnlineRequestResponse->returnStatus->explanation = soap_new_std__string(soap,-1);
        *(out.srmStatusOfBringOnlineRequestResponse->returnStatus->explanation) = "Not a BringOnline Request";
    } else {
        BringOnlineRequest& bo_req = static_cast<BringOnlineRequest&>(*(req.get()));
        // Process BringOnline Request
        process_request(bo_req);
        // Set Request Status
        out.srmStatusOfBringOnlineRequestResponse->returnStatus->statusCode = bo_req.status;
        if(false == bo_req.explanation.empty()){
            out.srmStatusOfBringOnlineRequestResponse->returnStatus->explanation = soap_new_std__string(soap,-1);
            *(out.srmStatusOfBringOnlineRequestResponse->returnStatus->explanation) = bo_req.explanation;
        }
        // Set File Status Array
        out.srmStatusOfBringOnlineRequestResponse->arrayOfFileStatuses = status_of_bringonline_files(soap,bo_req);
    }

    TRACE_LOG("Done: srm2__srmStatusOfBringOnlineRequest");
    return result;
}

//..............................................................................
// Copy Methods
//..............................................................................

/*
 * srm2__srmCopy
 */
int srm2__srmCopy(struct soap* soap, srm2__srmCopyRequest *srmCopyRequest, struct srm2__srmCopyResponse_ &out){
    TRACE_LOG("srm2__srmCopy called");
    int result = SOAP_OK;

    // Precondition Checks
    if((0 == srmCopyRequest) ||
       (0 == srmCopyRequest->arrayOfFileRequests) ||
       (true == srmCopyRequest->arrayOfFileRequests->requestArray.empty())){
        // Error
        return soap_sender_fault(soap,"Null request sent",0);
    }

    if (srm2_handle<SrmCopyTraits>(srmCopyRequest, out, soap))
      	return result;

    const std::vector<srm2__TCopyFileRequest * > & requestArray = srmCopyRequest->arrayOfFileRequests->requestArray;
    TRACE_LOG("Request to Copy " << (int)requestArray.size() << " Files");
    // Overwrite Mode
    if(srmCopyRequest->overwriteOption){
        switch(*(srmCopyRequest->overwriteOption)){
            case srm2__TOverwriteMode__NEVER :
                TRACE_LOG("\tOverwrite: Never");
            break;
            case srm2__TOverwriteMode__ALWAYS :
                TRACE_LOG("\tOverwrite: Always");
            break;
            case srm2__TOverwriteMode__WHEN_USCOREFILES_USCOREARE_USCOREDIFFERENT :
                TRACE_LOG("\tOverwrite: When Different");
            break;
        }
    }
    // Token
    if(0 != srmCopyRequest->targetSpaceToken){
        TRACE_LOG("\tToken: " << srmCopyRequest->targetSpaceToken);
    }
    // For Each File
    std::vector<srm2__TCopyFileRequest * >::const_iterator it;
    for(it = requestArray.begin(); it != requestArray.end(); ++it){
        const srm2__TCopyFileRequest *  file = (*it);
        if(0 != file){
            TRACE_LOG("\tSource " << file->sourceSURL);
            TRACE_LOG("\tTarget " << file->targetSURL);
        } else {
            TRACE_LOG("\tNull file");
        }
    }

    bool add_extra_info_details = false;
    std::stringstream extra_info_details;
    if (srmCopyRequest->sourceStorageSystemInfo){
        // check source space token
        TRACE_LOG("Reading sourceStorageSystemInfo extra info array");
        const std::vector<srm2__TExtraInfo * >& info = srmCopyRequest->sourceStorageSystemInfo->extraInfoArray;
        std::vector<srm2__TExtraInfo * >::const_iterator it;
        for (it = info.begin();it != info.end();++it){
            srm2__TExtraInfo * extra_info = (*it);
            const std::string& key = extra_info->key;
            if (key == "SourceSpaceToken"){
                TRACE_LOG("Reading parameter " << key);
                add_extra_info_details = true;
                std::string value = extra_info->value ? *(extra_info->value) : "<NULL>";
                extra_info_details << " [" << key << "=" << value << "]";
            } else {
                TRACE_LOG("Ignoring parameter " << key);
            }
        }
    }
    if (srmCopyRequest->targetStorageSystemInfo){
        // check copypinlifetime
        TRACE_LOG("Reading targetStorageSystemInfo extra info array");
        const std::vector<srm2__TExtraInfo * >& info = srmCopyRequest->targetStorageSystemInfo->extraInfoArray;
        std::vector<srm2__TExtraInfo * >::const_iterator it;
        for (it = info.begin();it != info.end();++it){
            srm2__TExtraInfo * extra_info = (*it);
            const std::string& key = extra_info->key;
            if (key == "CopyPinLifetime"){
                TRACE_LOG("Reading parameter " << key);
                add_extra_info_details = true;
                std::string value = extra_info->value ? *(extra_info->value) : "<NULL>";
                extra_info_details << " [" << key << "=" << value << "]";
            } else {
                TRACE_LOG("Ignoring parameter " << key);
            }
        }
    }

    // Create Response
    out.srmCopyResponse = soap_new_srm2__srmCopyResponse(soap,-1);
    out.srmCopyResponse->returnStatus = soap_new_srm2__TReturnStatus(soap,-1);
    out.srmCopyResponse->returnStatus->statusCode = srm2__TStatusCode__SRM_USCOREFAILURE;
    out.srmCopyResponse->returnStatus->explanation = 0;
    out.srmCopyResponse->remainingTotalRequestTime = 0;
    out.srmCopyResponse->arrayOfFileStatuses       = 0;

    // Create Request Object
    CopyRequest::Ptr req(new CopyRequest());
    req->status = srm2__TStatusCode__SRM_USCOREREQUEST_USCOREQUEUED;
    // Overwrite Mode
    if(srmCopyRequest->overwriteOption){
        switch(*(srmCopyRequest->overwriteOption)){
            case srm2__TOverwriteMode__NEVER :
                req->overwrite = false;
            break;
            case srm2__TOverwriteMode__ALWAYS :
            case srm2__TOverwriteMode__WHEN_USCOREFILES_USCOREARE_USCOREDIFFERENT :
                req->overwrite = true;
            break;
        }
    }
    // For each file
    for(it = requestArray.begin(); it != requestArray.end(); ++it){
        const srm2__TCopyFileRequest *  file = (*it);
        CopyFileRequest::Ptr f(new CopyFileRequest());
        if(0 != file){
            f->sourceSurl = file->sourceSURL;
            f->targetSurl = file->targetSURL;
            f->status = srm2__TStatusCode__SRM_USCOREREQUEST_USCOREQUEUED;
        } else {
            f->status = srm2__TStatusCode__SRM_USCOREFAILURE;
            f->explanation = "Null file";
        }
        // Store file
        req->fileRequests.push_back(f);
    }

    // Insert Request
    if(0 == store_request(req)){
        // Set Request Token
        out.srmCopyResponse->requestToken = soap_new_std__string(soap,-1);
        *(out.srmCopyResponse->requestToken) = req->id;
        // Set Request Status
        out.srmCopyResponse->returnStatus->statusCode = req->status;
        if(false == req->explanation.empty()){
            out.srmCopyResponse->returnStatus->explanation = soap_new_std__string(soap,-1);
            *(out.srmCopyResponse->returnStatus->explanation) = req->explanation;
        }
        // Set File Status Array
        out.srmCopyResponse->arrayOfFileStatuses = status_of_copy_files(soap,*req);
    } else {
        // Set error Message
        result = soap_receiver_fault(soap,"Failed to store request",0);
    }

    if (add_extra_info_details){
        if (0 == out.srmCopyResponse->returnStatus->explanation){
            out.srmCopyResponse->returnStatus->explanation = soap_new_std__string(soap,-1);
        }
        std::stringstream tmp;
        tmp << *(out.srmCopyResponse->returnStatus->explanation);
        tmp << extra_info_details.str();
        *(out.srmCopyResponse->returnStatus->explanation) = tmp.str();
    }

    TRACE_LOG("Done: srm2__srmCopy");
    return result;
}
/*
 * srm2__srmStatusOfCopyRequest
 */
int srm2__srmStatusOfCopyRequest(struct soap* soap, srm2__srmStatusOfCopyRequestRequest *srmStatusOfCopyRequestRequest, struct srm2__srmStatusOfCopyRequestResponse_ &out){
    TRACE_LOG("srm2__srmStatusOfCopyRequest called");
    int result = SOAP_OK;

    // Precondition Checks
    if(0 == srmStatusOfCopyRequestRequest){
        // Error
        return soap_sender_fault(soap,"Null request sent",0);
    }
    if(true == srmStatusOfCopyRequestRequest->requestToken.empty()){
        // Error
        return soap_sender_fault(soap,"Empty token sent",0);
    }
    TRACE_LOG("Get Status of Copy Request " << srmStatusOfCopyRequestRequest->requestToken);

    if (srm2_handle<SrmStatusOfCopyTraits>(srmStatusOfCopyRequestRequest, out, soap))
    	return result;

    // Create Response
    out.srmStatusOfCopyRequestResponse = soap_new_srm2__srmStatusOfCopyRequestResponse(soap,-1);
    out.srmStatusOfCopyRequestResponse->returnStatus = soap_new_srm2__TReturnStatus(soap,-1);
    out.srmStatusOfCopyRequestResponse->returnStatus->statusCode = srm2__TStatusCode__SRM_USCOREFAILURE;
    out.srmStatusOfCopyRequestResponse->returnStatus->explanation = 0;
    out.srmStatusOfCopyRequestResponse->remainingTotalRequestTime = 0;
    out.srmStatusOfCopyRequestResponse->arrayOfFileStatuses       = 0;
    // Find Request
    Request::Ptr req = get_request(srmStatusOfCopyRequestRequest->requestToken);
    if(0 == req.get()){
        // No such Request
        out.srmStatusOfCopyRequestResponse->returnStatus->statusCode = srm2__TStatusCode__SRM_USCOREINVALID_USCOREREQUEST;
        out.srmStatusOfCopyRequestResponse->returnStatus->explanation = soap_new_std__string(soap,-1);
        *(out.srmStatusOfCopyRequestResponse->returnStatus->explanation) = "Invalid request token";
    } else if(req->type != Request::T_COPY) {
        // Invalid request Type
        out.srmStatusOfCopyRequestResponse->returnStatus->statusCode = srm2__TStatusCode__SRM_USCOREINVALID_USCOREREQUEST;
        out.srmStatusOfCopyRequestResponse->returnStatus->explanation = soap_new_std__string(soap,-1);
        *(out.srmStatusOfCopyRequestResponse->returnStatus->explanation) = "Not a Copy Request";
    } else {
        CopyRequest& co_req = static_cast<CopyRequest&>(*(req.get()));
        // Process Copy Request
        process_request(co_req);
        // Set Request Status
        out.srmStatusOfCopyRequestResponse->returnStatus->statusCode = co_req.status;
        if(false == co_req.explanation.empty()){
            out.srmStatusOfCopyRequestResponse->returnStatus->explanation = soap_new_std__string(soap,-1);
            *(out.srmStatusOfCopyRequestResponse->returnStatus->explanation) = co_req.explanation;
        }
        // Set File Status Array
        out.srmStatusOfCopyRequestResponse->arrayOfFileStatuses = status_of_copy_files(soap,co_req);
    }

    TRACE_LOG("Done: srm2__srmStatusOfCopyRequest");
    return result;
}

//..............................................................................
// Release & Abort Methods
//..............................................................................

/*
 * srm2__srmReleaseFiles
 */
int srm2__srmReleaseFiles(struct soap* soap, srm2__srmReleaseFilesRequest *srmReleaseFilesRequest, struct srm2__srmReleaseFilesResponse_ &out){
    TRACE_LOG("srm2__srmReleaseFiles called");
    int result = SOAP_OK;

    // Precondition Checks
    if(0 == srmReleaseFilesRequest){
        // Error
        return soap_sender_fault(soap,"Null request sent",0);
    }

    if (srm2_handle<SrmReleaseFilesTraits>(srmReleaseFilesRequest, out, soap))
    	return result;

    // Create Response
    out.srmReleaseFilesResponse = soap_new_srm2__srmReleaseFilesResponse(soap,-1);
    out.srmReleaseFilesResponse->returnStatus = soap_new_srm2__TReturnStatus(soap,-1);
    out.srmReleaseFilesResponse->returnStatus->statusCode = srm2__TStatusCode__SRM_USCOREFAILURE;
    out.srmReleaseFilesResponse->returnStatus->explanation = 0;
    out.srmReleaseFilesResponse->arrayOfFileStatuses       = 0;

    if((0 == srmReleaseFilesRequest->requestToken) || (true == srmReleaseFilesRequest->requestToken->empty())){
        // Not Suported
        out.srmReleaseFilesResponse->returnStatus->statusCode = srm2__TStatusCode__SRM_USCORENOT_USCORESUPPORTED;
        out.srmReleaseFilesResponse->returnStatus->explanation = soap_new_std__string(soap,-1);
        *(out.srmReleaseFilesResponse->returnStatus->explanation) = "Cannot release without a token: Not Supported";
        return result;
    }
    TRACE_LOG("Release Request " << srmReleaseFilesRequest->requestToken);

    // Find Request
    Request::Ptr req = get_request(*(srmReleaseFilesRequest->requestToken));
    if(0 == req.get()){
        // No such Request
        out.srmReleaseFilesResponse->returnStatus->statusCode = srm2__TStatusCode__SRM_USCOREINVALID_USCOREREQUEST;
        out.srmReleaseFilesResponse->returnStatus->explanation = soap_new_std__string(soap,-1);
        *(out.srmReleaseFilesResponse->returnStatus->explanation) = "Invalid request token";
    } else if(req->type == Request::T_GET) {
        // Check if the request is the correct type
        GetRequest& get_req = static_cast<GetRequest&>(*(req.get()));
        // Complete Get Request
        complete_request(soap, get_req,*(out.srmReleaseFilesResponse));
    } else if(req->type == Request::T_BRING_ONLINE) {
        // Check if the request is the correct type
        BringOnlineRequest& bo_req = static_cast<BringOnlineRequest&>(*(req.get()));
        // Complete BringOnline Request
        complete_request(soap, bo_req,*(out.srmReleaseFilesResponse));
    } else if(req->type == Request::T_COPY) {
        // Copy not supported
        out.srmReleaseFilesResponse->returnStatus->statusCode = srm2__TStatusCode__SRM_USCOREINVALID_USCOREREQUEST;
        out.srmReleaseFilesResponse->returnStatus->explanation = soap_new_std__string(soap,-1);
        *(out.srmReleaseFilesResponse->returnStatus->explanation) = "Cannot release files of Copy request";
    } else {
        // Invalid request Type
        out.srmReleaseFilesResponse->returnStatus->statusCode = srm2__TStatusCode__SRM_USCOREINVALID_USCOREREQUEST;
        out.srmReleaseFilesResponse->returnStatus->explanation = soap_new_std__string(soap,-1);
        *(out.srmReleaseFilesResponse->returnStatus->explanation) = "Invalid request type";
    }
    TRACE_LOG("Done: srm2__srmReleaseFiles");
    return result;
}


/*
 * srm2__srmPutDone
 */
int srm2__srmPutDone(struct soap* soap, srm2__srmPutDoneRequest *srmPutDoneRequest, struct srm2__srmPutDoneResponse_ &out){
    TRACE_LOG("srm2__srmStatusOfPutRequest called");
    int result = SOAP_OK;

    // Precondition Checks
    if(0 == srmPutDoneRequest){
        // Error
        return soap_sender_fault(soap,"Null request sent",0);
    }
    if(true == srmPutDoneRequest->requestToken.empty()){
        // Error
        return soap_sender_fault(soap,"Empty token sent",0);
    }
    TRACE_LOG("Complete Put Request " << srmPutDoneRequest->requestToken);

    // Create Response
    out.srmPutDoneResponse = soap_new_srm2__srmPutDoneResponse(soap,-1);
    out.srmPutDoneResponse->returnStatus = soap_new_srm2__TReturnStatus(soap,-1);
    out.srmPutDoneResponse->returnStatus->statusCode = srm2__TStatusCode__SRM_USCOREFAILURE;
    out.srmPutDoneResponse->returnStatus->explanation = 0;
    out.srmPutDoneResponse->arrayOfFileStatuses       = 0;
    // Find Request
    Request::Ptr req = get_request(srmPutDoneRequest->requestToken);
    if(0 == req.get()){
        // No such Request
        out.srmPutDoneResponse->returnStatus->statusCode = srm2__TStatusCode__SRM_USCOREINVALID_USCOREREQUEST;
        out.srmPutDoneResponse->returnStatus->explanation = soap_new_std__string(soap,-1);
        *(out.srmPutDoneResponse->returnStatus->explanation) = "Invalid request token";
    } else if(req->type != Request::T_PUT) {
        // Invalid request Type
        out.srmPutDoneResponse->returnStatus->statusCode = srm2__TStatusCode__SRM_USCOREINVALID_USCOREREQUEST;
        out.srmPutDoneResponse->returnStatus->explanation = soap_new_std__string(soap,-1);
        *(out.srmPutDoneResponse->returnStatus->explanation) = "Not a Put Request";
    } else {
        // Check if the request is the correct type
        PutRequest& put_req = static_cast<PutRequest&>(*(req.get()));
        // Complete Put Request
        complete_request(soap, put_req,*(out.srmPutDoneResponse));
    }
    TRACE_LOG("Done: srm2__srmPutDone");
    return result;
}
/*
 * srm2__srmAbortRequest
 */
int srm2__srmAbortRequest(struct soap* soap, srm2__srmAbortRequestRequest *srmAbortRequestRequest, struct srm2__srmAbortRequestResponse_ &out){
    TRACE_LOG("srm2__srmAbortRequest called");
    int result = SOAP_OK;

    // Precondition Checks
    if(0 == srmAbortRequestRequest){
        // Error
        return soap_sender_fault(soap,"Null request sent",0);
    }
    if(true == srmAbortRequestRequest->requestToken.empty()){
        // Error
        return soap_sender_fault(soap,"Empty token sent",0);
    }
    TRACE_LOG("Abort Request " << srmAbortRequestRequest->requestToken);

    if (srm2_handle<SrmAbortRequestTraits>(srmAbortRequestRequest, out, soap))
      	return result;

    // Create Response
    out.srmAbortRequestResponse = soap_new_srm2__srmAbortRequestResponse(soap,-1);
    out.srmAbortRequestResponse->returnStatus = soap_new_srm2__TReturnStatus(soap,-1);
    out.srmAbortRequestResponse->returnStatus->statusCode = srm2__TStatusCode__SRM_USCOREFAILURE;
    out.srmAbortRequestResponse->returnStatus->explanation = 0;
    // Find Request
    Request::Ptr req = get_request(srmAbortRequestRequest->requestToken);
    if(0 == req.get()){
        // No such Request
        out.srmAbortRequestResponse->returnStatus->statusCode = srm2__TStatusCode__SRM_USCOREINVALID_USCOREREQUEST;
        out.srmAbortRequestResponse->returnStatus->explanation = soap_new_std__string(soap,-1);
        *(out.srmAbortRequestResponse->returnStatus->explanation) = "Invalid request token";
    } else if(req->type == Request::T_PUT) {
        PutRequest& put_req = static_cast<PutRequest&>(*(req.get()));
        // Abort Put Request
        abort_request(soap, put_req,*(out.srmAbortRequestResponse));
    } else if(req->type == Request::T_GET) {
        GetRequest& get_req = static_cast<GetRequest&>(*(req.get()));
        // Abort Get Request
        abort_request(soap, get_req,*(out.srmAbortRequestResponse));
    } else if(req->type == Request::T_COPY) {
        CopyRequest& co_req = static_cast<CopyRequest&>(*(req.get()));
        // Abort Copy Request
        abort_request(soap, co_req,*(out.srmAbortRequestResponse));
    } else if(req->type == Request::T_BRING_ONLINE) {
        BringOnlineRequest& bo_req = static_cast<BringOnlineRequest&>(*(req.get()));
        // Abort BringOnline Request
        abort_request(soap, bo_req,*(out.srmAbortRequestResponse));
    } else {
        // Invalid request Type
        out.srmAbortRequestResponse->returnStatus->statusCode = srm2__TStatusCode__SRM_USCOREINVALID_USCOREREQUEST;
        out.srmAbortRequestResponse->returnStatus->explanation = soap_new_std__string(soap,-1);
        *(out.srmAbortRequestResponse->returnStatus->explanation) = "Invalid Request Type";
    }
    TRACE_LOG("Done: srm2__srmAbortRequest");
    return result;
}

/*
 * srm2__srmAbortFiles
 */
int srm2__srmAbortFiles(struct soap* soap, srm2__srmAbortFilesRequest *srmAbortFilesRequest, struct srm2__srmAbortFilesResponse_ &out){
    TRACE_LOG("srm2__srmAbortFiles called");
    int result = SOAP_OK;

    // Precondition Checks
    if(0 == srmAbortFilesRequest){
        // Error
        return soap_sender_fault(soap,"Null request sent",0);
    }
    if(true == srmAbortFilesRequest->requestToken.empty()){
        // Error
        return soap_sender_fault(soap,"Empty token sent",0);
    }
    if((0 == srmAbortFilesRequest->arrayOfSURLs) ||
       (true == srmAbortFilesRequest->arrayOfSURLs->urlArray.empty())){
        // Error
        return soap_sender_fault(soap,"Empty surls list sent",0);
    }
    TRACE_LOG("Abort Request " << srmAbortFilesRequest->requestToken);

    if (srm2_handle<SrmAbortFilesTraits>(srmAbortFilesRequest, out, soap))
    	return result;

    // Create Response
    out.srmAbortFilesResponse = soap_new_srm2__srmAbortFilesResponse(soap,-1);
    out.srmAbortFilesResponse->returnStatus = soap_new_srm2__TReturnStatus(soap,-1);
    out.srmAbortFilesResponse->returnStatus->statusCode = srm2__TStatusCode__SRM_USCOREFAILURE;
    out.srmAbortFilesResponse->returnStatus->explanation = 0;
    // Find Request
    Request::Ptr req = get_request(srmAbortFilesRequest->requestToken);
    if(0 == req.get()){
        // No such Request
        out.srmAbortFilesResponse->returnStatus->statusCode = srm2__TStatusCode__SRM_USCOREINVALID_USCOREREQUEST;
        out.srmAbortFilesResponse->returnStatus->explanation = soap_new_std__string(soap,-1);
        *(out.srmAbortFilesResponse->returnStatus->explanation) = "Invalid request token";
    } else {
        out.srmAbortFilesResponse->arrayOfFileStatuses       = soap_new_srm2__ArrayOfTSURLReturnStatus(soap,-1);
        const std::vector<std::string>& urlArray = srmAbortFilesRequest->arrayOfSURLs->urlArray;
        // For Each File
        std::vector<std::string>::const_iterator it;
        for(it = urlArray.begin();it != urlArray.end(); ++it){
            // Create File Status
            srm2__TSURLReturnStatus * file = soap_new_srm2__TSURLReturnStatus(soap,-1);
            // Set SURL
            file->surl = (*it);
            // Set Status
            file->status = soap_new_srm2__TReturnStatus(soap,-1);
            file->status->statusCode = srm2__TStatusCode__SRM_USCOREFAILURE;
            file->status->explanation = 0;
            out.srmAbortFilesResponse->arrayOfFileStatuses->statusArray.push_back(file);
        }
        if(req->type == Request::T_PUT) {
            PutRequest& put_req = static_cast<PutRequest&>(*(req.get()));
            // Abort Put Request
            abort_files(soap, urlArray, put_req,*(out.srmAbortFilesResponse));
        } else if(req->type == Request::T_GET) {
            GetRequest& get_req = static_cast<GetRequest&>(*(req.get()));
            // Abort Get Request
            abort_files(soap, urlArray, get_req,*(out.srmAbortFilesResponse));
        } else if(req->type == Request::T_COPY) {
            CopyRequest& co_req = static_cast<CopyRequest&>(*(req.get()));
            // Abort Copy Request
            abort_files(soap, urlArray, co_req,*(out.srmAbortFilesResponse));
        } else if(req->type == Request::T_BRING_ONLINE) {
            BringOnlineRequest& bo_req = static_cast<BringOnlineRequest&>(*(req.get()));
            // Abort BringOnline Request
            abort_files(soap, urlArray, bo_req,*(out.srmAbortFilesResponse));
        } else {
            // Invalid request Type
            out.srmAbortFilesResponse->arrayOfFileStatuses          = 0;
            out.srmAbortFilesResponse->returnStatus->statusCode     = srm2__TStatusCode__SRM_USCOREINVALID_USCOREREQUEST;
            out.srmAbortFilesResponse->returnStatus->explanation    = soap_new_std__string(soap,-1);
            *(out.srmAbortFilesResponse->returnStatus->explanation) = "Invalid Request Type";
        }
    }
    TRACE_LOG("Done: srm2__srmAbortFiles");
    return result;
}

//..............................................................................
// Namespace Methods
//..............................................................................

namespace {
/* The strategy behind the async Ls test is:
 * - File name should follow the this-one-returns-SRM_REQUEST_QUEUED-on-ls.dat pattern
 * - We store the URL array from the request, until a srm2__srmRequest call
 * - What that is called, the surl-s are processed, and the corresponding array is
 *   removed from the store
 */

/**
 * Represents an asynchronous LS request
 */
struct AsyncLsRequest
{
	std::vector<std::string> urlArray;
};

/**
 * Store the details of the asynch Ls requests, indexed by the request token
 */
typedef std::map<std::string, AsyncLsRequest> Pending_Asynch_LS_Requests_type;
static Pending_Asynch_LS_Requests_type Pending_Asynch_LS_Requests;

/**
 * Fill in the detail field of an Ls response. Used in synch/asynch Ls calls.
 *
 * @param details The delaild field od a SOAP response object
 * @param urlArray The arrayOfSURLs->urlArray field of an ls request
 * @param returnStatus returnStatus field of an Ls response object
 * @param soap The SOAP environment
 * @param checkMatch Check the file pattern (and set the status code corresponding
 *                   to the file) only if this parameter is true.
 */
void srm2__srmLs_GenerateResponse(srm2__ArrayOfTMetaDataPathDetail& details,
								 const std::vector<std::string>& urlArray,
								 srm2__TReturnStatus& returnStatus,
								 struct soap* soap,
								 bool checkMatch = true)
{
	bool has_success = false;
	bool has_failed = false;
	// For each File
	for(std::vector<std::string>::const_iterator it = urlArray.begin();
	    it != urlArray.end();
	    ++it)
	{
		const std::string& surl = (*it);
		srm2__TMetaDataPathDetail * d = 0;

		srm2__TStatusCode code = srm2__TStatusCode__SRM_USCOREFAILURE;
		if ( checkMatch && FileStatusRule::instance().is_match(surl,"ls",code) ){
			TRACE_LOG("Returning status [" << code << "] from name [" << surl << "].");
			d = soap_new_srm2__TMetaDataPathDetail(soap,-1);
			d->path                 = surl;
			d->status               = soap_new_srm2__TReturnStatus(soap,-1);
			d->status->statusCode   = code;
			d->status->explanation  = soap_new_std__string(soap,-1);
			*(d->status->explanation) = "Got srmLs status from surl.";
			d->size                 = 0;
			d->createdAtTime        = 0;
			d->lastModificationTime = 0;
			d->fileStorageType      = 0;
			d->retentionPolicyInfo  = 0;
			d->fileLocality         = 0;
			d->arrayOfSpaceTokens   = 0;
			d->type                 = 0;
			d->lifetimeAssigned     = 0;
			d->lifetimeLeft         = 0;
			d->ownerPermission      = 0;
			d->groupPermission      = 0;
			d->otherPermission      = 0;
			d->checkSumType         = 0;
			d->checkSumValue        = 0;
			d->arrayOfSubPaths      = 0;
		} else {
			d = get_file_stat(soap,(*it));
		}
		if((0 == d->status) || (d->status->statusCode == srm2__TStatusCode__SRM_USCORESUCCESS)){
			has_success = true;
		} else {
			has_failed  = true;
		}

		details.pathDetailArray.push_back(d);
	}

	// Set Result
	if(true == has_success){
		if(true == has_failed){
			returnStatus.statusCode = srm2__TStatusCode__SRM_USCOREPARTIAL_USCORESUCCESS;
			returnStatus.explanation = soap_new_std__string(soap,-1);
			*(returnStatus.explanation) = "Some files failed";
		} else {
			returnStatus.statusCode = srm2__TStatusCode__SRM_USCORESUCCESS;
		}
	} else {
		returnStatus.explanation = soap_new_std__string(soap,-1);
		*(returnStatus.explanation) = "All files failed";
	}
}

} // unnamed namespace

/*
 * srm2__srmLs
 */
int srm2__srmLs(struct soap* soap, srm2__srmLsRequest *srmLsRequest, struct srm2__srmLsResponse_ &out){
    TRACE_LOG("srm2__srmLs called");
    int result = SOAP_OK;

    // Precondition Checks
    if(0 == srmLsRequest){
        // Error
        return soap_sender_fault(soap,"Null request sent",0);
    }
    if((0 == srmLsRequest->arrayOfSURLs)|| (true == srmLsRequest->arrayOfSURLs->urlArray.empty())){
        // Error
        return soap_sender_fault(soap,"Empty SURL array sent",0);
    }
    TRACE_LOG("Requested Ls for " << (int)srmLsRequest->arrayOfSURLs->urlArray.size() << " files");
    std::vector<std::string>::const_iterator it;
    for(it = srmLsRequest->arrayOfSURLs->urlArray.begin(); it != srmLsRequest->arrayOfSURLs->urlArray.end();++it){
        TRACE_LOG("File : " << (*it));
    }

    if (srm2_handle<SrmLsTraits>(srmLsRequest, out, soap))
    	return result;

    // Create Response
    out.srmLsResponse = soap_new_srm2__srmLsResponse(soap,-1);
    out.srmLsResponse->returnStatus = soap_new_srm2__TReturnStatus(soap,-1);
    out.srmLsResponse->returnStatus->statusCode  = srm2__TStatusCode__SRM_USCOREFAILURE;
    out.srmLsResponse->returnStatus->explanation = 0;
    out.srmLsResponse->details      = 0;

	out.srmLsResponse->requestToken = soap_new_std__string(soap,-1);
	static int token_count = 0;
	std::string token = "ls-token" + boost::lexical_cast<std::string>(token_count);
	++token_count;
	*(out.srmLsResponse->requestToken) = token;

    // Check unsupported features
    if((0 != srmLsRequest->allLevelRecursive) && (true == *(srmLsRequest->allLevelRecursive))){
        out.srmLsResponse->returnStatus->statusCode = srm2__TStatusCode__SRM_USCORENOT_USCORESUPPORTED;
        out.srmLsResponse->returnStatus->explanation = soap_new_std__string(soap,-1);
        *(out.srmLsResponse->returnStatus->explanation) = "Recursive listing is not supported";
    } else if((0 != srmLsRequest->numOfLevels) && (0 != *(srmLsRequest->numOfLevels))){
        out.srmLsResponse->returnStatus->statusCode = srm2__TStatusCode__SRM_USCORENOT_USCORESUPPORTED;
        out.srmLsResponse->returnStatus->explanation = soap_new_std__string(soap,-1);
        *(out.srmLsResponse->returnStatus->explanation) = "Multiple levels are not supported";
    } else if((0 != srmLsRequest->offset) && (0 != *(srmLsRequest->offset))){
        out.srmLsResponse->returnStatus->statusCode = srm2__TStatusCode__SRM_USCORENOT_USCORESUPPORTED;
        out.srmLsResponse->returnStatus->explanation = soap_new_std__string(soap,-1);
        *(out.srmLsResponse->returnStatus->explanation) = "Offset is not supported";
    } else if((0 != srmLsRequest->count) && (1 != *(srmLsRequest->count))){
        out.srmLsResponse->returnStatus->statusCode = srm2__TStatusCode__SRM_USCORENOT_USCORESUPPORTED;
        out.srmLsResponse->returnStatus->explanation = soap_new_std__string(soap,-1);
        *(out.srmLsResponse->returnStatus->explanation) = "Count is not supported";
    } else {
    	srm2__TStatusCode code = srm2__TStatusCode__SRM_USCOREFAILURE;
    	const std::string surl = *(srmLsRequest->arrayOfSURLs->urlArray.begin());

    	if (FileStatusRule::instance().is_match(surl, "ls", code) &&
    		code == srm2__TStatusCode__SRM_USCOREREQUEST_USCOREQUEUED)
    	{
			TRACE_LOG("Async srmLs called");
			out.srmLsResponse->returnStatus->statusCode = code;
			AsyncLsRequest req;
			req.urlArray = srmLsRequest->arrayOfSURLs->urlArray;
			Pending_Asynch_LS_Requests[token] = req ;
		} else {
			out.srmLsResponse->details = soap_new_srm2__ArrayOfTMetaDataPathDetail(soap,-1);

			srm2__srmLs_GenerateResponse(
					*(out.srmLsResponse->details),
					srmLsRequest->arrayOfSURLs->urlArray,
					*(out.srmLsResponse->returnStatus),
					soap
			);
    	}
    }

    TRACE_LOG("Done: srm2__srmLs");
    return result;
}

int srm2__srmStatusOfLsRequest(struct soap* soap, srm2__srmStatusOfLsRequestRequest *srmStatusOfLsRequestRequest, struct srm2__srmStatusOfLsRequestResponse_ &out)
{
    TRACE_LOG("srm2__srmStatusOfLsRequest");
    int result = SOAP_OK;
    std::string token = srmStatusOfLsRequestRequest->requestToken;

    if (srm2_handle<SrmStatusOfLsTraits>(srmStatusOfLsRequestRequest, out, soap))
    	return result;

    out.srmStatusOfLsRequestResponse = soap_new_srm2__srmStatusOfLsRequestResponse(soap,-1);
    out.srmStatusOfLsRequestResponse->returnStatus = soap_new_srm2__TReturnStatus(soap,-1);
    out.srmStatusOfLsRequestResponse->returnStatus->explanation = 0;
	out.srmStatusOfLsRequestResponse->details = soap_new_srm2__ArrayOfTMetaDataPathDetail(soap,-1);

	Pending_Asynch_LS_Requests_type::iterator pos = Pending_Asynch_LS_Requests.find(token);

    if (pos == Pending_Asynch_LS_Requests.end()) {
    	TRACE_LOG("Token " << token << " cannot be found");
    	out.srmStatusOfLsRequestResponse->returnStatus->statusCode = srm2__TStatusCode__SRM_USCOREINVALID_USCOREREQUEST;
		out.srmStatusOfLsRequestResponse->returnStatus->explanation = soap_new_std__string(soap,-1);
		*(out.srmStatusOfLsRequestResponse->returnStatus->explanation) = "Token cannot be found";
    } else {
    	// Token was found. We remove it from the list to signal that the
    	// request has been processed
    	TRACE_LOG("Token " << token << " found, request handled");
    	out.srmStatusOfLsRequestResponse->returnStatus->statusCode = srm2__TStatusCode__SRM_USCORESUCCESS;

		srm2__srmLs_GenerateResponse(
				*(out.srmStatusOfLsRequestResponse->details),
				pos->second.urlArray,
				*(out.srmStatusOfLsRequestResponse->returnStatus),
				soap,
				false
		);

    	Pending_Asynch_LS_Requests.erase(pos);
    }

    TRACE_LOG("Done: srm2__srmStatusOfLsRequest");
    return result;
}

int srm2__srmRm(struct soap* soap, srm2__srmRmRequest *srmRmRequest, struct srm2__srmRmResponse_ &out){
    TRACE_LOG("srm2__srmRm called");
    int result = SOAP_OK;

    // Precondition Checks
    if(0 == srmRmRequest){
        // Error
        return soap_sender_fault(soap,"Null request sent",0);
    }
    if((0 == srmRmRequest->arrayOfSURLs)|| (true == srmRmRequest->arrayOfSURLs->urlArray.empty())){
        // Error
        return soap_sender_fault(soap,"Empty SURL array sent",0);
    }
    TRACE_LOG("Request to remove " << (int)srmRmRequest->arrayOfSURLs->urlArray.size() << " files");
    std::vector<std::string>::const_iterator it;
    for(it = srmRmRequest->arrayOfSURLs->urlArray.begin(); it != srmRmRequest->arrayOfSURLs->urlArray.end();++it){
        TRACE_LOG("File : " << (*it));
    }

    // Try fors find a test handler. If not found, use the default one.
    if (srm2_handle<SrmRmTraits>(srmRmRequest, out, soap))
    	return result;

    // the default ("old") call handler
    // Create Response
    out.srmRmResponse = soap_new_srm2__srmRmResponse(soap,-1);
    out.srmRmResponse->returnStatus = soap_new_srm2__TReturnStatus(soap,-1);
    out.srmRmResponse->returnStatus->statusCode  = srm2__TStatusCode__SRM_USCOREFAILURE;
    out.srmRmResponse->returnStatus->explanation = 0;
    out.srmRmResponse->arrayOfFileStatuses = soap_new_srm2__ArrayOfTSURLReturnStatus(soap,-1);
    bool has_success = false;
    bool has_failed = false;
    // For each File
    for(it = srmRmRequest->arrayOfSURLs->urlArray.begin(); it != srmRmRequest->arrayOfSURLs->urlArray.end();++it){
        // Create File status object
        srm2__TSURLReturnStatus * s = soap_new_srm2__TSURLReturnStatus(soap,-1);
        s->surl = (*it);
        s->status = soap_new_srm2__TReturnStatus(soap,-1);
        s->status->statusCode  = srm2__TStatusCode__SRM_USCOREFAILURE;
        s->status->explanation = 0;
        // Get file path
        std::string file_error;
        std::string file_path = path_from_surl((*it));
        if(file_path.empty()){
            s->status->statusCode  = srm2__TStatusCode__SRM_USCOREINVALID_USCOREPATH;
            file_error = "Invalid SURL passed";
        } else {
            // Remove the file
            int r = unlink(file_path.c_str());
            if(0 != r){
                s->status->statusCode  = srm2__TStatusCode__SRM_USCOREINVALID_USCOREPATH;
                file_error = strerror(errno);
            } else {
                s->status->statusCode  = srm2__TStatusCode__SRM_USCORESUCCESS;
            }
        }
        if((0 == s->status) || (s->status->statusCode == srm2__TStatusCode__SRM_USCORESUCCESS)){
            has_success = true;
        } else {
            has_failed  = true;
        }
        out.srmRmResponse->arrayOfFileStatuses->statusArray.push_back(s);
    }

    // Set Result
    if(true == has_success){
        if(true == has_failed){
            out.srmRmResponse->returnStatus->statusCode = srm2__TStatusCode__SRM_USCOREPARTIAL_USCORESUCCESS;
            out.srmRmResponse->returnStatus->explanation = soap_new_std__string(soap,-1);
            *(out.srmRmResponse->returnStatus->explanation) = "Some files failed";
        } else {
            out.srmRmResponse->returnStatus->statusCode = srm2__TStatusCode__SRM_USCORESUCCESS;
        }
    } else {
        out.srmRmResponse->returnStatus->explanation = soap_new_std__string(soap,-1);
        *(out.srmRmResponse->returnStatus->explanation) = "All files failed";
    }

    TRACE_LOG("Done: srm2__srmRm");
    return result;
}

/*
 * srm2__srmMkdir
 */
int srm2__srmMkdir(struct soap* soap, srm2__srmMkdirRequest *srmMkdirRequest, struct srm2__srmMkdirResponse_ &out){
    TRACE_LOG("srm2__srmMkdir called");
    int result = SOAP_OK;

    if (srm2_handle<SrmMkDirTraits>(srmMkdirRequest, out, soap))
    	return result;

    // Create Response
    out.srmMkdirResponse = soap_new_srm2__srmMkdirResponse(soap,-1);
    out.srmMkdirResponse->returnStatus = soap_new_srm2__TReturnStatus(soap,-1);
    out.srmMkdirResponse->returnStatus->statusCode = srm2__TStatusCode__SRM_USCOREFAILURE;
    out.srmMkdirResponse->returnStatus->explanation = 0;

    // Check il pool exist
    if(0 != srmMkdirRequest){
        const std::string& surl = srmMkdirRequest->SURL;
        std::string dir_path = path_from_surl(surl);
        if(false == dir_path.empty()){

            // Create Path
            int r = create_dir_path(dir_path,false);
            if(r != 0){
                out.srmMkdirResponse->returnStatus->explanation = soap_new_std__string(soap,-1);
                *(out.srmMkdirResponse->returnStatus->explanation) = strerror(r);
                TRACE_LOG("Failed to create directory " << dir_path << ": " << out.srmMkdirResponse->returnStatus->explanation);
                switch(r){
                    case EEXIST:
                        TRACE_LOG("Directory already exists");
                        out.srmMkdirResponse->returnStatus->statusCode = srm2__TStatusCode__SRM_USCOREDUPLICATION_USCOREERROR;
                    break;
                    case EPERM:
                        TRACE_LOG("Cannot Write");
                        out.srmMkdirResponse->returnStatus->statusCode = srm2__TStatusCode__SRM_USCOREAUTHORIZATION_USCOREFAILURE;
                    break;
                    case ENOENT:
                    case ENOTDIR:
                    case EACCES:
                    case EFAULT:
                        TRACE_LOG("Invalid Path");
                        out.srmMkdirResponse->returnStatus->statusCode = srm2__TStatusCode__SRM_USCOREINVALID_USCOREPATH;
                    break;
                    default:
                        TRACE_LOG("Generic Failure");
                        out.srmMkdirResponse->returnStatus->statusCode = srm2__TStatusCode__SRM_USCOREFAILURE;
                    break;
                }
            } else {
                TRACE_LOG("Directory Created");

                srm2__TStatusCode code = srm2__TStatusCode__SRM_USCOREFAILURE;
                if ( FileStatusRule::instance().is_match(dir_path,"mkdir",code) ){
                    TRACE_LOG("Returning status from directory name.");
                    out.srmMkdirResponse->returnStatus->statusCode = code;
                    return result;
                }

                out.srmMkdirResponse->returnStatus->statusCode = srm2__TStatusCode__SRM_USCORESUCCESS;
            }
        } else {
            // Error, Invalid Path
            TRACE_LOG("Empty file path");
            result = soap_sender_fault(soap,"User sent an invalid SURL",0);
        }
    } else {
        // Invalid Request
        result = soap_sender_fault(soap,"User sent an null request object",0);
    }
    TRACE_LOG("Done: srm2__srmMkdir");
    return result;
}

/*
 * srm2__srmRmdir
 */
int srm2__srmRmdir(struct soap* soap, srm2__srmRmdirRequest *srmRmdirRequest, struct srm2__srmRmdirResponse_ &out){
    TRACE_LOG("srm2__srmRmdir called");
    int result = SOAP_OK;

    // Preconditions
    if(0 == srmRmdirRequest){
        // Invalid Request
        return soap_sender_fault(soap,"User sent an null request object",0);
    }
    if(true == srmRmdirRequest->SURL.empty()){
        // Invalid Request
        return soap_sender_fault(soap,"User sent an empty SURL",0);
    }

    if (srm2_handle<SrmRmDirTraits>(srmRmdirRequest, out, soap))
    	return result;

    TRACE_LOG("Request to remove dircetory " << srmRmdirRequest->SURL);

    // Create Response
    out.srmRmdirResponse = soap_new_srm2__srmRmdirResponse(soap,-1);
    out.srmRmdirResponse->returnStatus = soap_new_srm2__TReturnStatus(soap,-1);
    out.srmRmdirResponse->returnStatus->statusCode = srm2__TStatusCode__SRM_USCOREFAILURE;
    out.srmRmdirResponse->returnStatus->explanation = 0;

    bool recursive = false;
    if(0 != srmRmdirRequest->recursive){
        recursive = *(srmRmdirRequest->recursive);
    }
    // Get directory path
    std::string dir_path = path_from_surl(srmRmdirRequest->SURL);
    if(true == dir_path.empty()){
        // Error, Empty dir path
        TRACE_LOG("Empty directory path");
        result = soap_sender_fault(soap,"User sent an invalid SURL",0);
    } else {
        // Create Path
        int r = rm_dir_path(dir_path,recursive);
        if(r != 0){
            out.srmRmdirResponse->returnStatus->explanation = soap_new_std__string(soap,-1);
            *(out.srmRmdirResponse->returnStatus->explanation) = strerror(r);
            TRACE_LOG("Failed to delete directory " << dir_path << ": " << out.srmRmdirResponse->returnStatus->explanation);
            switch(r){
                case EEXIST:
                    TRACE_LOG("Directory does not exist");
                    out.srmRmdirResponse->returnStatus->statusCode = srm2__TStatusCode__SRM_USCOREINVALID_USCOREPATH;
                break;
                case ENOTEMPTY:
                    TRACE_LOG("Not empty");
                    out.srmRmdirResponse->returnStatus->statusCode = srm2__TStatusCode__SRM_USCORENON_USCOREEMPTY_USCOREDIRECTORY;
                break;
                case EPERM:
                case EROFS:
                    TRACE_LOG("Cannot Write");
                    out.srmRmdirResponse->returnStatus->statusCode = srm2__TStatusCode__SRM_USCOREAUTHORIZATION_USCOREFAILURE;
                break;
                case EINVAL:
                case ENOENT:
                case ENOTDIR:
                case EACCES:
                case EFAULT:
                    TRACE_LOG("Invalid Path");
                    out.srmRmdirResponse->returnStatus->statusCode = srm2__TStatusCode__SRM_USCOREINVALID_USCOREPATH;
                break;
                default:
                    TRACE_LOG("Generic Failure");
                    out.srmRmdirResponse->returnStatus->statusCode = srm2__TStatusCode__SRM_USCOREFAILURE;
                break;
            }
        } else {
            TRACE_LOG("Directory Deleted");
            out.srmRmdirResponse->returnStatus->statusCode = srm2__TStatusCode__SRM_USCORESUCCESS;
        }
    }
    TRACE_LOG("Done: srm2__srmRmdir");
    return result;
}

/*
 * srm2__srmGetSpaceTokens
 */
int srm2__srmGetSpaceTokens(struct soap* soap, srm2__srmGetSpaceTokensRequest *srmGetSpaceTokensRequest, struct srm2__srmGetSpaceTokensResponse_ &out){
    TRACE_LOG("srm2__srmGetSpaceTokens called");
    int result = SOAP_OK;

    // Preconditions
    if(0 == srmGetSpaceTokensRequest){
        // Invalid Request
        return soap_sender_fault(soap,"User sent an null request object",0);
    }

    if (srm2_handle<SrmGetSpaceTokensTraits>(srmGetSpaceTokensRequest, out, soap))
      	return result;

    // Create Response
    out.srmGetSpaceTokensResponse = soap_new_srm2__srmGetSpaceTokensResponse(soap,-1);
    out.srmGetSpaceTokensResponse->returnStatus = soap_new_srm2__TReturnStatus(soap,-1);
    out.srmGetSpaceTokensResponse->returnStatus->statusCode  = srm2__TStatusCode__SRM_USCOREFAILURE;
    out.srmGetSpaceTokensResponse->returnStatus->explanation = 0;
    out.srmGetSpaceTokensResponse->arrayOfSpaceTokens        = 0;

    if((0 == srmGetSpaceTokensRequest->userSpaceTokenDescription) || (true == srmGetSpaceTokensRequest->userSpaceTokenDescription->empty())){
        // Empty description not supported
        TRACE_LOG("Empty description not supported");
        out.srmGetSpaceTokensResponse->returnStatus->statusCode  = srm2__TStatusCode__SRM_USCORENOT_USCORESUPPORTED;
        out.srmGetSpaceTokensResponse->returnStatus->explanation = soap_new_std__string(soap,-1);
        *(out.srmGetSpaceTokensResponse->returnStatus->explanation) = "Empty Description not supported";
        return result;
    }

    TRACE_LOG("Request token for description " << (*srmGetSpaceTokensRequest->userSpaceTokenDescription));

    // Parse Token
    const char * const DESCRIPTION_PREFIX = "Description_for:";
    if(0 == srmGetSpaceTokensRequest->userSpaceTokenDescription->compare(0,strlen(DESCRIPTION_PREFIX),DESCRIPTION_PREFIX)){
        TRACE_LOG("description starts with prefix '" << DESCRIPTION_PREFIX << "'");
        std::string tokens = srmGetSpaceTokensRequest->userSpaceTokenDescription->substr(strlen(DESCRIPTION_PREFIX));
        out.srmGetSpaceTokensResponse->arrayOfSpaceTokens = soap_new_srm2__ArrayOfString(soap,-1);
        // get all the token names separated by ';'
        std::string expression("([^;]+);?");
        boost::regex tokens_re(expression);
        boost::match_results<std::string::const_iterator> what;
        boost::match_flag_type flags = boost::match_default | boost::match_not_dot_newline;
        std::string::const_iterator start = tokens.begin();
        std::string::const_iterator end = tokens.end();
        while (boost::regex_search(start,end,what,tokens_re,flags)){
            // what [0] = whole match
            // what [1] = the first group (token name without ';')
            std::string space_token = what[1];
            TRACE_LOG("Found token: " << space_token);
            out.srmGetSpaceTokensResponse->arrayOfSpaceTokens->stringArray.push_back(space_token);
            start += what[0].length();
        }

        // Set status
        out.srmGetSpaceTokensResponse->returnStatus->statusCode  = srm2__TStatusCode__SRM_USCORESUCCESS;

    } else {
        // Invalid description
        TRACE_LOG("invalid token description (prefix not found)");
        out.srmGetSpaceTokensResponse->returnStatus->statusCode  = srm2__TStatusCode__SRM_USCOREINVALID_USCOREREQUEST;
        out.srmGetSpaceTokensResponse->returnStatus->explanation = soap_new_std__string(soap,-1);
        *(out.srmGetSpaceTokensResponse->returnStatus->explanation) = "the Description does not refer to an existing space token";
    }

    TRACE_LOG("Done: srm2__srmGetSpaceTokens");
    return result;
}

/**
 * srm2__srmGetSpaceMetaData
 * @param soap
 * @param srmGetSpaceMetaDataRequest
 * @param out
 * @return
 */
int srm2__srmGetSpaceMetaData(
    struct soap*                                soap,
    srm2__srmGetSpaceMetaDataRequest *          srmGetSpaceMetaDataRequest,
    struct srm2__srmGetSpaceMetaDataResponse_ & out){

    TRACE_LOG("srm2__srmGetSpaceMetaData called");
    int result = SOAP_OK;

    // Precondition Checks
    if(0 == srmGetSpaceMetaDataRequest){
        // Error
        return soap_sender_fault(soap,"Null request sent",0);
    }
    if (0 == srmGetSpaceMetaDataRequest->arrayOfSpaceTokens){
        // Error
        return soap_sender_fault(soap,"Empty space tokens array sent",0);
    }

    if (srm2_handle<SrmGetSpaceMetaDataTraits>(srmGetSpaceMetaDataRequest, out, soap))
      	return result;

    const std::vector<std::string>& tokens = srmGetSpaceMetaDataRequest->arrayOfSpaceTokens->stringArray;
    if (tokens.empty()) {
        TRACE_LOG("Empty tokens array in srmGetSpaceMetaData request");
        out.srmGetSpaceMetaDataResponse = soap_new_srm2__srmGetSpaceMetaDataResponse(soap,-1);
        out.srmGetSpaceMetaDataResponse->returnStatus = soap_new_srm2__TReturnStatus(soap,-1);
        out.srmGetSpaceMetaDataResponse->returnStatus->statusCode  = srm2__TStatusCode__SRM_USCOREINVALID_USCOREREQUEST;
        out.srmGetSpaceMetaDataResponse->returnStatus->explanation = soap_new_std__string(soap,-1);
        *(out.srmGetSpaceMetaDataResponse->returnStatus->explanation) = "arrayOfSpaceTokens is empty.";
        out.srmGetSpaceMetaDataResponse->arrayOfSpaceDetails = 0;
        return result;
    }

    TRACE_LOG("Request to get space metadata for " << tokens.size() << " spaces");

    std::vector<std::string>::const_iterator it;
    for(it = tokens.begin(); it != tokens.end();++it){
        TRACE_LOG("Space : " << (*it));
    }

    // Create Response
    out.srmGetSpaceMetaDataResponse = soap_new_srm2__srmGetSpaceMetaDataResponse(soap,-1);
    out.srmGetSpaceMetaDataResponse->returnStatus = soap_new_srm2__TReturnStatus(soap,-1);
    out.srmGetSpaceMetaDataResponse->returnStatus->statusCode  = srm2__TStatusCode__SRM_USCOREFAILURE;
    out.srmGetSpaceMetaDataResponse->returnStatus->explanation = 0;
    out.srmGetSpaceMetaDataResponse->arrayOfSpaceDetails = 0;

    for(it = tokens.begin(); it != tokens.end();++it){
        if (SpaceMetaDataRule::instance().is_request_error(*it,out.srmGetSpaceMetaDataResponse->returnStatus->statusCode)){
            // matched request-level pattern for return code
            out.srmGetSpaceMetaDataResponse->returnStatus->explanation = soap_new_std__string(soap,-1);
            *(out.srmGetSpaceMetaDataResponse->returnStatus->explanation) = "Returning request status from token name.";
            return result;
        }
    }

    out.srmGetSpaceMetaDataResponse->arrayOfSpaceDetails = soap_new_srm2__ArrayOfTMetaDataSpace(soap,-1);
    bool has_success = false;
    bool has_failed = false;
    for(it = tokens.begin(); it != tokens.end();++it){
        // create a srm2__TMetaDataSpace object
        srm2__TMetaDataSpace * s = soap_new_srm2__TMetaDataSpace(soap,-1);
        s->spaceToken = (*it);
        s->status = soap_new_srm2__TReturnStatus(soap,-1);
        s->status->statusCode  = srm2__TStatusCode__SRM_USCOREFAILURE;
        s->status->explanation = 0;

        s->retentionPolicyInfo = 0;
        s->owner = 0;
        s->totalSize = 0;
        s->guaranteedSize = 0;
        s->unusedSize = 0;
        s->lifetimeAssigned = 0;
        s->lifetimeLeft = 0;


        if (SpaceMetaDataRule::instance().is_token_error(*it,s->status->statusCode)){
            TRACE_LOG("Returning token status from token name.");
            s->status->explanation = soap_new_std__string(soap,-1);
            *(s->status->explanation) = "Returning token status from token name.";
        } else if (SpaceMetaDataRule::instance().is_token_rule(*it,soap,*s)){
            TRACE_LOG("Applied token rule.");
            s->status->statusCode  = srm2__TStatusCode__SRM_USCORESUCCESS;
        } else {
            TRACE_LOG("No such token.");
            s->status->statusCode  = srm2__TStatusCode__SRM_USCOREINVALID_USCOREREQUEST;
            s->status->explanation = soap_new_std__string(soap,-1);
            *(s->status->explanation) = "No such token.";
        }

        if((0 == s->status) || (s->status->statusCode == srm2__TStatusCode__SRM_USCORESUCCESS)){
            has_success = true;
        } else {
            has_failed  = true;
        }
        out.srmGetSpaceMetaDataResponse->arrayOfSpaceDetails->spaceDataArray.push_back(s);
    }

    // Set Result
    if(true == has_success){
        if(true == has_failed){
            out.srmGetSpaceMetaDataResponse->returnStatus->statusCode = srm2__TStatusCode__SRM_USCOREPARTIAL_USCORESUCCESS;
            out.srmGetSpaceMetaDataResponse->returnStatus->explanation = soap_new_std__string(soap,-1);
            *(out.srmGetSpaceMetaDataResponse->returnStatus->explanation) = "Some tokens failed.";
        } else {
            out.srmGetSpaceMetaDataResponse->returnStatus->statusCode = srm2__TStatusCode__SRM_USCORESUCCESS;
        }
    } else {
        out.srmGetSpaceMetaDataResponse->returnStatus->explanation = soap_new_std__string(soap,-1);
        *(out.srmGetSpaceMetaDataResponse->returnStatus->explanation) = "All tokens failed.";
    }

    TRACE_LOG("Done: srm2__srmGetSpaceMetaData");
    return result;
}

// -----------------------------------------------------------------------------
// Not Implemented
// -----------------------------------------------------------------------------
int srm2__srmReserveSpace(struct soap* soap, srm2__srmReserveSpaceRequest *srmReserveSpaceRequest, struct srm2__srmReserveSpaceResponse_ &_param_1);
int srm2__srmStatusOfReserveSpaceRequest(struct soap* soap, srm2__srmStatusOfReserveSpaceRequestRequest *srmStatusOfReserveSpaceRequestRequest, struct srm2__srmStatusOfReserveSpaceRequestResponse_ &_param_2);
int srm2__srmReleaseSpace(struct soap* soap, srm2__srmReleaseSpaceRequest *srmReleaseSpaceRequest, struct srm2__srmReleaseSpaceResponse_ &_param_3);
int srm2__srmUpdateSpace(struct soap* soap, srm2__srmUpdateSpaceRequest *srmUpdateSpaceRequest, struct srm2__srmUpdateSpaceResponse_ &_param_4);
int srm2__srmStatusOfUpdateSpaceRequest(struct soap* soap, srm2__srmStatusOfUpdateSpaceRequestRequest *srmStatusOfUpdateSpaceRequestRequest, struct srm2__srmStatusOfUpdateSpaceRequestResponse_ &_param_5);
int srm2__srmChangeSpaceForFiles(struct soap* soap, srm2__srmChangeSpaceForFilesRequest *srmChangeSpaceForFilesRequest, struct srm2__srmChangeSpaceForFilesResponse_ &_param_7);
int srm2__srmStatusOfChangeSpaceForFilesRequest(struct soap* soap, srm2__srmStatusOfChangeSpaceForFilesRequestRequest *srmStatusOfChangeSpaceForFilesRequestRequest, struct srm2__srmStatusOfChangeSpaceForFilesRequestResponse_ &_param_8);
int srm2__srmExtendFileLifeTimeInSpace(struct soap* soap, srm2__srmExtendFileLifeTimeInSpaceRequest *srmExtendFileLifeTimeInSpaceRequest, struct srm2__srmExtendFileLifeTimeInSpaceResponse_ &_param_9);
int srm2__srmPurgeFromSpace(struct soap* soap, srm2__srmPurgeFromSpaceRequest *srmPurgeFromSpaceRequest, struct srm2__srmPurgeFromSpaceResponse_ &_param_10);
int srm2__srmSetPermission(struct soap* soap, srm2__srmSetPermissionRequest *srmSetPermissionRequest, struct srm2__srmSetPermissionResponse_ &_param_12);
int srm2__srmCheckPermission(struct soap* soap, srm2__srmCheckPermissionRequest *srmCheckPermissionRequest, struct srm2__srmCheckPermissionResponse_ &_param_13);
int srm2__srmGetPermission(struct soap* soap, srm2__srmGetPermissionRequest *srmGetPermissionRequest, struct srm2__srmGetPermissionResponse_ &_param_14);
int srm2__srmStatusOfLsRequest(struct soap* soap, srm2__srmStatusOfLsRequestRequest *srmStatusOfLsRequestRequest, struct srm2__srmStatusOfLsRequestResponse_ &_param_19);
int srm2__srmMv(struct soap* soap, srm2__srmMvRequest *srmMvRequest, struct srm2__srmMvResponse_ &_param_20);
int srm2__srmSuspendRequest(struct soap* soap, srm2__srmSuspendRequestRequest *srmSuspendRequestRequest, struct srm2__srmSuspendRequestResponse_ &_param_33);
int srm2__srmResumeRequest(struct soap* soap, srm2__srmResumeRequestRequest *srmResumeRequestRequest, struct srm2__srmResumeRequestResponse_ &_param_34);
int srm2__srmGetRequestSummary(struct soap* soap, srm2__srmGetRequestSummaryRequest *srmGetRequestSummaryRequest, struct srm2__srmGetRequestSummaryResponse_ &_param_35);
int srm2__srmExtendFileLifeTime(struct soap* soap, srm2__srmExtendFileLifeTimeRequest *srmExtendFileLifeTimeRequest, struct srm2__srmExtendFileLifeTimeResponse_ &_param_36);
int srm2__srmGetRequestTokens(struct soap* soap, srm2__srmGetRequestTokensRequest *srmGetRequestTokensRequest, struct srm2__srmGetRequestTokensResponse_ &_param_37);
int srm2__srmGetTransferProtocols(struct soap* soap, srm2__srmGetTransferProtocolsRequest *srmGetTransferProtocolsRequest, struct srm2__srmGetTransferProtocolsResponse_ &_param_38);
} // end namespace srm2
