/*
 *  Copyright (c) Members of the EGEE Collaboration. 2004.
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  GLite Data SRM Util CPP: Srm 2.2 Server (taken from the org.glite.data.io-resolve test)
 *
 *  Authors: Paolo Badino <paolo.badino@cern.ch>
 *  Version info: $Id: service.cpp,v 1.7 2009/06/21 17:09:31 molnarzs Exp $
 *  Release: $Name:  $
 *
 */
#include "glite/data/config/service/ServiceConfiguration.hxx"
#include "glite/data/config/service/ParamValue.hxx"
#include "log4cpp/Category.hh"
#include "log.h"

#include <signal.h>
#include <sys/wait.h>
#include <unistd.h>
#include <pthread.h>
#include <execinfo.h>

#ifndef SOAP_FMAC3S
#define SOAP_FMAC3S
#endif
#ifndef SOAP_FMAC4S
#define SOAP_FMAC4S
#endif

#include "glite/data/srm2/cpp/envH.h"
#include "glite/data/srm2/cpp/srm2.nsmap"

/*
 * Include CGSI GSOAP plugin
 */
extern "C"{
#include "cgsi_plugin.h"
}

/**
 * Global variable used by the server code
 */
std::string g_pool_path;
pthread_mutex_t g_lock = PTHREAD_MUTEX_INITIALIZER;

using namespace log4cpp;
using namespace glite::config;

#define QUEUE_LEN    100
#define NUM_THREADS  10

namespace glite{
namespace data {
namespace srm  {
namespace util {
namespace test {

class SrmService: public ServiceConfiguration {
public:

    typedef enum {
        SP_NONE = 0,
        SP_SSL,
        SP_GSI
    } SecureProtocol;

    /**
     * Initialize the Component
     */
    virtual int init(const Params& params);

    /**
     * Configure the Component
     */
    virtual int config(const Params& params);

    /**
     * Override start since the service should start in config, to the test
     * could be runned at the same time
     */
    virtual int start();

    /**
     * Finailize the Component
     */
    virtual int fini();

    /**
     * Return Singleton Instance
     */
    static SrmService * instance();

    /**
     * Finalize Singleton Instance
     */
    static void finalize();

    /**
     * process soap request
     */
    void * process(struct soap* soap);

private:
    /**
     * Constructor
     */
    SrmService();

    /**
     * Destructor
     */
    virtual ~SrmService(){}

    /**
     * Service Main Method
     */
    virtual int svc();

    /**
     * Callback to interrupt the Service Thread
     */
    virtual int svc_stop();

    /**
     * Enqueue request
     */
    int enqueue(SOAP_SOCKET s);

    /**
     * Dequeue request
     */
    SOAP_SOCKET dequeue();

private:
    /**
     * Singleton Instance
     */
    static SrmService * s_instance;

    /**
     * The Logger
     */
    mutable class Category& m_logger;

    /**
     * Run Flag
     */
    sig_atomic_t m_run;

    /**
     * Port
     */
    int m_port;

    /**
     * SecureServer
     */
    SecureProtocol m_secureProtocol;

    /**
     * EarlyStart
     */
    bool m_earlyStart;

    /**
     *  Soap request Queue
     */
    int m_head;
    int m_tail;
    SOAP_SOCKET      m_queue[QUEUE_LEN];
    pthread_mutex_t  m_queue_cs;
    pthread_cond_t   m_queue_cv;
};

} // End namespace test
} // End namespace util
} // End namespace srm
} // End namespace data
} // End namespace glite

using namespace glite::data::srm::util::test;

extern "C" {

/**
 * Create Component instance
 */
ComponentConfiguration * create_glite_component(){
    return SrmService::instance();
}

/**
 * Destroy Component instance
 */
void destroy_glite_component(ComponentConfiguration * component){
    SrmService::finalize();
}
} // End extern "C"

namespace {
const char * PORT_PROPERTY_NAME           = "Port";
const char * SECUREPROTOCOL_PROPERTY_NAME = "SecureProtocol";
const char * EARLYSTART_PROPERTY_NAME     = "EarlyStart";
const char * POOL_PROPERTY_NAME           = "PoolLocation";

int                        DEFAULTPORT            = 8044;
SrmService::SecureProtocol DEFAULTSECUREPROTOCOL  = SrmService::SP_NONE;
bool                       DEFAULTEARLYSTART      = false;
const char * const         DEFAULTPOOLLOCATION    = "/tmp/test/srm-util/srm22";
}
SrmService* SrmService::s_instance = 0;

namespace srm2{
int create_dir_path(const std::string& file_path, bool recursive);
}

/**
 * Return Singleton Instance
 */
SrmService * SrmService::instance(){
    if(0 == s_instance){
        s_instance = new SrmService();
    }
    return s_instance;
}

/**
 * Finalize Singleton Instance
 */
void SrmService::finalize(){
    if(0 != s_instance){
        delete s_instance;
        s_instance = 0;
    }
}

SrmService::SrmService():ServiceConfiguration("srm2-service-stub"),
   m_logger(Category::getInstance("srm2-service-stub")),m_run(0){
     m_port           = DEFAULTPORT;
     m_secureProtocol = DEFAULTSECUREPROTOCOL;
     m_earlyStart     = DEFAULTEARLYSTART;
     m_head = 0;
     m_tail = 0;
}

/**
 * Initialize the Component
 */
int SrmService::init(const Params& params){
    m_logger.log(Priority::DEBUG,"<%s> initialization completed",getName());
    return 0;
}

/**
 * Configure the Component
 */
int SrmService::config(const Params& params){

    Params::const_iterator it;
    if((it = params.find(PORT_PROPERTY_NAME)) != params.end()){
        ParamValue * pv = dynamic_cast<ParamValue *>((*it).second);
        if(0 == pv){
            m_logger.log(Priority::ERROR,"%s configuration Error. Invalid Param Format: %s ",getName(),(*it).first.c_str());
            return -1;
        }
        m_port = atoi(pv->getValue().c_str());
    }
    if((it = params.find(SECUREPROTOCOL_PROPERTY_NAME)) != params.end()){
        ParamValue * pv = dynamic_cast<ParamValue *>((*it).second);
        if(0 == pv){
            m_logger.log(Priority::ERROR,"%s configuration Error. Invalid Param Format: %s ",getName(),(*it).first.c_str());
            return -1;
        }
        const std::string& val = pv->getValue();
        if(0 == strcasecmp(val.c_str(),"ssl" )){
            m_secureProtocol = SrmService::SP_SSL;
        } else if(0 == strcasecmp(val.c_str(),"gsi" )){
            m_secureProtocol = SrmService::SP_GSI;
        } else {
            m_secureProtocol = SrmService::SP_NONE;
        }
    }
    if((it = params.find(EARLYSTART_PROPERTY_NAME)) != params.end()){
        ParamValue * pv = dynamic_cast<ParamValue *>((*it).second);
        if(0 == pv){
            m_logger.log(Priority::ERROR,"%s configuration Error. Invalid Param Format: %s ",getName(),(*it).first.c_str());
            return -1;
        }
        const std::string& val = pv->getValue();
        m_earlyStart = ((0 == strcasecmp(val.c_str(),"true" )) || (0 == strcasecmp(val.c_str(),"yes")))?true:false;
    }
    if((it = params.find(POOL_PROPERTY_NAME)) != params.end()){
        ParamValue * pv = dynamic_cast<ParamValue *>((*it).second);
        if(0 == pv){
            m_logger.log(Priority::ERROR,"%s configuration Error. Invalid Param Format: %s ",getName(),(*it).first.c_str());
            return -1;
        }
        g_pool_path = pv->getValue();
    } else {
        g_pool_path = DEFAULTPOOLLOCATION;
    }

    // Check il pool exist
    int r  = 0;
    if(0 != ::access(g_pool_path.c_str(),F_OK)){
        // Create dir path
        r = ::srm2::create_dir_path(g_pool_path,true);
        if(0 != r){
            m_logger.log(Priority::ERROR,"%s configuration Error. Failed to create pool",getName());
            return -1;
        }
    }

    m_logger.log(Priority::DEBUG,"<%s> configuration completed",getName());

    m_logger.log(Priority::DEBUG,"%s configured. Parameters are:",getName());
    m_logger.log(Priority::DEBUG,"Port                : %d",m_port);
    m_logger.log(Priority::DEBUG,"SecureProtocol      : %d",m_secureProtocol);
    m_logger.log(Priority::DEBUG,"Pool Location       : %s",g_pool_path.c_str());

    if(true == m_earlyStart){
        m_logger.log(Priority::DEBUG,"<%s> Start the thread",getName());
        return ServiceConfiguration::start();
    }
    return 0;
}

/*
 * Override start since the service should start in config, to the test
 * could be runned at the same time
 */
int SrmService::start(){
    if(false == m_earlyStart){
        return ServiceConfiguration::start();
    }
    return 0;
}

/**
 * Finalize the Component
 */
int SrmService::fini(){

    m_logger.log(Priority::DEBUG,"<%s> finalization completed",getName());

    return 0;
}

/**
 * Process queue
 */
void * process_queue(void* soap){
    struct soap * tsoap = (struct soap*)(soap);
    return SrmService::instance()->process(tsoap);
}

/**
 * Process queue
 */
void * SrmService::process(struct soap * soap){
    pthread_t id = pthread_self();
    m_logger.log(Priority::DEBUG,"Thread [%d] main",(int)id);
    while(m_run){
        soap->socket = dequeue();
        if(!soap_valid_socket(soap->socket)){
            m_logger.log(Priority::DEBUG,"Stopping thread [%d]", (int)id);
            break;
        }
        m_logger.log(Priority::DEBUG,"[%d] - Serving request", (int)id);
    pthread_mutex_lock(&g_lock);
        ::srm2::srm2_serve(soap);
    pthread_mutex_unlock(&g_lock);
        m_logger.log(Priority::DEBUG,"[%d] - Request served" , (int)id);
        soap_destroy(soap);
        soap_end(soap);
    }
    return NULL;
}

/**
 * Enqueue request
 */
int SrmService::enqueue(SOAP_SOCKET sock){
    int status = SOAP_OK;
    int next;
    pthread_mutex_lock(&m_queue_cs);
    next = m_tail + 1;
    if(next >= QUEUE_LEN){
        next = 0;
    }
    if(next == m_head){
        status = SOAP_EOM;
    } else {
        m_queue[m_tail] = sock;
        m_tail = next;
    }
    m_logger.log(Priority::DEBUG,"Request enqueued");
    pthread_cond_signal(&m_queue_cv);
    pthread_mutex_unlock(&m_queue_cs);
    return status;
}

/**
 * Dequeue request
 */
SOAP_SOCKET SrmService::dequeue(){
    SOAP_SOCKET sock = -1;

    pthread_mutex_lock(&m_queue_cs);
    while((m_run) && (m_head == m_tail)){
        struct timespec timeout;
        timeout.tv_sec  = time(0) + 1;
        timeout.tv_nsec = 0;
        int r = pthread_cond_timedwait(&m_queue_cv,&m_queue_cs,&timeout);
        if(r == ETIMEDOUT){
        //m_logger.log(Priority::DEBUG,"wait timed out");
        } else if(r == EINTR){
        //m_logger.log(Priority::DEBUG,"wait interrupted");
        } else if(r == EBUSY){
        //m_logger.log(Priority::DEBUG,"another thtead waiting");
        }
    }
    if(!m_run){
        m_logger.log(Priority::DEBUG,"service should stop");
    } else {
        sock = m_queue[m_head++];
        if(m_head >= QUEUE_LEN){
            m_head = 0;
        }
        m_logger.log(Priority::DEBUG,"Request dequeued");
    }
    pthread_mutex_unlock(&m_queue_cs);
    return sock;
}

void log_stack(int)
{
   const int stack_size = 25;
   void * array[stack_size];
   int nSize = backtrace(array, stack_size);
   char ** symbols = backtrace_symbols(array, nSize);

   for (int i = 0; i < nSize; ++i) {
	 	TRACE_LOG(symbols[i]);
   }
   free(symbols);
}

/**
 * TestService Main
 */
int SrmService::svc()
{
	//signal(SIGABRT, log_stack);
	//signal(SIGSEGV, log_stack);

    m_run = 1;

    // Init soap
    int m; // master socket
    struct soap soap;

    struct soap * soap_thr[NUM_THREADS];
    pthread_t tid[NUM_THREADS];

    soap_init(&soap);
    soap.namespaces = srm2_namespaces;

    // Set Security Plugin
    if(SrmService::SP_NONE != m_secureProtocol){
        m_logger.log(Priority::DEBUG,"%s Service requires secure protocol",getName());
        int flags = CGSI_OPT_DISABLE_NAME_CHECK;
        if(SrmService::SP_SSL == m_secureProtocol){
            m_logger.log(Priority::DEBUG,"%s Service requires SSL secure protocol",getName());
            flags |= CGSI_OPT_SSL_COMPATIBLE;
        }
        soap_register_plugin_arg (&soap, server_cgsi_plugin, &flags);
    }

    soap.accept_timeout = 1;
    m = soap_bind(&soap, "localhost", m_port, 100);
    if(!soap_valid_socket(m)) {
        soap_print_fault(&soap, stderr);
    } else {
        int i = 0;
        // starting thread pool
        pthread_mutex_init(&m_queue_cs,NULL);
        pthread_cond_init(&m_queue_cv,NULL);
        for(i = 0; i < NUM_THREADS;++i){
            soap_thr[i] = soap_copy(&soap);
            m_logger.log(Priority::DEBUG,"Starting thread %d",i);
            pthread_create(&tid[i],NULL,(void*(*)(void*))process_queue,(void*)soap_thr[i]);
            m_logger.log(Priority::DEBUG,"Thread %d started with id [%d]",i,(int)tid[i]);
        }
        m_logger.log(Priority::DEBUG,"%s Service Started",getName());
        while(m_run){
            int s = soap_accept(&soap);
            if (!soap_valid_socket(s)) {
                //check if it's a timeout
                if(soap.errnum){
                    m_logger.log(Priority::ERROR,"Accept failed");
                    soap_print_fault(&soap, stderr);
                    continue;
                } else {
                        // timeout
                    //m_logger.log(Priority::DEBUG,"Accept timeout");
                    continue;
                }
            }
            m_logger.log(Priority::DEBUG,"%s: accepted connection from IP=%d.%d.%d.%d socket=%d", getName(),
                         (int)((soap.ip >> 24)&0xFF), (int)((soap.ip >> 16)&0xFF), (int)((soap.ip >> 8)&0xFF), (int)(soap.ip&0xFF), (int)(s));
            while(enqueue(s) == SOAP_EOM){
                sleep(1);
            }
        }
        for(i = 0; i < NUM_THREADS;++i){
            m_logger.log(Priority::DEBUG,"Waiting for thread %d to complete",i);
            pthread_join(tid[i],NULL);
            soap_done(soap_thr[i]);
            free(soap_thr[i]);
        }

        pthread_mutex_destroy(&m_queue_cs);
        pthread_cond_destroy(&m_queue_cv);
        // cleanup everything and close socket
        soap_done(&soap);
    }
    m_logger.log(Priority::DEBUG,"Service %s Stopped",getName());

    return 0;
}



/**
 * Stop TestService Main
 */
int SrmService::svc_stop(){
    m_logger.log(Priority::DEBUG,"Stopping %s Service",getName());
    m_run = 0;

    return 0;
}
